#include "bp2strng.h"

#include <bppvm.h>
#include <bppclass.h>
#include <bppvalue.h>
#include <bppdict.h>
#include <bppvec.h>
#include <bppstrng.h>
#include <bppobjct.h>
#include <bppint.h>
#include <bppprint.h>
#include <bppbuf.h>

#include "bp2cldef.h"

static unsigned _strIndex = 0;


bool Bp2String_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   BppValue& strVal = MEMBER(_strIndex);
   if (argc == 1)
   {
      strVal = new BppString();
      return true;
   }
   BppValue savedSpVal = *sp; // save TOS value before function call
   BppString* s = bp[2];
	BppValue v;
	if (!s) {
		if (interpreter->doStrCast(v, bp[2]))
			s = v;
	}
   if (!s)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (argc == 2)
   {
      strVal = new BppString(*s);
   }
   else
   {
      // helper to prevent s from release
      BppValue savedStrVal = *sp;
      *sp = savedSpVal; // restore saved TOS value (last argument)
      ValPtr pv = bp;
      pv += 3;
      strVal = new BppString(s->c_str(),argc-2,pv);
   }
   *bp = pThis;
   sp = bp;
   return true;
}


bool Bp2String_OP_ADD(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppString* s = interpreter->getStringFromValueExtern(bp[2]);
   if (s == 0)
   {
      if (!bp[2].isIntegralType())
         return interpreter->badType(bp[2].getType(),BVT_STRING);
   }
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);

   BppObject* result = new BppObject(pThis->getClass());
   BppValue& resValue = (*(result->getMemberData()))[_strIndex];

   BppString* resString = new BppString(*str);
   resValue = resString;
   if ((s != 0))
      (*resString)+= s->c_str();
   else
      (*resString)+= (BppChar)(BppInt)bp[2];
   *bp = result;
   sp = bp;
   return true;
}

bool Bp2String_OP_ADD_R(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppString* s = interpreter->getStringFromValueExtern(bp[2]);
   if (s == 0)
   {
      if (!bp[2].isIntegralType())
         return interpreter->badType(bp[2].getType(),BVT_STRING);
   }
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);

   BppObject* result = new BppObject(pThis->getClass());
   BppValue& resValue = (*(result->getMemberData()))[_strIndex];

   BppString* resString = new BppString(*str);
   resValue = resString;
   if ((s != 0))
      resString->insert(0,*s);
   else
      resString->insert(0,(BppChar)(BppInt)bp[2]);
   *bp = result;
   sp = bp;
   return true;
}

bool Bp2String_OP_VREF(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppUInt idx = bp[2];
   if (idx >= (BppUInt) str->length())
      THROW(_T("subscript out of range"));
   *bp = (BppInt) (*str)[(size_t)idx];
   sp = bp;
   return true;
}

bool Bp2String_OP_VSET(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,3);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   CHKTYPE(bp[3],BVT_INTNUMBER);
   BppUInt idx = bp[2];
   if (idx >= (BppUInt) str->length())
      THROW(_T("subscript out of range"));
	BppInt v = bp[3];
   (*str)[(size_t)idx] = (BppChar) v;
   *bp = v;
   sp = bp;
   return true;
}

bool Bp2String_compare(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   BppString* s = bp[2];
	BppValue v;
	if (!s) {
		if (interpreter->doStrCast(v, bp[2]))
			s = v;
	}
   if (!s)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   *bp = (BppInt) str->compare(*s);
   sp = bp;
   return true;
}

bool Bp2String_icompare(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
	BppString* s = bp[2];
	BppValue v;
	if (!s) {
		if (interpreter->doStrCast(v, bp[2]))
			s = v;
	}
	if (!s)
		return interpreter->badType(strVal.getType(),BVT_STRING);
   *bp = (BppInt) str->icompare(*s);
   sp = bp;
   return true;
}

bool Bp2String_length(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   *bp = (BppUInt) str->length();
   sp = bp;
   return true;
}

bool Bp2String_c_str(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   *bp = BppValue((BppIntPtr) str->c_str(),true);
   sp = bp;
   return true;
}

bool Bp2String_c_buf(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   size_t l = str->length() +1;
   BppCharBuffer* cbuf = new BppCharBuffer(l);
   *bp = cbuf;
   _tcscpy_s(cbuf->data(),l,str->c_str());
   cbuf->adjustSize();
   sp = bp;
   return true;
}

bool Bp2String_a_buf(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   size_t l = (str->length() +1) * sizeof(char);
   BppBuffer* buf = new BppBuffer(l);
   *bp = buf;
   memcpy(buf->data(),str->a_str(),l);
   buf->setSize(l);
   sp = bp;
   return true;
}

bool Bp2String_w_buf(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   size_t l = (str->length() +1) * sizeof(wchar_t);
   BppBuffer* buf = new BppBuffer(l);
   *bp = buf;
   memcpy(buf->data(),str->w_str(),l);
   buf->setSize(l);
   sp = bp;
   return true;
}

bool Bp2String_clear(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   str->clear();
   sp->dispose();
   return true;
}

bool Bp2String_find(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,3);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   size_t pos = 0;
   if (argc == 3)
   {
      CHKTYPE(bp[3],BVT_INTNUMBER);
      pos = (size_t) (BppUInt) bp[3];
   }
   BppValue& arg1 = bp[2];
	size_t res = 0;
   if (arg1.isIntegralType())
   {
      res = str->find((BppChar)(BppInt)arg1,pos);
   }
   else
   {
      BppString* s = interpreter->getStringFromValueExtern(arg1);
      if (s == 0)
         return interpreter->badType(arg1.getType(),BVT_STRING);
      res = str->find(*s,pos);
   }
	*bp = (res == BPP_NNPOS) ? BPP_NPOS : BppUInt(res);
   sp = bp; 
   return true;
}

bool Bp2String_replace(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,3);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   BppValue arg1 = bp[2];
   BppValue arg2 = bp[3];
   if (arg1.isIntegralType())
   {
      CHKTYPE(arg2,BVT_INTNUMBER);
      str->replace((BppChar)(BppInt)arg1,(BppChar)(BppInt)arg2);
   }
   else
   {
      BppString* s1 = interpreter->getStringFromValueExtern(arg1);
      if (s1 == 0)
         return interpreter->badType(arg1.getType(),BVT_STRING);
      BppString* s2 = interpreter->getStringFromValueExtern(arg2);
      if (s2 == 0)
         return interpreter->badType(arg2.getType(),BVT_STRING);
      str->replace(*s1,*s2);
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2String_substr( BppVirtualMachine* vm, int argc, ValPtr& sp )
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,3);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   size_t start = (BppInt) bp[2];
   BppUInt len = BPP_NPOS;
   if (argc == 3)
   {
      CHKTYPE(bp[3],BVT_INTNUMBER);
      len = bp[3];
   }
   BppObject* result = new BppObject(pThis->getClass());
   BppValue& resValue = (*(result->getMemberData()))[_strIndex];

   BppString* resString = new BppString(str->substr(start,len));
   resValue = resString;
   *bp = result;
   sp = bp;
   return true;
}

bool Bp2String_erase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,3);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   size_t pos = (size_t) (BppInt) bp[2];
   BppUInt cnt = BPP_NPOS;
   if (argc == 3)
   {
      CHKTYPE(bp[3],BVT_INTNUMBER);
      cnt = bp[3];
   }
   str->erase(pos,cnt);
   *bp = pThis;
   sp = bp;
   return true;
}


bool Bp2String_insert(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   size_t pos = (size_t) (BppUInt) bp[2];
   BppValue arg2 = bp[3];
   if (arg2.isIntegralType())
   {
      size_t cnt = 1;
      if (argc == 4)
      {
         CHKTYPE(bp[4],BVT_INTNUMBER);
         cnt = (size_t) (BppUInt) bp[4];
      }
      str->insert(pos,(BppChar)(BppInt)arg2,cnt);
   }
   else
   {
      ARGCMAX(3);
      BppString* s = interpreter->getStringFromValueExtern(arg2);
      if (s == 0)
         return interpreter->badType(arg2.getType(),BVT_STRING);
      str->insert(pos,*s);
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2String_append( BppVirtualMachine* vm, int argc, ValPtr& sp )
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   BppValue arg = bp[2];
   if (arg.isIntegralType())
      (*str) += (BppChar)(BppInt)arg;
   else
   {
      BppString* s = interpreter->getStringFromValueExtern(arg);
      if (s == 0)
         return interpreter->badType(arg.getType(),BVT_STRING);
      (*str) += *s;
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2String_toString(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& strVal = MEMBER(_strIndex);
   BppString* str = strVal;
   if (str == 0)
      return interpreter->badType(strVal.getType(),BVT_STRING);
   *bp = new BppString(*str);
   sp = bp;
   return true;
}

/** 
Function is called from RegisterNativeModule function.
It initializes the String class and registers it in virtual machine <i>vm</i>.
@see RegisterNativeModule, BppVirtualMachine::registerNativeClass, BppClass::addMethod,
BppClass::addDataMember, BppClass::fixupDataDefs, BppClass::getMemberIndex
*/
void Bp2String_CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_T("String"),pClass)) return;
   pClass->addDataMember(_T("_str"));
   pClass->addMethod(_T("String"),&Bp2String_ctor);
   pClass->addMethod(_T("OP_ADD"),&Bp2String_OP_ADD);
   pClass->addMethod(_T("OP_ADD_R"),&Bp2String_OP_ADD_R);
   pClass->addMethod(_T("OP_VREF"),&Bp2String_OP_VREF);
   pClass->addMethod(_T("OP_VSET"),&Bp2String_OP_VSET);
   pClass->addMethod(_T("compare"),&Bp2String_compare);
   pClass->addMethod(_T("icompare"),&Bp2String_icompare);
   pClass->addMethod(_T("length"),&Bp2String_length);
   pClass->addMethod(_T("c_str"),&Bp2String_c_str);
   pClass->addMethod(_T("c_buf"),&Bp2String_c_buf);
   pClass->addMethod(_T("a_buf"),&Bp2String_a_buf);
   pClass->addMethod(_T("w_buf"),&Bp2String_w_buf);
   pClass->addMethod(_T("clear"),&Bp2String_clear);
   pClass->addMethod(_T("substr"),&Bp2String_substr);
   pClass->addMethod(_T("find"),&Bp2String_find);
   pClass->addMethod(_T("replace"),&Bp2String_replace);
   pClass->addMethod(_T("insert"),&Bp2String_insert);
   pClass->addMethod(_T("append"),&Bp2String_append);
   pClass->addMethod(_T("toString"),&Bp2String_toString);
   pClass->addMethod(_T("string"),&Bp2String_toString);
   pClass->fixupDataDefs();
   _strIndex = pClass->getMemberIndex(_T("_str"));
}

