#include "bp2struc.h"

#include <bppvm.h>
#include <bppclass.h>
#include <bppvalue.h>
#include <bppdict.h>
#include <bppvec.h>
#include <bppstrng.h>
#include <bppobjct.h>
#include <bppint.h>
#include <bppprint.h>
#include <bppbuf.h>
#include <bppdict.h>
#include "bp2cldef.h"

#define IDXPOS 0
#define IDXVAL 1
#define IDXSIZE 2
#define IDXPTRFLAG 3

static unsigned _dataDictIndex = 0;
static unsigned _dataIndex = 0;


bool Bp2Struct_ctor( BppVirtualMachine* vm, int argc, ValPtr& sp )
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   dictVal = new BppDictionary();
   BppValue& dataVal = MEMBER(_dataIndex);
   dataVal = new BppBuffer();
   *bp = pThis;
   sp = bp;
   return true;
}

// int bytepos = addInt(name,value,pos=-1);
bool Bp2Struct_addInt(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);
   
   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   BppInt v = bp[3];

   for (size_t s=datasize;s< (size_t)pos+sizeof(BppInt);++s) data->push_back(0);
   memcpy(data->data()+pos,&v,sizeof(BppInt));
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(v);
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 0); // no pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addIntPtr(name,value,pos=-1);
bool Bp2Struct_addIntPtr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }


   for (size_t s=datasize;s< (size_t)pos+sizeof(void*);++s) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back((BppIntPtr)bp[3]);
   void* v = (*entryVec)[IDXVAL].addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addInt8(name,value,pos=-1);
bool Bp2Struct_addInt8(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   char v = (char)(BppInt) bp[3];

   for (size_t s=datasize;s< (size_t)pos+sizeof(char);++s) data->push_back(0);
   memcpy(data->data()+pos,&v,sizeof(char));
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back((BppInt)v);
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 0); // no pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addInt16(name,value,pos=-1);
bool Bp2Struct_addInt16(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   short v = (short) (BppInt) bp[3];

   for (size_t s=datasize;s< (size_t)pos+sizeof(short);++s) data->push_back(0);
   memcpy(data->data()+pos,&v,sizeof(short));
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back((BppInt)v);
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 0); // no pointer
   *bp = pos;
   sp = bp;
   return true;

}

// int bytepos = addDouble(name,value,pos=-1);
bool Bp2Struct_addDouble(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   double v =  bp[3];

   for (size_t s=datasize;s< (size_t)pos+sizeof(double);++s) data->push_back(0);
   memcpy(data->data()+pos,&v,sizeof(double));
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(v);
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 0); // no pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addDoublePtr(name,value,pos=-1);
bool Bp2Struct_addDoublePtr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }


   for (size_t s=datasize;s< (size_t)pos+sizeof(void*);++s) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back((double)bp[3]);
   void* v = (*entryVec)[IDXVAL].addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addFloat(name,value,pos=-1);
bool Bp2Struct_addFloat(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   float v =  bp[3];

   for (size_t s=datasize;s< (size_t)pos+sizeof(float);++s) data->push_back(0);
   memcpy(data->data()+pos,&v,sizeof(float));
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(v);
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 0); // no pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addFloatPtr(name,value,pos=-1);
bool Bp2Struct_addFloatPtr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos =(BppInt) datasize;
   }


   for (size_t s=datasize;s< (size_t)pos+sizeof(void*);++s) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back((float)bp[3]);
   void* v = (*entryVec)[IDXVAL].addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addString(name,value,pos=-1);
bool Bp2Struct_addString(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   BppString* s = interpreter->getStringFromValueExtern(bp[3]);
   if (s == 0)
      return interpreter->badType(bp[3].getType(),BVT_STRING);

   for (size_t sz=datasize;sz< (size_t)pos+sizeof(void*);++sz) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(s);
   void* v = (*entryVec)[IDXVAL].addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addCBuffer(name,value,pos=-1);
bool Bp2Struct_addCBuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   BppCharBuffer* cb = bp[3];
   if (cb == 0)
      return interpreter->badType(bp[3].getType(),BVT_CHARBUFFER);

   for (size_t s=datasize;s< (size_t)pos+sizeof(void*);++s) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(bp[3]);
   void* v = (*entryVec)[IDXVAL].addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addBuffer(name,value,pos=-1);
bool Bp2Struct_addBuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   BppBuffer* b = bp[3];
   if (b == 0)
      return interpreter->badType(bp[3].getType(),BVT_BUFFER);

   for (size_t s=datasize;s< (size_t)pos+sizeof(void*);++s) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(bp[3]);
   void* v = (*entryVec)[IDXVAL].addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// int bytepos = addBuffer(name,value,pos=-1);
bool Bp2Struct_addStruct(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,4);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);

   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   if (dict->find(*name) != 0)
      THROW(_T("Struct member already defined: "),name->c_str());
   size_t datasize = data->size();
   BppInt pos = (BppInt) datasize;
   if (argc == 4)
   {
      CHKTYPE(bp[4],BVT_INTNUMBER);
      pos = bp[4];
      if ((pos < 0) || (pos > datasize))
         pos = (BppInt) datasize;
   }

   BppObject* o = bp[3];
   if (o == 0)
      return interpreter->badType(bp[3].getType(),BVT_OBJECT);
   if (!o->getClass()->isSubtypeOf(pThis->getClass()))
      THROW(_T("Object is not of type Struct: "),o->getClass()->getClassName()->c_str());

   for (size_t s=datasize;s< (size_t)pos+sizeof(void*);++s) data->push_back(0);
   BppVector* entryVec = new BppVector(4);
   (*dict)[*name] = entryVec;
   entryVec->push_back(pos);
   entryVec->push_back(bp[3]);
   o = (*entryVec)[IDXVAL];
   BppValue& structBuf = (*o->getMemberData())[_dataIndex];
   void* v = structBuf.addr();
   memcpy(data->data()+pos,&v,sizeof(void*));
   entryVec->push_back((BppInt)sizeof(v));
   entryVec->push_back((BppInt) 1); // pointer
   *bp = pos;
   sp = bp;
   return true;
}

// val = getValue(name)
bool Bp2Struct_getValue(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   BppString* name = interpreter->getStringFromValueExtern(bp[2]);
   if (name == 0)
      return interpreter->badType(bp[2].getType(),BVT_STRING);
   BppDictionaryEntry* pEntry = dict->find(*name);
   if ( pEntry == 0)
      THROW(_T("Struct member not found: "),name->c_str());
   BppVector* vec = pEntry->getValue();
   *bp = (*vec)[IDXVAL];
   sp = bp;
   return true;
}

// buf = getData();
bool Bp2Struct_getData(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,1);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);
   *bp = data;
   sp = bp;
   return true;
}

// object thisObj = update()
bool Bp2Struct_update( BppVirtualMachine* vm, int argc, ValPtr& sp )
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,1);
   BppValue& dataVal = MEMBER(_dataIndex);
   BppBuffer* data = dataVal;
   if (data == 0) return interpreter->badType(dataVal.getType(),BVT_BUFFER);
   BppValue& dictVal = MEMBER(_dataDictIndex);
   BppDictionary* dict = dictVal;
   if (dict == 0) return interpreter->badType(dictVal.getType(),BVT_DICTIONARY);
   for (BppDictionaryEntry* entry = dict->getFirst();
      entry != 0; entry = entry->getNext())
   {
      BppVector* entryVec = entry->getValue();
      if (entryVec == 0) continue;
      BppValue& entryVal = (*entryVec)[IDXVAL];
      BppInt isPtr = (*entryVec)[IDXPTRFLAG];
      if (isPtr == 0) // skip pointers
      {
         void* dest = entryVal.addr();
         size_t len = (size_t)(BppInt)(*entryVec)[IDXSIZE];
         BppByte* pdata = data->data();
         pdata += (BppInt) (*entryVec)[IDXPOS];
         memcpy(dest,pdata,len);
      }
   }
   *bp = pThis;
   sp = bp;
   return true;
}

void Bp2Struct_CreateClass( BppVirtualMachine& vm )
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_T("Struct"),pClass)) return;
   pClass->addDataMember(_T("_data"));
   pClass->addDataMember(_T("_dataDict"));
   pClass->addMethod(_T("Struct"),&Bp2Struct_ctor);
   pClass->addMethod(_T("addInt"),&Bp2Struct_addInt);
   pClass->addMethod(_T("addIntPtr"),&Bp2Struct_addIntPtr);
   pClass->addMethod(_T("addInt8"),&Bp2Struct_addInt8);
   pClass->addMethod(_T("addInt16"),&Bp2Struct_addInt16);
   pClass->addMethod(_T("addDouble"),&Bp2Struct_addDouble);
   pClass->addMethod(_T("addDoublePtr"),&Bp2Struct_addDoublePtr);
   pClass->addMethod(_T("addFloat"),&Bp2Struct_addFloat);
   pClass->addMethod(_T("addFloatPtr"),&Bp2Struct_addFloatPtr);
   pClass->addMethod(_T("addString"),&Bp2Struct_addString);
   pClass->addMethod(_T("addCBuffer"),&Bp2Struct_addCBuffer);
   pClass->addMethod(_T("addBuffer"),&Bp2Struct_addBuffer);
   pClass->addMethod(_T("addStruct"),&Bp2Struct_addStruct);
   pClass->addMethod(_T("getValue"),&Bp2Struct_getValue);
   pClass->addMethod(_T("getData"),&Bp2Struct_getData);
   pClass->addMethod(_T("update"),&Bp2Struct_update);
   pClass->fixupDataDefs();
   _dataIndex = pClass->getMemberIndex(_T("_data"));
   _dataDictIndex = pClass->getMemberIndex(_T("_dataDict"));
}

