#include "bp2vec.h"

#include <bppvm.h>
#include <bppclass.h>
#include <bppvalue.h>
#include <bppdict.h>
#include <bppvec.h>
#include <bppstrng.h>
#include <bppobjct.h>
#include <bppint.h>
#include <bppprint.h>
#include <bppbuf.h>

#include "bp2cldef.h"

static unsigned _vecIndex = 0;

bool Bp2Vector_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,3);
   BppValue& vecVal = MEMBER(_vecIndex);
   if (argc == 1)
   {
      vecVal = new BppVector();
      return true;
   }
   if (argc == 2)
   {
      BppVector* pv = bp[2];
      if (pv == 0)
      {
         BppObject* pobj = bp[2];
         if (pobj != 0)
         {
            BppClass* pclass = pobj->getClass();
            if (pclass->isSubtypeOf(pThis->getClass()))
               pv = (*(pobj->getMemberData()))[_vecIndex];
         } // if pobj
      }
      if (pv != 0)
      {
          vecVal = new BppVector(*pv);
          *bp = pThis;
          sp = bp;
          return true;
      }
   }
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppInt capacity = bp[2];
   bool init = false;
   if (argc == 3)
      init = (((BppInt)bp[3]) != 0);
   vecVal = new BppVector(capacity,init);
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Vector_OP_ADD(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* myvec = vecVal;
   if (myvec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   BppObject* result = new BppObject(pThis->getClass());
   BppValue& resValue = (*(result->getMemberData()))[_vecIndex];

   BppVector* resVec = new BppVector(*myvec);
   resValue = resVec;

   BppVector* pv = bp[2];
   if (pv == 0)
   {
      BppObject* pobj = bp[2];
      if (pobj != 0)
      {
         BppClass* pclass = pobj->getClass();
         if (pclass->isSubtypeOf(pThis->getClass()))
            pv = (*(pobj->getMemberData()))[_vecIndex];
      } // if pobj
   }
   if (pv != 0)
   {
      unsigned n = pv->size();
      resVec->setCapacity(resVec->size()+n);
      for (unsigned i = 0; i<n;++i)
         resVec->push_back((*pv)[i]);
   }
   else
      resVec->push_back(bp[2]);
   *bp = result;
   sp = bp;
   return true;
}

bool Bp2Vector_OP_ADD_R(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* myvec = vecVal;
   if (myvec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   BppObject* result = new BppObject(pThis->getClass());
   BppValue& resValue = (*(result->getMemberData()))[_vecIndex];

   BppVector* resVec = 0;
   BppVector* pv = bp[2];
   if (pv != 0)
   {
      resVec = new BppVector(*pv);
   }
   else
   {
      resVec = new BppVector(1);
      resVec->push_back(bp[2]);
   }
   unsigned n = myvec->size();
   resVec->setCapacity(resVec->size()+n);
   for (unsigned i = 0; i<n;++i)
      resVec->push_back((*myvec)[i]);
   resValue = resVec;
   *bp = result;
   sp = bp;
   return true;
}


bool Bp2Vector_OP_VREF(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppUInt idx = bp[2];
   if (idx >= (BppUInt) vec->size())
      THROW(_T("subscript out of range"));
   *bp = (*vec)[(size_t)idx];
   sp = bp;
   return true;
}

bool Bp2Vector_OP_VSET(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(3,3);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppUInt idx = bp[2];
   if (idx >= (BppInt) vec->size())
      THROW(_T("subscript out of range"));
   (*vec)[(size_t)idx] = bp[3];
   *bp = bp[3];
   sp = bp;
   return true;
}

bool Bp2Vector_size(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   *bp = (BppUInt) vec->size();
   sp = bp;
   return true;
}

bool Bp2Vector_capacity(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   *bp = (BppUInt) vec->capacity();
   sp = bp;
   return true;
}
/// Vector::setCapacity implementation function
bool Bp2Vector_setCapacity(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   *bp = (BppUInt) vec->setCapacity((size_t) (BppUInt) bp[2]);
   sp = bp;
   return true;
}
/// Vector::shrink implementation function
bool Bp2Vector_shrink(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   *bp = (BppUInt) vec->shrink();
   sp = bp;
   return true;
}

bool Bp2Vector_push_back(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   vec->push_back(bp[2]);
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Vector_pop_back(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   if (vec->size() == 0)
      THROW(_T("Can't pop value from empty vector."));
   *bp = *vec->last();
   vec->pop_back();
   sp = bp;
   return true;
}


bool Bp2Vector_push_front(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   vec->push_back(BppValue());
   unsigned n = vec->size();
   for (unsigned i=n-1;i>0;--i)
   {
      (*vec)[i] = (*vec)[i-1];
   }
   (*vec)[0] = bp[2];
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Vector_pop_front(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   unsigned n = vec->size();
   if (n == 0)
      THROW(_T("Can't pop value from empty vector."));
   *bp = (*vec)[0];
   for (unsigned i=0;i<n-1;++i)
   {
      (*vec)[i] = (*vec)[i+1];
   }   
   vec->pop_back();
   sp = bp;
   return true;
}

bool Bp2Vector_clear(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   vec->clear();
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Vector_insert(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMIN(3);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   BppUInt pos = bp[2];
   size_t size = vec->size();
   long n = argc-2; // number of values to insert
   if (pos > (BppUInt)size) pos = (BppUInt) size;
   vec->setCapacity(size+n);
   if (pos < (BppUInt)size)
   {
      for(BppUInt i=0;i<n;++i)
      {
         vec->push_back((*vec)[pos+i]);
         (*vec)[pos+i] = bp[int(3+i)];
      }
   }
   else
   {
      for(int i=0;i<n;++i)
      {
         vec->push_back(bp[3+i]);
      }
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Vector_slice(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,3);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppUInt pos = bp[2];
   BppUInt cnt = BPP_NPOS;
   if (argc == 3)
   {
      CHKTYPE(bp[3],BVT_INTNUMBER);
      cnt = bp[3];
   }
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   size_t size = vec->size();
   if (pos >= (BppUInt)size)
      THROW(_T("subscript out of range"));
   BppObject* result = new BppObject(pThis->getClass());
   BppValue& resValue = (*(result->getMemberData()))[_vecIndex];
	if (size > 0 && size >= pos) {
		if (cnt > (BppUInt)(size - pos))
			cnt = size - pos;


		BppVector* resVec = new BppVector((size_t)cnt);
		resValue = resVec;
		for (BppInt i = 0; i < cnt; ++i)
			resVec->push_back((*vec)[pos + i]);
	}
   *bp = result;
   sp = bp;
   return true;
}

bool Bp2Vector_erase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,3);
   CHKTYPE(bp[2],BVT_INTNUMBER);
   BppUInt pos = bp[2];
   BppUInt cnt = BPP_NPOS;
   if (argc == 3)
   {
      CHKTYPE(bp[3],BVT_INTNUMBER);
      cnt = bp[3];
   }
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   size_t size = vec->size();
   if (pos >= (BppUInt)size)
      THROW(_T("subscript out of range"));
   if (cnt > ((BppUInt)size-pos))
      cnt = BppUInt(size)-pos;

   if (cnt > 0)
   {
      BppUInt moveCnt = BppUInt(size) - (pos+cnt);
      for (BppUInt i = 0;i<moveCnt;++i)
         (*vec)[pos+i] = (*vec)[pos+cnt+i];
      while(cnt > 0)
      {
         vec->pop_back();
         --cnt;
      }
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Vector_vec(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppValue& vecVal = MEMBER(_vecIndex);
   BppVector* vec = vecVal;
   if (vec == 0)
      return interpreter->badType(vecVal.getType(),BVT_VECTOR);
   *bp = vec;
   sp = bp;
   return true;
}


void Bp2Vector_CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_T("Vector"),pClass)) return;
   pClass->addDataMember(_T("_vec"));
   pClass->addMethod(_T("Vector"),&Bp2Vector_ctor);
   pClass->addMethod(_T("OP_ADD"),&Bp2Vector_OP_ADD);
   pClass->addMethod(_T("OP_ADD_R"),&Bp2Vector_OP_ADD_R);
   pClass->addMethod(_T("OP_VREF"),&Bp2Vector_OP_VREF);
   pClass->addMethod(_T("OP_VSET"),&Bp2Vector_OP_VSET);
   pClass->addMethod(_T("size"),&Bp2Vector_size);
   pClass->addMethod(_T("capacity"),&Bp2Vector_capacity);
   pClass->addMethod(_T("setCapacity"),&Bp2Vector_setCapacity);
   pClass->addMethod(_T("shrink"),&Bp2Vector_shrink);
   pClass->addMethod(_T("push_back"),&Bp2Vector_push_back);
   pClass->addMethod(_T("pop_back"),&Bp2Vector_pop_back);
   pClass->addMethod(_T("push_front"),&Bp2Vector_push_front);
   pClass->addMethod(_T("pop_front"),&Bp2Vector_pop_front);
   pClass->addMethod(_T("clear"),&Bp2Vector_clear);
   pClass->addMethod(_T("insert"),&Bp2Vector_insert);
   pClass->addMethod(_T("slice"),&Bp2Vector_slice);
   pClass->addMethod(_T("erase"),&Bp2Vector_erase);
   pClass->addMethod(_T("vec"),&Bp2Vector_vec);
   pClass->addMethod(_T("vector"),&Bp2Vector_vec);
   pClass->fixupDataDefs();
   _vecIndex = pClass->getMemberIndex(_T("_vec"));
}
