#include "bp2wstruc.h"

#include <bppvm.h>
#include <bppclass.h>
#include <bppvalue.h>
#include <bppdict.h>
#include <bppvec.h>
#include <bppobjct.h>
#include <bppint.h>
#include <bppprint.h>
#include <bppbuf.h>

#include "bp2cldef.h"
#ifdef WIN32
#include <windows.h>

/* ===================== Bp2Point ========================== */
int Bp2Point::_ptIdx = 0;
const BppString Bp2Point::_className = _T("POINT");

bool Bp2Point::ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,3);
   BppBuffer* pb = 0;
   POINT* pp = 0;
   if (argc == 1)
   {
      pb = new BppBuffer(sizeof(POINT));
      pb->setSize(sizeof(POINT));
      pp = (POINT*)pb->data();
      pp->x = -1;
      pp->y = -1;
   }
   else if (argc == 2)
   {
      BppValue& arg = ARG(1);
      BppObject* p = arg;
      if (p != 0)
      {
         BppClass* pcl = p->getClass();
         if (pcl->isSubtypeOf(_className))
         {
            pb = OBJMEMBER(p,_ptIdx);
            if (pb == 0) THROW(_T("invalid member data"));
            pb = new BppBuffer(*pb); 
         }
      }
      if (pb == 0)
      {
         pb = new BppBuffer(sizeof(POINT));
         pb->setSize(sizeof(POINT));
         pp = (POINT*)pb->data();
         pp->x = (BppInt)arg;
         pp->y = -1;
      }
   }
   else
   {
      pb = new BppBuffer(sizeof(POINT));
      pb->setSize(sizeof(POINT));
      pp = (POINT*)pb->data();
      pp->x = (BppInt)ARG(1);
      pp->y = (BppInt)ARG(2);
   }
   MEMBER(_ptIdx) = pb;
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Point::getX(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_ptIdx);
   if (pb == 0) THROW(_T("Invalid data (_pt)"));
   *bp = ((POINT*) pb->data())->x;
   sp = bp;
   return true;
}

bool Bp2Point::setX(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_ptIdx);
   if (pb == 0) THROW(_T("Invalid data (_pt)"));
   ((POINT*) pb->data())->x = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Point::getY(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_ptIdx);
   if (pb == 0) THROW(_T("Invalid data (_pt)"));
   *bp = ((POINT*) pb->data())->y;
   sp = bp;
   return true;
}

bool Bp2Point::setY(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_ptIdx);
   if (pb == 0) THROW(_T("Invalid data (_pt)"));
   ((POINT*) pb->data())->y = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Point::OP_ADD(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppObject* p = ARG(1);
   if ((p == 0) || (!(p->getClass()->isSubtypeOf(_className))))
      THROW(_T("Invalid argument"));
   BppBuffer* pb = MEMBER(_ptIdx);
   BppBuffer* opb = OBJMEMBER(p,_ptIdx);
   if ((pb == 0) || (opb == 0)) THROW(_T("Invalid data (_pt)"));
   POINT* pp = (POINT*) pb->data();
   POINT* opp = (POINT*) opb->data();
   BppObject* pobj = new BppObject(pThis->getClass());
   BppBuffer* newpb = new BppBuffer(sizeof(POINT));
   newpb->setSize(sizeof(POINT));
   POINT* npp = (POINT*) newpb->data();
   OBJMEMBER(pobj,_ptIdx) = newpb;
   npp->x = pp->x + opp->x;
   npp->y = pp->y + opp->y;
   *bp = pobj;
   sp = bp;
   return true;
}

bool Bp2Point::OP_SUB(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppObject* p = ARG(1);
   if ((p == 0) || (!(p->getClass()->isSubtypeOf(_className))))
      THROW(_T("Invalid argument"));
   BppBuffer* pb = MEMBER(_ptIdx);
   BppBuffer* opb = OBJMEMBER(p,_ptIdx);
   if ((pb == 0) || (opb == 0)) THROW(_T("Invalid data (_pt)"));
   POINT* pp = (POINT*) pb->data();
   POINT* opp = (POINT*) opb->data();
   BppObject* pobj = new BppObject(pThis->getClass());
   BppBuffer* newpb = new BppBuffer(sizeof(POINT));
   newpb->setSize(sizeof(POINT));
   POINT* npp = (POINT*) newpb->data();
   OBJMEMBER(pobj,_ptIdx) = newpb;
   npp->x = pp->x - opp->x;
   npp->y = pp->y - opp->y;
   *bp = pobj;
   sp = bp;
   return true;
}

bool Bp2Point::equals(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppObject* p = ARG(1);
   if ((p == 0) || (!(p->getClass()->isSubtypeOf(_className))))
      THROW(_T("Invalid argument"));
   BppBuffer* pb = MEMBER(_ptIdx);
   BppBuffer* opb = OBJMEMBER(p,_ptIdx);
   if ((pb == 0) || (opb == 0)) THROW(_T("Invalid data (_pt)"));
   int res = memcmp(pb->data(),opb->data(),pb->size());
   *bp = (res == 0) ? (BppInt) 1 : (BppInt) 0;
   sp = bp;
   return true;
}

bool Bp2Point::ptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_ptIdx);
   if (pb == 0) THROW(_T("Invalid data (_pt)"));
   *bp = (BppIntPtr) pb->data();
   sp = bp;
   return true;
}

void Bp2Point::CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_className.c_str(),pClass)) return;
   pClass->addDataMember(_T("_pt"));
   pClass->addMethod(_className.c_str(),&ctor);
   pClass->addMethod(_T("getX"),&getX);
   pClass->addMethod(_T("setX"),&setX);
   pClass->addMethod(_T("getY"),&getY);
   pClass->addMethod(_T("setY"),&setY);
   pClass->addMethod(_T("OP_ADD"),&OP_ADD);
   pClass->addMethod(_T("OP_SUB"),&OP_SUB);
   pClass->addMethod(_T("equals"),&equals);
   pClass->addMethod(_T("ptr"),&ptr);
   pClass->fixupDataDefs();
   _ptIdx = pClass->getMemberIndex(_T("_pt"));
}

/* ====================== Bp2Size ============================= */

const BppString Bp2Size::_className = _T("SIZE");

//bool Bp2Size_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
//bool Bp2Size_getW(BppVirtualMachine* vm, int argc, ValPtr& sp);
//bool Bp2Size_setW(BppVirtualMachine* vm, int argc, ValPtr& sp);
//bool Bp2Size_getH(BppVirtualMachine* vm, int argc, ValPtr& sp);
//bool Bp2Size_setH(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Size::ok(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(Bp2Point::_ptIdx);
   if (pb == 0) THROW(_T("Invalid data (_pt)"));
   SIZE* pp = (SIZE*) pb->data();
   *bp = (BppInt) ((pp->cx > 0) && (pp->cy > 0) ? 1 : 0);
   sp = bp;
   return true;
}

void Bp2Size::CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   BppDictionaryEntry* pe = vm.classes().find(Bp2Point::_className);
   if (pe == 0) return;
   pClass->setBaseClass(pe->getValue());
   if (!vm.registerNativeClass(_className.c_str(),pClass)) return;
   pClass->addMethod(_className.c_str(),&Bp2Point::ctor);
   pClass->addMethod(_T("getW"),&Bp2Point::getX);
   pClass->addMethod(_T("setW"),&Bp2Point::setX);
   pClass->addMethod(_T("getH"),&Bp2Point::getY);
   pClass->addMethod(_T("setH"),&Bp2Point::setY);
   pClass->addMethod(_T("ok"),&ok);
   pClass->fixupDataDefs();
}

/* ========================== Bp2Rect =========================== */
int Bp2Rect::_rectIdx = 0;

const BppString Bp2Rect::_className = _T("RECT");

bool Bp2Rect::ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,5);
   BppBuffer* pb = new BppBuffer(sizeof(RECT));
   pb->setSize(sizeof(RECT));
   RECT& rect = *((RECT*)pb->data());
   MEMBER(_rectIdx) = pb;
   switch (argc)
   {
   case 1:
      rect.bottom = rect.top = rect.left = rect.right = -1;
      break;
   case 2:
      {
         BppObject* p = ARG(1);
         if (p != 0)
         {
            BppClass* pcl = p->getClass();
            if (pcl->isSubtypeOf(Bp2Rect::_className))
            {
               BppBuffer* opb = OBJMEMBER(p,_rectIdx);
               if (opb == 0) THROW(_T("invalid member data"));
               rect = *((RECT*)opb->data());
               break;
            }
         }
      }
      rect.left = (BppInt)ARG(1);
      rect.bottom = rect.top = rect.right = -1;
      break;
   case 3:
      rect.left = (BppInt)ARG(1);
      rect.top = (BppInt)ARG(2);
      rect.bottom = rect.right = -1;
      break;
   case 4:
      rect.left = (BppInt)ARG(1);
      rect.top = (BppInt)ARG(2);
      rect.right = (BppInt)ARG(3);
      rect.bottom = -1;
      break;
   case 5:
      rect.left = (BppInt)ARG(1);
      rect.top = (BppInt)ARG(2);
      rect.right = (BppInt)ARG(3);
      rect.bottom = (BppInt)ARG(4);
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Rect::getLeft(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   *bp = ((RECT*) pb->data())->left;
   sp = bp;
   return true;
}

bool Bp2Rect::setLeft(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   ((RECT*) pb->data())->left = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}
bool Bp2Rect::getTop(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   *bp = ((RECT*) pb->data())->top;
   sp = bp;
   return true;
}

bool Bp2Rect::setTop(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   ((RECT*) pb->data())->top = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Rect::getRight(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   *bp = ((RECT*) pb->data())->right;
   sp = bp;
   return true;
}
bool Bp2Rect::setRight(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   ((RECT*) pb->data())->right = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Rect::getBottom(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   *bp = ((RECT*) pb->data())->bottom;
   sp = bp;
   return true;
}

bool Bp2Rect::setBottom(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   ((RECT*) pb->data())->bottom = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Rect::getPos(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppClass* ptClass = 0;
   BppDictionaryEntry* pe = vm->classes().find(Bp2Point::_className);
   if (pe != 0) ptClass = pe->getValue();
   if (ptClass == 0)
      THROW (_T("POINT class not found."));
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));

   BppObject* pobj = new BppObject(ptClass);
   BppBuffer* newpb = new BppBuffer(sizeof(POINT));
   newpb->setSize(sizeof(POINT));
   POINT* npp = (POINT*) newpb->data();
   RECT* prect = (RECT*) pb->data();
   OBJMEMBER(pobj,Bp2Point::_ptIdx) = newpb;
   npp->x = prect->left;
   npp->y = prect->top;
   *bp = pobj;
   sp = bp;
   return true;
}

bool Bp2Rect::setPos(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   BppBuffer* opb = 0;
   BppObject* pobj = ARG(1);
   if (pobj != 0)
   {
      if (pobj->getClass()->isSubtypeOf(Bp2Point::_className))
      {
         opb = OBJMEMBER(pobj,Bp2Point::_ptIdx);
      }
   }
   if (opb == 0) THROW(_T("Invalid argument."));
   POINT* pp = (POINT*) opb->data();
   RECT* prect = (RECT*) pb->data();
   LONG dx = pp->x - prect->left;
   LONG dy = pp->y - prect->top;
   prect->left = pp->x;
   prect->top = pp->y;
   prect->right += dx;
   prect->bottom += dy;
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Rect::getSize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppClass* sizeClass = 0;
   BppDictionaryEntry* pe = vm->classes().find(Bp2Size::_className);
   if (pe != 0) sizeClass = pe->getValue();
   if (sizeClass == 0)
      THROW (_T("SIZE class not found."));
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));

   BppObject* pobj = new BppObject(sizeClass);
   BppBuffer* newpb = new BppBuffer(sizeof(POINT));
   newpb->setSize(sizeof(POINT));
   SIZE* nps = (SIZE*) newpb->data();
   RECT* prect = (RECT*) pb->data();
   OBJMEMBER(pobj,Bp2Point::_ptIdx) = newpb;
   nps->cx = prect->right - prect->left;
   nps->cy = prect->bottom - prect->top;
   *bp = pobj;
   sp = bp;
   return true;
}

bool Bp2Rect::setSize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   BppBuffer* opb = 0;
   BppObject* pobj = ARG(1);
   if (pobj != 0)
   {
      if (pobj->getClass()->isSubtypeOf(Bp2Size::_className))
      {
         opb = OBJMEMBER(pobj,Bp2Point::_ptIdx);
      }
   }
   if (opb == 0) THROW(_T("Invalid argument."));
   SIZE* ps = (SIZE*) opb->data();
   RECT* prect = (RECT*) pb->data();
   prect->right = prect->left + ps->cx;
   prect->bottom = prect->top + ps->cy;
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Rect::ok(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   RECT* pr = (RECT*) pb->data();
   *bp = (BppInt) ((pr->right > pr->left) && (pr->bottom > pr->top) ? 1 : 0);
   sp = bp;
   return true;
}

bool Bp2Rect::equals(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppObject* p = ARG(1);
   if ((p == 0) || (!(p->getClass()->isSubtypeOf(Bp2Rect::_className))))
      THROW(_T("Invalid argument"));
   BppBuffer* pb = MEMBER(Bp2Point::_ptIdx);
   BppBuffer* opb = OBJMEMBER(p,Bp2Point::_ptIdx);
   if ((pb == 0) || (opb == 0)) THROW(_T("Invalid data (_pt)"));
   int res = memcmp(pb->data(),opb->data(),pb->size());
   *bp = (res == 0) ? (BppInt) 1 : (BppInt) 0;
   sp = bp;
   return true;
}

bool Bp2Rect::ptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_rectIdx);
   if (pb == 0) THROW(_T("Invalid data (_rect)"));
   *bp = (BppIntPtr) pb->data();
   sp = bp;
   return true;
}

void Bp2Rect::CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_className.c_str(),pClass)) return;
   pClass->addDataMember(_T("_rect"));
   pClass->addMethod(_className.c_str(),&ctor);
   pClass->addMethod(_T("getLeft"),&getLeft);
   pClass->addMethod(_T("setLeft"),&setLeft);
   pClass->addMethod(_T("getTop"),&getTop);
   pClass->addMethod(_T("setTop"),&setTop);
   pClass->addMethod(_T("getRight"),&getRight);
   pClass->addMethod(_T("setRight"),&setRight);
   pClass->addMethod(_T("getBottom"),&getBottom);
   pClass->addMethod(_T("setBottom"),&setBottom);
   pClass->addMethod(_T("getPos"),&getPos);
   pClass->addMethod(_T("setPos"),&setPos);
   pClass->addMethod(_T("getSize"),&getSize);
   pClass->addMethod(_T("setSize"),&setSize);
   pClass->addMethod(_T("ok"),&ok);
   pClass->addMethod(_T("equals"),&equals);
   pClass->addMethod(_T("ptr"),&ptr);
   pClass->fixupDataDefs();
   _rectIdx = pClass->getMemberIndex(_T("_rect"));
}

/* ===================== Bp2Msg ========================== */

int Bp2Msg::_msgIdx = 0;
const BppString Bp2Msg::_className = _T("MSG");

bool Bp2Msg::ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,2);
   if (argc == 2)
   {
      bool ok = false;
      BppObject* p = ARG(1);
      if (p != 0)
      {
         BppClass* pcl = p->getClass();
         if (pcl->isSubtypeOf(_className))
         {
            BppBuffer* pb = OBJMEMBER(p,_msgIdx);
            if (pb != 0)
            {
               MEMBER(_msgIdx) = new BppBuffer(*pb); 
               ok = true;
            }
         }
      }
      if (!ok)  THROW(_T("MSG argument expected"));  
   }
   else
   {
      BppBuffer* pb = new BppBuffer(sizeof(MSG));
      pb->setSize(sizeof(MSG));
      MEMBER(_msgIdx) = pb; 
   }
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Msg::setHwnd(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->hwnd = (HWND) (BppIntPtr) ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getHwnd(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = (BppIntPtr)((MSG*) pb->data())->hwnd;
   sp = bp;
   return true;
}

bool Bp2Msg::setMsgId(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->message = (UINT) (BppUInt) ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getMsgId(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = (BppUInt)((MSG*) pb->data())->message;
   sp = bp;
   return true;
}

bool Bp2Msg::setWParam(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->wParam = (WPARAM) (BppUIntPtr) ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getWParam(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = (BppUIntPtr)((MSG*) pb->data())->wParam;
   sp = bp;
   return true;
}

bool Bp2Msg::setLParam(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->lParam = (LPARAM) (BppIntPtr) ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getLParam(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = (BppIntPtr)((MSG*) pb->data())->lParam;
   sp = bp;
   return true;
}

bool Bp2Msg::setTime(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->time = (DWORD) (BppInt) ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getTime(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = (BppInt)((MSG*) pb->data())->time;
   sp = bp;
   return true;
}

bool Bp2Msg::setPoint(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,3);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   POINT& pt = ((MSG*) pb->data())->pt;
   if (argc == 2)
   {
      BppBuffer* pb = 0;
      BppObject* p = ARG(1);
      if (p != 0)
      {
         BppClass* pcl = p->getClass();
         if (pcl->isSubtypeOf(Bp2Point::_className))
            pb = OBJMEMBER(p,Bp2Point::_ptIdx);
      }
      if (pb == 0)
         THROW(_T("POINT argument expected"));
      POINT* pp = (POINT*) pb->data();
      pt.x = pp->x;
      pt.y = pp->y;
   }
   else
   {
      pt.x = (BppInt)ARG(1);
      pt.y = (BppInt)ARG(2);
   }
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getPoint(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   POINT& pt = ((MSG*) pb->data())->pt;

   BppClass* ptClass = 0;
   BppDictionaryEntry* pe = vm->classes().find(Bp2Point::_className);
   if (pe != 0) ptClass = pe->getValue();
   if (ptClass == 0)
      THROW (_T("POINT class not found."));

   BppObject* pobj = new BppObject(ptClass);
   BppBuffer* newpb = new BppBuffer(sizeof(POINT));
   newpb->setSize(sizeof(POINT));
   POINT* npp = (POINT*) newpb->data();
   OBJMEMBER(pobj,Bp2Point::_ptIdx) = newpb;

   npp->x = pt.x;
   npp->y = pt.y;

   *bp = pobj;
   sp = bp;
   return true;
}

bool Bp2Msg::setX(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->pt.x = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getX(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = ((MSG*) pb->data())->pt.x;
   sp = bp;
   return true;
}

bool Bp2Msg::setY(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   ((MSG*) pb->data())->pt.y = (BppInt)ARG(1);
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Msg::getY(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = ((MSG*) pb->data())->pt.y;
   sp = bp;
   return true;
}

bool Bp2Msg::ptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_msgIdx);
   if (pb == 0) THROW(_T("Invalid data (_msg)"));
   *bp = (BppIntPtr) pb->data();
   sp = bp;
   return true;
}

void Bp2Msg::CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_className.c_str(),pClass)) return;
   pClass->addDataMember(_T("_msg"));
   pClass->addMethod(_className.c_str(),&ctor);
   pClass->addMethod(_T("getHwnd"),&getHwnd);
   pClass->addMethod(_T("setHwnd"),&setHwnd);
   pClass->addMethod(_T("getMsgId"),&getMsgId);
   pClass->addMethod(_T("setMsgId"),&setMsgId);
   pClass->addMethod(_T("getWParam"),&getWParam);
   pClass->addMethod(_T("setWParam"),&setWParam);
   pClass->addMethod(_T("getLParam"),&getLParam);
   pClass->addMethod(_T("setLParam"),&setLParam);
   pClass->addMethod(_T("getTime"),&getTime);
   pClass->addMethod(_T("setTime"),&setTime);
   pClass->addMethod(_T("getPoint"),&getPoint);
   pClass->addMethod(_T("setPoint"),&setPoint);
   pClass->addMethod(_T("getX"),&getX);
   pClass->addMethod(_T("setX"),&setX);
   pClass->addMethod(_T("getY"),&getY);
   pClass->addMethod(_T("setY"),&setY);
   pClass->addMethod(_T("ptr"),&ptr);
   pClass->fixupDataDefs();
   _msgIdx = pClass->getMemberIndex(_T("_msg"));
}

/* ===================== Bp2Accel ========================== */
int Bp2Accel::_accelIdx = 0;
const BppString Bp2Accel::_className = _T("ACCEL");

bool Bp2Accel::ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(1,4);
   WORD key = 0;
   WORD cmd = 0;
   BYTE flags = 0;
   switch (argc)
   {
   case 4:
      flags = (BYTE) ((BppInt) ARG(3)) & 0xFF;
   case 3:
      cmd = (WORD)((BppInt) ARG(2)) & 0xFFFF;
      key = (WORD) ((BppInt) ARG(1)) & 0xFFFF;
      break;
   case 2: // copy
      {
         bool ok = false;
         BppObject* p = ARG(1);
         if (p != 0)
         {
            BppClass* pcl = p->getClass();
            if (pcl->isSubtypeOf(_className))
            {
               BppBuffer* pb = OBJMEMBER(p,_accelIdx);
               if (pb != 0)
               {
                  ACCEL* pa = (ACCEL*) pb->data();
                  key = pa->key;
                  cmd = pa->cmd;
                  flags = pa->fVirt;
                  ok = true;
               }
            }
         }
         if (!ok)  THROW(_T("ACCEL argument expected"));  
      }
   } // end switch
   BppBuffer* pb = new BppBuffer(sizeof(ACCEL));
   pb->setSize(sizeof(ACCEL));
   MEMBER(_accelIdx) = pb; 
   ACCEL* pa = (ACCEL*) pb->data();
   pa->cmd = cmd;
   pa->fVirt = flags;
   pa->key = key;
   *bp = pThis;
   sp = bp;
   return true;
}

bool Bp2Accel::getFVirt(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   *bp = (BppInt)((ACCEL*) pb->data())->fVirt;
   sp = bp;
   return true;
}

bool Bp2Accel::setFVirt(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   ((ACCEL*) pb->data())->fVirt = (BYTE) ((BppInt) ARG(1)) & 0xFF;
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Accel::getKey(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   *bp = (BppInt)((ACCEL*) pb->data())->key;
   sp = bp;
   return true;
}

bool Bp2Accel::setKey(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   ((ACCEL*) pb->data())->key = (WORD) ((BppInt) ARG(1)) & 0xFFFF;
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Accel::getCmd(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   *bp = (BppInt)((ACCEL*) pb->data())->cmd;
   sp = bp;
   return true;
}

bool Bp2Accel::setCmd(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   CHKARGC(2,2);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   ((ACCEL*) pb->data())->cmd = (WORD) ((BppInt) ARG(1)) & 0xFFFF;
   bp->dispose();
   sp = bp;
   return true;
}

bool Bp2Accel::ptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
   CHKVM;
   CHKOBJECT;
   ARGCMAX(1);
   BppBuffer* pb = MEMBER(_accelIdx);
   if (pb == 0) THROW(_T("Invalid data (_accel)"));
   *bp = (BppIntPtr) pb->data();
   sp = bp;
   return true;
}

void Bp2Accel::CreateClass(BppVirtualMachine& vm)
{
   BppClass* pClass = new BppClass(vm);
   if (!vm.registerNativeClass(_className.c_str(),pClass)) return;
   pClass->addDataMember(_T("_accel"));
   pClass->addMethod(_className.c_str(),&ctor);
   pClass->addMethod(_T("getFVirt"),&getFVirt);
   pClass->addMethod(_T("setFVirt"),&setFVirt);
   pClass->addMethod(_T("getKey"),&getKey);
   pClass->addMethod(_T("setKey"),&setKey);
   pClass->addMethod(_T("getCmd"),&getCmd);
   pClass->addMethod(_T("setCmd"),&setCmd);
   pClass->addMethod(_T("ptr"),&ptr);
   pClass->fixupDataDefs();
   _accelIdx = pClass->getMemberIndex(_T("_accel"));
}


#endif // WIN32
