
#include <stdio.h>
#include <bppvm.h>
#include <bppvec.h>
#include <dictio.h>
#include <bppstrng.h>
#include <bppufunc.h>
#include <bppprint.h>
#ifdef GCC
#include <bppbuf.h>
#endif
#ifdef NOCONIO
 #undef NOCONIO
 #ifndef _CONSOLE
	#define NOCONIO
 #endif
#endif
#ifndef NOCONIO
#if defined(GCC) && !defined(WIN32)
  #define getch getchar
#else
  #include <conio.h>
#endif
#endif
#ifndef OUTNAME
  #define OUTNAME "bp2"
#endif

#ifndef RTNAME
#ifdef __unix__
  #define RTNAME "bp2r"
#else
  #define RTNAME "bp2r.exe"
#endif
#endif

#ifdef __OLDCPP__
#include <dos.h>

#endif 
static const BppChar* BANNER = _T("B++ v%d.%d - Copyright (c) 2005-%s, by Ralf-Erik Ebert\n")
_T("usage:\n")
_T(OUTNAME)
_T(" [-i][-d][-l][-v][-t][-c]")
_T("[-e]")
#ifdef WIN32
_T("[-E]")
#endif
_T("[-L rtfile]")
_T(" [-Dsymbol[=value]][..]\n")
_T(" [sourcefile [..]] [-r objfile [..]] [-o outfile] [# userarg [..]]\n")
_T("options:\n")
_T("    -i ... info (shows syntax and names of compiled functions)\n")
_T("    -d ... shows debug information (bytecode) during compilation\n")
_T("    -l ... includes source file and line information into bytecode\n")
_T("    -v ... includes local variables information into bytecode\n")
_T("    -t ... traces bytecodes during execution\n")
_T("    -c ... compiles only\n")
_T("    -e ... builds standalone exe file (compiles and links against ")_T(RTNAME) _T(") \n")
#ifdef WIN32
_T("    -E ... builds standalone exe file without console (compiles and links against default runtime)\n")
#endif
_T("    -L ... builds standalone exe file (compiles and links against runtime specified by rtname)\n")
_T("    -D ... defines a predefined symbol for compiler\n")
_T("    -r ... specifies precompiled files to be loaded\n")
_T("    -o ... name of output file (used only with -c, -e")
#ifdef WIN32
_T(", -E")
#endif
_T(" or -L)\n")
_T("     # ... marks begin of user arguments list\n"); 

enum ArgType
{
	AT_ANY = 0,
	AT_SOURCE,
	AT_LIB,
	AT_OUT,
	AT_LINK
};

// Wait callback fr debugging.
// On Win32 without console a message box is used, otherwise simply a character input.
static bool BPPAPI dbgWait(BppVirtualMachine*)
{
#ifdef NOCONIO
#ifdef WINCE
#define MB_SYSTEMMODAL MB_TOPMOST|MB_SETFOREGROUND
#endif
	return (MessageBox(GetActiveWindow(),_T("Press OK for next step\nor Cancel to leave step mode."),
	  _T("Single Step Mode"),MB_OKCANCEL|MB_SYSTEMMODAL) == IDOK);
#else
	return (getch() != 27);
#endif
}

#define CBUFSIZE 256
// file copy helper function 
unsigned long cpy_file(const BppChar* src, const BppChar* dest, bool append)
{
	unsigned long flen = 0;
	unsigned char cbuf[CBUFSIZE];
	FILE *fps, *fpd;
#ifdef GCC
    // when using GCC (i.e. un UNIX systems) entire file content is buffered.
	 BppBuffer buf(CBUFSIZE);
	fps = _tfopen(src,_T("rb"));
	if (fps == NULL)
	{
		_tprintf(_T("Unable to open file ('%s')\n"),src);
		return 0;
	}
	while(!feof(fps))
	{
		unsigned n = fread(cbuf,1,CBUFSIZE,fps);
	  //c = fgetc(fps);
	  if (!n) break;
	  for (unsigned i=0;i<n;++i) buf.push_back(cbuf[i]);
	  flen += n;
	  if (n<CBUFSIZE) break;
	}
	fclose(fps);
	fpd = _tfopen(dest,append ? _T("ab") : _T("wb"));
	if (fpd == NULL)
	{
		_tprintf(_T("Unable to open file ('%s')\n"),dest);
		return 0;
	}
	 if (!fwrite(buf.data(),flen,1,fpd))
	 {
	  _tprintf(_T("Error while writing '%s'.\n"),dest);
	  flen=0;
	 }
	fclose(fpd);
#else
	fps = _tfopen(src,_T("rb"));
	if (fps == NULL)
	{
		_tprintf(_T("Unable to open file ('%s')\n"),src);
		return 0;
	}
	fpd = _tfopen(dest,append ? _T("ab") : _T("wb"));
	if (fpd == NULL)
	{
		fclose(fps);
		_tprintf(_T("Unable to open file ('%s')\n"),dest);
		return 0;
	}
	while(!feof(fps))
	{
		unsigned n = fread(cbuf,1,CBUFSIZE,fps);
	  //c = fgetc(fps);
	  if (!n) break;
	  if (!fwrite(cbuf,n,1,fpd))
	  {
		 _tprintf(_T("Error while writing '%s'.\n"),dest);
		 break;
	  }
	  flen += n;
	  if (n<CBUFSIZE) break;
	}
	fclose(fps);
	fclose(fpd);
#endif
	return flen;
}

/* build exe from compiled file */
static bool build_exe(const BppChar* outname, const BppChar* rtname, bool noConsole = false)
{
	FILE* fp = 0;
	unsigned long flen;
	//const BppChar* rtname = _T(RTNAME);
#ifdef WIN32
	BppString rtnameStr = rtname;
	if (noConsole) {
		const BppChar* pr = _tcsrchr(rtname, _T('r'));
		size_t pos = rtnameStr.find(_T('r'));
		if (pos > 0 && pos != BPP_NNPOS) {
		  rtnameStr.insert(pos + 1, _T('w'));
		  rtname = rtnameStr.c_str();
		}
	}
#endif
	const BppChar* rtpath = searchPath(rtname);
	if (rtpath == NULL)
	{
			 FILE* p = _tfopen(rtname,_T("rb"));
		  if (p) {
			  fclose(p);
			  rtpath = rtname;
		  }
		  else {
		  _tprintf(_T("Unable to locate runtime file ('%s')\n"),rtname);
		  return false;
		  }
	}
	flen = cpy_file(rtpath,outname,false);
	cpy_file(_T("out.tmp"),outname,true);
	fp = _tfopen(outname,_T("r+b"));
	if (fp == 0)
	{
		_tprintf(_T("Open output file ('%s') failed.\n"),outname);
		return false;
	}
	fseek(fp,0,SEEK_END);
		fwrite(&flen,1,sizeof(BppInt),fp);
		fwrite("BPR",1,4,fp);
		fclose(fp);
	return true;
}

#ifdef STKLEN
// Sets stack size for Borland C++
extern unsigned _stklen = STKLEN;
#endif

int
#ifdef _WIN32_WCE_EMULATION
__cdecl
#endif
 _tmain(int argc, BppChar* argv[])
{
    
	BppFileStreamPrinter fprt(stdout); // create print output using strout
	BppVirtualMachine vm(&fprt);       // create instance of VM
	vm.setDbgWaitProc(dbgWait);        // install debug wait callback
	vm.cmdargc() = argc;			   // store command line in VM
	vm.cmdargv() = argv;
	bool showInfo = false;
	bool showDebug = false;
	bool doTrace = false;
	bool compileOnly = false;
	bool linkExe = false;
#ifdef WIN32
	bool noConsole = false;
#endif
	BppString outfile = _T("out");
	BppString rtfile = _T(RTNAME);
	BppVector moduleVec;
	BppVector srcVec;
	ArgType argtype = AT_SOURCE;
	if (argc < 2)
	{
	  _tprintf(BANNER,_BPP_VER / 100,_BPP_VER % 100,_T(__DATE__)+7);
	  return -1;
	}
	// parse arguments
	for (int i=1; i<argc; i++)
	{
	  const BppChar* arg = argv[i];
	  BppChar a0 = arg[0];
	  if (a0 == _T('#')) break; // rest of cmdline contains user args
	  if (a0 == _T('-')) // option
	  {
		 switch (arg[1])
		 {
		 case _T('d'):
			showDebug = true;
			break;
		 case _T('l'):
			vm.setGenerateLineInfo(true);
			break;
		 case _T('v'):
			vm.setGenerateLocalVarInfo(true);
			break;
		 case _T('c'):
			compileOnly = true;
			break;
#ifndef __unix__
		 case _T('E'):
#ifdef WIN32
			noConsole = true;
#endif
#endif
		 case _T('e'):
			compileOnly = true;
			linkExe = true;
			break;
		 case _T('t'):
			doTrace = true;
			break;
		 case _T('i'):
			showInfo = true;
			break;
		 case _T('r'):
			argtype = AT_LIB;
			break;
		 case _T('o'):
			argtype = AT_OUT;
			break;
		 case _T('L'):
			 argtype = AT_LINK;
			 compileOnly = true;
			 linkExe = true;
#ifdef WIN32
          noConsole = false;
#endif
			 break;
		 case _T('D'):
		 {
			 BppString strArg = arg+2;
			 long pos = strArg.find(_T('='));
			 if (pos > 0) {
				 BppString val = strArg.substr(pos + 1);
				 strArg.erase(pos);
				 vm.addPredefinedSymbol(strArg.c_str(),val.c_str());
			 }
			 else if (strArg.length() && pos<0l) {
				 vm.addPredefinedSymbol(strArg.c_str());
			 }
		 }
		 break;
		 default:
			_tprintf(_T("invalid option %s\n"),arg);
			_tprintf(BANNER,_BPP_VER / 100,_BPP_VER % 100,_T(__DATE__)+7);

			return -1;
		 }
		 continue;
	  }
	  switch (argtype) {
	  case AT_OUT:
		  outfile = arg;
		  argtype = AT_ANY;
		  break;
	  case AT_LIB:
	  {
		  BppString* s = new BppString(arg);
		  if (_tcschr(arg, _T('.')) == 0)
			  (*s) += _T(".bpm");
		  moduleVec.push_back(s);
	  }
			break;
	    case AT_LINK:
			 rtfile = arg;
			 argtype = AT_ANY;
			 break;
		 default:
			 {
				 BppString* s = new BppString(arg);
				 if (_tcschr(arg, _T('.')) == 0)
					 (*s) += _T(".bpp");
				 srcVec.push_back(s);
			 }
	  }
	}
	if (outfile.find(_T('.')) == BPP_NNPOS)
	{
		if (linkExe)
		{
#ifndef __unix__
			  outfile += _T(".exe");
#endif		   
		}
		else
			outfile += _T(".bpm");
	}
	// load precompiled modules
	unsigned cnt = moduleVec.size();
	unsigned j;
	for (j=0; j<cnt; j++)
	{
	  const BppChar* modulename = ((BppString*)moduleVec[j])->c_str();
	  const BppChar* fname = searchEnv(_T("BPPLIB"),modulename);
	  if (fname == 0)
		 fname = modulename;

	  if (!vm.readModule(fname))
		 return -1;
	}
	if (showInfo) _tprintf(BANNER,_BPP_VER / 100,_BPP_VER % 100,_T(__DATE__)+7);
	// compile source files
	cnt = srcVec.size();
	for (j=0; j<cnt; j++)
	  if (!vm.compileFile(((BppString*)srcVec[j])->c_str(),showInfo,showDebug))
		 return -1;
	if (compileOnly)
	{
		// write compiled modulr
		if (!vm.writeModule(linkExe ? _T("out.tmp") : outfile.c_str()))
		 return -1;
	    // build executable from compiled module
		if (linkExe && (!build_exe(outfile.c_str(),rtfile.c_str()
#ifdef WIN32
		 ,noConsole
#endif
	  )))
			return -1;
	  return 0;
	}
	// call main(argc,argv) as entry point function
	if (!vm.execute(_T("main"),argc,(const BppChar**)argv,doTrace))
	{
	  return -1;
	}
	// return result of main function
	return (int) (BppInt) vm.getResult();
}

#ifdef WINCE
int WINAPI WinMain(HINSTANCE instance, HINSTANCE prev_instance, BppChar* command_line, int show_command)
{    
	int    argc;
	BppChar** argv;    
	BppChar*  arg;
	int    index;
	int    result;    // count the arguments
	BppChar strchar = 0;
	argc = 1;
	arg  = command_line;
	
	while (arg[0] != 0) 
	{        
		while (arg[0] == _T(' ')) 
		{
			arg++;
	  }
		if ((arg[0] == _T('\'')) || (arg[0] == _T('"')))
		{
			strchar = arg[0];
			++arg;
		}
		if (arg[0] != 0) 
		{
			argc++;
			while ((arg[0] != 0) && (arg[0] !=  ((strchar == 0) ? _T(' ') : strchar))) 
			{
				arg++;
			}
			if (strchar != 0)
			{
				strchar = 0;
				++arg;
			}
		}
	}    
	
	// tokenize the arguments

	argv = (BppChar**)allocmem(argc * sizeof(BppChar**)); 
	arg = command_line;
	index = 1;    
	strchar = 0;
	while (arg[0] != 0) 
	{        
		while (arg[0] == _T(' ')) 
		{
			arg++;
		}        
		if ((arg[0] == _T('\'')) || (arg[0] == _T('"')))
		{
			strchar = arg[0];
			++arg;
		}
		if (arg[0] != 0) 
		{
			argv[index] = arg;
			index++;
			while ((arg[0] != 0) && (arg[0] !=  ((strchar == 0) ? _T(' ') : strchar))) 
			{
				arg++;
			}
			if ((arg[0] != 0) && (strchar != 0))
			{
				strchar = 0;
				arg[0] = 0;
				++arg;
			}
			if (arg[0] != 0) 
			{
				arg[0] = 0;    
				arg++;
			}
		}
	
	}        // put the program name into argv[0]

	BppChar filename[_MAX_PATH];
	
	GetModuleFileName(NULL, filename, _MAX_PATH);
	argv[0] = filename;    // call the user specified main function    
	
	result = _tmain(argc, argv);
	
	free(argv);
	return result;
}

#endif

#if (!defined(WIN32)) && defined(_UNICODE)
int main(int argc, char* argv[]) {
   struct Helper {
      wchar_t ** args;
	  BppBuffer* buffers;
	  int cnt;
	  Helper(int argc, char* argv[]) : cnt(argc) {
	     args = new wchar_t*[argc];
		 buffers = new BppBuffer[argc];
		 for (int i=0;i<cnt;++i) {
			 buffers[i] = str2wcsBuf(argv[i]);
			 args[i] = (wchar_t*) buffers[i].data();
		 }
	  }
	  ~Helper() {
	      delete [] args;
		  delete [] buffers;
	  }
   } helper(argc,argv);
   return _wmain(argc,helper.args);
   
}
#endif

