#ifndef BPPDEF_H
#define BPPDEF_H

#include <limits.h>
#include <stddef.h>

#define _BPP_VER 160 ///< Specifies version of B++ (100 * version number)

#ifdef __GNUC__
#ifndef GCC
#define GCC
#endif
#endif

#ifdef GCC
  #ifndef LLONG_MAX
	 #define LLONG_MAX __LONG_LONG_MAX__
  #endif
  #ifndef INT_MAX
	 #define INT_MAX __INT_MAX__
  #endif
  #ifndef SHRT_MAX
	 #define SHRT_MAX __SHRT_MAX__
  #endif
  #ifndef UINT_MAX
	 #define UINT_MAX __UINT_MAX__
  #endif
  #ifndef USHRT_MAX
	 #define USHRT_MAX __USHRT_MAX__
  #endif
  #ifndef SIZE_MAX
	 #if __WORDSIZE == 64
		#define SIZE_MAX UINT64_MAX
	#else
		#define SIZE_MAX UINT32_MAX
	#endif
  #endif
#endif
//#define EMULATE_I64
#ifndef WINCE
  #ifdef _WIN32_WCE
	 #define WINCE
  #endif
#endif

#ifndef LINUX
#if defined(__linux__) || defined(__linux)
#define LINUX
#endif
#endif

#ifndef UNIX
#if defined(LINUX) || defined(__unix__) || defined(__unix) || (defined(GCC) && !defined(WIN32))
#define UNIX
#endif
#endif

#ifdef _MSC_VER
  #if _MSC_VER <= 800 
	  typedef int bool;
	  #define true 1
	  #define false 0
	  #define __NOSTL__
	  #define NOCPPEXCEPTION 
	  #define asm __asm
  #endif
  #if _MSC_VER >= 1900 
	 #define HAVE_THREADLOCALE
  #endif
#endif
#ifndef NOASM
  #ifdef WINCE
	  #define NOASM
  #endif
  #ifdef GCC
	  #define NOASM
 //    #define __NOSTL__
  #endif
  #ifdef _WIN64
	 #define NOASM
	#ifndef WIN32
	  #define WIN32
	#endif
  #endif
#endif
#ifdef WINCE
// on Windows CE WIN32 must be defined too
#ifndef WIN32
  #define WIN32
#endif
	 typedef int ptrdiff_t;
	#define EDOM 33
	#define ERANGE 34
	#undef STRSAFE_NO_DEPRECATE
	#define DEPRECATE_SUPPORTED  
	// not all versions of WinCE SDK support C++ exceptions
  #define NOCPPEXCEPTION
  // CRT library for WinCE doesn't support conio.h functions
  #define NOCONIO
  // not all versions of WinCE SDK support STL
  #define __NOSTL__
#endif
#ifdef __TURBOC__

#if (__TURBOC__ < 0x0450)
#ifndef __OLDCPP__
#define __OLDCPP__
#endif
#ifndef USE_CAPI
#define NO_CAPI
#endif
#endif

	 enum bool
	 {
		false = 0,
		true = 1
	 };

#endif

#ifdef WIN32
	#ifndef _CONSOLE
		// console io is useable only in console applications
		#ifndef NOCONIO
			#define NOCONIO
		#endif
	#endif
  #include <windows.h>
	#ifndef BPP_STATIC
	  #ifdef BUILD_BPPLIB
		 #define BPPLIB_DLLEXP __declspec(dllexport)
	  #else
		 #define BPPLIB_DLLEXP __declspec(dllimport)
	  #endif
	#else
		#define BPPLIB_DLLEXP
	#endif
  #define BPPAPI WINAPI
#else // no WIN32
	#define BPPLIB_DLLEXP
	#define BPPAPI
   #define BPPCAPI
#endif 
#ifndef LLONG_MAX 
	 // Use int64 emulation if needed
	#ifndef EMULATE_I64
		#ifndef  WINCE
				#define EMULATE_I64
		#endif
	#endif
#endif


#include <bpptchar.h>
#ifndef __FUNCTION__
	#ifdef __func__
		#define __FUNCTION__ __func__
		#define __FUNC_HAS_ADDRESS
	#else 
		#define __FUNCTION__ ""
	#endif
#endif 

#ifndef __FUNC_HAS_ADDRESS
	#ifdef GCC
		#define __FUNC_HAS_ADDRESS
	#endif
#endif // !__FUNC_HAS_ADDRESS

#ifdef __FUNC_HAS_ADDRESS
#define T__FUNCTION BppString(__FUNCTION__).c_str()
#else
#define T__FUNCTION _T(__FUNCTION__)
#endif

#ifdef _DEBUG
#ifdef __FUNC_HAS_ADDRESS
#define T__DFUNCTION BppString::format(_T("%s Line %d:"),T__FUNCTION,__LINE__).c_str()
#else
#define	 LSTR(L) LSTR2(L)
#define LSTR2(L) _T(#L)
#define T__DFUNCTION T__FUNCTION _T(" Line ") LSTR(__LINE__)
#endif
#else
#define T__DFUNCTION _T("")
#endif

#ifndef _UNICODE
	typedef char BppChar;
	typedef unsigned char BppUChar;
#else
	typedef wchar_t BppChar;
#ifdef __OLDCPP__
	typedef unsigned char BppUChar;
#else
	typedef wchar_t BppUChar;
#endif
#endif

#ifndef __cpp_unicode_characters
	typedef unsigned short char16_t;
#if (USHRT_MAX < UINT_MAX)
	typedef unsigned int char32_t;
#else
	typedef unsigned long char32_t;
#endif
#endif 

#ifndef __cpp_char8_t
	typedef unsigned char char8_t;
#endif

	typedef char8_t BppChar8;
	typedef char16_t BppChar16;
	typedef char32_t BppChar32;
	typedef wchar_t BppWChar;

#ifdef NOCALLBACK
	#ifdef USECALLBACK
		#undef USECALLBACK
	#endif
#endif

#ifdef __OLDCPP__

#define NOCPPEXCEPTION

#else
   #ifndef NOCALLBACK
		#define USECALLBACK
	#endif
  // Check for 64 bit Platform
	#if ((SIZE_MAX > UINT_MAX) && (USHRT_MAX < UINT_MAX))
	#define BPP_I64
	#endif
#endif
#ifndef BPP_DLLWILD
// Define build specific suffix
#ifndef __PLFEXT
#ifdef BPP_I64
#define __PLFEXT _T("_64")
#else
#define __PLFEXT _T("")
#endif
#endif
#ifdef _UNICODE
  #ifdef _DEBUG
	 #define BPP_DLLWILD (L"ud" __PLFEXT)
  #else
	 #define BPP_DLLWILD (L"u" __PLFEXT)
  #endif
#else
  #ifdef _DEBUG
	 #define BPP_DLLWILD ("d" __PLFEXT)
  #else
	 #define BPP_DLLWILD __PLFEXT
  #endif
#endif
#endif
	typedef unsigned char BppByte;
	typedef char BppSByte;
	typedef unsigned short BppWord;
	typedef short BppShort;
	typedef unsigned short BppUShort;
#if (!defined(GCC)) && (INT_MAX == SHRT_MAX)
  typedef long BppInt;
  typedef unsigned long BppUInt;
  typedef unsigned BppNWord;
  typedef BppInt BppIntPtr;
  typedef BppUInt BppUIntPtr;
#else
  typedef int BppInt;
  typedef unsigned int BppUInt;
  typedef BppUInt BppNWord;
#endif
  typedef BppUInt BppDWord;
#ifdef GCC
  #define __int64 long long
  #define BppTime_T BppLong
#else
  #define BppTime_T time_t
#endif
#ifdef EMULATE_I64
	 #include <int64.h> 
	typedef Int64 BppLong;
	typedef UInt64 BppULong;
#else
  typedef __int64 BppLong;
  typedef unsigned __int64 BppULong;
  typedef __int64 _Int64;
  typedef unsigned __int64 _UInt64;
#endif
#ifdef GCC
	#ifdef BPP_I64
	  typedef __int64 BppIntPtr
	  typedef unsigned __int64 BppUIntPtr
#else
	 typedef int BppIntPtr;
	 typedef unsigned int BppUIntPtr;
#endif
	#include <stdint.h>
	template<class T> BppIntPtr INTPTRCAST(T v) { return BppIntPtr(v); }
	template<class T> BppIntPtr INTPTRCAST(T* v) { return BppIntPtr((uintptr_t)v); }
	template<class T> BppIntPtr UINTPTRCAST(T v) { return BppUIntPtr(v); }
	template<class T> BppIntPtr UINTPTRCAST(T* v) { return BppUIntPtr((uintptr_t)v); }
	//#define INTPTRCAST(p) ((BppIntPtr)(uintptr_t)(p))
#else
  typedef ptrdiff_t BppIntPtr;
#ifndef _UINTPTR_T_DEFINED
	#define _UINTPTR_T_DEFINED
	#ifdef BPP_I64
	  typedef BppULong  uintptr_t;
	#else
	  typedef BppUInt uintptr_t;
	#endif
	#endif  
  typedef uintptr_t BppUIntPtr;
	#define INTPTRCAST(p) ((BppIntPtr)(p))
    #define UINTPTRCAST(p) ((BppUIntPtr)(p))
#endif

/// Maximum value of uint (NPOS in B++ script)
#define BPP_NPOS (BppUInt(BppInt(-1)))
// Maximum value of size_t (native NPOS for internal use)
#define BPP_NNPOS (size_t(BppIntPtr(-1)))

class BppValue;
class ValPtr;
class BppVirtualMachine;
class BppObject;

enum BP_ERR_CODES
{
	BP_ERR_NOMEM = 1
};

/** Object notification types.
@see BppClass::notifyObjCreation, BppClass::notifyObjDelete, BppClass::notifyObjSerialization
*/
enum _BppObjectNotifyType
{
	BPP_OBJ_NONE = 0,
	BPP_OBJ_CREATE, ///< Object created
	BPP_OBJ_COPY, ///< Object copied
	BPP_OBJ_DELETE, ///< Object will be deleted
	BPP_OBJ_BEFORESERIALIZE, ///< Object will be serialized
	BPP_OBJ_AFTERDESERIALIZE ///< Object has been deserialized
};
typedef enum _BppObjectNotifyType BppObjectNotifyType;

#ifdef NOASM
typedef BppIntPtr (BPPAPI *DLLFUNC0)();
typedef BppIntPtr (BPPAPI *DLLFUNC1)(BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC2)(BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC3)(BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC4)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC5)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC6)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC7)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC8)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC9)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC10)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC11)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC12)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC13)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC14)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC15)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC16)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
#endif

/** Type of user defined native function.
@param vm Pointer to virtual machine calling the function.
@param argc Number of arguments passed to function.
@param sp Stack pointer (iterator to top of stack).
@return true on success; false throws exception in vm.
*/
typedef bool (*BPPUSERFUNC) (BppVirtualMachine* vm, int argc, ValPtr& sp);
#if (defined(GCC) && __cplusplus > 0) || defined(__cplusplus)
#if (defined(__cpp_rvalue_references)  || _MSC_VER >= 1600 || __cplusplus >= 201103)
  #define CPP11
  #include <utility>
  #define BPPMOVE(V) std::move((V))
  #define BPPNOEXCEPT noexcept
#else
	#define BPPMOVE(V) (V)
	#define BPPNOEXCEPT
#endif
/** Type of installable object notification callback function.
@param vm Pointer to virtual machine owning the object.
@param obj Related object of notification.
@param nativeObject Address of pointer to related native object (may be created or destroyed by callback)
@param reason Reason of notification
@note Notification callback may be installed by implementation of native class bindings as a helper for handling of native objects.
@see BppObjectNotifyType, BppClass:setNotifyProc, BppClass::getNotifyProc, Bpp_RegEx::createClass
*/
typedef void (BPPAPI * BppObjNotifyProc)(BppVirtualMachine* vm, BppObject* obj, void** nativeObject, BppObjectNotifyType reason);
typedef bool (BPPAPI *BppDbgWaitProc)(BppVirtualMachine* vm);
typedef bool (BPPAPI *BppBreakRequestProc)(BppVirtualMachine* vm);
typedef bool (BPPAPI *BppLineChangedProc)(BppVirtualMachine* vm, const char* srcName, BppInt lineNumber);
/** Type of registration function for native module.
@param vm Pointer to virtual machine to register native module for.
@note Each native B++-module (dynamic library) must implement and export a function 'RegisterNativeModule' 
of this type to allow loading by #import processing instruction loadNativeModule function.
@see BppVirtualMachine::loadNativeModule
*/
typedef void (BPPAPI *BppRegisterNativeModuleFunc)(BppVirtualMachine* vm);
typedef void (BPPAPI *BppPrinterCallback) (const char* text);
typedef int (BPPAPI *BppCharInputCallback) ();
typedef const char*  (BPPAPI *BppStringInputCallback) ();

#else
typedef void (BPPAPI *BppObjNotifyProc)void* pVm, void* pObj, void** nativeObject, BppObjectNotifyType reason);
typedef int (BPPAPI *BppDbgWaitProc)();
typedef int (BPPAPI *BppBreakRequestProc)();
typedef int (BPPAPI *BppLineChangedProc)(char* srcName, BppInt lineNumber);
typedef void (BPPAPI *BppRegisterNativeModuleFunc)(void* pVm);
typedef void (BPPAPI *BppPrinterCallback) (char* text);
#endif
#if defined(UNIX)
typedef void* HDLL;
typedef void* HMODULE;
typedef int BOOL;
typedef BppIntPtr (*FARPROC)();
#ifndef NO_DYNLIB
#include <dlfcn.h>
#define LoadLibraryA(lpFileName) dlopen(lpFileName, RTLD_NOW)
#define FreeLibrary(hModule) dlclose(hModule)
#define GetProcAddressA(hModule,lpProcName) dlsym(hModule,lpProcName)
#ifdef _UNICODE
#define LoadLibrary LoadLibraryW
#define GetProcAddress GetProcAddressW
#else
#define LoadLibrary LoadLibraryA
#define GetProcAddress GetProcAddressA
#endif // !UNICODE
#endif // !NO_DYNLIB
#endif // UNIX
#endif // BPPDEF_H
