#ifndef BPPDFUNC_H
#define BPPDFUNC_H

/** @file
Declares implementation functions for B++ predefined functions.
*/

#ifndef DOXY

// B++ default functions
#include <bppdef.h>
#include <bppvalue.h>
#include <bppdict.h>
// typedef bool (*BPPUSERFUNC) (BppVirtualMachine* vm, int argc, ValPtr& sp);
class BppVirtualMachine;
class ValPtr;
class BppString;
class BppDictionaryIO;
//enum BppValueType;

class BPPLIB_DLLEXP BppDefaultFunctions
{
   friend class BppDictionaryIO;
public:
   static void registerFunctions(BppVirtualMachine* vm);
   // object creation
   static bool Bpp_newvector(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_newstring(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_newbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_newcbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_newdict(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_newobject(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_clone(BppVirtualMachine* vm, int argc, ValPtr& sp);
#ifdef __MSDOS__
   static void registerDosFunctions(BppVirtualMachine* vm);
#endif   
   static bool Bpp_memsize(BppVirtualMachine* vm, int argc, ValPtr& sp);


   // I/O functions
   static bool Bpp_fopen(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_freopen(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_ftell(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_fseek(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_fflush(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_fclose(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_feof(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_getc(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_putc(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_gets(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fgets(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fputs(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_print(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_freadval(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fwriteval(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fread(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fwrite(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_kbhit(BppVirtualMachine* vm, int argc, ValPtr& sp);
#if (INT_MAX > SHRT_MAX)
   static bool Bpp_getwc(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_putwc(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fgetws(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_fputws(BppVirtualMachine* vm, int argc, ValPtr& sp);
#endif
   // string functions
   static bool Bpp_strlen(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strcmp(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_stricmp(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strsplit(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strfind(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strerase(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strinsert(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strreplace(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_strupr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_strlwr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_substr(BppVirtualMachine* vm, int argc, ValPtr& sp);

   // vector and buffer related functions
   static bool Bpp_vecsize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_size(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_shrink(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_T(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_CBuf(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_qsort(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_foreach(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_find(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_rfind(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_lbound(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_ubound(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_bsearch(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_clear(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_reserve(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_resize(BppVirtualMachine* vm, int argc, ValPtr& sp);
	// vector manipulation functions
	static bool Bpp_insert(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_insrange(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_swap(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_move(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_erase(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_slice(BppVirtualMachine* vm, int argc, ValPtr& sp);
   // dictionary functions
   static bool Bpp_dictcontains(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_dicterase(BppVirtualMachine* vm, int argc, ValPtr& sp);
//   static bool Bpp_dictclear(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_dictgetkeys(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_dictgetvalues(BppVirtualMachine* vm, int argc, ValPtr& sp);

   // program related functions
   static bool Bpp_compile(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_compilefn(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_litexpr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_loadmodule(BppVirtualMachine* vm, int argc, ValPtr& sp);

   // basic type conversion
   static bool Bpp_bool(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_sbyte(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_byte(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_short(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_ushort(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_int(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_uint(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_long(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_ulong(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_intptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_float(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_double(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_char(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_uchar(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_char8_t(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_char16_t(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_char32_t(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_wchar_t(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_string(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_buffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_charbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_vector(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_dictionary(BppVirtualMachine* vm, int argc, ValPtr& sp);

   // RTTI
   static bool Bpp_gettypename(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_gettype(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_getclass(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_getclassname(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_dynamic_cast(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_getsymbol(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_isconst(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_mkconst(BppVirtualMachine* vm, int argc, ValPtr& sp);
   // date and time
   static bool Bpp_time(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_timer(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_ctime(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_localtime(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool Bpp_strftime(BppVirtualMachine* vm, int argc, ValPtr& sp);
	// math functions
   static bool Bpp_sin(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_cos(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_tan(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_atan(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_sqrt(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_log(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_exp(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_abs(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_pow(BppVirtualMachine* vm, int argc, ValPtr& sp);

   // miscellanous funcs
   static bool Bpp_rand(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_srand(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_getargs(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_getusrargs(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_argcnt(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_arg(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_argv(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_call(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_errno(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_strerror(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_nerr(BppVirtualMachine* vm, int argc, ValPtr& sp);

   // system functions
   static bool Bpp_addr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_memmove(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_getenv(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_setenv(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_setlocale(BppVirtualMachine* vm, int argc, ValPtr& sp);
#ifdef USECALLBACK
   static bool Bpp_GetCallbackFunc(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetStaticCallbackFunc(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_SetStaticCallbackHandler(BppVirtualMachine* vm, int argc, ValPtr& sp);
#endif
#ifndef WINCE
   static bool Bpp_system(BppVirtualMachine* vm, int argc, ValPtr& sp);
#endif
#ifdef __MSDOS__
   static bool Bpp_setreg(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_getreg(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_segread(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_int86(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_int86x(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_inportb(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_outportb(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_inport(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_outport(BppVirtualMachine* vm, int argc, ValPtr& sp);

#endif // __MSDOS__
   // graphic bios functions
   static bool Bpp_setscrmode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_setpixel(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_line(BppVirtualMachine* vm, int argc, ValPtr& sp);
   // DLL access
#ifndef NO_DYNLIB
   static bool Bpp_LoadLibrary(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_FreeLibrary(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetProcAddress(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_CallLibFunc(BppVirtualMachine* vm, int argc, ValPtr& sp);
#endif
#ifdef WIN32
#ifndef NO_DYNLIB
   static bool Bpp_GetUser32Func(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetKernel32Func(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetGdi32Func(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetShell32Func(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetOle32Func(BppVirtualMachine* vm, int argc, ValPtr& sp);
#endif
   static bool Bpp_GetLastError(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool Bpp_GetSysErrorMsg(BppVirtualMachine* vm, int argc, ValPtr& sp);
#endif // WIN32

private:
	static bool Bpp_struplow(BppVirtualMachine* vm, int argc, ValPtr& sp, bool upper);
	static bool numcast(BppVirtualMachine* vm, int argc, ValPtr& sp, const BppChar* fname, BppValueType desttype);
   static void registerDictFunctions(BppVirtualMachine* vm);
   static void registerIOFunctions(BppVirtualMachine* vm);
   static void registerMathFunctions(BppVirtualMachine* vm);
   static void registerFiles(BppVirtualMachine* vm);
   static void registerSysVariables(BppVirtualMachine* vm);
	static void _registerSymbol(BppVirtualMachine* vm, const BppChar* name, BppSymbolType stype, const BppValue& val);
   BppDefaultFunctions();
   ~BppDefaultFunctions();
};

#endif // !DOXY

#endif // BPPDFUNC_H
