#ifndef BPPOBJCT_H
#define BPPOBJCT_H

#include <bpprobj.h>

class BppVector;
class BppClass;

/** Class for internal representation of a B++ class instance.
This class provides internal representaton of a B++ class istance, i.e.
a living object. It contains a vector of objects non-static data.
Please note that an object may be constructed as a copy of another one
where assignment is not possible.
*/
class BPPLIB_DLLEXP BppObject : public BppRefObject
{
public:
   /// initialized ctor
   BppObject(BppClass* cls, bool isDummy = false);
   /// copy ctor
   BppObject(const BppObject& cpy);
   /// dtor
   virtual ~BppObject();
   /// create clone of current object
   BppObject* clone() const;
   /// get class definition of this object
   BppClass* getClass() const BPPNOEXCEPT;
   /// get vector of non-static member data
   BppVector* getMemberData() const BPPNOEXCEPT;

   /// Assigns this object to static typed value.
   bool safeAssignToValue(BppValue& val);

   /// External native object reference
   void* _nativeObject;

private:
	BppObject() BPPNOEXCEPT;
   /// assignment forbidden
   BppObject& operator=(const BppObject&);
   /// member data
   BppVector* _memberData;
};

#endif // BPPOBJCT_H
