#ifndef BPPSCAN_H
#define BPPSCAN_H

#include <bppbuf.h>
#include <bppstrng.h>
#include <bppvalue.h>
#include <stdio.h>
enum BppToken
{
	_TMIN = 0,
	T_NOTOKEN = _TMIN,
	T_ERROR,
	T_EOF,
	T_STRING,
	T_IDENTIFIER,
	T_NUMBER,
	T_FNUMBER,
	T_CLASS,
	T_NAMESPACE,
	T_STATIC,
	T_IF,
	T_ELSE,
	T_WHILE,
	T_RETURN,
	T_FOR,
	T_BREAK,
	T_CONTINUE,
	T_DO,
	T_SWITCH,
	T_CASE,
	T_DEFAULT,
	T_NEW,
	T_DELETE,
	T_TRY,
	T_CATCH,
	T_THROW,
	T_VAR,
	T_CONST,
	T_MUTABLE,
	T_FUNCTION,
	T_OPERATOR,
	T_EVAL,
	T_PUBLIC,
	T_PROTECTED,
	T_PRIVATE,
	T_TRON,
	T_TROFF,
	T_TRSTEP,
	T_NIL,
	T_VOID,
	T_AUTO,
	T_YIELD,
	/* T_REMARK, */
	T_LE,	/* '<=' */
	T_EQ,	/* '==' */
	T_NE,	/* '!=' */
	T_GE,	/* '>=' */
	T_SHL,	/* '<<' */
	T_SHR,	/* '>>' */
	T_AND,	/* '&&' */
	T_OR,	/* '||' */
	T_INC,	/* '++' */
	T_DEC,	/* '--' */
	T_ADDEQ,	/* '+=' */
	T_SUBEQ,	/* '-=' */
	T_MULEQ,	/* '*=' */
	T_DIVEQ,	/* '/=' */
	T_REMEQ, // '%='
	T_BANDEQ, // &=
	T_BOREQ, // |=
	T_XOREQ, // ^=
	T_BNOTEQ, // ~=
	T_SHLEQ, // <<=
	T_SHREQ, // >>=
	//T_REFEQ, // =>
	//T_UNREFEQ, // =*
	T_CC,	/* '::' */
	T_MEMREF,	/* '->' */
	T_ELLIPSE, // ...
	T_LITEXPR, // 'litexp' or '#(' literal expression, following bracket is new token
	T_CONCAT, // concatenation of tokens in macro
	T_ANY,  // any other character
	_TMAX = T_ANY
};


class BppCharReader;
class BppStringCharReader;
class BooFileCharReader;
class BppCompiler;
class BppValue;

struct BppSavedToken
{
	BppSavedToken() : dblBuffer(0.0), intnumBuffer(0), charBuffer(0),
		ttype(T_NOTOKEN), numType(BVT_NULL), next(0) {}
	~BppSavedToken();
	BppCharBuffer stringBuffer;
	double dblBuffer;
	BppLong   intnumBuffer;
	BppChar charBuffer;
	BppToken ttype;
	BppValueType numType;
	BppSavedToken* next;
	BppSavedToken* clone() const;
};


class BPPLIB_DLLEXP BppScanner
{
	friend class BppCompiler;
	friend class BppStringCharReader;
	friend class BppFileCharReader;
public:
	BppScanner(FILE* fp, const BppChar* srcName = _T(""), long lnum = 0, unsigned lpos = 0);
	BppScanner(const BppChar* src, const BppChar* srcName = _T(""), long lnum = 0, unsigned lpos = 0);
	~BppScanner();
	BppToken getToken(int& errCnt, bool ignoreBuffer = false, bool skipEOL = true);
	const BppChar* getTokenName(BppToken tkn) const;
	const BppChar* getTokenString() const;
	double getDouble() const;
	BppSByte getSByte() const;
	BppByte getByte() const;
	BppShort getShort() const;
	BppUShort getUShort() const;
	BppInt getLong() const;
	BppUInt getULong() const;
	BppLong getLLong() const;
	BppULong getULLong() const;
	BppValue getNumberValue() const;
	BppChar getChar() const;
	BppToken parseError(const BppChar* msg);
	bool back(bool append = false, BppSavedToken** externalStack = 0);
	void push(FILE* fp, const BppChar* srcName = _T(""));
	void push(const BppChar* src, const BppChar* srcName = _T(""));
	const BppChar* getSourceName() const;
	long getLineNumber() const;
	unsigned getLinePos() const;
	BppValueType getNumberType() const;
	BppChar peekChar();
	BppUInt charConv(BppUInt src) const;
private:
	BppToken readToken(bool skipEOL);
	BppChar get();
	BppChar skipSpaces(bool skipEOL);
	BppToken readString(bool skipEOL);
	BppToken readChar();
	BppChar readLiteralChar();
	BppToken readIdentifier(BppChar firstChar);
	BppToken readNumber(BppChar firstChar, bool isCharcode = false);
	static bool isIdChar(BppUChar ch);
	static BppToken getKeywordTypeFromString(const BppChar* str);
	static BppToken getTokenTypeFromString(const BppChar* str);
	BppCharBuffer _stringBuffer;
	double _dblBuffer;
	BppValueType _numType;
	union {
		BppSByte _sbyteBuffer;
		BppByte _byteBuffer;
		BppShort _shortBuffer;
		BppUShort _ushortBuffer;
		BppInt   _intBuffer;
		BppUInt  _uintBuffer;
		_Int64   _longBuffer;
		_UInt64  _ulongBuffer;
	};
	BppChar _charBuffer[2];
	BppCharReader* _reader;
	BppSavedToken* _tokenStack; // stack of look ahead tokens
	BppToken _token;        /* current token */
	//bool     _getBack;     // get last token again, set by back(), reset by getToken() 
	BppChar _savedCh;      /* look ahead character */
	BppChar _lastCh;       /* last input character */
	BppCharBuffer _line;      /* last input line */
	unsigned _linepos;	   /* index in line */
	long _lnum;	           /* line number */
	BppScanner* _next;      // pointer to next (embedded) scanner
	BppString _srcName;    // source name (e.g. filename)
	BppString _funcName;   // name of currently compiled function
	BppString _callerDir;  // working directory when scanner is created
	static BppCompiler* _notifyCompiler;
	BppUShort _srcCodePage;
	BppUShort _internalCodePage;
#ifndef DOXY
	BppScanner(const BppScanner&);
	BppScanner& operator=(const BppScanner&);
#endif
};

#endif // BPPSCAN_H
