#ifndef BPPUFUNC_H
#define BPPUFUNC_H
#ifndef DOXY
#include <bppdef.h>
#endif
#include <bppstrng.h>
#include <time.h>
#include <stdio.h>
#if !defined(UNIX) && !defined(WINCE)
#include <conio.h>
#endif
class BppBuffer;

/// Searches for <tt>fname</tt> along path specified by PATH environment variable.
BPPLIB_DLLEXP const BppChar* searchPath(const BppChar* fname);
/// Searches for <tt>fname</tt> along path specified by environment variable <tt>varname</tt>.
BPPLIB_DLLEXP const BppChar* searchEnv(const BppChar* varname, const BppChar* fname);
/// Searches for <tt>fname</tt> along path specified by <tt>dirlist</tt>.
BPPLIB_DLLEXP const BppChar* searchDirList(const BppChar* dirList, const BppChar* fname);
/// Aborts program returning code specified by <tt>err</tt>.
BPPLIB_DLLEXP void errorExit(int err);
///  Allocates memory of <tt>blksize</tt> bytes.
BPPLIB_DLLEXP void* allocmem(size_t blksize);
///  Resizes memory block specified by <tt>p</tt> to <tt>blksize</tt> bytes.
BPPLIB_DLLEXP void* reallocmem(void* p, size_t blksize);
/// Frees (deallocates) memory block specified by <tt>p</tt>.
BPPLIB_DLLEXP void freemem(void* p);
/// Copies <tt>n</tt> bytes from address <tt>src</tt> to <tt>dst</tt>.
BPPLIB_DLLEXP BppIntPtr movemem(BppIntPtr dst, BppIntPtr src, BppUInt n);
/// Gets fully qualified directory part of file specified by <tt>filePath</tt>.
BPPLIB_DLLEXP const BppChar* getFileDir(const BppChar* filePath, bool stripSlash);
/// Duplicates character string <tt>src</tt>.
BPPLIB_DLLEXP BppChar* dupstr(const BppChar* src);
#ifdef WINCE
#ifdef _gettche
	#undef _gettche
#endif

BPPLIB_DLLEXP const BppChar* _tgetenv(const BppChar* varname);
BPPLIB_DLLEXP int _tputenv(const BppChar* envstr);

BPPLIB_DLLEXP int _gettche();
BPPLIB_DLLEXP time_t time(time_t* t);
BPPLIB_DLLEXP BppChar* _tctime(time_t* tt);

struct BPPLIB_DLLEXP tm
{
	int tm_sec, tm_min, tm_hour, tm_mday, tm_mon, tm_year, tm_wday, tm_yday, tm_isdst;
};

BPPLIB_DLLEXP tm* localtime( const time_t* t);

#endif // WINCE
#ifndef __OLDCPP__
struct _TmpLocale {
#ifdef HAVE_THREADLOCALE
	int _oldmode;
#endif
	BppString _old;
	_TmpLocale(const char* loc="C");
	~_TmpLocale();
};
#define C_LOCALE _TmpLocale __cLocale
#else
#define C_LOCALE
#endif

struct FlagHlp {
	bool& _flag;
	bool _savedFlag;
	FlagHlp(bool& flag, bool val) : _flag(flag), _savedFlag(flag) {
		_flag = val;
	}
	~FlagHlp() {
		_flag = _savedFlag;
	}
};
#ifndef _vsctprintf
#include <stdarg.h>
	BPPLIB_DLLEXP int _vscprintf(const char* format, va_list args);
 #if !(defined(WIN32) || defined(__OLDCPP__))
	BPPLIB_DLLEXP int _vscwprintf(const wchar_t* format, va_list args);
 #endif
 #ifndef _UNICODE
	  #define _vsctprintf _vscprintf
 #else
	  #define _vsctprintf _vscwprintf
 #endif
 #define emulate_vsctprintf
#endif

#ifndef _vstprintf_s
  #ifndef _UNICODE
	  #define _vstprintf_s(BUF,CNT,FORMAT,ARGS)  vsprintf(BUF,FORMAT,ARGS)
  #else
	  #define _vstprintf_s(BUF,CNT,FORMAT,ARGS)  vswprintf(BUF, CNT,FORMAT,ARGS)
  #endif
#endif 


#ifdef __OLDCPP__
	int snprintf(char* s, size_t n, const char* format, ...);
#endif

#if defined(UNIX)
   char* _fullpath(char* absPath,const char* relPath, size_t maxLength);
   wchar_t* _wfullpath(wchar_t* absPath,const wchar_t* relPath, size_t maxLength);
   char* strlwr(char* s);
   char* strupr(char* s);
   wchar_t* wcslwr(wchar_t* s);
   wchar_t* wcsupr(wchar_t* s);
   FILE* _wfopen(const wchar_t* fname, const wchar_t* mode);
   wchar_t* _wctime(const time_t* timer);
   wchar_t* wgetenv(const wchar_t* name);
   int wputenv(wchar_t* val);
   FILE* _wfreopen(const wchar_t* fname, const wchar_t* fmode, FILE* stream);
   wchar_t* _wsetlocale(int category, const wchar_t* locale);
   int _wsystem(const wchar_t* command);
#ifndef NO_DYNLIB
   void* LoadLibraryW(const wchar_t* lpFileName);
   void* GetProcAddressW(void* hModule,const wchar_t* lpProcName);
#endif
   long wtol(const wchar_t* str);
   int wtoi(const wchar_t* str);
   wchar_t* wstok(wchar_t* str, const wchar_t* delim);
   wchar_t* _wgetcwd(wchar_t* buffer, size_t bsize);
   int _wchdir(const wchar_t* dirname);
#endif

#ifndef __OLDCPP__
   enum BPP_UCPAGES {
	   BPP_UTF8 = 65001,
	   BPP_UTF16 = 1200,
	   BPP_UTF16be = 1201,
	   BPP_UTF32 = 65005,
	   BPP_UTF32be = 65006
   };
#else	  
   #define BPP_UTF8 65001u
   #define BPP_UTF16 1200u
   #define BPP_UTF16be 1201u
   #define BPP_UTF32 65005u
   #define BPP_UTF32be 65006u
#endif

#ifdef WIN32
#define DEF_CP 1252
#else 
#ifdef LINUX
  #define DEF_CP 65001
#else
  #define DEF_CP 437
#endif
#endif

#ifndef __OLDCPP__
   /// Converts single byte string to buffer containing wide character string.
   BppBuffer str2wcsBuf(const char* str);
   /// Converts wide character string to buffer containing single byte string.
   BppBuffer wcs2strBuf(const wchar_t* wcs);
#endif 
   /// Converts UTF-32le character to UTF-16le (UCS-2) character.
   BppUShort c32_to_c16(BppUInt c32);
   /// Converts UTF-32le character to UTF-8 character.
   BppUInt c32_to_u8(BppUInt c32);
   /// Converts UTF-32le character to single byte character.
   BppUChar c32_to_char(BppUInt c32, unsigned cp = 0);
   /// Converts UTF-16le (UCS-2) character to UTF-32le character.
   BppUInt c16_to_c32(BppUShort c16);
   /// ConvertsUTF-16le (UCS-2) character to UTF-8 character.
   BppUInt c16_to_u8(BppUShort c16);
   /// Converts UTF-16le (UCS-2) character to single byte character.
   BppUChar c16_to_char(BppUShort c16, unsigned cp = 0);
   /// Converts UTF-8 character to UTF-32le character.
   BppUInt u8_to_c32(BppUInt u8);
   /// Converts UTF-8 character to UTF-16le (UCS-2) character.
   BppUShort u8_to_c16(BppUInt u8);
   /// Converts UTF-8 character to single byte character.
   BppUChar u8_to_char(BppUInt u8, unsigned cp = 0);
   /// Converts single byte character to UTF-32le character.
   BppUInt char_to_c32(BppUChar c, unsigned cp = 0);
   /// Converts single byte character to UTF-16le (UCS-2) character.
   BppUShort char_to_c16(BppUChar c, unsigned cp = 0);
   /// Converts single byte character to UTF-8 character.
   BppUInt char_to_u8(BppUChar c, unsigned cp = 0);
   BppChar cbufToBppChar(BppUInt cbuf, unsigned csize, unsigned cp = 0);
   wchar_t cbufToWchar(BppUInt cbuf, unsigned csize, unsigned cp = 0);

   BppUShort getCpIdFromName(const BppChar* name, BppUShort defcp = DEF_CP);
   //const BppChar* getNameFromCpId(BppUShort cp = DEF_CP);

   BppUShort getActiveCp(BppUShort defcp = DEF_CP);

   unsigned detectFileCp(FILE* fp);

   bool keyPressed();
#endif // BPPUFUNC_H
