#ifndef INT64_H
#define INT64_H
#ifdef EMULATE_I64


typedef struct __Int64 {
	unsigned long _low;
	long _high;
} _Int64;

typedef struct __UInt64 {
	unsigned long _low;
	unsigned long _high;
} _UInt64;

#ifdef __cplusplus
class UInt64;

class Int64
{
	friend class UInt64;
public:
	inline Int64() :_high(0),_low(0){}
	inline Int64(long high, unsigned long low):_high(high),_low(low){}
	inline Int64(const _Int64& val):_high(val._high),_low(val._low){}
	inline Int64(const Int64& val):_high(val._high),_low(val._low){}
	Int64(const UInt64& val);
	inline Int64(long val):_high(val<0 ? -1 : 0),_low((unsigned long) val){}
	inline Int64(unsigned long val):_high(0),_low(val){}
	inline Int64(int val):_high(val<0 ? -1 : 0),_low((unsigned long) val){}
	inline Int64(unsigned int val):_high(0),_low(val){}
	inline Int64(short val):_high(val<0 ? -1 : 0),_low((unsigned long) val){}
	inline Int64(unsigned short val):_high(0),_low(val){}
	Int64(float val);
	Int64(double val);

	Int64& operator=(const Int64& val);
	Int64& operator++();
	const Int64 pinc(); // emulates postfix operator
	Int64& operator--();
	const Int64 pdec(); // emulates postfix operator
	
	Int64 operator-() const;
	Int64 operator~() const;
	Int64& operator+=(const Int64& val);
	Int64& operator+=(const UInt64& val);
	Int64& operator-=(const Int64& val);
	Int64& operator-=(const UInt64& val);
	Int64& operator*=(const Int64& val);
	Int64& operator*=(const UInt64& val);
	Int64& operator/=(const Int64& val);
	Int64& operator/=(const UInt64& val);
	Int64& operator%=(const Int64& val);
	Int64& operator%=(const UInt64& val);
	Int64& operator<<=(unsigned nbits);
	Int64& operator>>=(unsigned nbits);
	Int64& operator&=(const Int64& val);
	Int64& operator|=(const Int64& val);
	Int64& operator^=(const Int64& val);
	Int64 operator<<(unsigned nbits) const;
	Int64 operator>>(unsigned nbits) const;
	int compare(const Int64& cmp) const;
	inline bool operator<(const Int64& cmp) const { return compare(cmp) < 0;}
	inline bool operator<=(const Int64& cmp) const { return compare(cmp) <= 0;}
	inline bool operator>(const Int64& cmp) const { return compare(cmp) > 0;}
	inline bool operator>=(const Int64& cmp) const { return compare(cmp) != 0;}
	inline bool operator==(const Int64& cmp) const { return compare(cmp) == 0;}
	inline bool operator!=(const Int64& cmp) const { return compare(cmp) != 0;}
	inline operator _Int64() const { return *((_Int64*)this); }
	operator float() const;
	operator double() const;
	BppChar* toa(BppChar* buf = 0, int base = 10, bool upper=false) const;
	inline int toInt() const { return _low; }
	inline unsigned int toUInt() const { return _low; }
	inline long toLong() const { return _low; }
	inline unsigned long toULong() const { return _low; }
private:
	friend Int64 operator*(const Int64& l, const Int64& r);
	static bool _idiv(const Int64& dividend, const Int64& divisor, Int64& q, Int64& r);
	unsigned long _low;
	long _high;
};

Int64 operator+(const Int64& l, const Int64& r);
Int64 operator-(const Int64& l, const Int64& r);
Int64 operator*(const Int64& l, const Int64& r);
Int64 operator/(const Int64& l, const Int64& r);
Int64 operator%(const Int64& l, const Int64& r);
Int64 operator&(const Int64& l, const Int64& r);
Int64 operator|(const Int64& l, const Int64& r);
Int64 operator^(const Int64& l, const Int64& r);

class UInt64
{
	friend class Int64;
public:
	inline UInt64() :_high(0),_low(0) {}
	inline UInt64(unsigned long high, unsigned long low):_high(high),_low(low){}
	inline UInt64(const UInt64& val):_high(val._high),_low(val._low){}
	inline UInt64(const _UInt64& val):_high(val._high),_low(val._low){}
	inline UInt64(const Int64& val):_high(val._high),_low(val._low){}
	inline UInt64(long val):_high(val<0 ? (unsigned long)-1 : 0),_low((unsigned long) val){}
	inline UInt64(unsigned long val):_high(0),_low(val) {}
	inline UInt64(int val):_high(val<0 ? (unsigned long)-1 : 0),_low((unsigned long) val){}
	inline UInt64(unsigned int val):_high(0),_low(val) {}
	inline UInt64(short val):_high(val<0 ? (unsigned long)-1 : 0),_low((unsigned long) val){}
	inline UInt64(unsigned short val):_high(0),_low(val) {}
	UInt64(float val);
	UInt64(double val);

	UInt64& operator=(const UInt64& val);
	UInt64& operator++();
	const UInt64 pinc(); // emulates postfix operator
	UInt64& operator--();
	const UInt64 pdec(); // emulates postfix operator
	
	UInt64 operator-() const;
	UInt64 operator~() const;
	//UInt64& operator+=(const Int64& val);
	UInt64& operator+=(const UInt64& val);
	//UInt64& operator-=(const Int64& val);
	UInt64& operator-=(const UInt64& val);
	//UInt64& operator*=(const Int64& val);
	UInt64& operator*=(const UInt64& val);
	//UInt64& operator/=(const Int64& val);
	UInt64& operator/=(const UInt64& val);
	//UInt64& operator%=(const Int64& val);
	UInt64& operator%=(const UInt64& val);
	UInt64& operator<<=(unsigned nbits);
	UInt64& operator>>=(unsigned nbits);
	UInt64& operator&=(const UInt64& val);
	UInt64& operator|=(const UInt64& val);
	UInt64& operator^=(const UInt64& val);
	UInt64 operator<<(unsigned nbits) const;
	UInt64 operator>>(unsigned nbits) const;
	int compare(const UInt64& cmp) const;
	inline bool operator<(const UInt64& cmp) const { return compare(cmp) < 0;}
	inline bool operator<=(const UInt64& cmp) const { return compare(cmp) <= 0;}
	inline bool operator>(const UInt64& cmp) const { return compare(cmp) > 0;}
	inline bool operator>=(const UInt64& cmp) const { return compare(cmp) != 0;}
	inline bool operator==(const UInt64& cmp) const { return compare(cmp) == 0;}
	inline bool operator!=(const UInt64& cmp) const { return compare(cmp) != 0;}
	inline operator _UInt64() const { return *((const _UInt64*)this);}
	operator float() const;
	operator double() const;
	BppChar* toa(BppChar* buf = 0, int base = 10, bool upper=false) const;
	inline int toInt() const { return _low; }
	inline unsigned int toUInt() const { return _low; }
	inline long toLong() const { return _low; }
	inline unsigned long toULong() const { return _low; }
private:
	friend UInt64 operator*(const UInt64& l, const UInt64& r);
	static bool _div(const UInt64& dividend, const UInt64& divisor, UInt64& q, UInt64& r);
	unsigned long _low;
	unsigned long _high;
};

UInt64 operator+(const UInt64& l, const UInt64& r);
UInt64 operator-(const UInt64& l, const UInt64& r);
UInt64 operator*(const UInt64& l, const UInt64& r);
UInt64 operator/(const UInt64& l, const UInt64& r);
UInt64 operator%(const UInt64& l, const UInt64& r);
UInt64 operator&(const UInt64& l, const UInt64& r);
UInt64 operator|(const UInt64& l, const UInt64& r);
UInt64 operator^(const UInt64& l, const UInt64& r);


#else
typedef _Int64 Int64;
typedef _UInt64 UInt64;

#endif // !__cplusplus
Int64 _strtoInt64(const char* nptr, char** endptr, int base);
UInt64 _strtoUInt64(const char* nptr, char** endptr, int base);
#endif // EMULATE_I64
#endif
