
/** @file
Defines implementation functions for B++ predefined functions.
*/ 

#ifndef DOXY
#ifdef WIN32
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0601
//#include "bppwbgi.h"
#endif
#endif
#include "bppdef.h"

#include <stdlib.h>
#include <stdio.h>
#ifndef WINCE
 #include <time.h>
 #ifndef GCC
 #include <process.h>
 #else
	 #include <errno.h>
	extern char** environ;
 #endif


#endif
#include <locale.h>

#include "bppufunc.h"
#include "bppdfunc.h"
#include "bppbuf.h"
#include "bppvm.h"
#include "bppint.h"
#include "bppcomp.h"
#include "bppvalue.h"
#include "bppobjct.h"
#include "bppclass.h"
#include "bppstrng.h"
#include "bppvec.h"
#include "bppprint.h"
#include "bppfimpl.h"
#include "dictio.h"
#include "bppmsg.h"

#define ADDFUNC(NAME) _registerSymbol(vm,_T(#NAME),ST_SFUNCTION,(BPPUSERFUNC) &Bpp_##NAME)

	void BppDefaultFunctions::registerIOFunctions(BppVirtualMachine* vm)
	{
		ADDFUNC(fopen);
		ADDFUNC(freopen);
		ADDFUNC(fclose);
		ADDFUNC(feof);
		ADDFUNC(getc);
		ADDFUNC(ftell);
		ADDFUNC(fseek);
		ADDFUNC(fflush);
		ADDFUNC(putc);
		ADDFUNC(gets);
		ADDFUNC(fgets);
		ADDFUNC(fputs);
		ADDFUNC(print);
		ADDFUNC(freadval);
		ADDFUNC(fwriteval);
		ADDFUNC(fread);
		ADDFUNC(fwrite);
		ADDFUNC(kbhit);
#ifdef WIN32
		ADDFUNC(getwc);
		ADDFUNC(putwc);
		ADDFUNC(fgetws);
		ADDFUNC(fputws);
#endif
	}

	// program related functions
	// ok = compile(string src | FILE* fp)
	bool BppDefaultFunctions::Bpp_compile(BppVirtualMachine* vm, int argc, ValPtr& sp)
	{
		CHKVM;
		CHKARGC(1, 3);
		ValPtr bp = sp; bp -= argc;
		BppValue& v = bp[1];
		const BppChar* srcName = _T("");
		bool debug = false;
		if (argc == 3)
		{
			debug = bp[3].asBool();
		}
		if (argc >1)
		{
			BppString* s = bp[2];
			if (s == 0)
				THROW(FLIB_INVALID_ARG_TYPE);
			srcName = s->c_str();
		}
		bool ok = false;
		switch (v.getType())
		{
		case BVT_STRING:
		case BVT_STRINGREF:
		{
			BppString* s = v;
			if (s == 0) break;
			BppCompiler comp(*vm);
			if (debug) {
				comp.setGenerateLineInfo(true);
				comp.setGenerateLocalVarInfo(true);
			}
			ok = comp.compile(s->c_str(), srcName);
		}
		break;
		case BVT_FILE:
		{
			CHKVAR(v);
			BppCompiler comp(*vm);
			if (debug) {
				comp.setGenerateLineInfo(true);
				comp.setGenerateLocalVarInfo(true);
			}
			ok = comp.compile((FILE*)v, srcName);
		}
		break;
		default:
			THROW(FLIB_INVALID_ARG_TYPE);
			break;
		}
		*bp = BppValue::makeBool(ok);
		sp = bp;
		return true;
}

	// program related functions
	// ok = compilefn(string src,string args="",string rtype="",var ownerClass = null, string srcname = "")
	bool BppDefaultFunctions::Bpp_compilefn(BppVirtualMachine* vm, int argc, ValPtr& sp)
	{
		CHKVM;
		CHKARGC(1, 6);
		ValPtr bp = sp; bp -= argc;
		bool argok = true;
		const BppChar* srcName = _T("");
		const BppChar* rtype = _T("");
		const BppChar* params = _T("");
		BppString* s = 0;
		BppClass* pc = 0;
		bool debug = false;
		switch (argc)
		{
		case 6:
			debug = bp[6].asBool();
		case 5: // source name
			s = bp[5];
			if (!s) { argok = false; break; }
			srcName = s->c_str();
		case 4: // class or class name
		{
			BppValue& v = bp[4];
			switch (v.getType())
			{
			case BVT_CLASS:
			case BVT_CLASSREF:
				CHKVAR(v);
				pc = v;
				break;
			case BVT_STRING:
			case BVT_STRINGREF:
			{
				s = v;
				if (!s) { argok = false; break; }
				if (!s->length()) break;
				BppDictionaryEntry* pe = vm->classes().find(*s);
				if (!pe) THROW(FLIB_UNKNOWN_CLASS_NAME, s->c_str());
				pc = pe->getValue();
			}
			break;
			case BVT_NULL:
				break;
			default:
				argok = false;
			}
		}
		if (!argok) break;
		case 3: // result type
			s = bp[3];
			if (!s) { argok = false; break; }
			rtype = s->c_str();
		case 2: // formal parameter list
			s = bp[2];
			if (!s) { argok = false; break; }
			params = s->c_str();
		}
		s = bp[1];
		if (!(s && argok)) THROW(FLIB_INVALID_ARG_TYPE);
		BppCompiler comp(*vm);
		if (debug) {
			comp.setGenerateLineInfo(true);
			comp.setGenerateLocalVarInfo(true);
		}
		*bp = BppValue(comp.compileFunction(s->c_str(), params, rtype, srcName, pc), true);
		sp = bp;
		return true;
	}


	bool BppDefaultFunctions::Bpp_litexpr(BppVirtualMachine * vm, int argc, ValPtr & sp)
	{
		CHKVM;
		CHKARGC(1, 2);
		ValPtr bp = sp; bp -= argc;
		BppString* s = bp[1];
		BppString* sn = 0;
		if (!s) THROW(FLIB_INVALID_ARG_TYPE);
		if (argc > 1) {
			sn = bp[2];
			if (!sn) THROW(FLIB_INVALID_ARG_TYPE);
		}
		BppStringPrinter pr;
		BppCompiler comp(*vm, &pr);
		bool ok = comp.litExpr(s->c_str(), *bp, sn ? sn->c_str() : _T(""));
		sp = bp;
		if (!ok) THROW(pr.c_str());
		return ok;
	}

	// ok = loadmodule(filename)
	bool BppDefaultFunctions::Bpp_loadmodule(BppVirtualMachine* vm, int argc, ValPtr& sp)
	{
		CHKVM;
		CHKARGC(1, 1);
		BppString* filename = *sp;
		if (filename == 0)
			THROW(FLIB_INVALID_ARG_TYPE);
		BppDictionaryIO dict(*vm);
		bool ok = dict.readModule(*filename);
		--sp;
		*sp = BppValue::makeBool(ok);
		return true;
	}


#ifdef WIN32
	#include <windows.h>
#endif
#ifdef __linux__
#include <dlfcn.h>
#endif

#ifdef __MSDOS__
#include <dos.h>
#ifndef __TURBOCC__
	#include <conio.h>
#endif
#ifndef NO_DYNLIB
#include "usedll.h"
#endif
	static union REGS _regs;
	static struct SREGS _sregs;



#define SWITCHREGSET(NAME,PREFIX) if (_tcsicmp(#NAME,regname) == 0) {PREFIX##.##NAME = val; return true;}
static bool doSetReg(const BppChar* regname, BppWord val)
{
		  SWITCHREGSET(ax,_regs.x)
		  SWITCHREGSET(bx,_regs.x)
		  SWITCHREGSET(cx,_regs.x)
		  SWITCHREGSET(dx,_regs.x)
		  SWITCHREGSET(si,_regs.x)
		  SWITCHREGSET(di,_regs.x)
#ifndef __WATCOMC__
		  SWITCHREGSET(cflag,_regs.x)
#endif
#ifdef __TURBOC__
		  SWITCHREGSET(flags,_regs.x)
#endif
		  SWITCHREGSET(al,_regs.h)
		  SWITCHREGSET(ah,_regs.h)
		  SWITCHREGSET(bl,_regs.h)
		  SWITCHREGSET(bh,_regs.h)
		  SWITCHREGSET(cl,_regs.h)
		  SWITCHREGSET(ch,_regs.h)
		  SWITCHREGSET(dl,_regs.h)
		  SWITCHREGSET(dh,_regs.h)
		  SWITCHREGSET(ds,_sregs)
		  SWITCHREGSET(es,_sregs)
		  SWITCHREGSET(ss,_sregs)
		  SWITCHREGSET(cs,_sregs)
	return false;
}
// setreg(string regname1[, string regname2],int val)
bool BppDefaultFunctions::Bpp_setreg(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	if (!(bp[argc].isIntegralType()))
	  THROW(FLIB_INVALID_TYPE_OF_VALUE_ARG);
	BppString* s = bp[argc-1];
	if (s == 0)
	  THROW(FLIB_INVALID_TYPE_OF_REG_ARG);
	BppInt val = *sp;
	if (!doSetReg(s->c_str(),(BppWord) (val & 0xFFFF)))
					THROW(FLIB_INVALID_REGNAME);
	if (argc == 3)
	{
		s = bp[1];
		if (s == 0)
			THROW(FLIB_INVALID_TYPE_OF_REG_ARG);
		if (!doSetReg(s->c_str(),(BppWord) (val >> 16)))
			THROW(FLIB_INVALID_REGNAME);
	}
	sp = bp;
	sp->dispose();
	return true;
}

#define SWITCHREGGET(NAME,PREFIX) if (_tcsicmp(#NAME,regname) == 0) { val = (BppWord) PREFIX##.##NAME ; return true;}
static bool doGetReg(const BppChar* regname, BppWord& val)
{
		  SWITCHREGGET(ax,_regs.x)
		  SWITCHREGGET(bx,_regs.x)
		  SWITCHREGGET(cx,_regs.x)
		  SWITCHREGGET(dx,_regs.x)
		  SWITCHREGGET(si,_regs.x)
		  SWITCHREGGET(di,_regs.x)
#ifndef __WATCOMC__
		  SWITCHREGGET(cflag,_regs.x)
#endif
#ifdef __TURBOC__
		  SWITCHREGGET(flags,_regs.x)
#endif
		  SWITCHREGGET(al,_regs.h)
		  SWITCHREGGET(ah,_regs.h)
		  SWITCHREGGET(bl,_regs.h)
		  SWITCHREGGET(bh,_regs.h)
		  SWITCHREGGET(cl,_regs.h)
		  SWITCHREGGET(ch,_regs.h)
		  SWITCHREGGET(dl,_regs.h)
		  SWITCHREGGET(dh,_regs.h)
		  SWITCHREGGET(ds,_sregs)
		  SWITCHREGGET(es,_sregs)
		  SWITCHREGGET(ss,_sregs)
		  SWITCHREGGET(cs,_sregs)
	return false;
}
// intval = getreg(string regname1[, string regname2])
bool BppDefaultFunctions::Bpp_getreg(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	BppString* s = bp[1];
	if (s == 0)
	  THROW(FLIB_INVALID_TYPE_OF_VALUE_ARG);
	BppWord regval;
	BppDWord val=0;
	if (!doGetReg(s->c_str(),regval))
		 THROW(FLIB_INVALID_REGNAME);
	val = regval;
	if (argc == 2)
	{
		val = val << 16;
		s = bp[2];
		if (s == 0)
			THROW(FLIB_INVALID_TYPE_OF_REG_ARG);
		if (!doGetReg(s->c_str(),regval))
			THROW(FLIB_INVALID_REGNAME);
		val += regval;
	}
	*bp = (BppInt) val;
	sp = bp;
	return true;
}

// segread()
bool BppDefaultFunctions::Bpp_segread(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(0,0);
		  segread(&_sregs);
	sp->dispose();
	return true;
}

// int axval = int86(int intnr)
bool BppDefaultFunctions::Bpp_int86(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
	  THROW(FLIB_INVALID_ARG_TYPE);
		  BppUShort n =  *sp;
		  --sp;
	*sp = (BppInt) int86(n,&_regs,&_regs);
	return true;
}

// int axval = int86x(int intnr)
bool BppDefaultFunctions::Bpp_int86x(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
	  THROW(FLIB_INVALID_ARG_TYPE);
		  BppUShort n = *sp;
		  --sp;
	*sp = (BppInt) int86x(n,&_regs,&_regs,&_sregs);
	return true;
}

// inbyte = inportb(int port)
bool BppDefaultFunctions::Bpp_inportb(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
	  THROW(FLIB_INVALID_ARG_TYPE);
		  BppUShort port = *sp;
		  --sp;
#ifdef inportb
  #undef inportb
#endif
#ifndef __TURBOC__
	#ifdef _MSC_VER
		#define inportb _inp
	#else
		 #define inportb inp
	#endif
#endif
		  *sp = (BppByte) inportb((unsigned)port);
	return true;
}

// void outportb(port,byteval)
bool BppDefaultFunctions::Bpp_outportb(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= 2;    if (!sp->isIntegralType())

	if (!(bp[1].isIntegralType() && bp[2].isIntegralType()))
			 THROW(FLIB_INVALID_ARG_TYPE);
#ifdef outportb
  #undef outportb
#endif
#ifndef __TURBOC__
	#ifdef _MSC_VER
		#define outportb _outp
	#else
		 #define outportb outp
	#endif
#endif
						outportb((BppUShort) bp[1],(BppByte) bp[2]);
	sp = bp;
	sp->dispose();
	return true;
}

// inword = inport(int port)
bool BppDefaultFunctions::Bpp_inport(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
#ifdef inport
  #undef inport
#endif
#ifndef __TURBOC__
	#ifdef _MSC_VER
		#define inport _inpw
	#else
		 #define inport inpw
	#endif
#endif
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
	  THROW(FLIB_INVALID_ARG_TYPE);
		  BppUShort port = *sp;
	--sp;
	BppUShort v = inport(port);
	*sp = v;
	return true;
}

// void outportb(port,wordval)
bool BppDefaultFunctions::Bpp_outport(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
#ifdef outport
  #undef outport
#endif
#ifndef __TURBOC__
	#ifdef _MSC_VER
		#define outport _outpw
	#else
		 #define outport outpw
	#endif
#endif
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= 2;
	if (!(bp[1].isIntegralType() && bp[2].isIntegralType()))
	  THROW(FLIB_INVALID_ARG_TYPE);
		  outport((BppUShort) bp[1],(BppUShort) bp[2]);
	sp = bp;
	sp->dispose();
	return true;
}



#endif // __MSDOS__

#ifdef WIN32  // Borrowed from MSDN
// https://docs.microsoft.com/en-us/windows/console/clearing-the-screen

void cls(HANDLE hConsole)
{
	COORD coordScreen = { 0, 0 };    // home for the cursor 
	DWORD cCharsWritten;
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	DWORD dwConSize;

	// Get the number of character cells in the current buffer. 

	if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
	{
		return;
	}

	dwConSize = csbi.dwSize.X * csbi.dwSize.Y;

	// Fill the entire screen with blanks.

	if (!FillConsoleOutputCharacter(hConsole,        // Handle to console screen buffer 
		(TCHAR) ' ',     // Character to write to the buffer
		dwConSize,       // Number of cells to write 
		coordScreen,     // Coordinates of first cell 
		&cCharsWritten))// Receive number of characters written
	{
		return;
	}

	// Get the current text attribute.

	if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
	{
		return;
	}

	// Set the buffer's attributes accordingly.

	if (!FillConsoleOutputAttribute(hConsole,         // Handle to console screen buffer 
		csbi.wAttributes, // Character attributes to use
		dwConSize,        // Number of cells to set attribute 
		coordScreen,      // Coordinates of first cell 
		&cCharsWritten)) // Receive number of characters written
	{
		return;
	}

	// Put the cursor at its home coordinates.

	SetConsoleCursorPosition(hConsole, coordScreen);
}

#endif
// graphic bios functions
// int newmode =  setscrmode(int mode)
bool BppDefaultFunctions::Bpp_setscrmode(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 1);
	if (!sp->isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
#ifdef __MSDOS__
	union REGS regs;
	regs.h.ah = 0;
	regs.h.al = (BppByte)*sp;
	int86(0x10, &regs, &regs);
	regs.h.ah = 0x0F;
	int86(0x10, &regs, &regs);
	--sp;
	*sp = (BppByte)regs.h.al;;
#else
	BppInt mode = *sp;
	sp[-1] = mode;
	 --sp;
#ifdef WIN32
	 HWND console_handle = GetConsoleWindow();

		 //if (!console_handle) {
			// int driver;
			// int smode = mode;
			// if (mode <= 0 || mode == 7) {
			//	 /*if (chandle && chandle == console_handle)*/ {
			//		 closegraph();
			//		 return true;
			//	 }
			// }
			// detectgraph(&driver, &smode);
			// initgraph(&driver, &mode, 0,(void*)console_handle);
			//graphdefaults();
			////setbkcolor(15);
		 //}
		 
	if (console_handle) {
		HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE);

		if (hStdout) {
			cls(hStdout);
			CONSOLE_CURSOR_INFO info;
			if (GetConsoleCursorInfo(hStdout, &info))
			{
				info.bVisible = (mode == 0 || mode == 7) ? 1 : 0;
				SetConsoleCursorInfo(hStdout, &info);
			}
		}
	}
#endif
#endif
	return true;
}

// void setpixel(x, y, int color)
bool BppDefaultFunctions::Bpp_setpixel(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3, 3);
	ValPtr bp = sp; bp -= 3;
	if (!sp->isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
#ifdef __MSDOS__
	union REGS regs;
	regs.x.cx = (BppUShort)bp[1];
	regs.x.dx = (BppUShort)bp[2];
	regs.h.al = (BppByte)*sp;
	regs.h.ah = 0x0c;
	int86(0x10, &regs, &regs);
#else
#ifdef WIN32
	BppUInt color = *sp;
	HWND console_handle = GetConsoleWindow();
	if (console_handle) {
		HDC device_context = GetDC(console_handle);

		HPEN pen = CreatePen(PS_SOLID, 1, (COLORREF)color);
		SelectObject(device_context, pen);
		SetPixel(device_context, (BppUShort)bp[1], (BppUShort)bp[2], (COLORREF)color);

		ReleaseDC(console_handle, device_context);
	}
	//else
	//	putpixel((BppUShort)bp[1], (BppUShort)bp[2], color);

#endif

#endif
	sp = bp;
	sp->dispose();
	return true;
}

// void line(x1,y1,x2,y2,int color)
bool BppDefaultFunctions::Bpp_line(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(5, 5);
	if (!sp->isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
	ValPtr bp = sp; bp -= 5;
	BppShort x1 = bp[1];
	BppShort y1 = bp[2];
	BppShort x2 = bp[3];
	BppShort y2 = bp[4];
#ifdef  __MSDOS__
	union REGS regs, r2;
	int i;
	regs.h.ah = 0x0c;
	regs.h.al = (BppByte)*sp;

	register int dx = x2 - x1;
	register int dy = y2 - y1;
	register int dist, incx, incy, xerr, yerr;

	/* increment and sign for both directions */
	if (dx < 0)
	{
		incx = -1;
		dx = -dx;
	}
	else
		incx = (dx != 0) ? 1 : 0;

	if (dy < 0)
	{
		incy = -1;
		dy = -dy;
	}
	else
		incy = (dy != 0) ? 1 : 0;

	/* largest distance */
	dist = (dx>dy) ? dx : dy;

	/* init */
	regs.x.cx = x1;
	regs.x.dx = y1;
	xerr = dx;
	yerr = dy;

	/* draw pixels */
	for (i = 0; i<dist; i++)
	{
		int86(0x10, &regs, &r2);

		xerr += dx;
		yerr += dy;

		if (xerr>dist)
		{
			xerr -= dist;
			regs.x.cx += incx;
		}
		if (yerr>dist)
		{
			yerr -= dist;
			regs.x.dx += incy;
		}
	}

	regs.x.cx = x2;
	regs.x.dx = y2;
	int86(0x10, &regs, &r2);
#else
#ifdef WIN32
	BppUInt color = *sp;
	HWND console_handle = GetConsoleWindow();

	if (console_handle) {
		HDC device_context = GetDC(console_handle);

		HPEN pen = CreatePen(PS_SOLID, 1, (COLORREF)color);
		SelectObject(device_context, pen);
		MoveToEx(device_context, x1, y1, 0);
		LineTo(device_context, x2,y2);


		ReleaseDC(console_handle, device_context);
	}
	//else {
	//	setcolor(color);
	//	line(x1, y1, x2, y2);
	//}

#endif
#endif
	sp = bp;
	sp->dispose();
	return true;
}


	// miscellanous funcs
// rnd = rand()
bool BppDefaultFunctions::Bpp_rand(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
	*sp = (BppInt) rand();
	return true;
}

// srand(int seed)
bool BppDefaultFunctions::Bpp_srand(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	srand((unsigned) (BppInt) *sp);
	--sp;
	sp->dispose();
	return true;
}

bool BppDefaultFunctions::Bpp_ctime(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
	time_t tt = (BppTime_T) *sp;
	BppChar* p = _tctime(&tt);

	--sp;
	*sp = BppString(p);
	return true;
}

#define ltimefunc localtime
#define _timet time_t

// tmvec = localtime(time_t)
bool BppDefaultFunctions::Bpp_localtime(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
	_timet tt = (BppTime_T) *sp;
	tm* ptm = ltimefunc(&tt);
	BppVector* tmvec = new BppVector(9,true);
	int* p = (int*) ptm;
	for(int i=0;i<9;i++,p++)
		(*tmvec)[i] = (BppInt) (int) *p;
	--sp;
	*sp = tmvec;
	return true;
}

bool BppDefaultFunctions::Bpp_strftime(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 2);
	ValPtr bp = sp; bp -= 2;
	tm tmval;
	if (bp[2].isIntegralType()) {
		_timet tt = (BppTime_T)bp[2];
		tm* ptm = ltimefunc(&tt);
		tmval = *ptm;
	}
	else {
		BppVector* v = bp[2];
		if (!v || v->size() != 9) 
			THROW(FLIB_INVALID_ARG_TYPE, _T(" (vector[9] or int expected)"));
		int* p = (int*)&tmval;
		for (int i = 0; i < 9; i++, p++)
			*p = (int)(BppInt)(*v)[i];
	}
	BppString* fmt = bp[1];
	if (fmt == 0)
		THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	size_t bsize = 200 + fmt->length();
	BppCharBuffer cbuf(bsize, bsize);
	unsigned n = _tcsftime(cbuf.data(), bsize, fmt->c_str(), &tmval);
	if (n == 0) cbuf.clear();
	cbuf.adjustSize();
	bp[2].dispose();
	sp -= 2;
	*sp = new BppString(cbuf);
	return true;
}
// argvec = getargs()
bool BppDefaultFunctions::Bpp_getargs(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(0,0);
	BppVector* pvec = new BppVector();
	for (int i=0; i<vm->cmdargc(); i++)
		pvec->push_back(new BppString((const BppChar*) vm->cmdargv()[i]));
	*sp = pvec;
	return true;
}

// argvec = getusrargs()
bool BppDefaultFunctions::Bpp_getusrargs(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
	BppVector* pvec = new BppVector();
	bool found = false;
	for (int i=0; i<vm->cmdargc(); i++)
	{
		BppChar* p = vm->cmdargv()[i];
		if ((!found) && (p[0] == _T('#')))
		{
			found = true;
			p++;
			if (*p == 0) continue;
		}
		if (found)
			pvec->push_back(new BppString((const BppChar*) p));
	}
	*sp = pvec;
	return true;
}

bool BppDefaultFunctions::Bpp_errno(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(1);
#ifdef WINCE
	if (argc) { SetLastError((BppUInt)*sp); --sp; }
	*sp = (BppInt) GetLastError();

#else
	if (argc) { errno = (BppInt)*sp; --sp; }
	*sp = (BppInt) errno;
#endif
	return true;
}

bool BppDefaultFunctions::Bpp_strerror(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
#ifdef WINCE
	return Bpp_GetSysErrorMsg(vm,argc,sp);
#else
	CHKVM;
	ARGCMAX(1);
	int nr = errno;
	if (argc) { nr = (BppInt) *sp; --sp; }
#ifndef GCC
	if (nr >= 0 && nr < sys_nerr) 
#endif
		*sp = new BppString(_tcserror(nr));
#ifndef GCC
	else *sp = new BppString(FLIB_UNKNOWN_ERROR);
#endif
	return true;
#endif
}

bool BppDefaultFunctions::Bpp_nerr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	ARGCMAX(0);
	*sp = (BppInt)
#if defined(WINCE) || defined(GCC)
		0;
#else
		sys_nerr;
#endif
	return true;
}

// cnt = argcnt()
bool BppDefaultFunctions::Bpp_argcnt(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
	BppProcInfo* pinfo = vm->interpreter()->getProcInfo();
	BppInt cnt = -1;
	if (pinfo != 0)
	{
		cnt = pinfo->getNargs();
	}
	*sp = cnt;
	return true;
}

bool BppDefaultFunctions::Bpp_arg(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
	BppProcInfo* pinfo = interpreter->getProcInfo();
	BppInt cnt = -1;
	BppInt idx = *sp;
	if (pinfo != 0)
	{
		cnt = pinfo->getNargs();
	}
	if ((idx < 0) || (idx >= cnt))
		THROW(FLIB_INVALID_INDEX);
	--sp;
	*sp = pinfo->bp()[(int) (idx+1)];
	return true;
}

bool BppDefaultFunctions::Bpp_argv(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	ARGCMAX(0);
	BppProcInfo* pinfo = interpreter->getProcInfo();
	BppInt cnt = -1;
	if (pinfo != 0)
	{
		cnt = pinfo->getNargs();
	}
	BppVector* pvec = (cnt > 0) ? new BppVector((unsigned)cnt,true) : new BppVector();
	*sp = pvec;
	for (int i = 0; i < cnt; ++i)
		(*pvec)[i] = pinfo->bp()[i + 1];
	return true;
}

bool BppDefaultFunctions::Bpp_call(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= 2;
	const BppValue& func = bp[1];
	BppVector* args = 0;
	if (argc == 2)
	{
		args = bp[2];
		if (!args) THROW(FLIB_INVALID_SECOND_ARG, FLIB_VECTOR_EXPECTED_BR);
	}
	BppObject* pobj = func;
	bool ok = false;
	if (pobj) ok = interpreter->invokeMember(*bp, pobj, _T("OP_CALL"), args);
	else ok = interpreter->invokeFunc(*bp, func, args);
	sp = bp;
	return ok;
}


	// system functions
// ptr = addr(val)
bool BppDefaultFunctions::Bpp_addr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValueType vt = sp->getType();
	if (vt == BVT_NULL)
		THROW(_T("argument is NULL"));
	if (vt >= BVT_INTEGRAL_MIN && vt <= BVT_REAL_MAX)
		return Bpp_intptr(vm, argc, sp);
	else 
		sp[-1] = 
#ifdef GCC
	 (BppLong)
#else
	 (BppIntPtr) 
#endif

	sp->addr();
	--sp;
	return true;
}

bool BppDefaultFunctions::Bpp_memmove(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	static const BppChar* fname = _T("memmove");
	CHKVM;
	CHKARGC(3, 3);
	ValPtr bp = sp; bp -= 3;
	BppValue v;
	if (!interpreter->doNumCast(v, bp[1], BVT_INTPTR, fname)) return false;
	BppIntPtr dst = INTPTRCAST(v.ptr());
	if (!interpreter->doNumCast(v, bp[2], BVT_INTPTR, fname)) return false;
	BppIntPtr src = INTPTRCAST(v.ptr());
	if (!interpreter->doNumCast(v, bp[3], BVT_UINT, fname)) return false;
	BppUInt n = v;
#ifndef NOCPPEXCEPTION

	try
	{
#endif
	dst = movemem(dst, src, n);
#ifndef NOCPPEXCEPTION
	}
	catch (...)
	{
		return interpreter->returnError(_T("fatal error"),fname);
	}
#endif

	sp = bp;
	*sp = dst;
	return true;
}

bool BppDefaultFunctions::Bpp_getenv( BppVirtualMachine* vm, int argc, ValPtr& sp )
{
	CHKVM;
	ARGCMAX(1);
	if (argc > 0)
	{
		BppString* key = *sp;
		if (key == 0) THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
		const BppChar* presult = _tgetenv(key->c_str());
		--sp;
		(*sp) = new BppString(presult);
	}
	else
	{ 
		BppDictionary* dict = new BppDictionary(true);
		 (*sp) = dict;
#ifndef WINCE
#ifndef WIN32
		 char**
#else
		BppChar** 
#endif
			penv = environ;
		for(;(*penv);++penv)
		{
			BppString s = *penv;
			size_t pos = s.find(_T('='));
			if (pos>0 && pos != BPP_NNPOS)
			{
				(*dict)[s.substr(0,pos)] = new BppString(s.substr(pos+1));
			}
		}
#endif
	}
	return true;
}

bool BppDefaultFunctions::Bpp_setenv( BppVirtualMachine* vm, int argc, ValPtr& sp )
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= 2;
	BppString* key = bp[1];
	BppString* val = bp[2];
	if ((key == 0) || (val == 0)) THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	size_t l1 = key->length();
	bool succ = false;
	if (l1 > 0)
	{
		size_t l2 = val->length();
		BppChar* buf = (BppChar*) allocmem((l1+l2+2) * sizeof(BppChar));
		_tcscpy(buf,key->c_str());
		_tcscat(buf,_T("="));
		_tcscat(buf,val->c_str());
		if (_tputenv(buf) == 0) succ = true;
		freemem(buf);
	}
	sp = bp;
	(*sp) = BppValue::makeBool(succ);
	return true;
}

bool BppDefaultFunctions::Bpp_setlocale(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 2);
	ValPtr bp = sp; bp -= argc;
	BppValue v;
	if (!interpreter->doNumCast(v,bp[1],BVT_INT,_T("setlocale"))) return false;
	BppInt category = v;
	const BppChar* strloc = 0;
	if (sp->getType() != BVT_NULL) {
		v.dispose();	
		if (!interpreter->doStrCast(v,bp[2])) THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
		BppString* s = v;
		if (s) strloc = s->c_str();
	}
	strloc = _tsetlocale(category, strloc);
	sp = bp;
	*sp = new BppString(strloc);
	return true;
}


#ifndef WINCE
bool BppDefaultFunctions::Bpp_system(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= argc;
	BppString* s = bp[1];
	if (s == 0) THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	int r = _tsystem(s->c_str());
	bool ignoreErr = argc > 1 ? bp[2].asBool() : true;
	if (r == -1 && !ignoreErr) THROW(FLIB_SYSTEM_CALL_FAILED, s->c_str());
	sp = bp;
	*bp = (BppInt)r;
	return true;
}
#endif
#ifndef NO_DYNLIB
#ifdef __MSDOS__
	typedef HDLL HMODULE;
	typedef int BOOL;
	typedef long (huge *FARPROC)();
#endif // __MSDOS__
// int hinst = LoadLibrary(dllname)
bool BppDefaultFunctions::Bpp_LoadLibrary(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppString* s = *sp;
	if (s == 0)
	  THROW(FLIB_INVALID_ARG_TYPE);
	--sp;
	HMODULE h = 0;
#ifndef NOCPPEXCEPTION

	try
	{
#endif
		BppString _dllname = *s;
		_dllname.replace(_T("*"),BPP_DLLWILD);
		h = LoadLibrary(s->c_str());
#ifndef NOCPPEXCEPTION
	}
	catch (...)
	{
		h = 0;
	}
#endif
	*sp = INTPTRCAST(h);
	return true;
}

// ok = FreeLibrary(intptr hmodule)
bool BppDefaultFunctions::Bpp_FreeLibrary(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (!sp->isIntegralType())
	  THROW(FLIB_INVALID_ARG_TYPE);
	BOOL result = 0;
#ifndef NOCPPEXCEPTION
	try
	{
#endif
		result = FreeLibrary((HMODULE) (BppIntPtr) *sp);
#ifdef __MSDOS__
		result = (result< 0) ? 0 : 1;

#endif
#ifndef NOCPPEXCEPTION
	}
	catch (...)
	{
		result = 0;
	}
#endif
	--sp;
	*sp = (BppInt) result;
	return true;
}

#ifndef __MSDOS__
#ifdef _UNICODE
#define LIBREPLACE 'W'
#else
#define LIBREPLACE 'A'
#endif
#endif
// int procaddr = GetProcAddress(intptr hmodule, string name)
bool BppDefaultFunctions::Bpp_GetProcAddress(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp;
	bp -= 2;
	if (!bp[1].isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
	BppString* s = bp[2];
	if (s == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
#ifndef __MSDOS__
	BppString namestr(*s);
	namestr.replace(_T('?'),_T(LIBREPLACE));
	s = &namestr;
#endif
#ifndef NOCPPEXCEPTION
	try
	{
#endif
		*bp = INTPTRCAST(GetProcAddress((HMODULE) (BppIntPtr) bp[1], s->p_str()));
#ifndef NOCPPEXCEPTION
	}
	catch (...)
	{
		*bp = (BppIntPtr) 0;
	}
#endif
	sp = bp;
	return true;
}
#endif // NO_DYNLIB

#ifdef WIN32
#ifndef NO_DYNLIB
	#define _ole32dll _T("ole32.dll")
 #ifndef WINCE
	#define _user32dll _T("user32.dll")
	#define _shell32dll _T("shell32.dll")
	#define _kernel32dll _T("kernel32.dll")
	#define _gdi32dll _T("gdi32.dll")
 #else
	#define _user32dll _T("coredll.dll")
	#define _shell32dll _T("coredll.dll")
	#define _kernel32dll _T("coredll.dll")
	#define _gdi32dll _T("coredll.dll")
	#define _user32 _coredll
	#define _kernel32 _coredll
	#define _gdi32 _coredll
	#define _shell32 _coredll
 #endif

static struct LibHandles
{
#ifndef WINCE
	HMODULE _user32;
	HMODULE _kernel32;
	HMODULE _gdi32;
	HMODULE _shell32;
#else
	HMODULE _coredll;
#endif
	HMODULE _ole32;
	LibHandles():_ole32(0),
#ifndef WINCE
	_user32(0),_kernel32(0),_gdi32(0),_shell32(0)
#else
	_coredll(0)
#endif
	{
	}

	~LibHandles()
	{
		if (_ole32 != 0) FreeLibrary(_ole32);
#ifndef WINCE
		if (_user32 != 0) FreeLibrary(_user32);
		if (_shell32 != 0) FreeLibrary(_shell32);
		if (_gdi32 != 0) FreeLibrary(_gdi32);
		if (_kernel32 != 0) FreeLibrary(_kernel32);
#else
		if (_coredll != 0) FreeLibrary(_coredll);
#endif
	}

}
_libHandles;

#ifndef DOXY
static bool GetDefaultLibFunc(BppVirtualMachine* vm, int argc, ValPtr& sp, HMODULE& libHandle, const BppChar* libName)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp;
	bp -= argc;
	BppProcInfo* pi = interpreter->getProcInfo();
	BppClass* ownClass = pi ? pi->getClass() : 0;
	BppString* s = interpreter->getStringFromValue(ownClass,bp[1]);
	if (s == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	BppString namestr(*s);
	namestr.replace(_T('?'), _T(LIBREPLACE));
	s = &namestr;
	if (libHandle == 0)
	{
		libHandle = LoadLibrary(libName);
	}
#ifndef NOCPPEXCEPTION
	try
	{
#endif
		*bp = (BppIntPtr)GetProcAddress(libHandle, s->p_str());
#ifndef NOCPPEXCEPTION
	}
	catch (...)
	{
		*bp = (BppIntPtr)0;
	}
#endif
	sp = bp;
	return true;

}
#endif

bool BppDefaultFunctions::Bpp_GetUser32Func(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return GetDefaultLibFunc(vm, argc, sp, _libHandles._user32, _user32dll);
}

bool BppDefaultFunctions::Bpp_GetKernel32Func(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return GetDefaultLibFunc(vm, argc, sp, _libHandles._kernel32, _kernel32dll);
}

bool BppDefaultFunctions::Bpp_GetGdi32Func(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return GetDefaultLibFunc(vm, argc, sp, _libHandles._gdi32, _gdi32dll);
}

bool BppDefaultFunctions::Bpp_GetShell32Func(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return GetDefaultLibFunc(vm, argc, sp, _libHandles._shell32, _shell32dll);
}

bool BppDefaultFunctions::Bpp_GetOle32Func(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return GetDefaultLibFunc(vm, argc, sp, _libHandles._ole32, _ole32dll);
}
#endif // !NO_DYNLIB

bool BppDefaultFunctions::Bpp_GetLastError(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
	*sp = (BppUInt) GetLastError();
	return true;
}

bool BppDefaultFunctions::Bpp_GetSysErrorMsg(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(1);
	ValPtr bp = sp;
	bp -= argc;
	DWORD errcode = (argc > 0) ? ((DWORD) (BppUInt) bp[1]) : GetLastError();

	LPVOID lpMsgBuf = 0;
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		errcode,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL
		);
	if (lpMsgBuf != 0)
	{
	  *bp = new BppString((const BppChar*) lpMsgBuf);
	  LocalFree(lpMsgBuf);
	}
	else
		*bp = new BppString();
	sp = bp;
	return true;
}

#endif // WIN32
#ifndef NO_DYNLIB
#ifdef NOASM
static BppIntPtr CallLibFuncCE(FARPROC proc,int argc, BppIntPtr* argv)
{
	switch (argc)
	{
	case 0:
		return ((DLLFUNC0)proc)();
	case 1:
		return ((DLLFUNC1)proc)(argv[0]);
	case 2:
		return ((DLLFUNC2)proc)(argv[0],argv[1]);
	case 3:
		return ((DLLFUNC3)proc)(argv[0],argv[1],argv[2]);
	case 4:
		return ((DLLFUNC4)proc)(argv[0],argv[1],argv[2],argv[3]);
	case 5:
		return ((DLLFUNC5)proc)(argv[0],argv[1],argv[2],argv[3],argv[4]);
	case 6:
		return ((DLLFUNC6)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5]);
	case 7:
		return ((DLLFUNC7)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6]);
	case 8:
		return ((DLLFUNC8)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7]);
	case 9:
		return ((DLLFUNC9)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8]);
	case 10:
		return ((DLLFUNC10)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9]);
	case 11:
		return ((DLLFUNC11)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9],argv[10]);
	case 12:
		return ((DLLFUNC12)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9],argv[10],argv[11]);
	case 13:
		return ((DLLFUNC13)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9],argv[10],argv[11],argv[12]);
	case 14:
		return ((DLLFUNC14)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9],argv[10],argv[11],argv[12],argv[13]);
	case 15:
		return ((DLLFUNC15)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9],argv[10],argv[11],argv[12],argv[13],argv[14]);
	case 16:
		return ((DLLFUNC16)proc)(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],
			argv[8],argv[9],argv[10],argv[11],argv[12],argv[13],argv[14],argv[15]);
	}
	return 0;
}

#endif

#ifdef __WATCOMC__
#define asm _asm
#endif
// result = CallLibFunc(procadr,...)
bool BppDefaultFunctions::Bpp_CallLibFunc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMIN(1);
#ifdef NOASM
	ARGCMAX(17);
#endif
	ValPtr vp = sp;
	vp -= argc;
	if (!vp[1].isIntegralType())
		THROW(FLIB_INVALID_ARG_TYPE);
	BppIntPtr* pargs =(BppIntPtr*) allocmem(sizeof(BppIntPtr) *argc);
#ifndef NOASM
	for (int i=0; i<argc;i++)
		pargs[argc-(i+1)] = (BppIntPtr) vp[i+1].ptr();
	FARPROC proc = (FARPROC) pargs[argc-1];
#else
	for (int i=0; i<argc;i++)
		pargs[i] = INTPTRCAST(vp[i + 1].ptr());
	FARPROC proc = (FARPROC) pargs[0];
#endif
	int procargc = argc-1;
#ifndef __MSDOS__
	BppIntPtr result = 0;
#else
	struct RESSTRUCT
	{
		int lw;
		int hw;
	} result = { 0, 0 };
	int rl, rh;
#endif
#ifndef __MSDOS__
 #ifndef NOASM
		  BppInt savedsp;
		  try
		  {
			  __asm
			  {
				  mov eax, esp
				  mov savedsp, eax
				  mov ecx, procargc
				  mov edx, pargs
				  cmp ecx, 0
				  je docall
nxtarg:
				  mov eax, [edx]
				  push eax
				  add edx, 4
				  dec ecx
				  jnz nxtarg
docall:
				  call proc
				  mov result, eax
				  mov eax, savedsp
				  mov esp, eax
			  }
		  }
		  catch (...)
		  {
		  }
  #else
	  BppIntPtr* p =  pargs;
	  p++;
	  result =  (BppIntPtr) CallLibFuncCE(proc,argc-1,p);
  #endif
#else
	int savedsp;
	asm{
							push es
							push di
							mov ax, sp
							mov savedsp, ax
			mov cx, procargc
								les di,pargs
			cmp cx, 0
			je docall
#ifdef __TURBOC__
		}
nxtarg:
	  asm {
#else
nxtarg:
#endif
								mov ax, es:[di]
								add di, 2
								mov bx, es:[di]
								add di, 2
								push bx
			push ax
			dec cx
			jnz nxtarg
#ifdef __TURBOC__
	 }
docall:
	 asm {
#else
docall:
#endif
#ifdef __WATCOMC__
								call dword ptr proc
#else
								call dword ptr proc
#endif
								mov rl, ax
								mov rh, dx
								mov ax, savedsp
								mov sp, ax
								pop di
								pop es
					 }
					 result.lw = rl;
		result.hw = rh;
#endif // __MSDOS__
		freemem(pargs);
  *vp = *((BppIntPtr*) &result);
  sp = vp;
  return true;
}

#endif // !NO_DYNLIB

#ifndef DOXY


static const BppString _OP_CALL = _T("OP_CALL");



// qsort helpers
// variables are initialized by Bpp_qsort function
struct QSort {
	// comparer callback function value - may be object, bytecode or native function
	BppValue cmp_func;
	BppObject* pobj;
	BppValue cmp_obj;
	int argcnt;
	// temporary pointer to stack pointer
	ValPtr* cmpSP;
	// temporary VM pointer
	BppVirtualMachine* pvm;
	// flag for success, set by doCpmpare function
	bool cmpOk;
	BppVector args;
	BppValueType ftype;
	QSort(BppVirtualMachine* pVm, ValPtr* sp)
		: pobj(0), argcnt(2), cmpSP(sp), pvm(pVm), cmpOk(true), args(2, true), ftype(BVT_NULL) {}
	void setFunc(const BppValue& f) {
		pobj = f;
		if (pobj) {
			cmp_obj = f;
			args.insert(0, cmp_obj);
			++argcnt;
			BppDictionaryEntry* pe = pobj->getClass()->findMember(0, _OP_CALL);
			if (pe) cmp_func = pe->getValue();
		}
		else cmp_func = f;
		ftype = cmp_func.getType();
	}
	// compares two values using cmp_func
	int doCompare(const BppValue& l, const BppValue& r)
	{
		if (ftype == BVT_NULL) return l.compare(r);
		BppValue result;
		cmpOk = false;
		args[argcnt - 2] = l;
		args[argcnt - 1] = r;
		cmpOk = pvm->interpreter()->invokeFunc(result, cmp_func, &args);
		return (BppInt)result;
	}
	int cmp(BppUInt l, BppUInt r)
	{
		if (ftype == BVT_NULL)
			return l < r ? -1 : r < l ? 1 : 0;
		return doCompare(l, r);
	}

	void qusort(BppBuffer& arr, size_t i1, size_t i2)
	{
		if (i1 == i2) return;
		size_t l = i1;
		size_t r = i2;
		BppByte p = arr[(i1 + i2) >> 1];
		do {
			while (cmpOk && cmp(arr[r], p) > 0) --r;
			while (cmpOk && cmp(arr[l], p) < 0) ++l;
			if ((!cmpOk) || (l > r)) break;
			arr.swap(r, l);
		} while (++l <= --r);
		if (cmpOk && (i1 < r)) qusort(arr, i1, r);
		if (cmpOk && (l < i2)) qusort(arr, l, i2);
	}
	void qusort(BppCharBuffer& arr, size_t i1, size_t i2)
	{
		if (i1 == i2) return;
		size_t l = i1;
		size_t r = i2;
		BppChar p = arr[(i1 + i2) >> 1];
		do {
			while (cmpOk && cmp(arr[r], p) > 0) --r;
			while (cmpOk && cmp(arr[l], p) < 0) ++l;
			if ((!cmpOk) || (l > r)) break;
			arr.swap(r, l);
		} while (++l <= --r);
		if (cmpOk && (i1 < r)) qusort(arr, i1, r);
		if (cmpOk && (l < i2)) qusort(arr, l, i2);
	}

	void qusort(BppVector& arr, size_t i1, size_t i2)
	{
		if (i1 == i2) return;
		if (i1 + 1 == i2) {
			if (doCompare(arr[i1], arr[i2]) > 0) {
				if (cmpOk) arr.swap(i1, i2);
			}
			return;
		}
		size_t l = i1;
		size_t r = i2;
		BppValue  p = arr[(i1 + i2) >> 1];
		do {
			while (cmpOk && doCompare(arr[r], p) > 0) --r;
			while (cmpOk && doCompare(arr[l], p) < 0) ++l;
			if ((!cmpOk) || (l > r)) break;
			arr.swap(r, l);
		} while (++l <= --r);
		if (r - i1 < i2 - l)
		{
			if (cmpOk && (i1 < r)) qusort(arr, i1, r);
			if (cmpOk && (l < i2)) qusort(arr, l, i2);
		}
		else
		{
			if (cmpOk && (l < i2)) qusort(arr, l, i2);
			if (cmpOk && (i1 < r)) qusort(arr, i1, r);
		}
	}

	size_t bfind(const BppBuffer& arr, size_t l, size_t r, BppByte val) {
		if (r == l)
			return (cmp(arr[r], val) < 0) ? r + 1 : r;
		size_t i = (l + r) >> 1;
		int res = cmp(val, arr[i]);
		if (res == 0) return i;
		if (res < 0) return bfind(arr, l, i, val);
		return bfind(arr, ++i, r, val);
	}
	size_t lbound(const BppBuffer& arr, size_t l, size_t r, BppByte val) {
		size_t res = bfind(arr, l, r, val);
		while (res > l && cmp(arr[res - 1], arr[res]) == 0) --res;
		return res;
	}
	size_t ubound(const BppBuffer& arr, size_t l, size_t r, BppByte val) {
		size_t res = bfind(arr, l, r, val);
		while (res <= r) {
			++res;
			if (res > r || cmp(arr[res - 1], arr[res]) < 0) break;
		}
		return res;
	}
	size_t bfind(const BppCharBuffer& arr, size_t l, size_t r, BppChar val) {
		if (r == l)
			return (cmp(arr[r], val) < 0) ? r + 1 : r;
		size_t i = (l + r) >> 1;
		int res = cmp(val, arr[i]);
		if (res == 0) return i;
		if (res < 0) return bfind(arr, l, i, val);
		return bfind(arr, ++i, r, val);
	}
	size_t lbound(const BppCharBuffer& arr, size_t l, size_t r, BppChar val) {
		size_t res = bfind(arr, l, r, val);
		while (res > l && cmp(arr[res - 1], arr[res]) == 0) --res;
		return res;
	}
	size_t ubound(const BppCharBuffer& arr, size_t l, size_t r, BppChar val) {
		size_t res = bfind(arr, l, r, val);
		while (res <= r) {
			++res;
			if (res > r || cmp(arr[res - 1], arr[res]) < 0) break;
		}
		return res;
	}
	size_t bfind(const BppVector& arr, size_t l, size_t r, const BppValue& val) {
		if (r == l)
			return (doCompare(arr[r], val) < 0) ? r + 1 : r;
		size_t i = (l + r) >> 1;
		int res = doCompare(val, arr[i]);
		if (res == 0) return i;
		if (res < 0) return bfind(arr, l, i, val);
		return bfind(arr, ++i, r, val);
	}
	size_t lbound(const BppVector& arr, size_t l, size_t r, const BppValue& val) {
		size_t res = bfind(arr, l, r, val);
		while (res > l && doCompare(arr[res - 1], arr[res]) == 0) --res;
		return res;
	}
	size_t ubound(const BppVector& arr, size_t l, size_t r, const BppValue& val) {
		size_t res = bfind(arr, l, r, val);
		while (res <= r) {
			++res;
			if (res > r || doCompare(arr[res - 1], arr[res]) < 0) break;
		}
		return res;
	}
};



#endif 
// null qsort(var container, var comparer)
bool BppDefaultFunctions::Bpp_qsort(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= argc;
	QSort _sorter(vm, &sp);
	CHKVAR(bp[1]);
	if (argc == 2) _sorter.setFunc(bp[2]);
	BppValueType vt = maskWeakRef(bp[1].getType());
	switch (vt)
	{
	case BVT_VECTOR:
	{
		BppVector* vec = bp[1];
		if (!vec) THROW(FLIB_INVALID_CONTAINER);
		if (vec->size() > 0) _sorter.qusort(*vec, 0, vec->size() - 1);
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		if (buf->size() > 0) _sorter.qusort(*buf, 0, buf->size() - 1);
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		if (buf->size() > 0) _sorter.qusort(*buf, 0, buf->size() - 1);
	}
	break;
	default:
		THROW(FLIB_INVALID_CONTAINER_TYPE);
	}
	sp = bp;
	sp->dispose();
	return _sorter.cmpOk;
}

// uint lbound(var container,uint startpos, uint endpos, var val, var comparer = null)
// uint lbound(var container,var val, var comparer = null)
bool BppDefaultFunctions::Bpp_lbound(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 5);
	ValPtr bp = sp; bp -= argc;
	QSort _sorter(vm, &sp);
	if (argc == 3) _sorter.setFunc(bp[3]);
	else if (argc == 5) _sorter.setFunc(bp[5]);
	BppValueType vt = maskWeakRef(bp[1].getType());
	switch (vt)
	{
	case BVT_VECTOR:
	{
		BppVector* vec = bp[1];
		if (!vec) THROW(FLIB_INVALID_CONTAINER);
		if (vec->size() == 0) {
			*bp = BppUInt(0);
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(vec->size() - 1);
		*bp = (BppUInt)_sorter.lbound(*vec, l, r, bp[(argc > 3) ? 4 : 2]);
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		if (buf->size() == 0) {
			*bp = BppUInt(0);
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(buf->size() - 1);
		*bp = (BppUInt)_sorter.lbound(*buf, l, r, bp[(argc > 3) ? 4 : 2]);
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		if (buf->size() == 0) {
			*bp = BppUInt(0);
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(buf->size() - 1);
		*bp = (BppUInt)_sorter.lbound(*buf, l, r, BppCharVal(bp[(argc > 3) ? 4 : 2]));
	}
	break;
	default:
		THROW(FLIB_INVALID_CONTAINER);
	}
	sp = bp;
	return _sorter.cmpOk;
}

// uint ubound(var container,uint startpos, uint endpos, var val)
// uint ubound(var container,var val)
bool BppDefaultFunctions::Bpp_ubound(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 5);
	ValPtr bp = sp; bp -= argc;
	QSort _sorter(vm, &sp);
	if (argc == 3) _sorter.setFunc(bp[3]);
	else if (argc == 5) _sorter.setFunc(bp[5]);
	BppValueType vt = maskWeakRef(bp[1].getType());
	switch (vt)
	{
	case BVT_VECTOR:
	{
		BppVector* vec = bp[1];
		if (!vec) THROW(FLIB_INVALID_CONTAINER);
		if (vec->size() == 0) {
			*bp = BppUInt(0);
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(vec->size() - 1);
		*bp = (BppUInt)_sorter.ubound(*vec, l, r, bp[(argc > 3) ? 4 : 2]);
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		if (buf->size() == 0) {
			*bp = BppUInt(0);
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(buf->size() - 1);
		*bp = (BppUInt)_sorter.ubound(*buf, l, r, bp[(argc > 3) ? 4 : 2]);
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		if (buf->size() == 0) {
			*bp = BppUInt(0);
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(buf->size() - 1);
		*bp = (BppUInt)_sorter.ubound(*buf, l, r, BppCharVal(bp[(argc > 3) ? 4 : 2]));
	}
	break;
	default:
		THROW(FLIB_INVALID_CONTAINER_TYPE);
	}
	sp = bp;
	return _sorter.cmpOk;
}

// uint bsearch(var container,uint startpos, uint endpos, var val, var comparer = null)
// uint bsearch(var container,var val, var comparer = null)
bool BppDefaultFunctions::Bpp_bsearch(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 5);
	ValPtr bp = sp; bp -= argc;
	QSort _sorter(vm, &sp);
	if (argc == 3) _sorter.setFunc(bp[3]);
	else if (argc == 5) _sorter.setFunc(bp[5]);
	BppValueType vt = maskWeakRef(bp[1].getType());
	switch (vt)
	{
	case BVT_VECTOR:
	{
		BppVector* vec = bp[1];
		if (!vec) THROW(FLIB_INVALID_CONTAINER);
		size_t n = vec->size();
		if (n == 0) {
			*bp = BPP_NPOS;
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(vec->size() - 1);
		BppValue& val = bp[(argc > 3) ? 4 : 2];
		BppUInt res = (BppUInt)_sorter.lbound(*vec, l, r, val);
		if (res >= n) *bp = BPP_NPOS;
		else {
			*bp = (_sorter.doCompare((*vec)[res], val) == 0) ? res : BPP_NPOS;
		}
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		size_t n = buf->size();
		if (n == 0) {
			*bp = BPP_NPOS;
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(n - 1);
		BppByte val = bp[(argc > 3) ? 4 : 2];
		BppUInt res = (BppUInt)_sorter.lbound(*buf, l, r, val);
		if (res >= n) *bp = BPP_NPOS;
		else {
			*bp = (_sorter.cmp((*buf)[res], val) == 0) ? res : BPP_NPOS;
		}
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* buf = bp[1];
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		size_t n = buf->size();
		if (n == 0) {
			*bp = BPP_NPOS;
			break;
		}
		BppUInt l = (argc > 3) ? BppUInt(bp[2]) : 0U;
		BppUInt r = (argc > 3) ? BppUInt(bp[3]) : BppUInt(n - 1);
		BppUInt val = bp[(argc > 3) ? 4 : 2];
		BppUInt res = (BppUInt)_sorter.lbound(*buf, l, r, val);
		if (res >= n) *bp = BPP_NPOS;
		else {
			*bp = (_sorter.cmp((*buf)[res], val) == 0) ? res : BPP_NPOS;
		}
	}
	break;
	default:
		THROW(FLIB_INVALID_CONTAINER_TYPE);
	}
	sp = bp;
	return _sorter.cmpOk;
}

#ifdef USECALLBACK

static BppIntPtr _callCallbackObject(BppIntPtr pCallback, BppVector* args = 0) {
	BppObject* pobj = (BppObject*)(void*)pCallback;
	if (!pobj) return 0;
	BppClass* pclass = pobj->getClass();
	if (!pclass) return 0;
	BppVirtualMachine& vm = pclass->getVM();
	BppValue result;
	if (!vm.invokeMember(result, pobj, _T("OP_CALL"), args))
		return 0;
	return result;
}

#define MAXSCBARGS 5
#define MAXSCBINSTANCES 5
#define REGCBSIZE (MAXSCBARGS * MAXSCBINSTANCES)

static BppVector _registeredCallbacks(REGCBSIZE,true);

static BppIntPtr BPPAPI CBFUNC0(BppIntPtr pCallback) {
	return _callCallbackObject(pCallback);
}

static BppIntPtr BPPAPI SCBFUNC0_0() {
	return CBFUNC0(_registeredCallbacks[0]);
}
static BppIntPtr BPPAPI SCBFUNC0_1() {
	return CBFUNC0(_registeredCallbacks[1]);
}
static BppIntPtr BPPAPI SCBFUNC0_2() {
	return CBFUNC0(_registeredCallbacks[2]);
}
static BppIntPtr BPPAPI SCBFUNC0_3() {
	return CBFUNC0(_registeredCallbacks[3]);
}
static BppIntPtr BPPAPI SCBFUNC0_4() {
	return CBFUNC0(_registeredCallbacks[4]);
}

static BppIntPtr BPPAPI CBFUNC1(BppIntPtr arg0, BppIntPtr pCallback) {
	BppVector args(1, true);
	args[0] = arg0;
	return _callCallbackObject(pCallback, &args);
}
static BppIntPtr BPPAPI SCBFUNC1_0(BppIntPtr arg0) {
	return CBFUNC1(arg0,_registeredCallbacks[5]);
}
static BppIntPtr BPPAPI SCBFUNC1_1(BppIntPtr arg0) {
	return CBFUNC1(arg0, _registeredCallbacks[6]);
}
static BppIntPtr BPPAPI SCBFUNC1_2(BppIntPtr arg0) {
	return CBFUNC1(arg0, _registeredCallbacks[7]);
}
static BppIntPtr BPPAPI SCBFUNC1_3(BppIntPtr arg0) {
	return CBFUNC1(arg0, _registeredCallbacks[8]);
}
static BppIntPtr BPPAPI SCBFUNC1_4(BppIntPtr arg0) {
	return CBFUNC1(arg0, _registeredCallbacks[9]);
}

static BppIntPtr BPPAPI CBFUNC2(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr pCallback) {
	BppVector args(2, true);
	args[0] = arg0;
	args[1] = arg1;
	return _callCallbackObject(pCallback, &args);
}
static BppIntPtr BPPAPI SCBFUNC2_0(BppIntPtr arg0, BppIntPtr arg1) {
	return CBFUNC2(arg0, arg1, _registeredCallbacks[10]);
}
static BppIntPtr BPPAPI SCBFUNC2_1(BppIntPtr arg0, BppIntPtr arg1) {
	return CBFUNC2(arg0, arg1, _registeredCallbacks[11]);
}
static BppIntPtr BPPAPI SCBFUNC2_2(BppIntPtr arg0, BppIntPtr arg1) {
	return CBFUNC2(arg0, arg1, _registeredCallbacks[12]);
}
static BppIntPtr BPPAPI SCBFUNC2_3(BppIntPtr arg0, BppIntPtr arg1) {
	return CBFUNC2(arg0, arg1, _registeredCallbacks[13]);
}
static BppIntPtr BPPAPI SCBFUNC2_4(BppIntPtr arg0, BppIntPtr arg1) {
	return CBFUNC2(arg0, arg1, _registeredCallbacks[14]);
}

static BppIntPtr BPPAPI CBFUNC3(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr pCallback) {
	BppVector args(3, true);
	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	return _callCallbackObject(pCallback, &args);
}
static BppIntPtr BPPAPI SCBFUNC3_0(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2) {
	return CBFUNC3(arg0, arg1, arg2, _registeredCallbacks[15]);
}
static BppIntPtr BPPAPI SCBFUNC3_1(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2) {
	return CBFUNC3(arg0, arg1, arg2, _registeredCallbacks[16]);
}
static BppIntPtr BPPAPI SCBFUNC3_2(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2) {
	return CBFUNC3(arg0, arg1, arg2, _registeredCallbacks[17]);
}
static BppIntPtr BPPAPI SCBFUNC3_3(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2) {
	return CBFUNC3(arg0, arg1, arg2, _registeredCallbacks[18]);
}
static BppIntPtr BPPAPI SCBFUNC3_4(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2) {
	return CBFUNC3(arg0, arg1, arg2, _registeredCallbacks[19]);
}

static BppIntPtr BPPAPI CBFUNC4(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3, BppIntPtr pCallback) {
	BppVector args(4, true);
	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	args[3] = arg3;
	return _callCallbackObject(pCallback, &args);
}
static BppIntPtr BPPAPI SCBFUNC4_0(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3) {
	return CBFUNC4(arg0, arg1, arg2, arg3, _registeredCallbacks[20]);
}
static BppIntPtr BPPAPI SCBFUNC4_1(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3) {
	return CBFUNC4(arg0, arg1, arg2, arg3, _registeredCallbacks[21]);
}
static BppIntPtr BPPAPI SCBFUNC4_2(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3) {
	return CBFUNC4(arg0, arg1, arg2, arg3, _registeredCallbacks[22]);
}
static BppIntPtr BPPAPI SCBFUNC4_3(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3) {
	return CBFUNC4(arg0, arg1, arg2, arg3, _registeredCallbacks[23]);
}
static BppIntPtr BPPAPI SCBFUNC4_4(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3) {
	return CBFUNC4(arg0, arg1, arg2, arg3, _registeredCallbacks[24]);
}

static BppIntPtr BPPAPI CBFUNC5(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3, BppIntPtr arg4,
	BppIntPtr pCallback) {
	BppVector args(5, true);
	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	args[3] = arg3;
	args[4] = arg4;
	return _callCallbackObject(pCallback, &args);
}

static BppIntPtr BPPAPI CBFUNC6(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3, BppIntPtr arg4,
	BppIntPtr arg5, BppIntPtr pCallback) {
	BppVector args(6, true);
	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	args[3] = arg3;
	args[4] = arg4;
	args[5] = arg5;
	return _callCallbackObject(pCallback, &args);
}

static BppIntPtr BPPAPI CBFUNC7(BppIntPtr arg0, BppIntPtr arg1, BppIntPtr arg2, BppIntPtr arg3, BppIntPtr arg4,
	BppIntPtr arg5, BppIntPtr arg6, BppIntPtr pCallback) {
	BppVector args(7, true);
	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	args[3] = arg3;
	args[4] = arg4;
	args[5] = arg5;
	args[6] = arg6;
	return _callCallbackObject(pCallback, &args);
}



bool BppDefaultFunctions::Bpp_GetCallbackFunc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp; bp -= argc;
	BppIntPtr result = 0;
	BppUInt funcargs = bp[1];
	switch (funcargs) {
	case 1 :
		result = INTPTRCAST(&CBFUNC0);
		break;
	case 2:
		result = INTPTRCAST(&CBFUNC1);
		break;
	case 3:
		result = INTPTRCAST(&CBFUNC2);
		break;
	case 4:
		result = INTPTRCAST(&CBFUNC3);
		break;
	case 5:
		result = INTPTRCAST(&CBFUNC4);
		break;
	case 6:
		result = INTPTRCAST(&CBFUNC5);
		break;
	case 7:
		result = INTPTRCAST(&CBFUNC6);
		break;
	case 8:
		result = INTPTRCAST(&CBFUNC7);
		break;
	default:
		THROW(BppString::format(FLIB_INVALID_CALLBACKARGNUM_FMT, funcargs).c_str());
	};
	*bp = result;
	sp = bp;
    return true;
}

// intptr GetStaticCallbackFunc(uint numArgs, uint iInst)
bool BppDefaultFunctions::Bpp_GetStaticCallbackFunc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2, 2);
	ValPtr bp = sp; bp -= argc;
	BppIntPtr result = 0;
	BppUInt funcargs = bp[1];
	BppUInt instidx = bp[2];
	switch (funcargs) {
	case 0:
		switch (instidx) {
		case 0:
			result = INTPTRCAST(&SCBFUNC0_0);
			break;
		case 1: 
			result = INTPTRCAST(&SCBFUNC0_1);
			break;
		case 2:
			result = INTPTRCAST(&SCBFUNC0_2);
			break;
		case 3:
			result = INTPTRCAST(&SCBFUNC0_3);
			break;
		case 4:
			result = INTPTRCAST(&SCBFUNC0_4);
			break;
		default:
			THROW(BppString::format(FLIB_INVALID_CALLBACKINSTANCENUM_FMT, instidx).c_str());
		}
		break;
	case 1:
		switch (instidx) {
		case 0:
			result = INTPTRCAST(&SCBFUNC1_0);
			break;
		case 1:
			result = INTPTRCAST(&SCBFUNC1_1);
			break;
		case 2:
			result = INTPTRCAST(&SCBFUNC1_2);
			break;
		case 3:
			result = INTPTRCAST(&SCBFUNC1_3);
			break;
		case 4:
			result = INTPTRCAST(&SCBFUNC1_4);
			break;
		default:
			THROW(BppString::format(FLIB_INVALID_CALLBACKINSTANCENUM_FMT, instidx).c_str());
		}
		break;
	case 2:
		switch (instidx) {
		case 0:
			result = INTPTRCAST(&SCBFUNC2_0);
			break;
		case 1:
			result = INTPTRCAST(&SCBFUNC2_1);
			break;
		case 2:
			result = INTPTRCAST(&SCBFUNC2_2);
			break;
		case 3:
			result = INTPTRCAST(&SCBFUNC2_3);
			break;
		case 4:
			result = INTPTRCAST(&SCBFUNC2_4);
			break;
		default:
			THROW(BppString::format(FLIB_INVALID_CALLBACKINSTANCENUM_FMT, instidx).c_str());
		}
		break;
	case 3:
		switch (instidx) {
		case 0:
			result = INTPTRCAST(&SCBFUNC3_0);
			break;
		case 1:
			result = INTPTRCAST(&SCBFUNC3_1);
			break;
		case 2:
			result = INTPTRCAST(&SCBFUNC3_2);
			break;
		case 3:
			result = INTPTRCAST(&SCBFUNC3_3);
			break;
		case 4:
			result = INTPTRCAST(&SCBFUNC3_4);
			break;
		default:
			THROW(BppString::format(FLIB_INVALID_CALLBACKINSTANCENUM_FMT, instidx).c_str());
		}
		break;
	case 4:
		switch (instidx) {
		case 0:
			result = INTPTRCAST(&SCBFUNC4_0);
			break;
		case 1:
			result = INTPTRCAST(&SCBFUNC4_1);
			break;
		case 2:
			result = INTPTRCAST(&SCBFUNC4_2);
			break;
		case 3:
			result = INTPTRCAST(&SCBFUNC4_3);
			break;
		case 4:
			result = INTPTRCAST(&SCBFUNC4_4);
			break;
		default:
			THROW(BppString::format(FLIB_INVALID_CALLBACKINSTANCENUM_FMT, instidx).c_str());
		}
		break;
	default:
		THROW(BppString::format(FLIB_INVALID_CALLBACKARGNUM_FMT, funcargs).c_str());
	};
	*bp = result;
	sp = bp;
	return true;
}

// void SetStaticCallbackHandler(uint numArgs, uint iInst, var handler)
bool BppDefaultFunctions::Bpp_SetStaticCallbackHandler(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3, 3);
	ValPtr bp = sp; bp -= argc;
	BppUInt numArgs = bp[1];
	BppUInt iInst = bp[2];
	const BppValue& hdl = bp[3];
	if (numArgs >= MAXSCBARGS)
		THROW(BppString::format(FLIB_INVALID_CALLBACKARGNUM_FMT, numArgs).c_str());
	if (iInst >= MAXSCBINSTANCES)
		THROW(BppString::format(FLIB_INVALID_CALLBACKINSTANCENUM_FMT, iInst).c_str());
	_registeredCallbacks[MAXSCBARGS * numArgs + iInst] = hdl;
	bp->dispose();
	sp = bp;
	return true;
}
#endif // USECALLBACK
#endif //!DOXY

