
/** @file
Defines implementation functions for B++ predefined functions.
*/

#ifndef DOXY
#include <bppdef.h>
#ifdef __linux__
#include <wctype.h>
#include <wchar.h>
#include <locale.h>
#endif
#ifndef WINCE
#ifndef __linux__
#include <conio.h>
#endif
#include <time.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "bppufunc.h"
#include "bppdfunc.h"
#include "bppvm.h"
#include "bppint.h"
#include "bppvalue.h"
#include "bppbuf.h"
#include "bppclass.h"
#include "bppobjct.h"
#include "bppstrng.h"
#include "bppvec.h"
#include "bppprint.h"
#include "bppcomp.h"
#include "dictio.h"
#include "bppfimpl.h"
#include "bppmsg.h"
#ifndef WINCE
#ifdef __MSDOS__
	#include <dos.h>
	#ifdef __TURBOC__
		#include <alloc.h>
	#else
		#include <malloc.h>
	#endif

#endif // __MSDOS__
#endif
#ifdef WIN32
	#include <windows.h>
#endif

const BppChar* FIMPL_MISSING_INTERPRETER = _T("missing interpreter\n");
const BppChar* FIMPL_TOO_FEW_ARGS = _T("too few arguments");
const BppChar* FIMPL_TOO_MANY_ARGS = _T("too many arguments");
const BppChar* FIMPL_INVALID_THIS_PTR = _T("invalid 'this' pointer");
const BppChar* FIMPL_CONST_ARG_NOT_ALLOWED = FLIB_CONST_ARG_NOT_ALLOWED;

void BppDefaultFunctions::_registerSymbol(BppVirtualMachine* vm,const BppChar* name,BppSymbolType stype,const BppValue& val)
{
	BppDictionaryEntry& de = vm->symbols()[name]; de.setValue(val,false,true); de.setSymbolType(stype);
}

#define ADDFILE(NAME) _registerSymbol(vm,_T(#NAME),ST_SDATA,(FILE*) NAME)
#define ADDFUNC(NAME) _registerSymbol(vm,_T(#NAME),ST_SFUNCTION,(BPPUSERFUNC) &Bpp_##NAME)
#define ADDFUNCN(NAME,FUNC) _registerSymbol(vm,_T(#NAME),ST_SFUNCTION,(BPPUSERFUNC) &FUNC)


void BppDefaultFunctions::registerFiles(BppVirtualMachine* vm)
{
	ADDFILE(stdin);
	ADDFILE(stdout);
	ADDFILE(stderr);
}

void BppDefaultFunctions::registerSysVariables(BppVirtualMachine* vm)
{
	static const BppChar* OSTYPE=
#ifdef WIN32
#ifndef WINCE
		LN_WIN32
#else
		LN_WINCE
#endif
#else
#ifdef __MSDOS__
		LN_DOS
#else
#ifdef LINUX
		LN_LINUX
#else
#ifdef UNIX
		LN_UNIX
#else
		_T("")
#endif
#endif
#endif
#endif
		;
	BppValue vos = new BppString(OSTYPE);
	vos.setFlags(BVF_CONST);
	_registerSymbol(vm, SYMN__OSTYPE, ST_SDATA,vos);
	BppValue vuc = 
      (BppInt)
#ifdef _UNICODE
			1
#else
			0
#endif
			;
	vuc.setFlags(BVF_CONST);
	_registerSymbol(vm, SYMN__UNICODE, ST_SDATA, vuc);
	BppValue vps = (BppUInt)sizeof(BppIntPtr);
	vps.setFlags(BVF_CONST);
	_registerSymbol(vm, SYMN__PTRSIZE, ST_SDATA, vps);
}

void BppDefaultFunctions::registerDictFunctions(BppVirtualMachine* vm)
{
	ADDFUNC(newdict);
	ADDFUNC(dictcontains);
	ADDFUNC(dicterase);
	ADDFUNCN(dictclear,Bpp_clear);
	ADDFUNC(dictgetkeys);
	ADDFUNC(dictgetvalues);
}

void BppDefaultFunctions::registerMathFunctions(BppVirtualMachine* vm)
{
	ADDFUNC(sin);
	ADDFUNC(cos);
	ADDFUNC(tan);
	ADDFUNC(atan);
	ADDFUNC(sqrt);
	ADDFUNC(log);
	ADDFUNC(exp);
	ADDFUNC(abs);
	ADDFUNC(pow);
}


void BppDefaultFunctions::registerFunctions(BppVirtualMachine* vm)
{
	registerFiles(vm);
	registerSysVariables(vm);
	ADDFUNC(newvector);
	ADDFUNC(newstring);
	ADDFUNC(newbuffer);
	ADDFUNC(newcbuffer);
	ADDFUNC(newobject);
	ADDFUNC(clone);

	// I/O functions
	registerIOFunctions(vm);

	// string functions
	ADDFUNC(strlen);
	ADDFUNC(strcmp);
	ADDFUNC(stricmp);
	ADDFUNC(strsplit);
	ADDFUNC(strfind);
	ADDFUNC(strerase);
	ADDFUNC(strinsert);
	ADDFUNC(strreplace);
	ADDFUNC(substr);
	ADDFUNC(strupr);
	ADDFUNC(strlwr);

	// vector and buffer related functions
	ADDFUNC(vecsize);
	ADDFUNC(insert);
	ADDFUNCN(vinsert, Bpp_insert);
	ADDFUNC(insrange);
	ADDFUNC(erase);
	ADDFUNCN(verase,Bpp_erase);
	ADDFUNC(reserve);
	ADDFUNC(resize);
	ADDFUNC(swap);
	ADDFUNCN(vswap, Bpp_swap);
	ADDFUNC(move);
	ADDFUNCN(vmove, Bpp_move);
	ADDFUNC(slice);
	ADDFUNCN(vslice, Bpp_slice);
	ADDFUNC(clear);
	ADDFUNCN(vclear,Bpp_clear);
	ADDFUNC(size);
	ADDFUNC(shrink);
	ADDFUNCN(Vec,Bpp_T);
	ADDFUNC(T);
	ADDFUNC(CBuf);
	ADDFUNC(qsort);
	ADDFUNC(lbound);
	ADDFUNC(ubound);
	ADDFUNC(bsearch);
	ADDFUNC(foreach);
	ADDFUNC(find);
	ADDFUNC(rfind);

	// program related functions
	ADDFUNC(compile);
	ADDFUNC(compilefn);
	ADDFUNC(litexpr);
	ADDFUNC(loadmodule);

	// basic type conversion
	ADDFUNC(bool);
	ADDFUNC(sbyte);
	ADDFUNC(byte);
	ADDFUNC(short);
	ADDFUNC(ushort);
	ADDFUNC(int);
	ADDFUNC(uint);
	ADDFUNC(long);
	ADDFUNC(ulong);
	ADDFUNC(intptr);
	ADDFUNC(uintptr);
	ADDFUNC(float);
	ADDFUNC(double);
	ADDFUNC(char);
	ADDFUNC(uchar);
	ADDFUNC(char8_t);
	ADDFUNC(char16_t);
	ADDFUNC(char32_t);
	ADDFUNC(wchar_t);
	ADDFUNC(string);
	ADDFUNC(buffer);
	ADDFUNC(charbuffer);
	ADDFUNC(vector);
	ADDFUNC(dictionary);

	// RTTI
	ADDFUNC(gettypename);
	ADDFUNC(gettype);
	ADDFUNC(getclassname);
	ADDFUNC(getclass);
	ADDFUNC(dynamic_cast);
	ADDFUNC(getsymbol);
	ADDFUNC(isconst);
	ADDFUNC(mkconst);

	// date and time
	ADDFUNC(time);
	ADDFUNC(timer);
	// math functions
	registerMathFunctions(vm);

	// miscellanous funcs
	ADDFUNC(rand);
	ADDFUNC(srand);
	ADDFUNC(ctime);
	ADDFUNC(localtime);
	ADDFUNC(strftime);
	ADDFUNC(getargs);
	ADDFUNC(getusrargs);
	ADDFUNC(argcnt);
	ADDFUNC(arg);
	ADDFUNC(argv);
	ADDFUNC(call);

	// system functions
	ADDFUNC(addr);
	ADDFUNC(getenv);
	ADDFUNC(setenv);
	ADDFUNC(memsize);
#ifndef NO_DYNLIB
	ADDFUNC(LoadLibrary);
	ADDFUNC(FreeLibrary);
	ADDFUNC(GetProcAddress);
	ADDFUNC(CallLibFunc);
#endif
#ifdef USECALLBACK
	ADDFUNC(GetCallbackFunc);
	ADDFUNC(GetStaticCallbackFunc);
	ADDFUNC(SetStaticCallbackHandler);
#endif
   ADDFUNC(errno);
	ADDFUNC(nerr);
	ADDFUNC(strerror);
	ADDFUNC(setlocale);
	ADDFUNC(memmove);
#ifndef WINCE
	ADDFUNC(system);
#endif
#ifdef WIN32
#ifndef NO_DYNLIB
	ADDFUNC(GetUser32Func);
	ADDFUNC(GetKernel32Func);
	ADDFUNC(GetGdi32Func);
	ADDFUNC(GetShell32Func);
	ADDFUNC(GetOle32Func);
	ADDFUNC(GetLastError);
#endif
	ADDFUNC(GetSysErrorMsg);
#ifndef NO_DYNLIB
	vm->registerNativeFunction(_T("GetCoreFunc") ,(BPPUSERFUNC) &Bpp_GetUser32Func);
#endif
#endif
	registerDictFunctions(vm);
	// graphic bios functions
	ADDFUNC(setscrmode);
	ADDFUNC(setpixel);
	ADDFUNC(line);
#ifdef __MSDOS__
	registerDosFunctions(vm);
#endif
}

#ifdef __MSDOS__

void BppDefaultFunctions::registerDosFunctions(BppVirtualMachine* vm)
{

	ADDFUNC(setreg);
	ADDFUNC(getreg);
	ADDFUNC(segread);
	ADDFUNC(int86);
	ADDFUNC(int86x);
	ADDFUNC(inportb);
	ADDFUNC(outportb);
	ADDFUNC(inport);
	ADDFUNC(outport);

}
#endif
#undef ADDFUNC


// vec = newvector(capacity=0,init=true)
bool BppDefaultFunctions::Bpp_newvector(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(2);
	ValPtr bp = sp; bp-=argc;
	bool init = true;
	switch (argc)
	{
	case 0:
		*bp = new BppVector();
		return true;
	case 2:
		init = bp[2].asBool();
	default:
		*bp = new BppVector((BppUInt) bp[1],init);
		break;
	}
	sp = bp;
	return true;
}
// str = newstring(string|buffer|charbuffer)
// str = newstring(size,fillchar=' ')
// str = newstring(fmtstring, ... )
bool BppDefaultFunctions::Bpp_newstring(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMIN(1);
	ValPtr bp = sp; bp -= argc;
	switch (maskWeakRef(bp[1].getType()))
	{
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				if (argc > 1)
				{
					ValPtr pv = bp;
					pv += 2;
					*bp = new BppString(s->c_str(),argc-1,pv);
				}
				else
					*bp = new BppString(s->c_str());
			}
			else
				*bp = new BppString();
		}
		break;
	case BVT_BUFFER:
		{
			ARGCMAX(1);
			BppBuffer* b = bp[1];

		 BppString* s = b != 0 ? new BppString(*b) : new BppString();
			*bp = s;
		}
		break;
	case BVT_CHARBUFFER:
		{
			ARGCMAX(1);
			BppCharBuffer* b = bp[1];
		 BppString* s = b != 0 ? new BppString(*b) : new BppString();
			*bp = s;
		}
		break;
	case BVT_INT:
	case BVT_UINT:
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_INTPTR:
	case BVT_UINTPTR:
		{
			ARGCMAX(2);
			BppUInt size = bp[1];
			BppChar fill = _T(' ');
			if (argc == 2) fill = (BppChar) (BppUInt) bp[2];
			BppString* s = new BppString();
			*bp = s;
			for (BppUInt i=0;i<size;i++) (*s) += fill;
		}
		break;
	default:
		THROW(FLIB_INVALID_VALUE_TYPE);
		break;
	}
	sp = bp;
	return true;
}

// str = newbuffer(string|buffer|charbuffer)
// str = newbuffer(size,fillbyte=0)
bool BppDefaultFunctions::Bpp_newbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	switch (maskWeakRef(bp[1].getType()))
	{
	case BVT_BUFFER:
		{
			BppBuffer* b = bp[1];
			if (b != 0)
				*bp = new BppBuffer(*b);
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				size_t size =  s->length()+1;
				BppBuffer* b = new BppBuffer(size,size);
            BppByte* pb = b->data();
            const BppChar* pcb = s->c_str();
				for (size_t i=0;i<size;i++,pb++,pcb++)
					*pb = (BppByte)(*pcb);
				*bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* cb = bp[1];
			if (cb != 0)
			{
				size_t size = cb->size();
				BppBuffer* b = new BppBuffer(size,size);
            BppByte* pb = b->data();
            BppChar* pcb = cb->data();
            for (size_t i=0;i<size;pcb++,pb++,i++)
					*pb = (BppByte)*pcb;
				*bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_INT:
	case BVT_UINT:
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_INTPTR:
		{
			BppUInt size = bp[1];
			BppBuffer* b = new BppBuffer((size_t) size,(size_t) size);
			*bp = b;
         if (b && argc == 2) memset(b->data(), (BppByte)bp[2], size);
		}
		break;
	default:
		THROW(FLIB_INVALID_VALUE_TYPE);
		break;
	}
	sp = bp;
	return true;
}

// str = newcbuffer(string|buffer|charbuffer)
// str = newcbuffer(size,fillchar=0)
bool BppDefaultFunctions::Bpp_newcbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	switch (maskWeakRef(bp[1].getType()))
	{
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* b = bp[1];
			if (b != 0)
				*bp = new BppCharBuffer(*b);
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				size_t size =  s->length()+1;
				BppCharBuffer* b = new BppCharBuffer(size,size);
				memcpy(b->data(), s->c_str(), size * sizeof(BppChar));
				b->push_back(0);
				*bp = b;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_BUFFER:
		{
			BppBuffer* b = bp[1];
			if (b != 0)
			{
				size_t size = b->size();
				BppCharBuffer* cb = new BppCharBuffer(size,size);
				for (size_t i = 0; i < size; i++) 
					(*cb)[i] = (BppChar)(*b)[i];
				*bp = cb;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_INT:
	case BVT_UINT:
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_INTPTR:
		{
			BppUInt size = bp[1];
			BppChar fill = 0;
			if (argc == 2) fill = (BppChar) (BppInt) bp[2];
			BppCharBuffer* b = new BppCharBuffer((size_t) size,(size_t)size);
			*bp = b;
			if (fill) 
				for (BppUInt i = 0; i < size; ++i) (*b)[i] = fill;
		}
		break;
	default:
		THROW(FLIB_INVALID_ARG_TYPE);
		break;
	}
	sp = bp;
	return true;
}

// dict = newdict([bool sorted = false])
bool BppDefaultFunctions::Bpp_newdict(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(1);
	bool sorted = false;
	if (argc == 1)
		sorted = sp->asBool();
	sp -= argc;
	*sp = new BppDictionary(sorted);
	return true;
}

bool BppDefaultFunctions::Bpp_newobject(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	ARGCMIN(1);
	ValPtr bp = sp;
	bp -= argc;
	BppClass* pclass = bp[1];
	if (!pclass) return interpreter->badType(bp[1].getType(), BVT_CLASS);
	bp[1] = new BppObject(pclass); 
	bool ok =  interpreter->send(argc-1, pclass->getClassName()->c_str());
	*bp = *sp;
	sp = bp;
	return ok;
}

bool BppDefaultFunctions::Bpp_clone(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue v(*sp,true);
	*(--sp) = BPPMOVE(v);
	return true;
}
#ifdef __MSDOS__
#define malloc farmalloc
#define free farfree
#endif

bool BppDefaultFunctions::Bpp_memsize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
	BppUInt maxsize = BPP_NPOS;
	BppUInt currsize = BPP_NPOS >> 1;	
	BppUInt minsize = 0;
	while (1) {
		void* p = malloc(currsize);
		if (p) {
			free(p);
			if ((maxsize - minsize) < 16)
				break;
			minsize = currsize;
			currsize += (maxsize - currsize) >> 1;
		}
		else {
			if ((maxsize - minsize) < 16) {
				currsize = minsize;
				break;
			}
			maxsize = currsize;
			currsize -= (currsize - minsize) >> 1;
		}
	}
	*sp = currsize;// (BppInt)coreleft();
	return true;
}
#ifdef __MSDOS__
#undef malloc
#undef free
#endif

//#ifdef __MSDOS__
//
//#ifndef __TURBOC__
//#define  coreleft _memmax
//#endif
//// size = memsize()
//bool BppDefaultFunctions::Bpp_memsize(BppVirtualMachine* vm, int argc, ValPtr& sp)
//{
//	CHKVM;
//	ARGCMAX(0);
//		  *sp = (BppInt) coreleft();
//	return true;
//}
//
//#endif

	// I/O functions
// fp = fopen(name,mode)
bool BppDefaultFunctions::Bpp_fopen(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp;
	bp-=2;
	BppString* filename = bp[1];
	BppString* mode = bp[2];
	if ((filename == 0) || (mode == 0))
		THROW(FLIB_INVALID_ARG_TYPE);
	FILE *fp = _tfopen(filename->c_str(),mode->c_str());
	if (fp != 0)
		*bp = fp;
	else
		bp->dispose();
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_freopen(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(3, 3);
	ValPtr bp = sp;
	bp -= argc;
	BppString* filename = bp[1];
	BppString* mode = bp[2];
	FILE* fp0 = bp[3];
	if ((filename == 0) || (mode == 0) || (fp0 ==0))
		THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(bp[3]);
	FILE *fp = _tfreopen(filename->c_str(), mode->c_str(),fp0);
	if (fp != 0)
		*bp = fp;
	else
		bp->dispose();
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_ftell(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp;
	bp -= argc;
	FILE* fp0 = bp[1];
	if (!fp0)
		THROW(FLIB_INVALID_ARG_TYPE);
	BppInt c = ftell(fp0);
	*bp = c;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_fseek(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(3, 3);
	ValPtr bp = sp;
	bp -= argc;
	FILE* fp0 = bp[1];
	BppInt offs = bp[2];
	BppInt mode = bp[3];
	if (fp0 == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(bp[1]);
	BppInt r = fseek(fp0, offs, mode);
	*bp = r;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_fflush(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 1);
	FILE* fp = *sp;
	sp[-1] = (BppInt)fflush(fp);
	--sp;
	return true;
}

// done = fclose(fp)
bool BppDefaultFunctions::Bpp_fclose(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (sp->getType() != BVT_FILE) THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(*sp);
	sp[-1] = (BppInt) fclose(*sp);
	 --sp;
	 return true;
}

// end = feof(fp)
bool BppDefaultFunctions::Bpp_feof(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 1);
	if (sp->getType() != BVT_FILE) THROW(FLIB_INVALID_ARG_TYPE);
	sp[-1] = (BppInt)feof((FILE*)*sp);
	--sp;
	return true;
}

// ch = getc([fp])
bool BppDefaultFunctions::Bpp_getc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(1);
	if (argc == 1)
	{
		if (sp->getType() != BVT_FILE)
			THROW(FLIB_INVALID_ARG_TYPE);
		CHKVAR(*sp);
		ValPtr bp = sp; --bp;
		BppValue* pv = bp;
		BppInt l = (BppInt) getc((FILE*)(*sp));
		*pv = l;
		sp=bp;
	}
	else
		*sp = (BppInt) vm->getCurrentInput()->getChar();
	 return true;
}

// ch = putc(ch,fp)
bool BppDefaultFunctions::Bpp_putc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2, 2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	if ((!bp[1].isIntegralType()) || (fp == 0)) THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(bp[2]);
	*bp = (BppInt)putc((int)(BppInt)bp[1], fp);
	sp = bp;
	return true;
}
#if !(defined(WINCE) || defined(__linux__))
#ifndef _gettche // watcom hack
	#define _gettche getche
#endif
#endif

// str = gets()
bool BppDefaultFunctions::Bpp_gets(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(0,0);
	BppString* s = new BppString();
	*s = vm->getCurrentInput()->getString();
	*sp = s;
	return true;
}

// str = fgets(fp)
bool BppDefaultFunctions::Bpp_fgets(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	FILE* fp = *sp;
	if (fp == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(*sp);
	BppChar c = 0;
	BppString* s = new BppString();
	while (((c = (BppChar) getc(fp)) != _T('\n')) &&(!feof(fp)))
	{
		(*s) += c;
	}
	--sp;
	*sp = s;
	return true;
}

// result = fputs(str,fp)
bool BppDefaultFunctions::Bpp_fputs(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	BppString* s = bp[1];
	if ((s == 0) || (fp == 0))
		THROW(FLIB_INVALID_ARG);
	CHKVAR(bp[2]);
	*bp = (BppInt) fputs(s->a_str(),fp);
	 sp = bp;
	 return true;
}


// var = freadval(fp)
bool BppDefaultFunctions::Bpp_freadval(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	FILE* fp = *sp;
	if (fp == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(*sp);
	BppDictionaryIO io(*vm);
	--sp;
	if (!io.readSingleValue(fp,sp[0]))
		sp[0].dispose();
	return true;
}

// ok = fwriteval(value, fp)
bool BppDefaultFunctions::Bpp_fwriteval(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	if (fp == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(bp[2]);
	BppDictionaryIO io(*vm);
	*bp = BppValue::makeBool(io.writeSingleValue(fp,bp[1]));
	sp = bp;
	return true;
}

// int fread(buffer,[[unsigned size=1,]unsigned cnt,]FILE* fp)
bool BppDefaultFunctions::Bpp_fread(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,4);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[argc];
	if (fp == 0) THROW(FLIB_INVALID_ARG_TYPE, FLIB_FILE_EXPECTED_BR);
	CHKVAR(bp[argc]);
	BppBuffer* pbuf = bp[1];
	if (pbuf == 0) THROW(FLIB_INVALID_ARG_TYPE, FLIB_BUFFER_EXPECTED_BR);
	CHKVAR(bp[1]);
	size_t cnt = 0;
	bool fixedsize = (argc > 2) ? true : false;
	if (fixedsize)
	{
		size_t size = 1;
		BppValue& vcnt = bp[argc-1];
		if (!vcnt.isIntegralType())
			THROW(FLIB_INVALID_ARG_TYPE, FLIB_INT_EXPECTED_BR);
		cnt = (size_t) (BppUInt) vcnt;
		if (argc == 4)
		{
			BppValue& vsize = bp[2];
			if (!vsize.isIntegralType())
				THROW(FLIB_INVALID_ARG_TYPE, FLIB_INT_EXPECTED_BR);
			size = (size_t) (BppUInt) vsize;
	  }
		size_t bsize = pbuf->size();
		size_t requestedSize = bsize + (cnt * size);
		if (requestedSize > pbuf->capacity())
			pbuf->setCapacity(requestedSize);
		cnt = fread(&(pbuf->data()[bsize]),size,cnt,fp);
		pbuf->setSize(bsize + (cnt * size));
	}
	else
	{
		int c;
		while ((c = fgetc(fp)) != EOF)
		{
			pbuf->push_back((BppByte) c);
			++cnt;
		}
	}
	*bp = (BppUInt) cnt;
	sp = bp;
	return true;
}


// int fwrite(buffer,[[unsigned size=1,]unsigned cnt,]FILE* fp)
bool BppDefaultFunctions::Bpp_fwrite(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,4);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[argc];
	if (fp == 0) THROW(FLIB_INVALID_ARG_TYPE, FLIB_FILE_EXPECTED_BR);
	CHKVAR(bp[argc]);
	BppBuffer* pbuf = bp[1];
	if (pbuf == 0) THROW(FLIB_INVALID_ARG_TYPE, FLIB_BUFFER_EXPECTED_BR);
	size_t size = 1;
	size_t cnt = 0;
	bool fixedsize = (argc > 2) ? true : false;
	if (fixedsize)
	{
		BppValue& vcnt = bp[argc-1];
		if (!vcnt.isIntegralType())
			THROW(FLIB_INVALID_ARG_TYPE, FLIB_INT_EXPECTED_BR);
		cnt = (size_t) (BppUInt) vcnt;
		if (argc == 4)
		{
			BppValue& vsize = bp[2];
			if (!vsize.isIntegralType())
				THROW(FLIB_INVALID_ARG_TYPE, FLIB_INT_EXPECTED_BR);
			size = (size_t) (BppUInt) vsize;
	  }
		cnt = fwrite(pbuf->data(),size,cnt,fp);
	}
	else
	{
		cnt = pbuf->size();
		cnt = fwrite(pbuf->data(),size,cnt,fp);
	}
	*bp = (BppUInt) cnt;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_kbhit(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
	*sp = keyPressed();
	return true;
}

#if (INT_MAX > SHRT_MAX)
bool BppDefaultFunctions::Bpp_getwc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 1);
	if (sp->getType() != BVT_FILE) THROW(FLIB_INVALID_ARG_TYPE);
	ValPtr bp = sp; --bp;
	BppValue* pv = bp;
	CHKVAR(*sp);
	BppInt l = (BppInt)fgetwc((FILE*)(*sp));
	*pv = l;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_putwc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2, 2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	CHKVAR(bp[2]);
	if ((!bp[1].isIntegralType()) || (fp == 0)) THROW(FLIB_INVALID_ARG_TYPE);
	*bp = (BppInt)fputwc((wchar_t)(BppUInt)bp[1], fp);
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_fgetws(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	FILE* fp = *sp;
	if (fp == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	CHKVAR(*sp);
	BppChar c = 0;
	BppString* s = new BppString();
	while (((c = (BppChar)cbufToBppChar(fgetwc(fp),sizeof(BppChar))) != L'\n') &&(!feof(fp)))
	{
		(*s) += c;
	}
	--sp;
	*sp = s;
	return true;
}

bool BppDefaultFunctions::Bpp_fputws(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	CHKVAR(bp[2]);
	BppString* s = bp[1];
	if ((s == 0) || (fp == 0))
		THROW(FLIB_INVALID_ARG);
	*bp = (BppInt) fputws(s->w_str(),fp);
	 sp = bp;
	 return true;
}
#endif

bool BppDefaultFunctions::Bpp_print(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ValPtr bp = sp; bp -= argc;
	for (int i=1;i<=argc;i++)
	{
		BppValue& v = bp[i];
		if (v.isCharacterType()) {
			BppValue tmp = BppValue::makeChar(v);
			vm->printer()(_T("%c"), (BppUInt)tmp);
		}
		else
			switch (v.getType())
		{
		case BVT_INT:
	  case BVT_SHORT:
	  case BVT_SBYTE:
#if (INT_MAX == SHRT_MAX)
		  vm->printer()(_T("%ld"),(BppInt) v);
#else
		  vm->printer()(_T("%d"),(BppInt) v);
#endif
			break;
	  case BVT_UINT:
	  case BVT_USHORT:
	  case BVT_BYTE:
#if (INT_MAX == SHRT_MAX)
		  vm->printer()(_T("%lu"),(BppUInt) v);
#else
		  vm->printer()(_T("%u"),(BppUInt) v);
#endif
			break;
	  case BVT_BOOL:
		  vm->printer()(v.asBool() ? _T("true") : _T("false"));
		  break;
	  case BVT_INTPTR:
			vm->printer()(_T("%p"),(BppIntPtr) v);
		  break;
	  case BVT_LONG:
#ifdef EMULATE_I64
		  {
			  BppChar buf[65];
			  ((BppLong)v).toa(buf);
			  vm->printer()(buf);
		  }
#else
			vm->printer()(_T("%lld"),(BppLong) v);
#endif
		 break;
	  case BVT_ULONG:
#ifdef EMULATE_I64
		  {
			  char buf[65];
			  ((BppULong)v).toa(buf);
			  vm->printer()(buf);
		  }
#else
			vm->printer()(_T("%llu"),(BppULong) v);
#endif
		  break;
		case BVT_FLOAT:
		case BVT_DOUBLE:
			vm->printer()(_T("%g"),(double) v);
			break;
		default:
		{
			BppProcInfo* pi = interpreter->getProcInfo();
			BppClass* ownClass = pi ? pi->getClass() : 0;
			BppString* s = interpreter->getStringFromValue(ownClass,v);
#ifdef _UNICODE
			if (s != 0) {
			
				vm->printer()(_T("%ls"), s->c_str());
			}
			else
				vm->printer()(_T("(%ls:%p)"), getValueTypeName(v.getType()), v.ptr());
			break;
#else
			if (s != 0)
				vm->printer()(_T("%s"), s->c_str());
			else
				vm->printer()(_T("(%s:%p)"), getValueTypeName(v.getType()), v.ptr());
			break;
#endif
		}
		}
	}
	sp = bp;
	sp->dispose();
	return true;
}

	// string functions
// len = strlen(s)
bool BppDefaultFunctions::Bpp_strlen(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppString* s = *sp;
	if (s == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	--sp;
	*sp = (BppUInt) s->length();
	return true;
}

// result = strcmp(s1,s2)
bool BppDefaultFunctions::Bpp_strcmp(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppString* s1 = bp[1];
	BppString* s2 = bp[2];
	if ((s1 == 0) || (s2 == 0))
		THROW(FLIB_INVALID_ARG_TYPE);
	*bp = (BppInt) s1->compare(*s2);
	sp = bp;
	return true;
}

// result = stricmp(s1,s2)
bool BppDefaultFunctions::Bpp_stricmp(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppString* s1 = bp[1];
	BppString* s2 = bp[2];
	if ((s1 == 0) || (s2 == 0))
		THROW(FLIB_INVALID_ARG_TYPE);
	*bp = (BppInt) s1->icompare(*s2);
	sp = bp;
	return true;
}

// vec = strsplit(str,delimit,storeempty=false,storedelimit=false)
bool BppDefaultFunctions::Bpp_strsplit(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,4);
	ValPtr bp = sp; bp -= argc;
	BppString* s1 = bp[1];
	BppString* s2 = bp[2];
	if ((s1 == 0) || (s2 == 0))
		THROW(FLIB_INVALID_ARG_TYPE);
	bool emptytok = argc > 2 ? bp[3].asBool() : false;
	bool storedelim = argc > 3 ? bp[4].asBool() : false;
	BppVector* vec = new BppVector();
	s1->split(*vec, *s2, emptytok, storedelim);
	vec->shrink();
	*bp = vec;
	sp = bp;
	return true;
}

// uint pos = strfind(string str,string s, uint offset=0);
// uint pos = strfind(string str, long c, int offset = 0)
bool BppDefaultFunctions::Bpp_strfind(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
	  THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	BppString* s = bp[2];
	BppChar c = (s == 0) ? (BppChar) (BppInt) bp[2] : 0;
	if ((s == 0) && (c == 0))
		THROW(FLIB_INVALID_ARG, FLIB_STRING_OR_CHAR_EXPECTED_BR);
	size_t pos = (argc == 3) ? (size_t) (BppUInt) bp[3] : 0;
	size_t res = 0;
	if (s != 0)
	  res = str->find(*s,pos);
	else
		res = str->find(c,pos);
	*bp = (res == BPP_NNPOS) ? BPP_NPOS : BppUInt(res);
	sp = bp;
	return true;
}

// string modifiedstring = strerase(string str, uint pos, uint cnt=-NPOS)
// modifies str
bool BppDefaultFunctions::Bpp_strerase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppValue strval((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppString* str = strval;
	if (str == 0)
		THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	size_t pos = (size_t) (BppUInt) bp[2];
	size_t cnt = (argc == 3) ? size_t((BppUInt) bp[3]) : BPP_NNPOS;
	str->erase(pos,cnt);
	*bp = str;
	sp = bp;
	return true;
}
// string modifiedstring = strinsert(string str, uint pos, string insertion)
// string modifiedstring = strinsert(string str, uint pos , int c, uint cnt=1)
bool BppDefaultFunctions::Bpp_strinsert(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,4);
	ValPtr bp = sp; bp -= argc;
	BppValue strval((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppString* str = strval;
	if (str == 0)
		THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	BppUInt pos = bp[2];
	BppString* s = bp[3];
	if (s != 0)
	{
		ARGCMAX(3);
		str->insert((size_t)pos,*s);
	}
	else
	{
		size_t cnt = (argc == 4) ? (size_t) (BppUInt) bp[4] : 1;
		str->insert((size_t) pos,(BppChar) (BppInt) bp[3],cnt);
	}
	*bp = str;
	sp = bp;
	return true;
}

// string modifiedstring = strreplace(string str, string s, string replacement)
// string modifiedstring = strreplace(string str, int c, int replacement)
bool BppDefaultFunctions::Bpp_strreplace(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,3);
	ValPtr bp = sp; bp -= argc;
	BppValue strval((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppString* str = strval;
	if (str == 0)
		THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	BppString* s = bp[2];
	if (s != 0)
	{
		BppString* rep = bp[3];
		if (rep == 0)
			THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
		str->replace(*s,*rep);
	}
	else
	{
		str->replace((BppChar)(BppInt)bp[2],(BppChar)(BppInt)bp[3]);
	}
	*bp = str;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_struplow(BppVirtualMachine * vm, int argc, ValPtr & sp, bool upper)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp;
	bp -= argc;
	BppValue strval((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppString* str = strval;
	if (str == 0)
		THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	--sp;

	if (upper) str->upper();
	else str->lower();
	*sp = str;
	return true;
}

bool BppDefaultFunctions::Bpp_strupr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	return Bpp_struplow(vm, argc, sp, true);
}

bool BppDefaultFunctions::Bpp_strlwr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	return Bpp_struplow(vm, argc, sp, false);
}

// string sub = substr(string str,uint start, int len=-1)
bool BppDefaultFunctions::Bpp_substr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
		THROW(FLIB_INVALID_ARG_TYPE, FLIB_STRING_EXPECTED_BR);
	BppUInt pos = bp[2];
	size_t len = (argc == 3) ? size_t((BppUInt) bp[3]) : BPP_NNPOS;
	*bp = new BppString(str->substr((size_t) pos,len));
	sp = bp;
	return true;
}

	// vector and buffer related functions
// uint size = vecsize(vec)
bool BppDefaultFunctions::Bpp_vecsize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppVector* v = *sp;
	if (v == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	--sp;
	*sp = (BppUInt) v->size();
	return true;
}

// size = size(vec|buf|cbuf|dict)
bool BppDefaultFunctions::Bpp_size(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	size_t s = 0;
	BppValueType vtype = maskWeakRef(sp->getType());
	if (!sp->addr()) vtype = BVT_NULL;
	switch (vtype)
	{
	case BVT_VECTOR:
		s = ((BppVector*)(*sp))->size();
		break;
	case BVT_BUFFER:
		s = ((BppBuffer*)(*sp))->size();
		break;
	case BVT_CHARBUFFER:
		s = ((BppCharBuffer*)(*sp))->size();
		break;
	case BVT_DICTIONARY:
		s = ((BppDictionary*)(*sp))->count();
		break;
	case BVT_STRING:
		s = ((BppString*)(*sp))->length();
		break;
	default:
			  THROW(FLIB_INVALID_ARG_TYPE);
	}
	--sp;
	*sp = (BppUInt) s;
	return true;
}

// uint
bool BppDefaultFunctions::Bpp_shrink(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 1);
	size_t s = 0;
	BppValueType vtype = maskWeakRef(sp->getType());
	if (!sp->addr()) vtype = BVT_NULL;
	CHKVAR(*sp);
	switch (vtype)
	{
	case BVT_VECTOR:
		s = ((BppVector*)(*sp))->shrink();
		break;
	case BVT_BUFFER:
		s = ((BppBuffer*)(*sp))->shrink();
		break;
	case BVT_CHARBUFFER:
		s = ((BppCharBuffer*)(*sp))->shrink();
		break;
	default:
		THROW(FLIB_INVALID_ARG_TYPE);
	}
	--sp;
	*sp = (BppUInt)s;
	return true;
}

//uint pos =  find(vec cont, var searchVal, int pos=0)
bool BppDefaultFunctions::Bpp_find(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppUInt pos = 0;
	if (argc == 3) pos = bp[3];
	BppVector* pvec = bp[1];
	size_t res = 0;
	if (pvec) 
		res = pvec->find(bp[2],pos);
	else {
		BppBuffer* pbuf = bp[1];
		if (pbuf) res = pbuf->find(bp[2], pos);
		else {
			BppCharBuffer* pcbuf = bp[1];
			if (pcbuf) res = pcbuf->find((BppChar)(BppUInt)bp[2], pos);
			else {
				BppString* s = bp[1];
				if (s) {
					BppString* s2 = bp[2];
					if (s2) res = s->find(*s2, pos);
					else res = s->find((BppChar)(BppUInt)bp[2], pos);
				}
			    else THROW(FLIB_INVALID_ARG_TYPE);
			}
		}
	}

	*bp = (res == BPP_NNPOS) ? BPP_NPOS : BppUInt(res);
	sp = bp;
	return true;
}

//int pos =  rfind(vec cont, var searchVal, int pos=-1)
bool BppDefaultFunctions::Bpp_rfind(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	size_t pos = BPP_NNPOS;
	if (argc == 3) pos = (BppUInt) bp[3];
	BppVector* pvec = bp[1];
	size_t res = 0;
	if (pvec)
		res = pvec->rfind(bp[2],pos);
	else {
		BppBuffer* pbuf = bp[1];
		if (pbuf) res = pbuf->rfind(bp[2], pos);
		else {
			BppCharBuffer* pcbuf = bp[1];
			if (pcbuf) res = pcbuf->rfind((BppChar)(BppUInt)bp[2], pos);
			else THROW(FLIB_INVALID_ARG_TYPE);
		}
	}
	*bp = (res == BPP_NNPOS) ? BPP_NPOS : BppUInt(res);
	sp = bp;
	return true;
}


// uint foreach(var container, var func, ...)
bool BppDefaultFunctions::Bpp_foreach(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	static const BppString _OP_CALL = _T("OP_CALL");
	CHKVM;
	ARGCMIN(2);
	ValPtr bp = sp; bp -= argc;
	BppValue vfunc = bp[2];
	BPPUSERFUNC pufunc = vfunc;
	BppObject* pfuncobj = pufunc ? 0 : (BppObject*)vfunc;
	BppVector* pfunc = 0;
	if (!(pufunc || pfuncobj)) {
		if (maskWeakRef(vfunc.getType()) == BVT_BYTECODE)
			pfunc = vfunc;
		if (!pfunc) THROW(FLIB_INVALID_FUNCTION_ARG);
	}
	BppValue& vcont = bp[1];
	BppValue resval;
	BppObject* contobj = vcont;
	if (contobj) {
		BppVector fargs(1, true);
		fargs.addRef();
		fargs[0] = vfunc;
		for (int i = 3; i <= argc; ++i) fargs.push_back(bp[i]);
		if (!interpreter->invokeMember(resval, contobj, SYMN_foreach, &fargs)) return false;
		*bp = BPPMOVE(resval);
		sp = bp;
		return true;
	}
	BppValueType vt = maskWeakRef(vcont.getType());
	int argoff = 0;
	BppDictionaryEntry* method = 0;
	if (pfuncobj) {
		method = pfuncobj->getClass()->findMember(0, _OP_CALL);
		if (method) {
			++argoff;
			vfunc = method->getValue();
		}
	}
	BppVector fargs(argc+argoff,true);
	fargs.addRef();
	if (method) fargs[0] = pfuncobj;
	for (int i = 3; i <= argc; ++i) fargs[argoff+i-1] = (const BppValue&) bp[i];
	bool ok = true;
	BppUInt loopcnt = 0;
	switch (vt)
	{
	case BVT_VECTOR:
	{
		BppVector* vec = vcont;
		if (!vec) THROW(FLIB_INVALID_CONTAINER);
		unsigned n = vec->size();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff + 1] = (*vec)[i];
			ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_DICTIONARY:
	{
		BppDictionary* dict = vcont;
		if (!dict) THROW(FLIB_INVALID_CONTAINER);
		for (BppDictionaryEntry* i = dict->getFirst(); i; i = i->getNext())
		{
			fargs[argoff] = i->getKey();
			fargs[argoff + 1] = i->getValue();
			ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* buf = vcont;
		if (!buf) THROW(FLIB_INVALID_CONTAINER);
		size_t n = buf->size();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff + 1] = (*buf)[i];
			ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* buf = vcont;
		size_t n = buf->size();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff+1] = (*buf)[i];
			ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_STRING:
	{
		BppString* buf = vcont;
		size_t n = buf->length();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff+1] = (*buf)[i];
			ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	default:
		THROW(FLIB_INVALID_CONTAINER_TYPE);
	}
	sp = bp;
	*bp = loopcnt;
	return true;
}

// var res = insert(var cont, int pos, var insertval, uint cnt = 1)
// string res = insert(string str, int pos, string insertval)
bool BppDefaultFunctions::Bpp_insert(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,4);
	ValPtr bp = sp; bp -= argc;
	BppUInt pos = bp[2];
	BppUInt cnt = (argc == 4) ?  (BppUInt) bp[4] : 1;
	BppValue contVal((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppValue insVal((const BppValue&)bp[3], (bp[3].getFlags() & BVF_CONST) != 0 ? true : false);
	BppVector* vec = contVal;
	if (vec) {
		vec->insert(pos, insVal, cnt);
		*bp = vec;
	}
	else {
		BppBuffer* buf = contVal;
		if (buf) {
			buf->insert(pos, insVal, cnt);
			*bp = buf;			
		}
		else {
			BppCharBuffer* cbuf = contVal;
			if (cbuf) {
				cbuf->insert(pos, (BppChar)(BppUInt)BppValue::makeChar(insVal), cnt);
				*bp = cbuf;
			}
			else {
				BppString* str = contVal;
				if (str) {
					BppString* s = (argc == 3) ? (BppString*)(insVal) : 0;
					if (s) str->insert(pos, *s);
					else str->insert(pos, (BppChar)(BppUInt)BppValue::makeChar(insVal), cnt);
					*bp = str;
				}
				else THROW(FLIB_INVALID_ARG_TYPE);
			}
		}
	}
	sp = bp;
	return true;
}

// var res=insrange(var cont, uint pos, var srccont, uint startpos = 0, uint endpos=NPOS)
bool BppDefaultFunctions::Bpp_insrange(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(3, 5);
	ValPtr bp = sp; bp -= argc;
	BppUInt pos = bp[2];
	BppUInt startpos = (argc > 3) ? (BppUInt)bp[4] : 0;
	BppUInt endpos = (argc == 5) ? (BppUInt)bp[5] : BPP_NPOS;
	int dist = startpos > endpos ? -1 : 1;
	BppValueType T1 = maskWeakRef(bp[1].getType());
	if (T1 < BVT_STRING || T1 > BVT_VECTOR) THROW(FLIB_INVALID_ARG_TYPE_DEST, FLIB_SEQUENCE_CONTAINER_EXPECTED_BR);
	BppValueType T2 = maskWeakRef(bp[3].getType());
	if (T2 != T1) return interpreter->badType(T2, T1);
	if (!bp[1].addr()) THROW(FLIB_DEST_IS_NULL);
	if (!bp[3].addr()) THROW(FLIB_SOURCE_IS_NULL);
	BppValue contVal((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppValue& srcVal = bp[3];
	bool constsrc = srcVal.getFlags() & BVF_CONST;
	size_t i = 0;
	int j = 0;
	switch (T1) {
	case BVT_VECTOR:
	{
		BppVector* pd = contVal;
		size_t nd = pd->size();
		BppVector* ps = srcVal;
		size_t ns = ps->size();
		if (startpos < ns) {
			BppVector* pbuf = 0;
			if (endpos >= ns) endpos = ns - 1;
			size_t cnt = (startpos > endpos ? startpos - endpos : endpos - startpos) + 1;
			if (pos > nd) pos = nd;
			if (pd == ps && (startpos >= pos || endpos >= pos)) {
				pbuf = new BppVector(*ps);
				ps = pbuf;
			}
			pd->insert(pos, BppValue(), cnt);
			for (i = 0, j = 0; i < cnt; ++i, j += dist) {
				if (constsrc)
					(*pd)[pos + i] = BppValue((*ps)[startpos + j], true);
				else
				  (*pd)[pos + i] = (*ps)[startpos + j];
			}
			if (pbuf) delete pbuf;
		}
		*bp = pd;
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* pd = contVal;
		size_t nd = pd->size();
		BppBuffer* ps = srcVal;
		size_t ns = ps->size();
		if (startpos < ns) {
			BppBuffer* pbuf = 0;
			if (endpos >= ns) endpos = ns - 1;
			size_t cnt = (startpos > endpos ? startpos - endpos : endpos - startpos) + 1;
			if (pos > nd) pos = nd;
			if (pd == ps && (startpos >= pos || endpos >= pos)) {
				pbuf = new BppBuffer(*ps);
				ps = pbuf;
			}
			pd->insert(pos, 0, cnt);
			for (i = 0, j = 0; i < cnt; ++i, j += dist) (*pd)[pos + i] = (*ps)[startpos + j];
			if (pbuf) delete pbuf;
		}
		*bp = pd;
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* pd = contVal;
		size_t nd = pd->size();
		BppCharBuffer* ps = srcVal;
		size_t ns = ps->size();
		if (startpos < ns) {
			BppCharBuffer* pbuf = 0;
			if (endpos >= ns) endpos = ns - 1;
			size_t cnt = (startpos > endpos ? startpos - endpos : endpos - startpos) + 1;
			if (pos > nd) pos = nd;
			if (pd == ps && (startpos >= pos || endpos >= pos)) {
				pbuf = new BppCharBuffer(*ps);
				ps = pbuf;
			}
			pd->insert(pos, _T(' '), cnt);
			for (i = 0, j = 0; i < cnt; ++i, j += dist) (*pd)[pos + i] = (*ps)[startpos + j];
			if (pbuf) delete pbuf;
		}
		*bp = pd;
	}
	break;
	case BVT_STRING:
	{
		BppString* pd = contVal;
		size_t nd = pd->length();
		BppString* ps = srcVal;
		size_t ns = ps->length();
		if (startpos < ns) {
			BppString* pbuf = 0;
			if (endpos >= ns) endpos = ns - 1;
			size_t cnt = (startpos > endpos ? startpos - endpos : endpos - startpos) + 1;
			if (pos > nd) pos = nd;
			if (pd == ps && (startpos >= pos || endpos >= pos)) {
				pbuf = new BppString(*ps);
				ps = pbuf;
			}
			pd->insert(pos, _T(' '), cnt);
			for (i = 0, j = 0; i < cnt; ++i, j += dist) (*pd)[pos + i] = (*ps)[startpos + j];
			if (pbuf) delete pbuf;
		}
		*bp = pd;
	}
	break;
	default:
		THROW(FLIB_INVALID_ARG_TYPE_DEST);
	}
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_swap(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3, 3);
	ValPtr bp = sp; bp -= argc;
	BppUInt pos1 = bp[2];
	BppUInt pos2 = bp[3];
	BppBuffer* pb = 0;
	BppCharBuffer* pcb = 0;
	size_t s = 0;
	BppValue contVal((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppVector* vec = contVal;
	if (!vec) {
		pb = contVal;
		if (!pb) {
			pcb = contVal;
			if (!pcb) THROW(FLIB_INVALID_ARG_TYPE);
			s = pcb->size();
		}
		else s = pb->size();
	}
	else s = vec->size();
	if (pos1 >= s || pos2 >= s) THROW(FLIB_INDEX_OUT_OF_RANGE);
	if (vec) {
		vec->swap(pos1, pos2);
		*bp = vec;
	}
	else if (pb) {
		pb->swap(pos1, pos2);
		*bp = pb;
	}
	else {
		pcb->swap(pos1, pos2);
		*bp = pcb;
	}
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_move(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,3);
	ValPtr bp = sp; bp -= argc;
	BppUInt pos1 = bp[2];
	BppUInt pos2 = bp[3];
	BppBuffer* pb = 0;
	BppCharBuffer* pcb = 0;
	size_t s = 0;
	BppValue contVal((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppVector* vec = contVal;
	if (!vec) {
		pb = contVal;
		if (!pb) {
			pcb = contVal;
			if (!pcb) THROW(FLIB_INVALID_ARG_TYPE);
			s = pcb->size();
		}
		else s = pb->size();
	}
	else s = vec->size();
	if (pos1 >= s || pos2 >= s) THROW (FLIB_INDEX_OUT_OF_RANGE);
	if (vec) {
		vec->move(pos1, pos2);
	   *bp = vec;
	}
	else if (pb) {
		pb->move(pos1, pos2);
		*bp = pb;
	}
	else {
		pcb->move(pos1, pos2);
		*bp = pcb;
	}
	sp = bp;
	return true;
}


// var resvec = erase(var cont, uint pos = 0, uint cnt = NPOS)
// dict resdict = erase(dictionary dict, string key)
bool BppDefaultFunctions::Bpp_erase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,3);
	ValPtr bp = sp; bp -= argc;
	BppValue contVal((const BppValue&)bp[1], (bp[1].getFlags() & BVF_CONST) != 0 ? true : false);
	BppDictionary* pdict = contVal;
	if (pdict) {
		ARGCMAX(2);
		if (argc == 2) {
			BppString* key = bp[2];
			if (key == 0)
				THROW(FLIB_INVALID_ARG_TYPE);
			pdict->erase(*key);
		}
		else
			pdict->clear();
		*bp = pdict;
	}
	size_t pos = (argc > 1) ? (BppUInt) bp[2] : (BppUInt)0;
  size_t cnt = (argc == 3) ? (size_t)(BppUInt) bp[3] : BPP_NNPOS;
	BppVector* vec = contVal;
	if (vec) {
		vec->erase(pos, cnt);
		*bp = vec;
	}
	else {
		BppCharBuffer* cb = contVal;
		if (cb) {
			cb->erase(pos, cnt);
			*bp = cb;
		}
		else {
			BppBuffer* b = contVal; 
			if (b){
				b->erase(pos, cnt);
				*bp = b;
			}
			else {
				BppString* s = contVal;
				if (s) {
					s->erase(pos, cnt);
					*bp = s;
				}
				else THROW(FLIB_INVALID_ARG_TYPE);
			}
		}
	}
	sp = bp;
	return true;
}

// vector resvec = slice(vector vec, uint from = 0, cnt = -1)
bool BppDefaultFunctions::Bpp_slice(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,3);
	ValPtr bp = sp; bp -= argc;
	size_t pos = (argc > 1) ? (size_t)(BppUInt) bp[2] : size_t(0);
	size_t cnt = (argc == 3) ? (size_t)(BppUInt) bp[3] : BPP_NNPOS;
	BppVector* vec = bp[1];
	if (vec) *bp = new BppVector(vec->slice(pos, cnt));
	else {
		BppBuffer* buf = bp[1];
		if (buf) *bp = new BppBuffer(buf->slice(pos, cnt));
		else {
			BppCharBuffer* cbuf = bp[1];
			if (cbuf) *bp = new BppCharBuffer(cbuf->slice(pos, cnt));
		   else THROW(FLIB_INVALID_ARG_TYPE);
		}
	}
	sp = bp;
	return true;
}

// void clear(var container)
bool BppDefaultFunctions::Bpp_clear(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	ValPtr bp = sp; bp -= argc;
	CHKVAR(bp[1]);
	BppVector* vec = bp[1];
	if (vec) {
		vec->clear();
	}
	else {
		BppCharBuffer* cb = bp[1];
		if (cb) cb->clear();
		else {
			BppBuffer* b = bp[1];
			if (b) b->clear();
			else {
				BppDictionary* d = bp[1];
				if (d) d->clear();
				else {
					BppString* s = bp[1];
					if (s) s->clear();
					else THROW(FLIB_INVALID_ARG_TYPE);
				}
			}
		}
	}
	bp->dispose();
	sp = bp;
	return true;
}

// uint reserve(var container, uint capacity)
bool BppDefaultFunctions::Bpp_reserve(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 2);
	ValPtr bp = sp; bp -= argc;
	if (!bp[2].isIntegralType()) THROW(FLIB_INVALID_ARG_TYPE, FLIB_ARG_CAPACITY_BR);
	CHKVAR(bp[1]);
	BppVector* vec = bp[1];
	if (vec) {
		*bp = BppUInt(vec->setCapacity((BppUInt)bp[2]));
	}
	else {
		BppBuffer* b = bp[1];
		if (b) *bp = BppUInt(b->setCapacity((BppUInt)bp[2]));
		else {
			BppCharBuffer* cb = bp[1];
			if (cb) *bp = BppUInt(cb->setCapacity((BppUInt)bp[2]));
			else THROW(FLIB_INVALID_ARG_TYPE, FLIB_ARG_CONTAINER_BR);
		}
	}
	sp = bp;
	return true;
}

//bool resize(size_t size, const BppValue& fill = BppValue()
bool BppDefaultFunctions::Bpp_resize(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(2, 3);
	ValPtr bp = sp; bp -= argc;
	if (!bp[2].isIntegralType()) THROW(FLIB_INVALID_ARG_TYPE, FLIB_ARG_SIZE_BR);
	bool ok = false;
	CHKVAR(bp[1]);
	BppVector* vec = bp[1]; 
	if (vec) {
		if (argc == 3) ok = vec->resize((BppUInt)bp[2], bp[3]);
		else ok = vec->resize((BppUInt)bp[2]);
	}
	else {
		BppBuffer* b = bp[1];
		if (b) {
			if (argc == 3) ok = b->resize((BppUInt)bp[2], bp[3]);
			else ok = b->resize((BppUInt)bp[2]);
		}
		else {
			BppCharBuffer* cb = bp[1];
			if (cb) {
				if (argc == 3) ok = cb->resize((BppUInt)bp[2], (BppChar)(BppInt)bp[3]);
				else ok = cb->resize((BppUInt)bp[2]);
			}
			else THROW(FLIB_INVALID_ARG_TYPE, FLIB_ARG_CONTAINER_BR);
		}
	}
	*bp = BppInt(ok ? 1 : 0);
	sp = bp;
	return true;
}


// bool found = dictcontains(dict,string key)
bool BppDefaultFunctions::Bpp_dictcontains(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppDictionary* dict = bp[1];
	BppString* key = bp[2];
	if ((dict == 0) || (key == 0))
		THROW(FLIB_INVALID_ARG_TYPE);
	*bp = BppValue::makeBool(dict->find(*key) != 0);
	sp = bp;
	return true;
}

// bool deleted = dicterase(dict,string key)
bool BppDefaultFunctions::Bpp_dicterase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	CHKVAR(bp[1]);
	BppDictionary* dict = bp[1];
	BppString* key = bp[2];
	if ((dict == 0) || (key == 0))
		THROW(FLIB_INVALID_ARG_TYPE);
	*bp = BppValue::makeBool(dict->erase(*key));
	sp = bp;
	return true;
}


// keyvec = dictgetkeys(dict)
bool BppDefaultFunctions::Bpp_dictgetkeys(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppDictionary* dict = *sp;
	if (dict == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	BppVector* result = new BppVector();
		for (BppDictionaryEntry* pe = dict->getFirst(); pe != 0; pe = pe->getNext())
			result->push_back(new BppString(*(pe->getKey())));
	--sp;
	*sp = result;
	return true;
}

// keyvec = dictgetvalues(dict)
bool BppDefaultFunctions::Bpp_dictgetvalues(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 1);
	BppDictionary* dict = *sp;
	unsigned char flags = sp->getFlags() & BVF_CONST;
	if (dict == 0)
		THROW(FLIB_INVALID_ARG_TYPE);
	BppVector* result = new BppVector();
	for (BppDictionaryEntry* pe = dict->getFirst(); pe != 0; pe = pe->getNext()) result->push_back(pe->getValue());
	--sp;
	*sp = result;
	sp->setFlags(flags);
	return true;
}

// v = Vec(...)

// v = T(...)
bool BppDefaultFunctions::Bpp_T(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	ValPtr bp = sp; bp -= argc;
	BppVector* v = new BppVector((unsigned)argc);
	for(int i=1;i<=argc;i++)
		v->push_back(bp[i]);
	*bp = v;
	sp = bp;
	return true;
}

// cb = CBuf(...) // argtypes: integral, buffer, charbuffer, string; other ignored
bool BppDefaultFunctions::Bpp_CBuf(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	ValPtr bp = sp; bp -= argc;
	BppCharBuffer* cb = new BppCharBuffer((unsigned)argc);
	for(int i=1;i<=argc;i++)
	{
		BppValue& v = bp[i];
		BppValueType vt = maskWeakRef(v.getType());
		switch (vt)
		{
	  case BVT_BYTE:
	  case BVT_SBYTE:
			cb->push_back((BppChar) (BppByte) v);
		 break;
	  case BVT_SHORT:
	  case BVT_USHORT:
	  case BVT_INT:
	  case BVT_UINT:
	  case BVT_LONG:
	  case BVT_ULONG:
	  case BVT_INTPTR:
			cb->push_back((BppChar) (BppUShort) v);
		 break;
		case BVT_BUFFER:
			{
				BppBuffer* b = v;
				if (b != 0)
				{
					size_t n = b->size();
					for (size_t j=0; j<n; j++)
						cb->push_back((BppChar)(*b)[j]);
				}
			}
			break;
		case BVT_STRING:
			{
				BppString* s = v;
				if (s != 0)
				{
					size_t n =  s->length();
					for (size_t j=0; j<n; j++)
						cb->push_back((BppChar)((*s)[j]));
				}
			}
			break;

		case BVT_CHARBUFFER:
			{
				BppCharBuffer* pcb = v;
				if (pcb != 0)
					(*cb) += (*pcb);
			}
			break;
		}
	}
	*bp = cb;
	sp = bp;
	return true;
}



	// basic type conversion
#ifndef DOXY
static const BppString s_bool = TN_bool;
static const BppString s_int = TN_int;
static const BppString s_buffer = TN_buffer;
static const BppString s_charbuffer = TN_charbuffer;
static const BppString s_vector = TN_vector;
static const BppString s_dictionary = TN_dictionary;
#endif
// bool result = bool(val)
bool BppDefaultFunctions::Bpp_bool(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	bp -= argc;
	bool result = v.asBool();
	BppValueType vt = maskWeakRef(v.getType());
	if (vt == BVT_OBJECT || vt == BVT_STRONGOBJECT)
	{
		BppObject*pobj = v;
		if (pobj != 0)
		{
			BppProcInfo* pi = interpreter->getProcInfo();
			BppClass* ownClass = pi ? pi->getClass() : 0;
			BppClass* pclass = pobj->getClass();
			BppDictionaryEntry* pe = pclass->findMember(ownClass, s_bool);
			if ((pe == 0) || (pe->getSymbolType() != ST_FUNCTION))
				pe = pclass->findMember(ownClass, s_int);
			if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
			{
				*(++sp) = pe->getValue();
				if (interpreter->send(1, 0))
				{
					result = sp->asBool();
				};
			}
		} // if pobj
	}

	sp = bp;
	*sp = BppValue::makeBool(result);
	return true;
}



bool BppDefaultFunctions::numcast(BppVirtualMachine* vm, int argc, ValPtr& sp, const BppChar* fname, BppValueType desttype)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp;
	const BppValue& src = *bp;
	BppValue& dest = * (--bp);
	bool ok = interpreter->doNumCast(dest, src, desttype, fname);
	sp = bp;
	 return ok;
}

bool BppDefaultFunctions::Bpp_sbyte(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("sbyte"), BVT_SBYTE);
}

bool BppDefaultFunctions::Bpp_byte(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("byte"), BVT_BYTE);
}

bool BppDefaultFunctions::Bpp_short(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("short"), BVT_SHORT);
}

bool BppDefaultFunctions::Bpp_ushort(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("ushort"), BVT_USHORT);
}

bool BppDefaultFunctions::Bpp_int(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("int"), BVT_INT);
}

bool BppDefaultFunctions::Bpp_uint(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("uint"), BVT_UINT);
}

bool BppDefaultFunctions::Bpp_long(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("long"), BVT_LONG);
}

bool BppDefaultFunctions::Bpp_ulong(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("ulong"), BVT_ULONG);
}

bool BppDefaultFunctions::Bpp_intptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("intptr"), BVT_INTPTR);
}

bool BppDefaultFunctions::Bpp_uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("intptr"), BVT_UINTPTR);
}



// result = float(numericval|string)
bool BppDefaultFunctions::Bpp_float(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("float"), BVT_FLOAT);
}

// result = double(numericval|string)
bool BppDefaultFunctions::Bpp_double(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("double"), BVT_DOUBLE);
}

bool BppDefaultFunctions::Bpp_char(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("char"), BVT_CHAR);
}

bool BppDefaultFunctions::Bpp_uchar(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("uchar"), BVT_UCHAR);
}

bool BppDefaultFunctions::Bpp_char8_t(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("char8_t"), BVT_CHAR8);
}

bool BppDefaultFunctions::Bpp_char16_t(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("char16_t"), BVT_CHAR16);
}

bool BppDefaultFunctions::Bpp_char32_t(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("char32_t"), BVT_CHAR32);
}

bool BppDefaultFunctions::Bpp_wchar_t(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("wchar_t"), BVT_WCHAR);
}

// result = string(int|double|float|string|charbuffer|object|character type)
bool BppDefaultFunctions::Bpp_string(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	const BppValue& src = *sp;
	ValPtr bp = sp;
	--bp;
	bool ok = interpreter->doStrCast(*bp, src);
	sp = bp;
	return ok;
}

// result = buffer(int|double|float|buffer|charbuffer|string|object)
bool BppDefaultFunctions::Bpp_buffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = mkWcharInternal(maskWeakRef(v.getType()));
   size_t intsize = 0;
	switch (vt)
	{
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_CHAR:
	case BVT_UCHAR:
      intsize = sizeof(BppByte);
		break;
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_CHAR16:
		{
         intsize = sizeof(BppShort);
		}
		break;
	case BVT_INT:
	case BVT_UINT:
   case BVT_FLOAT:
   case BVT_CHAR32:
   case BVT_CHAR8:
#ifndef BPP_I64
   case BVT_INTPTR:
   case BVT_UINTPTR:
#endif
      intsize = sizeof(BppInt);
		break;
#ifdef BPP_INT64
   case BVT_INTPTR:
   case BVT_UINTPTR:
#endif
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_DOUBLE:
      intsize = sizeof(double);
		break;
	case BVT_BUFFER:
		*bp = v;
		break;
	case BVT_STRING:
		{
			BppString* s = v;
			if (s != 0)
			{
				size_t bytes = (s->length()+1) * sizeof(BppChar);
				BppBuffer* b = new BppBuffer(bytes,bytes);
			memcpy(b->data(),s->c_str(),bytes);
				*bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* cb = v;
			if (cb != 0)
			{
			 size_t bytes = (cb->size()+1) * sizeof(BppChar);
				BppBuffer* b = new BppBuffer(bytes,bytes);
				BppChar c = 0;
				memcpy(b->data(),cb->data(),bytes-sizeof(BppChar));
				memcpy(b->data()+(bytes-sizeof(BppChar)),&c,sizeof(BppChar));
			  *bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
	{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_buffer);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_BUFFER || vt == BVT_BUFFERREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(FLIB_INVALID_ARG_TYPE);
		break;
	}
   if (intsize) {
      BppBuffer* b = new BppBuffer(intsize,intsize);
      *bp = b;
      BppByte* pd = (BppByte*)&v._v;
      memcpy(b->data(), pd,intsize);
   }
	sp = bp;
	return true;
}

// result = buffer(int|double|float|buffer|charbuffer|string|object)
bool BppDefaultFunctions::Bpp_charbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	if (vt <= _BVTVALMAX) {
		if (!interpreter->doStrCast(*bp, v))
		return false;
	}
	else {
		switch (vt)
		{
		case BVT_CHARBUFFER:
			*bp = v;
			break;
		case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				BppCharBuffer* b = new BppCharBuffer(s->c_str());
				*bp = b;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
		case BVT_BUFFER:
		{
			BppBuffer* b = bp[1];
			if (b != 0)
			{
				size_t size = b->size();
				BppCharBuffer* cb = new BppCharBuffer(size + 1, size + 1);
				*bp = cb;
				BppChar* pcd = cb->data();
				BppByte* pb = b->data();
				for (size_t i = 0; i < size; ++pcd, ++pb, i++)
					*pcd = (BppChar)*pb;
				*pcd = _T('\0');
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
		case BVT_OBJECT:
		case BVT_STRONGOBJECT:
		{
			BppObject* pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass, s_charbuffer);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
					*(++sp) = pe->getValue();
					if (interpreter->send(1, 0))
					{
						*bp = *sp;
						vt = bp->getType();
						if (vt == BVT_CHARBUFFER || vt == BVT_CHARBUFFERREF)
							break;
					}
				}
			} // if pobj
		}
		default:
			sp = bp;
			THROW(FLIB_INVALID_ARG_TYPE);
			break;
		}
	}
	sp = bp;
	return true;
}

// result = vector(vector|buffer|charbuffer|dictionary|object)
bool BppDefaultFunctions::Bpp_vector(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	switch (vt)
	{
	case BVT_VECTOR:
		*bp = v;
		break;
	case BVT_DICTIONARY:
		return Bpp_dictgetvalues(vm,argc,sp);
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* b = v;
			if (b)
			{
			  size_t size = b->size();
			  BppVector* vec = new BppVector(size,size ? true : false);
			  for (size_t i = 0; i < size; ++i) {
				  (*vec)[i] = (*b)[i];
				  (*vec)[i].v_exttype._type = BVT_TCHAR;
			  }
			  *bp = vec;
			}
			else *bp = new BppVector();
		}
		break;
	case BVT_BUFFER:
		{
			BppBuffer* b = v;
			if (b)
			{
			  size_t size = b->size();
			  BppVector* vec = new BppVector(size,size ? true : false);
			  for (size_t i=0;i<size;++i)
				  (*vec)[i] = (*b)[i];
				  *bp = vec;
			}
			else *bp = new BppVector();
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
	{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_vector);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_VECTOR || vt == BVT_VECTORREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(FLIB_INVALID_ARG_TYPE);
		break;
	}
	sp = bp;
	return true;
}

// result = vector(vector|dictionary|object)
bool BppDefaultFunctions::Bpp_dictionary(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	switch (vt)
	{
	case BVT_DICTIONARY:
		*bp = v;
		break;
	case BVT_VECTOR:
		{
			BppDictionary* dict = new BppDictionary();
			*bp = dict;
			BppVector* vec = v;
			if (vec)
			{
			  size_t size = vec->size();
			  for (size_t i=0;i<size;++i)
			  {
				  (*dict)[BppString::format(_T("%d"),i)] = (*vec)[i];
			  }
				  *bp = dict;
			}
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
		{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_dictionary);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_DICTIONARY || vt == BVT_DICTIONARYREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(FLIB_INVALID_ARG_TYPE);
		break;
	}
	sp = bp;
	return true;
}


	// RTTI
// name = gettypename(typeid)
bool BppDefaultFunctions::Bpp_gettypename(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	BppValue& v = bp[1];
	if (!v.isIntegralType())
		  THROW(FLIB_INVALID_ARG_TYPE);
	bool prefix = false;
	if (argc > 1)
	{
		prefix = bp[2].asBool();
	}
	BppInt vt = v;
	if ((vt<_BVTMIN)|| (vt>_BVTREFMAX))
		  THROW(FLIB_INVALID_ARG_VALUE);
	sp = bp;
	*sp = new BppString(getValueTypeName((BppValueType) vt,prefix));
	return true;
}

bool BppDefaultFunctions::Bpp_gettype(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	--sp;
	*sp = (BppInt) v.getType();
	return true;
}

bool BppDefaultFunctions::Bpp_getclass(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 1);
	BppClass* c = *sp;
	if (!c) {
		BppObject* o = *sp;
		if (o) c = o->getClass();
		if (!c) {
			BppString* s = *sp;
			if (!s)
				THROW(FLIB_INVALID_ARG_TYPE);
			BppDictionaryEntry* pe = vm->classes().find(*s);
			if (pe) {
				c = pe->getValue();
			}
			
		}
	}
	--sp;
	if (c) *sp = c;
	else sp->dispose();
	return true;
}

bool BppDefaultFunctions::Bpp_getclassname(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	BppClass* c = bp[1];
	if (!c) {
		BppObject* o = bp[1];
		if (!o)
			THROW(FLIB_INVALID_ARG_TYPE);
		c = o->getClass();
	}
	bool fullpath = false;
	if (argc == 2) fullpath = bp[2].asBool();
	if (fullpath)
		*bp = c->getClassPath();
	else
		*bp = *c->getClassName();
	bp->setFlags(BVF_CONST);
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_dynamic_cast(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppClass* pclass = bp[1];
	if (!pclass)
	{
		BppString* s = bp[1];
		if (s != 0)
		{
			BppDictionaryEntry* de = vm->classes().find(*s);
			if (de != 0)
				pclass = de->getValue();
		}
	}
	if (!pclass)
	{
		bp->dispose();
		sp = bp;
		THROW(FLIB_FIRST_ARG_NOT_CLASS);
	}
	BppObject* pobj = bp[2];
	if (!pobj)
		bp->dispose();
	else
	{
		if (pobj->getClass()->isSubtypeOf(pclass))
			*bp = bp[2];
		else
			bp->dispose();
	}
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_getsymbol(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= argc;
	BppString* pname = bp[1];
	if (!pname) THROW(FLIB_FIRST_ARG_NOT_STRING);
	bool findclass = false;
	if (argc == 2) {
		const BppValue& v = bp[2];
		if (!v.isIntegralType()) return interpreter->badType(v.getType(), BVT_INT);
		findclass = v.asBool();
	}
	const BppDictionary& table = findclass ? vm->classes() : vm->symbols();
	BppDictionaryEntry* pe = table.find(*pname);
	if (pe) *bp = pe->getValue();
	else bp->dispose();
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_isconst(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp; bp -= argc;
	*bp = BppValue::makeBool((bp[1].getFlags() & BVF_CONST) != 0);
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_mkconst(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= argc;
	bool constflag = (argc == 2) ? bp[2].asBool() : true;
	*bp = bp[1];
	if (constflag)
		bp->setFlags(bp->getFlags() | BVF_CONST);
	else
		bp->setFlags(bp->getFlags() & ~BVF_CONST);
	sp = bp;
	return true;
}


// intval = time();
bool BppDefaultFunctions::Bpp_time(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
#ifdef EMULATE_I64
	*sp =  (BppInt)time(0);
#else
	*sp = (BppLong)time(0);
#endif
	return true;
}

#ifndef CLK_TCK 
#define CLK_TCK CLOCKS_PER_SEC
#endif
// int mselapsed = timer()
bool BppDefaultFunctions::Bpp_timer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
#ifdef WINCE
	*sp = (BppUInt) GetTickCount();
#else
	*sp = (BppUInt) ((clock() * 1000L) / CLK_TCK);
#endif
	return true;
}

#ifdef _WIN32_WCE_EMULATION
  typedef double (__cdecl *mathfunc) (double);
#else
  typedef double (*mathfunc) (double);
#endif

#ifndef DOXY

static bool mfunc(mathfunc mf, BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	double d = *sp;
	if (d == DBL_MAX) {
	  if (!interpreter->doNumCast(*sp,*sp,BVT_DOUBLE, TN_double))
		  return false;
	 d = *sp;
	}
	--sp;
	*sp = mf(d);
	return true;
}
#endif

	// math functions
bool BppDefaultFunctions::Bpp_sin(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&sin,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_cos(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&cos,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_tan(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&tan,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_atan(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&atan,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_sqrt(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&sqrt,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_log(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&log,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_exp(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&exp,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_abs(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	switch (sp->getType())
	{
	case BVT_BYTE:
	case BVT_USHORT:
	case BVT_UINT:
	case BVT_ULONG:
		sp[-1] = *sp;
		break;
	case BVT_SBYTE:
		{
			BppSByte x = *sp;
			sp[-1] = (BppSByte) (x<0 ? -x : x);
		}
		break;
	case BVT_SHORT:
		{
			BppShort x = *sp;
			sp[-1] = (BppShort) (x<0 ? -x : x);
		}
		break;
	case BVT_INT:
		{
			BppInt x = *sp;
			sp[-1] = x<0 ? -x : x;
		}
		break;
	case BVT_LONG:
		{
			BppLong x = *sp;
			sp[-1] = x<BppLong(0L) ? -x : x;
		}
		break;
	case BVT_INTPTR:
		{
			BppIntPtr x = *sp;
			sp[-1] = x<BppIntPtr(0L) ? -x : x;
		}
		break;
	case BVT_FLOAT:
		{
			float x = *sp;
			sp[-1] = x<0.0f ? -x : x;
		}
		break;
	case BVT_DOUBLE:
		{
			double x = *sp;
			sp[-1] = x<0.0 ? -x : x;
		}
		break;
	default:
		 THROW(FLIB_INVALID_ARG_TYPE);
	}
	--sp;
	return true;
}

bool BppDefaultFunctions::Bpp_pow(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= 2;
	double arg1 = bp[1];
	if (arg1 == DBL_MAX)
	{
		BppValue v;
		if (!interpreter->doNumCast(v, bp[1], BVT_DOUBLE, TN_double))
				return false;
		arg1 = v;
	}
	double arg2 = bp[2];
	if (arg2 == DBL_MAX)
	{
		BppValue v = bp[2];
		if (!interpreter->doNumCast(v, bp[2], BVT_DOUBLE, TN_double))
			return false;
		arg2 = v;
	}
	*bp = (double)pow(arg1, arg2); // exp(arg2 * log(arg1));
	 sp = bp;
	 return true;
}


#endif // !DOXY



