
#include "bppinput.h"
#include "bppufunc.h"
/**
Argumant values are simply assigned to menber variables, whithout any chaecking.
@param fp Pointer to opened file (or FILE variable representing a character device).
@param autoClose Flag specifies, if fclose(fp) should be called in destructor.
*/
BppFileStreamInput::BppFileStreamInput( FILE* fp, bool autoClose /*= false*/ )
:_fp(fp),_autoClose(autoClose)
{
}

/**
Constructor tries to open file in specified mode using fopen function.
@param fileName Name (or path) of file to open
@param mode Open mode for file.
@note File opened by this constructor is always closed in destructor.
*/
BppFileStreamInput::BppFileStreamInput( const BppChar* fileName, const BppChar* mode /*= _T("rt")*/ )
:_fp(0),_autoClose(true)
{
   if (fileName && mode)
   {
      _fp = _tfopen(fileName,mode);
   }
}

BppFileStreamInput::~BppFileStreamInput()
{
   if (_autoClose && (0 != _fp))
      fclose(_fp);
}

int BppFileStreamInput::getChar()
{
    if (!_fp) return _TEOF;
    return fgetc(_fp);
}

BppString BppFileStreamInput::getString()
{
   BppString s;
   if (_fp)
   {
      BppUInt c = 0;
      while (((c = _gettc(_fp)) != _T('\n')) && c != -1 &&(!feof(_fp)))
      {
         s += (BppChar) c;
      }
   }
   return s;
}


BppCallbackInput::BppCallbackInput( BppCharInputCallback charCallback /*= 0*/,BppStringInputCallback stringCallback /*= 0*/ )
:_charCallback(charCallback),_stringCallback(stringCallback)
{
}

int BppCallbackInput::getChar()
{
   if (_charCallback) return _charCallback();
   return _TEOF;
}

BppString BppCallbackInput::getString()
{
   if (_stringCallback) return _stringCallback();
   return (const BppChar*)0;
}

#ifdef WIN32
UINT BppWinMsgInput::_stringInputMsg = RegisterWindowMessage(_T("BppStringInputMsg"));
UINT BppWinMsgInput::_charInputMsg = RegisterWindowMessage(_T("BppCharInputMsg"));


BppWinMsgInput::BppWinMsgInput( HWND hwnd )
:_hwnd(hwnd),_chrBuffer(0)
{
}

UINT BppWinMsgInput::getStringInputMsg()
{
   return _stringInputMsg;
}

BppString BppWinMsgInput::getString()
{
   return _strBuffer;
}

UINT BppWinMsgInput::getCharInputMsg()
{
   return _charInputMsg;
}

int BppWinMsgInput::getChar()
{
   return _chrBuffer;
}

void BppWinMsgInput::setChar( char chr )
{
   _chrBuffer = (int) (unsigned char) chr;
}

void BppWinMsgInput::setChar( wchar_t chr )
{
   _chrBuffer = (int) chr;
}

void BppWinMsgInput::setString( const char* str )
{
   _strBuffer = BppString(str);
}

void BppWinMsgInput::setString( const wchar_t* str )
{
   _strBuffer = BppString(str);
}
#endif
