#include "bppprint.h"
#include <stdarg.h>
#include <string.h>
#include "bppstrng.h"
#include "bppufunc.h"

int BppDummyPrinter::operator()(const BppChar*, ...)
{
   return 0;
}

/**
@param fp Pointer to already opened file descriptor.
@param autoClose d <tt>true</tt>, file <i>fp</i> will be closed by destructor.
@note This constructor may be used for output to <tt>strout</tt> or <tt>stderr</tt> too.
*/
BppFileStreamPrinter::BppFileStreamPrinter(FILE *fp, bool autoClose)
:_fp(fp), _autoClose(autoClose)
{
}

/**
@param fileName Name (including path) of file to create for output.
@param mode Open mode for autput file. Possible values are same as for <tt>fopen</tt>.
*/
BppFileStreamPrinter::BppFileStreamPrinter(const BppChar *fileName, const BppChar* mode)
:_fp(0), _autoClose(true)
{
   if (fileName && mode)
   {
      _fp = _tfopen(fileName,mode);
   }
}

/**
Dtor closes output file in required.
*/
BppFileStreamPrinter::~BppFileStreamPrinter()
{
   if (_autoClose && (_fp != 0))
      fclose(_fp);
}

int BppFileStreamPrinter::operator ()(const BppChar *format, ...)
{
   if (!(format && _fp)) return -1;
   va_list args;

   // retrieve the variable arguments
   va_start( args, format );
   int result =  _vftprintf( _fp, format, args );
   va_end(args); 
   return result;
}


BppCallbackPrinter::BppCallbackPrinter(BppPrinterCallback callback)
:_callback(callback)
{
}

int BppCallbackPrinter::operator()(const BppChar* format, ...)
{
   if (!_callback) return 0;
   if (!format) return -1;
   va_list args;

   // retrieve the variable arguments
   va_start( args, format );

   int cnt =  _vsctprintf( format, args );
   va_end(args);
   if (cnt < 1) return cnt;
   BppChar* pbuf = (BppChar*) allocmem((cnt+1) * sizeof(BppChar));
   va_start(args, format);
   _vstprintf_s(pbuf,cnt+1,format,args);
    va_end(args);   
#ifdef _UNICODE
   BppString s((const BppChar*) pbuf);
   _callback(s.a_str());
#else
   _callback(pbuf);
#endif
   freemem(pbuf);
 //  va_end(args);                           
   return cnt;
}

BppStringPrinter::BppStringPrinter() 
{
}

int BppStringPrinter::operator()(const BppChar* format, ...)
{
   if (!format) return -1;
   va_list args;

   // retrieve the variable arguments
   va_start( args, format );

   int cnt =  _vsctprintf( format, args );
   va_end(args);                           
   if (cnt < 1) return cnt;

   BppChar* pbuf = (BppChar*) allocmem((cnt+2) * sizeof(BppChar));
   va_start(args, format);
   _vstprintf_s(pbuf,cnt+1,format,args);
   va_end(args);

   _buffer += (const BppChar*) pbuf; 
   freemem(pbuf);
   return cnt;
}

const BppChar* BppStringPrinter::c_str() const
{
   return _buffer.c_str();
}

const char* BppStringPrinter::a_str() const
{
   return _buffer.a_str();
}
   
const wchar_t* BppStringPrinter::w_str() const
{
   return _buffer.w_str();
}

void BppStringPrinter::clear()
{
   _buffer.clear();
}

#ifdef WIN32

UINT BppWinMsgPrinter::_printStringMsg = RegisterWindowMessage(_T("BppPrintStringMsg"));

BppWinMsgPrinter::BppWinMsgPrinter( HWND hwnd )
:_hwnd(hwnd)
{
}

UINT BppWinMsgPrinter::getPrintStringMsg()
{
   return _printStringMsg;
}

int BppWinMsgPrinter::operator() (const BppChar* format, ... )
{
   if (!format) return -1;
   va_list args;

   // retrieve the variable arguments
   va_start( args, format );

   int cnt =  _vsctprintf( format, args );
   va_end(args);
   if (cnt < 1) return cnt;

   BppChar* pbuf = (BppChar*) allocmem((cnt+1) * sizeof(BppChar));
   va_start(args, format);

   _vstprintf_s(pbuf,cnt,format,args);
   va_end(args);                           

   BppString buffer = pbuf; 
   freemem(pbuf);
   SendMessage(_hwnd,_printStringMsg,(WPARAM) buffer.a_str(),(LPARAM) buffer.w_str());
   return cnt;
}

#endif // WIN32
