#include "bppdef.h"
#include "bppscan.h"
#include "bppstrng.h"
#include "bppcomp.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <float.h>
#include <math.h>
#include <errno.h>
#include "bppufunc.h"
#include "bppvalue.h"
#include "bppmsg.h"
#ifdef __TURBOC__
#include <dir.h>
#else
	#ifndef WINCE
	   #ifdef WIN32
		#include <direct.h>
		#else
		#include <unistd.h>
		#endif
	#endif
#endif

#define DEFAULTBUFSIZE 50
class BppCharReader
{
public:
	BppCharReader(BppScanner* scanner) : _scanner(scanner) {}
	virtual ~BppCharReader() {}
	virtual BppUInt get() =0;
	BppScanner* _scanner;
};

class BppStringCharReader : public BppCharReader
{
public:
	BppStringCharReader(const BppChar* src, BppScanner* scanner)
		:BppCharReader(scanner), _src(src)
	{
		p = (BppChar*) _src.c_str();
	}
	virtual BppUInt get()
	{

		BppChar r = *p;
		if (r != 0)
		{
			if (sizeof(BppChar) == 1 && _scanner->_srcCodePage == BPP_UTF8 && (r & 0xc0) == 0xc0) {
				BppUInt c32 = r;
				int fbytes = 1;
				if (r & 0x20) {
					++fbytes;
					if (r & 0x10) ++fbytes;

				}
				for (int i = 0; i < fbytes; ++i) {
					r = *(++p);
					if (r == 0) 
						return BppInt(_TEOF);
					c32 += r << (8 * (i + 1));
				}
				p++;
				return _scanner->charConv(c32);
			}
			else { 
				p++;
				return r;//_scanner->charConv(r);
			}
		}
		return BppInt(_TEOF);
	}
private:
	BppString _src;
	BppChar* p;
};

class BppFileCharReader : public BppCharReader
{
public:
	BppFileCharReader(FILE* src, BppScanner* scanner)
		:BppCharReader(scanner), fp(src)
	{
	}
	virtual ~BppFileCharReader()
	{
		fclose(fp);
	}

	virtual BppUInt get()
	{
		if (feof(fp)) return (BppInt)_TEOF;
		BppUInt result = 0;
		switch (_scanner->_srcCodePage) {
		case BPP_UTF16:
			result = (unsigned)getc(fp);
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp)) << 8;
			break;
		case BPP_UTF16be:
			result = ((unsigned)getc(fp)) << 8;
			if (feof(fp)) return (BppInt)_TEOF;
			result += (unsigned)getc(fp);
			break;
		case BPP_UTF32:
			result = (unsigned)getc(fp);
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp)) << 8;
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp)) << 16;
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp)) << 24;
			break;
		case BPP_UTF32be:
			result = ((unsigned)getc(fp)) << 24;
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp)) << 16;
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp)) << 8;
			if (feof(fp)) return (BppInt)_TEOF;
			result += ((unsigned)getc(fp));
			break;
		case BPP_UTF8:
		{
			unsigned  c = getc(fp);
			result = c;
			if ((c & 0xc0) == 0xc0) {
				int fbytes = 1;
				if (c & 0x20) {
					++fbytes;
					if (c & 0x10) ++fbytes;
				}
				for (int i = 0; i < fbytes; ++i) {
					if (feof(fp)) 
						return (BppInt)_TEOF;
					c = getc(fp);
					if (c == 0) 
						return BppInt(_TEOF);
					result += c << (8 * (i + 1));
				}
			}
		}
		break;
		default:
			result = getc(fp);
		}
		if ((result & _TEOF) == _TEOF)
			return (BppInt)_TEOF;
		return _scanner->charConv(result);
	}
private:
	FILE* fp;
};


/* token name table, order is same as in BppToken enum*/
static const BppChar *t_names[] = 
{
	_T("<notoken>"),
	_T("<error>"),
	_T("<eof>"),
	_T("<string>"),
	_T("<ident>"),
	_T("<long>"),
	_T("<float>"),
	_T("class"),
	_T("namespace"),
	_T("static"),
	_T("if"),
	_T("else"),
	_T("while"),
	_T("return"),
	_T("for"),
	_T("break"),
	_T("continue"),
	_T("do"),
	_T("switch"),
	_T("case"),
	_T("default"),
	_T("new"),
	_T("delete"),
	_T("try"),
	_T("catch"),
	_T("throw"),
	_T("var"),
	_T("const"),
	_T("mutable"),
	_T("function"),
	_T("operator"),
	_T("eval"),
	_T("public"),
	_T("protected"),
	_T("private"),
	_T("TRON"),
	_T("TROFF"),
	_T("TRSTEP"),
	_T("null"),
	_T("void"),
	_T("auto"),
	_T("yield"),
	_T("<="),
	_T("=="),
	_T("!="),
	_T(">="),
	_T("<<"),
	_T(">>"),
	_T("&&"),
	_T("||"),
	_T("++"),
	_T("--"),
	_T("+="),
	_T("-="),
	_T("*="),
	_T("/="),
	_T("%="),
	_T("&="),
	_T("|="),
	_T("^="),
	_T("~="),
	_T("<<="),
	_T(">>="),
	_T("::"),
	_T("->"),
	_T("..."),
	_T("litexp"),
	_T("##"),
	_T("<any>"),
	NULL
};

/* keyword table, alphabetical ordered */
struct KTTYPE { const BppChar* kt_keyword; BppToken kt_token;};
static KTTYPE ktab[] = 
{
	{ _T("TROFF"),    T_TROFF    },
	{ _T("TRON"),     T_TRON     },
	{ _T("TRSTEP"),   T_TRSTEP   },
	{ _T("auto"),     T_AUTO      },
	{ _T("break"),	   T_BREAK    },
	{ _T("case"),     T_CASE     },
	{ _T("catch"), 	T_CATCH	  },
	{ _T("class"),	   T_CLASS    },
	{ _T("const"),	   T_CONST 	  },
	{ _T("continue"), T_CONTINUE },
	{ _T("default"),  T_DEFAULT  },
	{ _T("delete"),	T_DELETE	  },
	{ _T("do"),       T_DO       },
	{ _T("else"),	   T_ELSE     },
	{ _T("for"),	   T_FOR      },
	{ _T("function"),	T_FUNCTION },
	{ _T("if"),		   T_IF       },
	{ _T("litexp"),     T_LITEXPR }, 
	{ _T("mutable"),     T_MUTABLE },
	{ _T("namespace"), T_NAMESPACE },
	{ _T("new"),	   T_NEW      },
	{ _T("null"),     T_NIL      },
	{ _T("operator"),	T_OPERATOR },
	{ _T("private"),	T_PRIVATE },
	{ _T("protected"), T_PROTECTED  },
	{ _T("public"),	T_PUBLIC	  },
	{ _T("return"),	T_RETURN	  },
	{ _T("static"),	T_STATIC	  },
	{ _T("switch"),   T_SWITCH   },
	{ _T("throw"),	   T_THROW	  },
	{ _T("try"),	   T_TRY  	  },
	{ _T("var"),	   T_VAR 	  },
	{ _T("void"),	   T_VOID      },
	{ _T("while"),	   T_WHILE    },
	{ _T("yield"),	   T_YIELD   }
};

static const size_t ktabsize = sizeof(ktab) / sizeof(KTTYPE);

static bool chFileDir(const BppChar* filePath)
{
	const BppChar* dirpath = getFileDir(filePath,false);
	 if (!dirpath) return false;
	 return (_tchdir(dirpath) == 0);
}

BppSavedToken::~BppSavedToken() { if (next) delete next; }

BppSavedToken* BppSavedToken::clone() const {
	BppSavedToken* tkn = new BppSavedToken(*this);
	if (next) tkn->next = next->clone();
	return tkn;
}

BppCompiler* BppScanner::_notifyCompiler = 0;

BppScanner::BppScanner(FILE* fp, const BppChar* srcName, long lnum, unsigned lpos)
:_dblBuffer(0),_numType(BVT_NULL),_longBuffer(BppLong(0)),_tokenStack(0),_token(T_NOTOKEN),_savedCh(0),_lastCh(0), _lnum(lnum),
_linepos(lpos),_line(DEFAULTBUFSIZE),_stringBuffer(DEFAULTBUFSIZE),_srcName(srcName),_next(0), _srcCodePage(0),
_internalCodePage(0)
{
	_charBuffer[0] = 0;
	_charBuffer[1] = 0;
#ifdef __MSDOS__
	_internalCodePage = getActiveCp();
#else
	#ifdef _UNICODE 
		if (sizeof(BppChar) == 4)
			_internalCodePage = BPP_UTF32;
		else
			_internalCodePage = BPP_UTF16;
	#else 
		_internalCodePage = getActiveCp();
	#endif
#endif
	_srcCodePage = detectFileCp(fp);
	_reader = new BppFileCharReader(fp,this);
	_stringBuffer.addRef();
	_callerDir.addRef();
	BppChar* pdir = _tgetcwd(NULL,0);
	if (pdir)
	{
		_callerDir = pdir;
		free/*mem*/(pdir);
	}
	chFileDir(srcName);
}

BppScanner::BppScanner(const BppChar* src, const BppChar* srcName, long lnum, unsigned lpos)
:_dblBuffer(0),_numType(BVT_NULL),_longBuffer(BppLong(0)),_tokenStack(0),_token(T_NOTOKEN),_savedCh(0),_lastCh(0), _lnum(lnum),
_linepos(lpos),_line(DEFAULTBUFSIZE),_stringBuffer(DEFAULTBUFSIZE),_srcName(srcName), _next(0),_srcCodePage(0),
_internalCodePage(0)
{
	_charBuffer[0] = 0;
	_charBuffer[1] = 0;
#ifdef __MSDOS__
	_internalCodePage = getActiveCp();
	_srcCodePage = _internalCodePage;
#else
	#ifdef _UNICODE 
		if (sizeof(BppChar) == 4)
			_internalCodePage = BPP_UTF32;
		else
			_internalCodePage = BPP_UTF16;
	#else 
		_internalCodePage = getActiveCp();
    #endif
	#ifdef LINUX
		_srcCodePage = DEF_CP;
	#else
		_srcCodePage = getActiveCp();
	#endif
#endif
	_reader = new BppStringCharReader(src,this);
	_stringBuffer.addRef();
	_callerDir.addRef();
	
	BppChar* pdir = _tgetcwd(NULL,0);
	if (pdir)
	{
		_callerDir = pdir;
		freemem(pdir);
	}
}

void BppScanner::push(FILE* fp, const BppChar* srcName)
{
	if (_next != 0)
		_next->push(fp,srcName);
	else
		_next = new BppScanner(fp,srcName);
}

void BppScanner::push(const BppChar* src, const BppChar* srcName)
{
	if (_next != 0)
		_next->push(src,srcName);
	else
		_next = new BppScanner(src,srcName);
}

BppScanner::~BppScanner()
{
	if (_next != 0)
		delete _next;
	_next = 0;
	if (_callerDir.length() > 0)
		_tchdir(_callerDir.c_str()); 
	if (_notifyCompiler) _notifyCompiler->onScannerDelete(this);
	delete _reader;
	delete _tokenStack;
}

bool BppScanner::back(bool append, BppSavedToken** externalStack)
{
	if (_next != 0)
		return _next->back(append,externalStack);
	if (_token == T_NOTOKEN) return false;
	BppSavedToken*& tokenStack = externalStack ? *externalStack : _tokenStack;
	BppSavedToken* tmp = new BppSavedToken();
	tmp->ttype = _token;
	if (_token == T_ANY) tmp->charBuffer = _charBuffer[0];
	if (_token == T_NUMBER || _token == T_FNUMBER) {
		tmp->dblBuffer = _dblBuffer;
		tmp->intnumBuffer = _longBuffer;
	    tmp->numType = _numType;
	}
	if (_token <= T_AUTO) tmp->stringBuffer = _stringBuffer.data();
	//tmp->stringBuffer.push_back(_T('\0'));
	//tmp->stringBuffer.adjustSize();
	//tmp->stringBuffer.setCapacity(tmp->stringBuffer.size() + 1);
	 if (!tokenStack)
	  tokenStack = tmp;
	else
	{
		if (append)
		{
			BppSavedToken* p = tokenStack;
			while (p->next) p=p->next;
			p->next = tmp;
		}
		else
		{
			tmp->next = tokenStack;
			_tokenStack = tmp;
		}
	}
	return true;
}

BppToken BppScanner::getToken(int& errCnt, bool ignoreBuffer, bool skipEOL)
{
	if (_next != 0)
	{
		BppToken result = _next->getToken(errCnt,ignoreBuffer, skipEOL );
		if (result != T_EOF) return result;
		delete _next;
		_next = 0;
	}
	if ((!ignoreBuffer) && _tokenStack)
	{
		BppSavedToken* p = _tokenStack;
		_tokenStack = p->next;
		p->next = 0;
		_token = p->ttype;
		_charBuffer[0] = p->charBuffer;
		_dblBuffer =  p->dblBuffer;
		_longBuffer = p->intnumBuffer;
		_stringBuffer = p->stringBuffer;
		_numType = p->numType;
		delete p;
	}
	else
		_token = readToken(skipEOL);
	if (_charBuffer[0] == _T('#') && _token == T_ANY && _notifyCompiler) {
		if (_notifyCompiler->notifyPreproc(errCnt)) {
			return getToken(errCnt, ignoreBuffer, skipEOL);
		}
	}
	if (_token == T_IDENTIFIER && _notifyCompiler && _notifyCompiler->notifyToken(_token,errCnt))
		return getToken(errCnt, ignoreBuffer, skipEOL);
	return _token;
} 

const BppChar* BppScanner::getSourceName() const
{
	if (_next != 0)
		return _next->getSourceName();
	return _srcName.c_str();
}

long BppScanner::getLineNumber() const
{
	if (_next != 0)
		return _next->getLineNumber();
	return _lnum;
}

unsigned BppScanner::getLinePos() const
{
	if (_next != 0)
		return _next->getLinePos();
	return _linepos;
}

BppToken BppScanner::readToken(bool skipEOL)
{
	BppChar ch,ch2;

	/* check the next character */
	while (1)
	{
		switch (ch = skipSpaces(skipEOL)) 
		{
		case _TEOF:	return (T_EOF);
		case _T('"'):	return (readString(skipEOL));
		case _T('\''):	return (readChar());
		case _T('<'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_LE);
			case _T('<'):
				switch (ch = get())
				{
				case _T('='):
					return (T_SHLEQ);
				default:
					_savedCh = ch;
					return (T_SHL);
				}
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('<');
				return (T_ANY);
			}
		case _T('='):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_EQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('=');
				return (T_ANY);
			}
		case _T('!'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_NE);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('!');
				return (T_ANY);
			}
		case _T('>'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_GE);
			case _T('>'):
				switch (ch = get())
				{
				case _T('='):
					return (T_SHREQ);
				default:
					_savedCh = ch;
					return (T_SHR);
				}
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('>');
				return (T_ANY);
			}
		case _T('&'):	
			switch (ch = get()) 
			{
			case _T('&'):
				return (T_AND);
			case _T('='):
				return (T_BANDEQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('&');
				return (T_ANY);
			}
		case _T('|'):	
			switch (ch = get()) 
			{
			case _T('|'):
				return (T_OR);
			case _T('='):
				return (T_BOREQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('|');
				return (T_ANY);
			}
		case _T('^'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_XOREQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('^');
				return (T_ANY);
			}
		case _T('~'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_BNOTEQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('~');
				return (T_ANY);
			}
		case _T('+'):	
			switch (ch = get()) 
			{
			case _T('+'):
				return (T_INC);
			case _T('='):
				return (T_ADDEQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('+');
				return (T_ANY);
			}
		case _T('-'):	
			switch (ch = get()) 
			{
			case _T('-'):
				return (T_DEC);
			case _T('='):
				return (T_SUBEQ);
			case _T('>'):
				return (T_MEMREF);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('-');
				return (T_ANY);
			}
		case _T('*'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_MULEQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('*');
				return (T_ANY);
			}
		case _T('%'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_REMEQ);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('%');
				return (T_ANY);
			}
		case _T('@'):
			while ((ch = get()) != _TEOF)
				if (ch == _T('\n')) {
					_savedCh = ch;
					break;
				}
			break;
		case _T('#'):
			if ((ch=get()) == _T('!'))
			{
				while ((ch = get()) != _TEOF)
					if (ch == _T('\n')) {
						_savedCh = ch;
						break;
					}
				break;
			}
			_savedCh = ch; 
			_charBuffer[0] = _T('#');
			if (ch == _T('(')) return T_LITEXPR;
			if (ch == _T('#')) {
				_savedCh = _T('\0');
				return T_CONCAT;
			}
			return T_ANY;
		case _T('/'):	
			switch (ch = get()) 
			{
			case _T('='):
				return (T_DIVEQ);
			case _T('/'):
				while ((ch = get()) != EOF)
					if (ch == _T('\n')) {
						_savedCh = ch;
						break;
					}
				break;
			case _T('*'):
				ch = ch2 = _TEOF;
				for (; (ch2 = get()) != _TEOF; ch = ch2)
					if (ch == _T('*') && ch2 == _T('/'))
						break;
				break;
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('/');
				return (T_ANY);
			}
			break;
		case _T(':'):	
			switch (ch = get()) 
			{
			case _T(':'):
				return (T_CC);
			default:
				_savedCh = ch;
				_charBuffer[0] = _T(':');
				return (T_ANY);
			}
		case _T('.'): 
			switch (ch = get())
			{
			case _T('.'):
				if ((ch = get())!= _T('.')) return parseError(SCN_INVALID_TOKEN_PTPT);
				return T_ELLIPSE;
			default:
				_savedCh = ch;
				_charBuffer[0] = _T('.');
				return (T_ANY);
			}
		default:	
			if (_istdigit((BppUChar) ch))
				return (readNumber(ch));
			else if (isIdChar(ch))
			{
			 if (ch == _T('R'))
			 {
				 if ((ch = get()) == _T('"')) // read raw string
				 {
							BppString endtoken = _T(")");
					 while (isIdChar(ch = get())) endtoken += ch;
					 endtoken += _T('"');
					 size_t lend = endtoken.length();
					 if (ch != _T('(')) 
						 return parseError(SCN_INVALID_RAW_STRINGLIT);
					 _stringBuffer.clear();
					 size_t cnt = 0;
					 const BppChar* pend = endtoken.c_str();
					 while ((ch = get()) != _TEOF)
					 {
						 _stringBuffer.push_back(ch);
						 if (++cnt < lend) continue;
						 if (_tcsncmp(_stringBuffer.data()+(cnt-lend),pend,lend) == 0)
						 {
									  _stringBuffer[cnt-lend] = 0;
							 _stringBuffer.adjustSize();
							 return (T_STRING);
						 }
					 }
					 return parseError(SCN_EOF_IN_RAW_STRINGLIT);
				 }
				 else
				 {
					 _savedCh = ch;
					 ch = _T('R');
				 }
			 }
				BppToken t = readIdentifier(ch);
				return t;
			}
			else {
				_charBuffer[0] = ch;
				return (T_ANY);
			}
		}
	}
}

const BppChar* BppScanner::getTokenName(BppToken tkn) const
{
	if (_next != 0)
		return _next->getTokenName(tkn);

	if ((tkn >= T_NOTOKEN) && (tkn < T_ANY))
		return t_names[tkn];
	else if (tkn == T_ANY)
	{
		if (_charBuffer[0] == 0) return SCN_SPECIAL_CHARACTER;
		return _charBuffer;
	}
	return SCN_INVALID_TOKEN;
}

const BppChar* BppScanner::getTokenString() const
{
	if (_next != 0)
		return _next->getTokenString();
	return (const BppChar*) _stringBuffer.data();
}

double BppScanner::getDouble() const
{
	if (_next != 0)
		return _next->getDouble();
	return _dblBuffer;
}

BppSByte BppScanner::getSByte() const
{
	if (_next != 0)
		return _next->getSByte();
	return _sbyteBuffer;
}

BppByte BppScanner::getByte() const
{
	if (_next != 0)
		return _next->getByte();
	return _byteBuffer;
}

BppShort BppScanner::getShort() const
{
	if (_next != 0)
		return _next->getShort();
	return _shortBuffer;
}

BppUShort BppScanner::getUShort() const
{
	if (_next != 0)
		return _next->getUShort();
	return _ushortBuffer;
}

BppInt BppScanner::getLong() const
{
	if (_next != 0)
		return _next->getLong();
	return _intBuffer;
}

BppUInt BppScanner::getULong() const
{
	if (_next != 0)
		return _next->getULong();
	return _uintBuffer;
}

BppLong BppScanner::getLLong() const
{
	if (_next != 0)
		return _next->getLLong();
	return _longBuffer;
}

BppULong BppScanner::getULLong() const
{
	if (_next != 0)
		return _next->getULLong();
	return _ulongBuffer;
}

BppValue BppScanner::getNumberValue() const
{
	BppValueType nt = mkWcharInternal(getNumberType());
	if ((nt & BIST_CHAR) != 0) {
		return BppValue::makeChar((BppUInt)getLong(), nt);
	}
	switch (getNumberType())
	{
	case BVT_SBYTE:
		return getSByte();
	case BVT_BYTE:
		return getByte();
	case BVT_SHORT:
		return getShort();
	case BVT_USHORT:
		return getUShort();
	case BVT_UINT:
		return getULong();
	case BVT_LONG:
		return getLLong();
	case BVT_ULONG:
		return getULLong();
	case BVT_FLOAT:
		return (float) getDouble();
	case BVT_DOUBLE:
		return getDouble();
	}
	return getLong();
}

BppValueType BppScanner::getNumberType() const 
{ 
	if (_next != 0)
		return _next->getNumberType();
	return _numType; 
}

BppChar BppScanner::peekChar()
{
	if (_next != 0)
		return _next->peekChar();
	if (_savedCh == _T('\0')) _savedCh = get();
	return _savedCh;
}

BppUInt BppScanner::charConv(BppUInt src) const
{
	if (_next != 0)
		return _next->charConv(src);
	if (_srcCodePage == _internalCodePage) return src;
	switch (_internalCodePage) {
#ifndef __MSDOS__
	case BPP_UTF16:
		switch (_srcCodePage) {
		case BPP_UTF32:
			return c32_to_c16(src);
		case BPP_UTF8:
			return u8_to_c16(src);
		default:
			return char_to_c16(src, _srcCodePage);
		}
	case BPP_UTF32:
		switch (_srcCodePage) {
		case BPP_UTF16:
			return c16_to_c32(src);
		case BPP_UTF8:
			return u8_to_c32(src);
		default:
			return char_to_c32(src, _srcCodePage);
		}
#endif
	case BPP_UTF8:
		switch (_srcCodePage) {
		case BPP_UTF16:
			return c16_to_u8(src);
		case BPP_UTF32:
			return c32_to_u8(src);
		default:
			return char_to_u8(src, _srcCodePage);
		}
	default:
		switch (_srcCodePage) {
		case BPP_UTF16:
			return c16_to_char(src, _internalCodePage);
		case BPP_UTF32:
			return c32_to_char(src, _internalCodePage);
		case BPP_UTF8:
			return u8_to_char(src, _internalCodePage);
		}
	}
	return c16_to_char(char_to_c16(src,_srcCodePage), _internalCodePage);
}

BppChar BppScanner::getChar() const
{
	if (_next != 0)
		return _next->getChar();
	return _charBuffer[0];
}

BppChar BppScanner::get()
{
	if (_next != 0)
		return _next->get();
	BppChar ch;

	/* check for a lookahead character */
	if ((ch = _savedCh) != _T('\0'))
		_savedCh = _T('\0');

	/* check for a buffered character */
	else {
		while ((ch = _line[_linepos++]) == _T('\0')) 
		{
			/* check for being at the end of file */
			if (_lastCh == _TEOF)
				return (_lastCh);

			/* read the next line */
			_line.clear();
			_linepos=0 ;
			while ((_lastCh = _reader->get()) != _TEOF && _lastCh != _T('\n'))
			{
				_line.push_back(_lastCh);
				_linepos++;
			}
			_line.push_back(_T('\n'));
			_line.push_back(_T('\0'));
			_linepos = 0;
			++_lnum;
		}
	}

	/* return the current character */
	return (ch);
}


BppChar BppScanner::skipSpaces(bool skipEOL)
{
	BppUChar ch;
	while ((ch = get()) != _T('\0') && _istspace(ch) && (skipEOL || ch !=_T('\n')));
	if (ch == _T('\\'))
	{
		BppChar c = get();
		if (c == _T('\n')) return skipSpaces(skipEOL);
		_savedCh = c;
	}
	return (ch);
}

BppToken BppScanner::readString(bool skipEOL)
{
	BppChar ch;
	bool bslash = false;
	_stringBuffer.clear();
	/* get the string */
	while (1)
	{
		ch = get();
		_savedCh = ch;
		bslash = (ch == _T('\\'));
		ch = readLiteralChar();
		if ((ch == _TEOF) || ((!bslash) && (ch == _T('"'))))
		{
			if (ch == _T('"'))
			{
				ch = skipSpaces(skipEOL);
				if (ch == _T('"')) continue;
				_savedCh = ch;
			}
			break;
		}
		if (ch == 0)
		{
			BppCharBuffer b = _stringBuffer;
			BppValueType chtype = _savedCh == _T('u') ? BVT_CHAR16 : 
				(_savedCh == _T('U') ? BVT_CHAR32 : BVT_TCHAR);
		 if (chtype != BVT_TCHAR || _savedCh == _T('x') || _istdigit(_savedCh))
		 {
			 ch = _T('0');
			 if (_savedCh == _T('0')) _savedCh = 0;
			 else if (chtype != BVT_TCHAR) _savedCh = _T('x');
		 }
		 else {
			 ch = _savedCh;
			 _savedCh = 0;
		 }
			if (readNumber(ch,true) != T_NUMBER)
				return parseError(SCN_NUMBERLIT_EXPECTED);
			switch (chtype) {
			case BVT_CHAR16:
				ch = cbufToBppChar(_intBuffer, 2);
				break;
			case BVT_CHAR32:
				ch = cbufToBppChar(_intBuffer, 4);
				break;
			default:
				ch = (BppChar)_intBuffer;
			}
			_stringBuffer = b;
		}
		_stringBuffer.push_back(ch);
	}
	if (ch == _TEOF) _savedCh = ch;
	_stringBuffer.push_back(_T('\0'));
	_stringBuffer.adjustSize();
	return (T_STRING);
}

/// character constant literal
BppToken BppScanner::readChar()
{
	_charBuffer[0] = readLiteralChar();
	if (_charBuffer[0] == 0)
	{
		 BppChar ch = _savedCh;  
		 BppValueType  chtype = _savedCh == _T('u') ? BVT_CHAR16 :
			 (_savedCh == _T('U') ? BVT_CHAR32 : BVT_TCHAR);
		 if (chtype != BVT_TCHAR || _savedCh == _T('x') || _istdigit(_savedCh)) {
			ch = _T('0');
			if (_savedCh == _T('0')) _savedCh = 0;
			else if (chtype != BVT_TCHAR) _savedCh = _T('x');
		}
		else {
			  _savedCh = 0;
		}
		 if (readNumber(ch,true) != T_NUMBER)
			 return parseError(SCN_NUMBERLIT_EXPECTED);
		 _numType = chtype;
	}
	else
	{
		_numType = BVT_TCHAR;
		_intBuffer = _charBuffer[0];
	}
	if (get() != _T('\''))
		return parseError(SCN_CLOSING_SINGLEQUOTE_EXPECTED);
	return (T_NUMBER);
}

BppChar BppScanner::readLiteralChar()
{
	BppChar ch = get();
	if (ch == _T('\\'))
	{
		switch (ch = get())
		{
		case _T('n'):  ch = _T('\n'); break;
		case _T('r'):  ch = _T('\r'); break;
		case _T('t'):  ch = _T('\t'); break;
		case _T('v'):  ch = _T('\v'); break;
		case _T('f'):  ch = _T('\f'); break;
		case _T('a'):  ch = _T('\a'); break;
		case _T('b'):  ch = _T('\b'); break;
		case _T('\n'): return readLiteralChar();
		case _T('x'):
		case _T('u'):
		case _T('U'):
			_savedCh = ch; ch = 0; break;
		case _TEOF:  ch = _T('\\'); _savedCh = _TEOF; break;
		default:
			if (ch >= _T('0') && ch <= _T('7'))
			{
				_savedCh = ch;
				ch = 0;
			}
			break;
		}
	}
	return (ch);
}

BppToken BppScanner::readIdentifier(BppChar firstChar)
{
	BppChar ch;
	/* get the identifier */
	_stringBuffer.clear();
	_stringBuffer.push_back(firstChar);
	while ((ch = get()) != _TEOF && isIdChar(ch))
		_stringBuffer.push_back(ch);
	_savedCh = ch;
	_stringBuffer.push_back(_T('\0'));
	_stringBuffer.adjustSize();

	/* check to see if it is a keyword */
	const BppChar* pt = (const BppChar*)_stringBuffer.data();
	BppToken kt = getKeywordTypeFromString(pt);
	if (kt!= T_NOTOKEN) return kt;
	if (_tcscmp(pt, _T("__FILE__")) == 0)
	{
		_stringBuffer = _srcName.c_str(); //.clear();
		//size_t n = _srcName.length();
		//pt = _srcName.c_str();
		//for (size_t i = 0; i < n; ++i)
		//	_stringBuffer.push_back(pt[i]);
		//_stringBuffer.push_back(_T('\0'));
		//_stringBuffer.adjustSize();
		return T_STRING;
	}
	if (_tcscmp(pt, _T("__func__")) == 0 || _tcscmp(pt, _T("__FUNCTION__")) == 0)
	{
		_stringBuffer = _funcName.c_str(); // .clear();
		return T_STRING;
	}
	if (_tcscmp(pt, _T("__LINE__")) == 0)
	{
		_intBuffer = _lnum;
		_dblBuffer = 0;
		BppString nstr = BppString::format(_T("%d"), _lnum);
		_stringBuffer = nstr.c_str(); // .clear();
		return T_NUMBER;
	}
	return (T_IDENTIFIER);
}
#ifdef EMULATE_I64
	#define _ULONGBUFFER (*((BppULong*)&_ulongBuffer))
	#define _LONGBUFFER (*((BppLong*)&_longBuffer))
#else 
	#define _ULONGBUFFER _ulongBuffer
	#define _LONGBUFFER _longBuffer
#endif

BppToken BppScanner::readNumber(BppChar firstChar, bool isCharcode)
{
	bool xnum = false;
	bool bnum = false;
	_numType = BVT_NULL;
	BppUChar ch;
	BppToken ttype = T_NUMBER;
	/* get the number */
	_stringBuffer.clear();
	_stringBuffer.push_back(firstChar);
	_intBuffer = 0;
	_dblBuffer = 0;
	BppChar lastChar = (BppChar)_totlower(firstChar); // stores character read before
	while ((ch = get()) != _TEOF)
	{
		if ((!isCharcode) && ch == _T('\''))
		{
			// if char before is a valid digit, ignore apostroph
			if (!(_istdigit(lastChar) || (xnum && _istxdigit(lastChar)))) break;
			lastChar = ch;
			continue;
		}
		lastChar = (BppChar)_totlower(ch); // save current char (lower case)
		if (!_istdigit(ch))  // current character is not a digit
		{
			if (!(xnum || bnum)) // if not binary nor hex number detected
			{
				// check for floating point number
				if (ch == _T('.'))
				{
					if (ttype == T_FNUMBER)  // only one decimal point allowed
						break;
					ttype = T_FNUMBER;
				}
				else if (lastChar == _T('e')) // exponent
				{
					ttype = T_FNUMBER;
					_stringBuffer.push_back(ch);
					lastChar = (BppChar)_totlower(ch = get());
					if (!(_istdigit(ch) || (ch == _T('-')) || (ch == _T('+'))))
						break; // exponent must start with digit or sign
				}
				else // check for hex or binary prefixes
				{
					if ((lastChar == _T('x')) && (_stringBuffer.size() < 3 && (_stringBuffer[_stringBuffer.size() - 1] == _T('0'))))
						xnum = true;
					else if ((lastChar == _T('b')) && (_stringBuffer.size() < 3 && (_stringBuffer[_stringBuffer.size() - 1] == _T('0'))))
						bnum = true;
					else
						break;
				}
			}
			else if (bnum && (ch != '0') && (ch != '1')) // check valid binary digits
				break;
			else if (!_istxdigit(ch))
				break;
		}
		_stringBuffer.push_back(ch);
	}
	_savedCh = ch;
	_stringBuffer.push_back(_T('\0'));
	_stringBuffer.adjustSize();
	int bytes = -1; // bytes needed to store number
	bool isU = false; // assume unsigned not specified
	bool isS = false; // assume signed not specified
	if (!isCharcode) // parse suffixes if not character code
	{
		// number must not end with quotation
		if (ch == _T('\''))  return parseError(SCN_INVALID_QUOTED_NUMBERLIT);

		// check suffixes
		if (lastChar == _T('l'))  bytes = 8;
		if (ttype == T_NUMBER) {
			switch (lastChar) {
			case _T('u'):
				isU = true;
				break;
			case _T('s'):
				isS = true;
				break;
			}
		}
		else if (ttype == T_FNUMBER && lastChar == _T('f')) bytes = 4;
		if (isU || isS || bytes >= 0)
		{
			get(); // skip buffered char
			if (ttype == T_NUMBER) // first suffix char was l or u
			{
				lastChar = (BppChar)_totlower(ch = get());
				if (bytes == 8) // first suffix was l
				{
					if (_tcschr(_T("01248l"), lastChar)) {
						bytes = lastChar == _T('l') ? 8 : lastChar - _T('0');
						lastChar = (BppChar)_totlower(ch = get());
					}
				}
				if (bytes >= 0) {
					switch (lastChar) {
					case _T('u'):
						isU = true;
						break;
					case _T('s'):
						isS = true;
						break;
					default:
						_savedCh = ch; // push char
					}
				}
				else if (isU || isS) // first suffix was u or s
				{
					if (lastChar == 'l') {
						bytes = 8;
						lastChar = (BppChar)_totlower(ch = get()); // read size in bytes if specified
						if (_tcschr(_T("01248l"), lastChar)) {
							bytes = lastChar == _T('l') ? 8 : lastChar - _T('0');
						}
						else _savedCh = ch; // push char
					}
					else _savedCh = ch; // push char
				}
			}
		}
	}
	// convert to number
	static const BppULong uint8max_i64 = BppULong((unsigned long)0xFFUL);
	static const BppULong int8max_i64 = BppULong((unsigned long)0x7FUL);
	static const BppULong uint16max_i64 = BppULong((unsigned long)0xFFFFUL);
	static const BppULong int16max_i64 = BppULong((unsigned long)0x7FFFUL);
	static const BppULong uint32max_i64 = BppULong((unsigned long)0xFFFFFFFFUL);
	static const BppULong int32max_i64 = BppULong((unsigned long)0x7FFFFFFFUL);
	static const BppLong zero = BppLong(0);
	BppChar* endp = 0;
	if (ttype == T_NUMBER)
	{
		if (bnum)
			_ulongBuffer = _tcstoui64(_stringBuffer.data() + 2, &endp, 2);
		else
			_ulongBuffer = _tcstoui64(_stringBuffer.data(), &endp, 0);
		_numType = BVT_LONG;
		bool rangeerr = false;
#ifndef EMULATE_I64
	#ifndef WINCE
			rangeerr = (errno == ERANGE);
	#endif
#endif
		if (!rangeerr) {
			if (isS && firstChar != _T('0')) {
				switch (bytes) {
				case 1:
					rangeerr = _ULONGBUFFER > int8max_i64;
					break;
				case 2:
					rangeerr = _ULONGBUFFER > int16max_i64;
					break;
				case 4:
					rangeerr = _ULONGBUFFER > int32max_i64;
					break;
				default:
					rangeerr = (_LONGBUFFER < zero);
				}
			}
			else {
				switch (bytes) {
				case 1:
					rangeerr = _ULONGBUFFER > uint8max_i64;
					break;
				case 2:
					rangeerr = _ULONGBUFFER > uint16max_i64;
					break;
				case 4:
					rangeerr = _ULONGBUFFER > uint32max_i64;
					break;
				}
			}
		}
		if (rangeerr) return parseError(SCN_INTLIT_OUT_OF_RANGE);
		if (bytes == 0) {
			bytes = 8;
			if (_ULONGBUFFER <= (isS ? int8max_i64 : uint8max_i64)) bytes = 1;
			else if (_ULONGBUFFER <= (isS ? int16max_i64 : uint16max_i64)) bytes = 2;
			else if (_ULONGBUFFER <= (isS ? int32max_i64 : uint32max_i64)) bytes = 4;
		}
		if (isU)
		{
			switch (bytes) {
			case 1:
				_numType = BVT_BYTE;
				break;
			case 2:
				_numType = BVT_USHORT;
				break;
			case 4:
				_numType = BVT_UINT;
				break;
			case 8:
				_numType = BVT_ULONG;
				break;
			default:
				if (_ULONGBUFFER <= uint32max_i64)
					_numType = BVT_UINT;
				else
					_numType = BVT_ULONG;
			}
		}
		else if (isS) {
			switch (bytes) {
			case 1:
				_numType = BVT_SBYTE;
				break;
			case 2:
				_numType = BVT_SHORT;
				break;
			case 4:
				_numType = BVT_INT;
				break;
			case 8:
				_numType = BVT_LONG;
				break;
			default:
				if (_ULONGBUFFER <= int32max_i64)
					_numType = BVT_INT;
			}
		}
		else {
			switch (bytes) {
			case 1:
				_numType = _ULONGBUFFER <= int8max_i64 ? BVT_SBYTE : BVT_BYTE;
				break;
			case 2:
				_numType = _ULONGBUFFER <= int16max_i64 ? BVT_SHORT : BVT_USHORT;
				break;
			case 4:
				_numType = _ULONGBUFFER <= int32max_i64 ? BVT_INT : BVT_UINT;
				break;
			case 8:
				_numType = (_LONGBUFFER < zero) ? BVT_ULONG : BVT_LONG;
				break;
			default:
				if (_ULONGBUFFER <= int32max_i64)
					_numType = BVT_INT;
				else _numType = (_LONGBUFFER < zero) ? BVT_ULONG : BVT_LONG;
			}
		}
	}
	else // FNUMBER
	{
		_dblBuffer = _tcstod(_stringBuffer.data(), &endp);
		_numType = (bytes == 4) ? BVT_FLOAT : BVT_DOUBLE;
		if (
#ifndef __OLDCPP__
	#ifndef WINCE
			(errno == ERANGE) ||
	#endif
#endif
			(bytes == 4 && (fabs(_dblBuffer) > FLT_MAX)))
			return parseError(SCN_REALLIT_OUT_OF_RANGE);
	}
	if (endp && *endp != _T('\0'))
		return parseError(SCN_INVALID_NUMBERLIT);
	return (ttype);
}


bool BppScanner::isIdChar(BppUChar ch)
{

	return (_istupper(ch)
		|| _istlower(ch)
		|| _istdigit(ch)
		|| ch == _T('_')
	  || ch == _T('$'));
}


static int cmpKtItems(const void* pkey, const void* pitem) {
	return _tcscmp((const BppChar*)pkey, ((const KTTYPE*)pitem)->kt_keyword);
}
BppToken BppScanner::getKeywordTypeFromString(const BppChar* str)
{
	KTTYPE* pr = (KTTYPE*) bsearch(str, ktab, ktabsize, sizeof(KTTYPE), cmpKtItems);
	if (pr) return pr->kt_token;
	return T_NOTOKEN;
}

BppToken BppScanner::getTokenTypeFromString(const BppChar* str)
{
		for (int i = 0; t_names[i] != NULL; ++i)
		{
			if (_tcscmp(t_names[i], str) == 0)
				return (BppToken)i;
		}
	return T_NOTOKEN;
}

BppToken BppScanner::parseError(const BppChar* msg)
{
	if (_next != 0)
		return _next->parseError(msg);
	_stringBuffer.setCapacity(_tcslen(msg)+ _srcName.length() + _line.size() + 60);
	/* redisplay the line with the error */
		_sntprintf(_stringBuffer.data(), _stringBuffer.capacity(),SCN_PARSEERR_FMT5,msg,_srcName.c_str(),_lnum,_linepos,_line.data());

	_stringBuffer.adjustSize();
	return T_ERROR;
}

