
#include "bppstruc.h"
#include "bppufunc.h"
#include "bppdfunc.h"
#include "bppbuf.h"
#include "bppvec.h"
#include "bppfimpl.h"
#include "bppvalue.h"
#include "bppint.h"
#include "bppvm.h"
#include "bppprint.h"
#include "bppclass.h"
#include "bppmsg.h"
#include"bppobjct.h"

unsigned Bpp_StructHelper::_idxDataBuffer = 0;
unsigned Bpp_StructHelper::_idxOffs = 0;
unsigned Bpp_StructHelper::_idxDesc = 0;

static BppClass* phlpClass = 0;
static BppClass* pahlpClass = 0;
static BppClass* pcstructClass = 0;

BppObject * Bpp_StructHelper::_createInstance(BppClass& cls, BppBuffer * buf, BppUInt offs, BppDictionary * desc)
{
	if (!buf || !desc) return 0;
	BppObject* pThis = new BppObject(&cls);
	MEMBER(_idxDataBuffer) = buf;
   MEMBER(_idxOffs) = offs;
	MEMBER(_idxDesc) = desc;
	return pThis;
}



static const BppString _type = _T("type");
static const BppString _func = _T("func");
static const BppString _size = _T("size");
static const BppString _arrsize = _T("arrsize");
static const BppString _offs = _T("offs");
static const BppString _desc = _T("desc");
static const BppString _maxelsize = _T("maxelsize");

bool Bpp_StructHelper::op_pref(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER);
	BppUInt offs = MEMBER(_idxOffs);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppString* s = ARG(1);
	BppDictionaryEntry* pmdesc = pdesc->find(*s);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED,s->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, s->c_str());
	BppValueType vt = BppValueType(BppInt((*mdesc)[_type].getValue()));
	BppUInt boffs = (*mdesc)[_offs].getValue();
	BppUInt barrsize = (*mdesc)[_arrsize].getValue();
	if (barrsize) {
		BppObject* result = Bpp_ArrayHelper::_createInstance(*pahlpClass, pbuf, offs + boffs,mdesc);
		if (!result) THROW(STRUC_INVALID_ITEM_DESC, s->c_str());
		*bp = result;
		sp = bp;
		return true;
	}
	if (vt != BVT_DICTIONARY) {
		BppUInt bsize = (*mdesc)[_size].getValue();
		BppBuffer* buf = new BppBuffer(bsize, bsize);
		BppVector args;
		args.push_back(buf);
		memcpy(buf->data(), pbuf->data() + offs + boffs, bsize);
		if (!interpreter->invokeFunc(*bp, (*mdesc)[_func].getValue(), &args)) return false;
		sp = bp;
		return true;
	}
	BppDictionary* pdict = (*mdesc)[_func].getValue();
	if (!pdict) THROW(STRUC_INVALID_ITEM_DESC, s->c_str());
	BppObject* result = _createInstance(*pThis->getClass(), pbuf, offs + boffs, (*pdict)[_desc].getValue());
	if (!result) THROW(STRUC_INVALID_ITEM_DESC,s->c_str());
	*bp = result;
	sp = bp;
	return true;
}

bool Bpp_StructHelper::op_pset(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 3);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER);
	BppUInt offs = MEMBER(_idxOffs);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppString* s = ARG(1);
	BppDictionaryEntry* pmdesc = pdesc->find(*s);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, s->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, s->c_str());
	BppValueType vt = BppValueType(BppInt((*mdesc)[_type].getValue()));
	BppUInt boffs = (*mdesc)[_offs].getValue();
	BppUInt barrsize = (*mdesc)[_arrsize].getValue();
	if (barrsize) {
		BppVector* vvec = ARG(2);
		if (!vvec) THROW(STRUC_VECTOR_EXPECTED, _T("val"));
		BppObject* ahlp = Bpp_ArrayHelper::_createInstance(*pahlpClass, pbuf, offs + boffs, mdesc);
		if (!ahlp) THROW(STRUC_INVALID_ITEM_DESC, s->c_str());
		BppValue vsetfunc = (BPPUSERFUNC)Bpp_ArrayHelper::op_vset;
		BppVector args(3, true);
		args[0] = ahlp;
		BppUInt n = vvec->size();
		for (BppUInt i = 0; i < n;++i) {
			args[1] = i;
			args[2] = (*vvec)[i];
			BppValue v;
			if (!interpreter->invokeFunc(v, vsetfunc, &args)) THROW(STRUC_SET_VALUE_FAILED, s->c_str());
		}
		*bp = vvec;
		sp = bp;
		return true;
	}
	if (vt != BVT_DICTIONARY) {
		BppUInt bsize = (*mdesc)[_size].getValue();
		BppVector args;
		args.push_back(ARG(2));
		if (!interpreter->invokeFunc(*bp, (*mdesc)[_func].getValue(), &args)) return false;
		memcpy(pbuf->data() + offs + boffs,bp->v_ptr(), bsize);
		sp = bp; // *bp contains ARG(2), converted into required type
		return true;
	}
	BppDictionary* vdict = ARG(2);
	if (!vdict) THROW(STRUC_DICT_EXPECTED, _T("val"));
	BppDictionary* pdict = (*mdesc)[_func].getValue();
	if (!pdict) THROW(STRUC_INVALID_ITEM_DESC, s->c_str());
	BppObject* phlp = _createInstance(*pThis->getClass(), pbuf, offs + boffs, (*pdict)[_desc].getValue());
	if (!phlp) THROW(STRUC_INVALID_ITEM_DESC, s->c_str());
	BppValue psetfunc = (BPPUSERFUNC) Bpp_StructHelper::op_pset;
	BppVector args(3, true);
	args[0] = phlp;
	for (BppDictionaryEntry* pe = vdict->getFirst(); pe; pe = pe->getNext()) {
		args[1] = pe->getKey();
		args[2] = pe->getValue();
		BppValue v;
		if (!interpreter->invokeFunc(v, psetfunc, &args)) THROW (STRUC_SET_VALUE_FAILED,pe->getKey()->c_str());
	}
	*bp = vdict;
	sp = bp;
	return true;
}

bool Bpp_StructHelper::getData(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppValue preffunc = (BPPUSERFUNC)Bpp_StructHelper::op_pref;
	BppVector args(2, true);
	args[0] = pThis;
	BppValue result = new BppDictionary();
	BppDictionary* presult = result;
	for (BppDictionaryEntry* pe = pdesc->getFirst(); pe; pe = pe->getNext()) {
		args[1] = pe->getKey();
		BppValue v;
		if (!interpreter->invokeFunc(v, preffunc, &args)) THROW(STRUC_GET_VALUE_FAILED, pe->getKey()->c_str());
		BppObject* po = v;
		if (po) {
			if (!interpreter->invokeMember(v, po, _T("getData"))) THROW(STRUC_GET_DATA_FAILED, pe->getKey()->c_str());
		}
		(*presult)[pe->getKey()]->setValue( BPPMOVE(v));
	}
	*bp = BPPMOVE(result);
	sp = bp;
	return true;
}

bool Bpp_StructHelper::offsetOf(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppValue vname;
	if (!interpreter->doStrCast(vname, ARG(1))) return false;
	BppString* s = vname;
	BppUInt offs = MEMBER(_idxOffs);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppDictionaryEntry* pmdesc = pdesc->find(*s);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, s->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, s->c_str());
	BppUInt boffs = (*mdesc)[_offs].getValue(); 
	*bp = offs + boffs;
	sp = bp;
	return true;
}

bool Bpp_StructHelper::sizeOf(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppValue vname;
	if (!interpreter->doStrCast(vname, ARG(1))) return false;
	BppString* s = vname;
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppDictionaryEntry* pmdesc = pdesc->find(*s);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, s->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, s->c_str());
	BppUInt bsize = (*mdesc)[_size].getValue();
	BppUInt basize = (*mdesc)[_arrsize].getValue();
	if (basize) bsize *= basize;
	*bp = bsize;
	sp = bp;
	return true;
}

bool Bpp_StructHelper::intptr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(_T("invalid buffer"), _T("__SubstructProxy::intptr"));
	BppUInt offs = MEMBER(_idxOffs);
	*bp = INTPTRCAST(pbuf->data()) + offs;
	sp = bp;
	return true;
}

bool Bpp_StructHelper::uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(_T("invalid buffer"), _T("__SubstructProxy::uintptr"));
	BppUInt offs = MEMBER(_idxOffs);
	*bp = UINTPTRCAST(pbuf->data()) + offs;
	sp = bp;
	return true;
}

static const BppString __SubstructProxy = _T("__SubstructProxy");
static const BppString __SubArrayProxy = _T("__SubArrayProxy");

void Bpp_StructHelper::createClass(BppVirtualMachine & vm)
{
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(__SubstructProxy.c_str(), pClass)) return;
	pClass->addDataMember(_T("_sub buf"), BppValue(BVT_BUFFER, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), false));
	pClass->addDataMember(_T("_sub offs"), BppValue(BVT_UINT, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), true));
	pClass->addDataMember(_T("_sub desc"), BppValue(BVT_DICTIONARY, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), false));
	pClass->addMethod(_T("OP_PREF"), op_pref);
	pClass->addMethod(_T("OP_PSET"), op_pset);
	pClass->addMethod(_T("getData"), getData);
	pClass->addMethod(_T("offsetOf"), offsetOf);
	pClass->addMethod(_T("sizeOf"), sizeOf);
	pClass->addMethod(_T("intptr"), intptr);
	pClass->addMethod(_T("uintptr"), intptr);
	pClass->fixupDataDefs();
	_idxDataBuffer = pClass->getMemberIndex(_T("_sub buf"));
	_idxOffs = pClass->getMemberIndex(_T("_sub offs"));
	_idxDesc = pClass->getMemberIndex(_T("_sub desc"));
	phlpClass = pClass;
}

unsigned Bpp_ArrayHelper::_idxDataBuffer = 0;
unsigned Bpp_ArrayHelper::_idxOffs = 0;
unsigned Bpp_ArrayHelper::_idxDesc = 0;

bool Bpp_ArrayHelper::op_vref(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER);
	BppUInt offs = MEMBER(_idxOffs);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppValue vidx;
	if (!interpreter->doNumCast(vidx, ARG(1),BVT_UINT)) return false;
	BppUInt idx = vidx;
	BppUInt arrsize = (*pdesc)[_arrsize].getValue();
	if (idx >= arrsize) THROW(STRUC_INVALID_INDEX);
	BppValueType vt = BppValueType(BppInt((*pdesc)[_type].getValue()));
	BppUInt bsize = (*pdesc)[_size].getValue();
	if (vt != BVT_DICTIONARY) {
		BppBuffer* buf = new BppBuffer(bsize, bsize);
		BppVector args;
		args.push_back(buf);
		memcpy(buf->data(), pbuf->data() + offs + idx*bsize, bsize);
		if (!interpreter->invokeFunc(*bp, (*pdesc)[_func].getValue(), &args)) return false;
		sp = bp;
		return true;
	}
	BppDictionary* pdict = (*pdesc)[_func].getValue();
	if (!pdict) THROW(STRUC_INVALID_ITEM_DESC);
	BppObject* result = Bpp_StructHelper::_createInstance(*phlpClass, pbuf, offs + bsize*idx, (*pdict)[_desc].getValue());
	if (!result) THROW(STRUC_INVALID_ITEM_DESC);
	*bp = result;
	sp = bp;
	return true;
}


bool Bpp_ArrayHelper::op_vset(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 3);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER);
	BppUInt offs = MEMBER(_idxOffs);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppValue vidx;
	if (!interpreter->doNumCast(vidx, ARG(1), BVT_UINT)) return false;
	BppUInt idx = vidx;
	BppUInt arrsize = (*pdesc)[_arrsize].getValue();
	if (idx >= arrsize) THROW(STRUC_INVALID_INDEX);
	BppValueType vt = BppValueType(BppInt((*pdesc)[_type].getValue()));
	BppUInt bsize = (*pdesc)[_size].getValue();
	if (vt != BVT_DICTIONARY) {
		BppVector args;
		args.push_back(ARG(2));
		if (!interpreter->invokeFunc(*bp, (*pdesc)[_func].getValue(), &args)) return false;
		memcpy(pbuf->data() + offs + idx*bsize, bp->v_ptr(), bsize);
		sp = bp; // *bp contains ARG(2), converted into required type
		return true;
	}
	BppDictionary* vdict = ARG(2);
	if (!vdict) THROW(STRUC_DICT_EXPECTED, _T("val"));
	BppDictionary* pdict = (*pdesc)[_func].getValue();
	if (!pdict) THROW(STRUC_INVALID_ITEM_DESC);
	BppObject* phlp = Bpp_StructHelper::_createInstance(*phlpClass, pbuf, offs + idx*bsize, (*pdict)[_desc].getValue());
	if (!phlp) THROW(STRUC_INVALID_ITEM_DESC);
	BppValue psetfunc = (BPPUSERFUNC)Bpp_StructHelper::op_pset;
	BppVector args(3, true);
	args[0] = phlp;
	for (BppDictionaryEntry* pe = vdict->getFirst(); pe; pe = pe->getNext()) {
		args[1] = pe->getKey();
		args[2] = pe->getValue();
		BppValue v;
		if (!interpreter->invokeFunc(v, psetfunc, &args)) THROW(STRUC_SET_VALUE_FAILED, pe->getKey()->c_str());
	}
	*bp = vdict;
	sp = bp;
	return true;
}

bool Bpp_ArrayHelper::getData(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER);
	BppUInt offs = MEMBER(_idxOffs);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppUInt arrsize = (*pdesc)[_arrsize].getValue();
	BppValueType vt = BppValueType(BppInt((*pdesc)[_type].getValue()));
	BppValue result = new BppVector(arrsize,true);
	BppVector* presult = result;
	const BppValue& func = (*pdesc)[_func].getValue();
	BppUInt elsize = (*pdesc)[_size].getValue();
	if (vt != BVT_DICTIONARY) {
		BppBuffer* buf = new  BppBuffer(elsize, elsize);
		BppVector args;
		args.push_back(buf);
		for (BppUInt i = 0; i < arrsize; ++i) {
			memcpy(buf->data(), pbuf->data() + offs + i*elsize, elsize);
			if (!interpreter->invokeFunc((*presult)[i], func, &args)) return false;
		}
	}
	else {
		BppDictionary* pdict = func;
		if (!pdict) THROW(STRUC_INVALID_ITEM_DESC);
		for (BppUInt i = 0; i < arrsize; ++i) {
			BppObject* phlp = Bpp_StructHelper::_createInstance(*phlpClass, pbuf, offs + i*elsize, (*pdict)[_desc].getValue());
			if (!phlp) THROW(STRUC_INVALID_ITEM_DESC);
			if (!interpreter->invokeMember((*presult)[i], phlp, _T("getData"))) THROW(STRUC_GET_DATA_FAILED);
		}
	}
	*bp = BPPMOVE(result);
	sp = bp;
	return true;
}

bool Bpp_ArrayHelper::offsetOf(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppValue vidx;
	if (!interpreter->doNumCast(vidx, ARG(1),BVT_UINT)) return false;
	BppUInt idx = vidx;
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppUInt arrsize = (*pdesc)[_arrsize].getValue();
	if (idx >= arrsize) THROW(STRUC_INVALID_INDEX);
	BppUInt offs = MEMBER(_idxOffs);
	BppUInt elsize = (*pdesc)[_size].getValue();
	*bp = offs + elsize * idx;
	sp = bp;
	return true;
}

bool Bpp_ArrayHelper::size(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	*bp = (*pdesc)[_arrsize].getValue();
	sp = bp;
	return true;
}

bool Bpp_ArrayHelper::elSize(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppDictionary* pdesc = MEMBER(_idxDesc);
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	*bp = (*pdesc)[_size].getValue();
	sp = bp;
	return true;
}

bool Bpp_ArrayHelper::intptr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER, _T("__SubArrayProxy::intptr"));
	BppUInt offs = MEMBER(_idxOffs);
	*bp = INTPTRCAST(pbuf->data()) + offs;
	sp = bp;
	return true;
}

bool Bpp_ArrayHelper::uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER, _T("__SubArrayProxy::uintptr"));
	BppUInt offs = MEMBER(_idxOffs);
	*bp = UINTPTRCAST(pbuf->data()) + offs;
	sp = bp;
	return true;
}

void Bpp_ArrayHelper::createClass(BppVirtualMachine & vm)
{
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(__SubArrayProxy.c_str(), pClass)) return;
	pClass->addDataMember(_T("_sub buf"), BppValue(BVT_BUFFER, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), false));
	pClass->addDataMember(_T("_sub offs"), BppValue(BVT_UINT, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), true));
	pClass->addDataMember(_T("_sub desc"), BppValue(BVT_DICTIONARY, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), false));
	pClass->addMethod(_T("OP_VREF"), op_vref);
	pClass->addMethod(_T("OP_VSET"), op_vset);
	pClass->addMethod(_T("getData"), getData);
	pClass->addMethod(_T("offsetOf"), offsetOf);
	pClass->addMethod(_T("size"), size);
	pClass->addMethod(_T("elSize"), elSize);
	pClass->addMethod(_T("intptr"), intptr);
	pClass->addMethod(_T("uintptr"), intptr);
	pClass->fixupDataDefs();
	_idxDataBuffer = pClass->getMemberIndex(_T("_sub buf"));
	_idxOffs = pClass->getMemberIndex(_T("_sub offs"));
	_idxDesc = pClass->getMemberIndex(_T("_sub desc"));
	pahlpClass = pClass;
}

BppObject * Bpp_ArrayHelper::_createInstance(BppClass & cls, BppBuffer * buf, BppUInt offs, BppDictionary * desc)
{
	if (!buf || !desc) return 0;
	BppObject* pThis = new BppObject(&cls);
	MEMBER(_idxDataBuffer) = buf;
	MEMBER(_idxOffs) = offs;
	MEMBER(_idxDesc) = desc;
	return pThis;
}


unsigned  Bpp_CStruct::_idxBufDesc = 0;
unsigned  Bpp_CStruct::_idxDataBuffer = 0;
static const BppString _bufDescriptors = _T("_bufDescriptors");

bool Bpp_CStruct::create(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 3);
	bool isdynamic = argc == 3 ? bp[3].asBool() : false;
	if (!createBuffer(vm, pThis, isdynamic)) return false;
	if (argc > 1) {
		return applyData(vm, argc, sp);
	}
	sp = bp;
	return true;
}

bool Bpp_CStruct::getStructDesc(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	THROW(STRUC_PURE_VCALL_GETDESC,pThis->getClass()->getClassName()->c_str());
}

bool Bpp_CStruct::getStructPacking(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	*bp = BppUInt(0);
	sp = bp;
	return true;
}

bool Bpp_CStruct::createBuffer(BppVirtualMachine* vm, BppObject* pThis, bool isdynamic)
{
	BppClass* pClass = pThis->getClass();
	if (!pClass) return false;
	if (!pcstructClass) return false;
	const BppString& clname = *pClass->getClassName();
	BppDictionaryEntry* pd = pcstructClass->findMemberPL(BVF_PRIVATE, _bufDescriptors, false);
	if (!pd) return false;
	BppDictionary* bufDescriptors = pd->getValue();
	if (!bufDescriptors) return false;
	BppDictionaryEntry& descEntry = (*bufDescriptors)[clname];
	BppDictionary* pbufdesc = descEntry.getValue();
	if (!pbufdesc) {
	   BppInterpreter* interpreter = vm->interpreter();
		BppValue structdescval;
		if (!interpreter->invokeMember(structdescval, pThis, _T("getStructDesc"))) return false;
		BppDictionary* pstructdesc = structdescval;
		if (!pstructdesc) THROW(STRUC_MISSING_STRUCT_DESC);
		BppValue alignval;
		if (!interpreter->invokeMember(alignval, pThis, _T("getStructPacking"))) return false;
		BppUInt packing = alignval;
		BppValue resval;
		if (buildBufDesc(vm, resval, *pstructdesc,packing,false))
		   pbufdesc = resval;
		if (!pbufdesc) THROW(STRUC_CREATE_BUFFER_DESC_FAILED);
		MEMBER(_idxBufDesc) = pbufdesc;
		BppUInt sz = (*pbufdesc)[_size].getValue();
		MEMBER(_idxDataBuffer) = new BppBuffer(sz, sz);
		if (!isdynamic) descEntry.setValue(BPPMOVE(resval));
	}
	else {
		MEMBER(_idxBufDesc) = pbufdesc;
		BppUInt sz = (*pbufdesc)[_size].getValue();
		MEMBER(_idxDataBuffer) = new BppBuffer(sz, sz);
	}
	return true;
}


// aufgerufen von buildBufDesc
// cont ist
bool Bpp_CStruct::buildDescItem(BppVirtualMachine* vm, BppDictionaryEntry& item, BppDictionary& cont, BppUInt packing, bool isUnion)
{
	static BppInt ptrsize = sizeof(BppIntPtr);
	BppString* pkey = item.getKey();
	if (pkey->length() == 0) return false;
	bool anonymous = (*pkey)[0] == _T('-');
	bool structelement = false;
	BppDictionary* desc = cont[_desc].getValue();
	if (!desc) return false;
	if (desc->find(*pkey)) return false;

	BppDictionary* itemdesc = new BppDictionary();
	const BppValue& v = item.getValue();
	BppInt valtype = v.getType();	
	if (v.isIntegralType()) valtype = v;

	BppValue itemDescVal = itemdesc;
	(*itemdesc)[_type] = valtype;
	(*itemdesc)[_offs] = BppUInt(0);
	(*itemdesc)[_size] = BppUInt(0);
	(*itemdesc)[_arrsize] = BppUInt(0);
	(*itemdesc)[_func];
	BppDictionary* pdict = 0;
	bool startsUnion = false;
	if (valtype == BVT_VECTOR) {
		BppVector* vec = v;
		if (vec->size() == 2) {
			BppUInt arrsize = (*vec)[0];
			(*itemdesc)[_arrsize] = arrsize;
			startsUnion = (arrsize == 0); 
			if ((*vec)[1].isIntegralType())
			  valtype = (*vec)[1];
			else {
				valtype = (*vec)[1].getType();
				pdict = (*vec)[1];
			}

			(*itemdesc)[_type] = valtype;
		}
	}
	BppUInt siz = 0;
	BppUInt submaxelsize = 0;
	switch (valtype) {
	case BVT_SBYTE:
	case BVT_BYTE:
		siz = 1;
		(*itemdesc)[_func] = (valtype == BVT_SBYTE) ? &BppDefaultFunctions::Bpp_sbyte : &BppDefaultFunctions::Bpp_byte;
		break;
	case BVT_BOOL:
		siz = 1;
		(*itemdesc)[_func] = &BppDefaultFunctions::Bpp_bool;
		break;
	case BVT_CHAR:
	case BVT_UCHAR:
		siz = 1;
		(*itemdesc)[_func] = (valtype == BVT_CHAR) ? &BppDefaultFunctions::Bpp_char : &BppDefaultFunctions::Bpp_uchar;
		break;
	case BVT_CHAR8:
	case BVT_CHAR32:
		siz = 4;
		(*itemdesc)[_func] = (valtype == BVT_CHAR8) ? &BppDefaultFunctions::Bpp_char8_t : &BppDefaultFunctions::Bpp_char32_t;
		break;
	case BVT_WCHAR:
		siz = sizeof(wchar_t);
		(*itemdesc)[_func] = &BppDefaultFunctions::Bpp_wchar_t;
		break;
	case BVT_SHORT:
	case BVT_USHORT:
		siz = 2;
		(*itemdesc)[_func] = (valtype == BVT_SHORT) ? &BppDefaultFunctions::Bpp_short : &BppDefaultFunctions::Bpp_ushort;
		break;
	case BVT_CHAR16:
		siz = 2;
		(*itemdesc)[_func] = &BppDefaultFunctions::Bpp_char16_t;
		break;
	case BVT_INT:
	case BVT_UINT:
		siz = 4;
		(*itemdesc)[_func] = (valtype == BVT_INT) ? &BppDefaultFunctions::Bpp_int : &BppDefaultFunctions::Bpp_uint;
		break;
	case BVT_INTPTR:
	case BVT_UINTPTR:
		siz = ptrsize;
		(*itemdesc)[_func] = (valtype == BVT_INTPTR) ?  &BppDefaultFunctions::Bpp_intptr : &BppDefaultFunctions::Bpp_uintptr;
		break;
	case BVT_LONG:
	case BVT_ULONG:
		siz = 8;
		(*itemdesc)[_func] = (valtype == BVT_LONG) ? &BppDefaultFunctions::Bpp_long : &BppDefaultFunctions::Bpp_ulong;
		break;
	case BVT_FLOAT:
		siz = 4;
		(*itemdesc)[_func] = &BppDefaultFunctions::Bpp_float;
		break;
	case BVT_DOUBLE:
		siz = 8;
		(*itemdesc)[_func] = &BppDefaultFunctions::Bpp_double;
		break;
	default:
	{
		if (!pdict) 
		   pdict = item.getValue();
		if (pdict)
		{
			BppValue resval;
			if (!buildBufDesc(vm, resval, *pdict,packing,startsUnion)) return false;
			BppDictionary* subdesc = resval;
			if (!subdesc) return false;
			siz = (*subdesc)[_size].getValue();
			submaxelsize = (*subdesc)[_maxelsize].getValue();
			(*itemdesc)[_func] = subdesc;
			(*itemdesc)[_type] = BppInt(BVT_DICTIONARY);
			structelement = true;
		}
		else return false;
	}

	}
	BppDictionaryEntry& emaxelsize = cont[_maxelsize];
	BppUInt maxelsize = emaxelsize.getValue();
	if (submaxelsize > 0) {
		if (submaxelsize > maxelsize) emaxelsize.setValue(submaxelsize);
	}
	else {
		if (siz > maxelsize) emaxelsize.setValue(siz);
		submaxelsize = siz;
	}
	(*itemdesc)[_size].setValue(siz);
	BppUInt arrsize = (*itemdesc)[_arrsize].getValue();
	if (arrsize) siz *= arrsize;
	if (isUnion) {
		BppUInt csize = cont[_size].getValue();
		if (csize < siz) 
			cont[_size].setValue(siz);
	}
	else {
		BppUInt offs = cont[_size].getValue();
		BppUInt mod = packing ? packing : submaxelsize;
		BppUInt rem = offs % mod;
		if (rem) offs += (mod - rem);
		(*itemdesc)[_offs].setValue(offs);
		cont[_size].setValue(offs + siz);
	}

	if (anonymous && structelement) {
		BppDictionary* ifunc = (*itemdesc)[_func].getValue();
		if (!ifunc) return false;
		BppDictionary* subdesc = (*ifunc)[_desc].getValue();
		if (!subdesc) return false;
		BppUInt offs = (*itemdesc)[_offs].getValue();
		for (BppDictionaryEntry* ei = subdesc->getFirst(); ei; ei = ei->getNext()) {
			BppDictionary* desci = ei->getValue();
			if (!desci)
				return false;
			BppDictionary* idesc = desci->clone();
			(*idesc)[_offs].setValue(BppUInt((*idesc)[_offs].getValue()) + offs);
			(*desc)[ei->getKey()]->setValue(idesc);
		}
		return true;
	}
	(*desc)[pkey]->setValue(BPPMOVE(itemDescVal));
	return true;
}

// aufgerufen aus createBuffer (isUnion immer false)
// und aus buildDescItem mit isUnion=startsUnion (fr untergeordnete Strukturen)
// structDesc ist die Sutukturbeschreibung aus getStrucDesc bzw. ein untergeordnetes Element davon; 
bool Bpp_CStruct::buildBufDesc(BppVirtualMachine* vm, BppValue& result, BppDictionary& structDesc, BppUInt packing, bool isUnion)
{
	BppDictionary* pcont = new BppDictionary();
	result = pcont;
	(*pcont)[_size] = BppUInt(0);
	(*pcont)[_maxelsize] = BppUInt(0);
	(*pcont)[_desc] = new BppDictionary();
	BppDictionaryEntry* pe = 0;
	for (pe = structDesc.getFirst(); pe; pe = pe->getNext()) {
		if (!buildDescItem(vm, *pe, *pcont,packing,isUnion)) break;
	}
	if (pe) {
		CHKVM;
		THROW(STRUC_INVALID_MEMBER_DEF, pe->getKey()->c_str());
	}
	BppUInt siz = (*pcont)[_size].getValue();
	BppUInt maxelsize = (*pcont)[_maxelsize].getValue();
	BppUInt mod = packing ? packing : maxelsize;
	BppUInt rem = siz % mod;
	if (rem) (*pcont)[_size].setValue(siz+mod-rem);
	return true;
}

bool Bpp_CStruct::intptr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER,_T("CStruct::intptr"));
	*bp = INTPTRCAST(pbuf->data());
	sp = bp;
	return true;
}

bool Bpp_CStruct::uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER, _T("CStruct::uintptr"));
	*bp = UINTPTRCAST(pbuf->data());
	sp = bp;
	return true;
}

bool Bpp_CStruct::getDataBuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER, _T("CStruct::getDataBuffer"));
	*bp = pbuf;
	sp = bp;
	return true;
}

bool Bpp_CStruct::getDataSize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER, _T("CStruct::getDataSize"));
	*bp = (BppUInt) pbuf->size();
	sp = bp;
	return true;
}

bool Bpp_CStruct::op_pref(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	BppDictionary* pdesc = MEMBER(_idxBufDesc);
	if (pdesc) {
		BppDictionaryEntry* p = pdesc->find(_desc);
		pdesc = p ? (BppDictionary*) p->getValue() : (BppDictionary*)0;
	}
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppString* name = ARG(1);
	if (!name) return interpreter->badType(ARG(1).getType(), BVT_STRING);
	BppDictionaryEntry* pmdesc = pdesc->find(*name);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, name->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, name->c_str());
	BppUInt offs = (*mdesc)[_offs].getValue();
	BppUInt barrsize = (*mdesc)[_arrsize].getValue();
	if (barrsize) {
		BppObject* result = Bpp_ArrayHelper::_createInstance(*pahlpClass, pbuf, offs, mdesc);
		if (!result) THROW(STRUC_INVALID_ITEM_DESC, name->c_str());
		*bp = result;
		sp = bp;
		return true;
	}
	if (BppUInt((*mdesc)[_type].getValue()) != BVT_DICTIONARY) {
		BppUInt bsize = (*mdesc)[_size].getValue();
		BppBuffer* buf = new BppBuffer(bsize, bsize);
		BppVector args;
		args.push_back(buf);
		memcpy(buf->data(), pbuf->data() + offs, bsize);
		if (!interpreter->invokeFunc(*bp, (*mdesc)[_func].getValue(), &args)) return false;
		sp = bp;
		return true;
	}
	BppDictionary* pdict = (*mdesc)[_func].getValue();
	if (!pdict) THROW(STRUC_INVALID_ITEM_DESC, name->c_str());
	if (!phlpClass) THROW(STRUC_SUB_PROXY_NOT_DEFINED, __SubstructProxy.c_str());

	BppObject* result = Bpp_StructHelper::_createInstance(*phlpClass, pbuf, offs, (*pdict)[_desc].getValue());
	if (!result) THROW(STRUC_INVALID_ITEM_DESC, name->c_str());
	*bp = result;
	sp = bp;
	return true;
}

bool Bpp_CStruct::op_pset(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 3);
	CHKVAR(bp[1]);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	BppDictionary* pdesc = MEMBER(_idxBufDesc);
	if (pdesc) {
		BppDictionaryEntry* p = pdesc->find(_desc);
		pdesc = p ? (BppDictionary*)p->getValue() : (BppDictionary*)0;
	}
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppString* name = ARG(1);
	BppDictionaryEntry* pmdesc = pdesc->find(*name);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, name->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, name->c_str());
	BppUInt vt = (*mdesc)[_type].getValue();
	BppUInt boffs = (*mdesc)[_offs].getValue();
	BppUInt barrsize = (*mdesc)[_arrsize].getValue();
	if (barrsize) {
		BppVector* vvec = ARG(2);
		if (!vvec) THROW(STRUC_VECTOR_EXPECTED, _T("val"));
		BppObject* ahlp = Bpp_ArrayHelper::_createInstance(*pahlpClass, pbuf, boffs, mdesc);
		if (!ahlp) THROW(STRUC_INVALID_ITEM_DESC, name->c_str());
		BppValue vsetfunc = (BPPUSERFUNC)Bpp_ArrayHelper::op_vset;
		BppVector args(3, true);
		args[0] = ahlp;
		BppUInt n = vvec->size();
		for (BppUInt i = 0; i < n; ++i) {
			args[1] = i;
			args[2] = (*vvec)[i];
			BppValue v;
			if (!interpreter->invokeFunc(v, vsetfunc, &args)) THROW(STRUC_SET_VALUE_FAILED, name->c_str());
		}
		*bp = vvec;
		sp = bp;
		return true;
	}
	if (vt != BVT_DICTIONARY) {
		BppUInt bsize = (*mdesc)[_size].getValue();
		BppVector args;
		args.push_back(ARG(2));
		BppValue v = *bp;
		if (!interpreter->invokeFunc(v, (*mdesc)[_func].getValue(), &args)) return false;
		*bp = v;
		memcpy(pbuf->data() + boffs, v.v_ptr(), bsize);
		sp = bp;
		return true;
	}
	BppDictionary* vdict = ARG(2);
	if (!vdict) THROW(STRUC_DICT_EXPECTED, _T("val"));
	BppDictionary* pdict = (*mdesc)[_func].getValue();
	if (!pdict) THROW(STRUC_INVALID_ITEM_DESC, name->c_str());
	if (!phlpClass) THROW(STRUC_SUB_PROXY_NOT_DEFINED, __SubstructProxy.c_str());
	BppObject* phlp = Bpp_StructHelper::_createInstance(*phlpClass, pbuf, boffs, (*pdict)[_desc].getValue());
	if (!phlp) THROW(STRUC_INVALID_ITEM_DESC, name->c_str());
	BppValue psetfunc = (BPPUSERFUNC)Bpp_StructHelper::op_pset;
	BppVector args(3, true);
	args[0] = phlp;
	for (BppDictionaryEntry* pe = vdict->getFirst(); pe; pe = pe->getNext()) {
		args[1] = pe->getKey();
		args[2] = pe->getValue();
		BppValue v;
		if (!interpreter->invokeFunc(v, psetfunc, &args)) THROW(STRUC_SET_VALUE_FAILED, pe->getKey()->c_str());
	}
	*bp = vdict;
	sp = bp;
	return true;
}

bool Bpp_CStruct::getData(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppDictionary* pdesc = MEMBER(_idxBufDesc);
	if (pdesc) {
		BppDictionaryEntry* p = pdesc->find(_desc);
		pdesc = p ? (BppDictionary*)p->getValue() : (BppDictionary*)0;
	}
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppValue preffunc = (BPPUSERFUNC)Bpp_CStruct::op_pref;
	BppVector args(2, true);
	args[0] = pThis;
	BppValue result = new BppDictionary();
	BppDictionary* presult = result;
	for (BppDictionaryEntry* pe = pdesc->getFirst(); pe; pe = pe->getNext()) {
		args[1] = pe->getKey();
		BppValue v;
		if (!interpreter->invokeFunc(v, preffunc, &args)) THROW(STRUC_GET_VALUE_FAILED, pe->getKey()->c_str());
		BppObject* po = v;
		if (po) {
			if (!interpreter->invokeMember(v, po, _T("getData"))) THROW(STRUC_GET_DATA_FAILED, pe->getKey()->c_str());
		}
		(*presult)[pe->getKey()]->setValue(BPPMOVE(v));
	}
	*bp = BPPMOVE(result);
	sp = bp;
	return true;
}

bool Bpp_CStruct::applyData(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 3);
	CHKVAR(bp[1]);
	BppBuffer* pbuf = MEMBER(_idxDataBuffer);
	if (!pbuf) THROW(STRUC_INVALID_BUFFER, _T("CStruct::applyData"));
	BppValue dataval(BVT_DICTIONARY, BVF_STRONGTYPED, false);
	if (!dataval.assign(ARG(1))) return false;
	BppDictionary* pdata = dataval;
	if (pdata) {
		BppVector args(3, true);
		args[0] = pThis;
		BppValue dummyResult;
		for (BppDictionaryEntry* pe = pdata->getFirst(); pe; pe = pe->getNext()) {
			args[1] = pe->getKey();
			args[2] = pe->getValue();
			if (!interpreter->invokeFunc(dummyResult, &op_pset, &args)) return false;
		}
	}
	sp = bp;
	return true;
}

bool Bpp_CStruct::offsetOf(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppValue vname;
	if (!interpreter->doStrCast(vname, ARG(1))) return false;
	BppString* s = vname;
	BppDictionary* pdesc = MEMBER(_idxBufDesc);
	if (pdesc) {
		BppDictionaryEntry* p = pdesc->find(_desc);
		pdesc = p ? (BppDictionary*)p->getValue() : (BppDictionary*)0;
	}
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppDictionaryEntry* pmdesc = pdesc->find(*s);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, s->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, s->c_str());
	BppUInt boffs = (*mdesc)[_offs].getValue();
	*bp = boffs;
	sp = bp;
	return true;
}

bool Bpp_CStruct::sizeOf(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppValue vname;
	if (!interpreter->doStrCast(vname, ARG(1))) return false;
	BppString* s = vname;
	BppDictionary* pdesc = MEMBER(_idxBufDesc);
	if (pdesc) {
		BppDictionaryEntry* p = pdesc->find(_desc);
		pdesc = p ? (BppDictionary*)p->getValue() : (BppDictionary*)0;
	}
	if (!pdesc) THROW(STRUC_INVALID_ITEM_DESC);
	BppDictionaryEntry* pmdesc = pdesc->find(*s);
	if (!pmdesc)  THROW(STRUC_MEMBER_NOT_CONTAINED, s->c_str());
	BppDictionary* mdesc = pmdesc->getValue();
	if (!mdesc) THROW(STRUC_INVALID_MEMBER_DESC, s->c_str());
	BppUInt bsize = (*mdesc)[_size].getValue();
	BppUInt barrsize = (*mdesc)[_arrsize].getValue();
	if (barrsize) bsize *= barrsize;
	*bp = bsize;
	sp = bp;
	return true;
}

void Bpp_CStruct::createClass(BppVirtualMachine & vm)
{
	Bpp_StructHelper::createClass(vm);
	Bpp_ArrayHelper::createClass(vm);
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(_T("CStruct"), pClass)) return;
	pClass->addDataMember(_T("_bufDesc"), BppValue(BVT_DICTIONARY, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), false));
	pClass->addDataMember(_T("_dataBuffer"), BppValue(BVT_BUFFER, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), false));
	pClass->addStaticDataMember(_bufDescriptors.c_str(), BppValue(BVT_DICTIONARY, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), true));
	pClass->addMethod(_T("intptr"), intptr);
	pClass->addMethod(_T("uintptr"), uintptr);
	pClass->addMethod(_T("getDataBuffer"), getDataBuffer);
	pClass->addMethod(_T("getDataSize"), getDataSize);
	pClass->addMethod(_T("OP_PREF"), op_pref);
	pClass->addMethod(_T("OP_PSET"), op_pset);
	pClass->addMethod(_T("applyData"), applyData);
	pClass->addMethod(_T("getData"), getData);
	pClass->addMethod(_T("offsetOf"), offsetOf);
	pClass->addMethod(_T("sizeOf"), sizeOf);
	pClass->addMethod(_T("create"), create,BVF_PROTECTED);
	pClass->addMethod(_T("getStructDesc"), getStructDesc, BVF_PROTECTED);
	pClass->addMethod(_T("getStructPacking"), getStructPacking, BVF_PROTECTED);
	pClass->fixupDataDefs();
	_idxDataBuffer = pClass->getMemberIndex(_T("_dataBuffer"));
	_idxBufDesc = pClass->getMemberIndex(_T("_bufDesc"));
	pcstructClass = pClass;
}
