
#include "bppufunc.h"
#include "bppstrng.h"
#include "bppbuf.h"
#ifdef __OLDCPP__
#include <dir.h>
#include <stdarg.h>
#include <dos.h>
#endif
#ifdef GCC
#include <wctype.h>
#include <wchar.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
//#include <unicode/convutf.h>
#ifdef BPPCHKMEM
#include <set>
#include <debugapi.h>
static std::set<void*>& RSET() {
	static std::set<void*> rset;
	return rset;
}
#endif
#ifndef _MAX_PATH
#if (defined(GCC) && !defined(WIN32))
	#define _MAX_PATH PATH_MAX
#else
 #define _MAX_PATH 256
#endif
#endif

void errorExit(int err)
{
	exit(err);
}


void* allocmem(size_t blksize)
{
#ifdef BPPCHKMEM
	if (blksize == 0)
		DebugBreak();
#endif
	void* result = malloc(blksize);
	if (!result)
		errorExit(BP_ERR_NOMEM);
#ifdef BPPCHKMEM
	RSET().insert(result);
#endif
	return result;
}

void* reallocmem(void* p, size_t blksize)
{
#ifdef BPPCHKMEM
	std::set<void*>::iterator it = RSET().find(p);
	if (it == RSET().end())
		DebugBreak();
	RSET().erase(p);

	if (blksize == 0)
		DebugBreak();
#endif	
	void* result = realloc(p,blksize);
	if (!result)
		errorExit(BP_ERR_NOMEM);
#ifdef BPPCHKMEM
	RSET().insert(result);
#endif
	return result;
}

BppChar* dupstr(const BppChar* src)
{
#ifdef BPPCHKMEM
	size_t n = _tcslen(src);
	BppChar* result = (BppChar*)allocmem((n + 1) * sizeof(BppChar));
	_tcsncpy(result, src,n+1);
	return result;
#else
	return _tcsdup(src);
#endif
}

void freemem(void* p)
{
#ifdef BPPCHKMEM
	std::set<void*>::iterator it = RSET().find(p);
	if (it == RSET().end())
		DebugBreak();
	RSET().erase(p);
#endif
	free(p);
}

BPPLIB_DLLEXP BppIntPtr movemem(BppIntPtr dst, BppIntPtr src, BppUInt n)
{
	if (n==0 || dst == src) return dst;
#ifdef __MSDOS__
	char huge * _dst = (char huge*) dst;
	char huge * _src = (char huge*) src;
	if (dst > src) {
		for (_dst += n-1, _src += n-1; n > 0; --_dst, --_src, --n) *_dst = *_src;
	}
	else {
		for (; n > 0; ++_dst, ++_src, --n) *_dst = *_src;
	}
	return dst;
#else
	return INTPTRCAST(memmove((void*) dst, (const void*) src, n));
#endif
}



const BppChar* getFileDir(const BppChar* filePath, bool stripSlash)
{
	static BppChar buf[_MAX_PATH];
	BppChar* dirpath = _tfullpath(buf,filePath,_MAX_PATH);
	 if (!dirpath) return 0;
	 int i = _tcslen(dirpath)-2;
	for (;i>1;--i)
	{
		BppChar c = dirpath[i];
		if (c == _T('/') || c == _T('\\') || c == _T(':'))
		{
			dirpath[i+(stripSlash ? 0 : 1)] = _T('\0');
			break;
		}
	}
	 return buf;
}


const BppChar* searchPath(const BppChar* fname)
{
	if (fname == 0) return 0;
#ifdef __OLDCPP__
	return searchpath(fname);
#else
		const BppChar* result = searchEnv(_T("Path"),fname);
	  if (result == 0)
		  result = searchEnv(_T("PATH"),fname);
	  return result;
#endif
}

const BppChar* searchEnv(const BppChar* varname, const BppChar* fname)
{
  if ((varname == 0) || (fname == 0)) return 0;
#if defined(__OLDCPP__)  || (defined(GCC) && (!defined(WIN32)))
	BppChar* envval = _tgetenv(varname);
	return searchDirList(envval,fname);
#else
  #ifndef WINCE
		static BppChar buf[_MAX_PATH +1];
		_tsearchenv(fname,varname,buf);
	  if (_tcslen(buf) == 0) return 0;
	  return buf;
  #else
	const BppChar* envval = _tgetenv(varname);
	return searchDirList(envval,fname);
  #endif // WINCE
#endif
}

/**
Function tries to find file <i>fname</i> in list of directories to use for search.
@param dirList List of absolute or relative folder paths to use for search in. Items are separated by semicolon (;).
@param fname Name of file to search for.
@return Path of <i>fname</i> on success, NULL otherwise.
@see searchEnv, searchPath
*/
const BppChar* searchDirList(const BppChar* dirList, const BppChar* fname)
{
#ifdef UNIX
#define DIRDELIMIT _T("/")
#else
#define DIRDELIMIT _T("\\")
#endif
	static BppChar buf[_MAX_PATH * 2];
	if ((dirList == 0) || (fname == 0)) return 0;
	//int dListLen = _tcslen(dirList);
	BppChar* dirbuf = dupstr(dirList);
//   _tcscpy(dirbuf,dirList);
	bool found = false;
	for (BppChar* token = _tcstok(dirbuf,_T(";"));(!found) && (token != 0);token = _tcstok(NULL,_T(";")))
	{
		_tcsncpy(buf,token,_MAX_PATH);
	  int buflen = _tcslen(buf);
	  if ((buflen > 0) && (buf[buflen-1] != _T('\\')) && (buf[buflen-1] != _T('/')))
		  _tcscat(buf,DIRDELIMIT);
		_tcscat(buf,fname);
		FILE* fp = _tfopen(buf,_T("r"));
		if (fp != 0)
		{
			fclose(fp);
			found = true;
		}
	}
	freemem(dirbuf);
	if (!found)
		return 0;
	return buf;
}

#ifndef __OLDCPP__
_TmpLocale::_TmpLocale(const char* loc)  {
#ifdef HAVE_THREADLOCALE
		_oldmode = _configthreadlocale(_ENABLE_PER_THREAD_LOCALE);
#endif
		char* pold = setlocale(LC_ALL,0);
		if (pold) {
			_old = pold;
			//_old = strdup(pold);
		}
		if (loc) {
#ifdef UNIX
			if ((strlen(loc) == 0) || strchr(loc, '.')) {
				setlocale(LC_ALL, loc);
				return;
			}
			const char* plang = getenv("LANG");
			if (plang) plang = strchr(plang,'.');
			if (!plang || strlen(plang) == 0) {
				setlocale(LC_ALL, loc);
				return;
			}

			char* pnew = (char*)allocmem(strlen(loc) + strlen(plang) + 2);
			sprintf(pnew, "%s%s", loc, plang);
			setlocale(LC_ALL, pnew);
			freemem(pnew);
#else
			setlocale(LC_ALL, loc);
#endif
		}
	}
_TmpLocale::~_TmpLocale() {
	if (_old.length() > 0) {
		setlocale(LC_ALL, _old.a_str());
		//freemem(_old);
	}
#ifdef HAVE_THREADLOCALE
		if (_oldmode > 0) _configthreadlocale(_oldmode);
#endif
	}

#endif

#ifdef WINCE
	int _gettche()
	{
		int result = _gettchar();
//      if (result)
//         _puttchar(result);
		return result;
	}


const BppChar* _tgetenv(const BppChar* varname)
{
	static BppChar buf[4096];
	unsigned long bufsize = sizeof(buf);
	unsigned long type;
	 buf[0] = 0;
	HKEY hkResult;
	 if (RegOpenKeyEx(HKEY_CURRENT_USER,_T("Environment"),0,0,&hkResult) == ERROR_SUCCESS)
	{
		 RegQueryValueEx(hkResult,varname,0,&type,(unsigned char*)buf,&bufsize);
		RegCloseKey(hkResult);
	}
	return buf;
}

int _tputenv(const BppChar* envstr)
{
	HKEY hkResult;
	DWORD dwDisp;
	BppChar* buf = dupstr(envstr);
	int result = 0;
	BppChar* p = _tcschr(buf, _T('='));
	if (p != 0)
	{
		 *p = 0;
		p++;
		 result = RegCreateKeyEx(HKEY_CURRENT_USER,_T("Environment"),0,_T("REG_SZ"),0,0,0,&hkResult,&dwDisp);
		if (result == 0)
		{
			if ((*p) == 0) // remove entry
			{
				result = RegDeleteValue(hkResult,buf);
			}
			else
			{
				result = RegSetValueEx(hkResult,buf,0,REG_SZ,(const unsigned char*)p,(_tcslen(p)+1) * sizeof(BppChar));
			}
		}
		RegCloseKey(hkResult);
	}
	else result = -1;
	 freemem(buf);
	return result;
}

union __FTIME
{
	FILETIME ft;
	unsigned __int64 fti;
};
static SYSTEMTIME __reftime = { 1970,1,4,1,0,0,0,0};

time_t time(time_t* t)
{
	static bool initialized = false;
	static __FTIME __refftime;
	if (!initialized)
	{
		SystemTimeToFileTime(&__reftime,&__refftime.ft);
		initialized = true;
	}
	SYSTEMTIME currentTime;
	GetSystemTime(&currentTime);
	__FTIME currentFTime;
	SystemTimeToFileTime(&currentTime,&currentFTime.ft);
	if (t != 0)
	{
	  *t = (time_t) ((currentFTime.fti - __refftime.fti) / 10000000);
	  return *t;
	}
	return (time_t) ((currentFTime.fti - __refftime.fti) / 10000000);
}

BppChar* _tctime(time_t* tt)
{
	static BppChar* __weekdays[] = {_T("Sun"),_T("Mon"),_T("Tue"),_T("Wed"),_T("Thu"),_T("Fri"),_T("Sat")};
	static BppChar* __months[] = {_T(""),_T("Jan"),_T("Feb"),_T("Mar"),_T("Apr"),_T("May"),_T("Jun"),
		_T("Jul"),_T("Aug"),_T("Sep"),_T("Oct"),_T("Nov"),_T("Dec")};
	static BppChar buf[50];
	static bool initialized = false;
	static __FTIME __refftime;
	if (!tt) return 0;
	if (!initialized)
	{
		SystemTimeToFileTime(&__reftime,&__refftime.ft);
		initialized = true;
	}
	SYSTEMTIME currentTime;
	__FTIME currentFTime;
	currentFTime.fti = *tt;
	currentFTime.fti *= 10000000;
	currentFTime.fti += __refftime.fti;
	FileTimeToLocalFileTime(&currentFTime.ft,&currentFTime.ft);
	if (FileTimeToSystemTime(&currentFTime.ft,&currentTime) == 0) return 0;
	_stprintf(buf,_T("%s %s %02u %02u:%02u:%02u %04u\n"),
		__weekdays[currentTime.wDayOfWeek],
		__months[currentTime.wMonth],
		(unsigned) currentTime.wDay,
		(unsigned) currentTime.wHour,
		(unsigned) currentTime.wMinute,
		(unsigned) currentTime.wSecond,
		(unsigned) currentTime.wYear);
	return (BppChar*) buf;
}

tm* localtime( const time_t* t)
{
	static tm _tm;
	static bool initialized = false;
	static __FTIME __refftime;
	if (!t) return 0;
	if (!initialized)
	{
		SystemTimeToFileTime(&__reftime,&__refftime.ft);
		initialized = true;
	}
	if (t == 0) return 0;
	SYSTEMTIME currentTime;
	__FTIME currentFTime;
	currentFTime.fti = ((*t) * 10000000) + __refftime.fti;
	 if (FileTimeToLocalFileTime(&currentFTime.ft,&currentFTime.ft) == 0) return 0;
	 if (FileTimeToSystemTime(&currentFTime.ft,&currentTime) == 0) return 0;
	_tm.tm_sec = currentTime.wSecond;
	_tm.tm_min = currentTime.wMinute;
	_tm.tm_hour = currentTime.wHour;
	_tm.tm_mday = currentTime.wDay;
	_tm.tm_mon = currentTime.wMonth;
	_tm.tm_year = currentTime.wYear;
	_tm.tm_wday = currentTime.wDayOfWeek;
	_tm.tm_yday = -1;
	_tm.tm_isdst = -1;
	return &_tm;
}

#endif // WINCE

#ifdef emulate_vsctprintf

#ifdef WINCE
	#define NULDEVICE "\\Temp\\NUL"
	#define NULDEVICEW L"\\Temp\\NUL"
#else
	#ifdef __MSDOS__
	  #define  NULDEVICE "NUL"
	  #define  NULDEVICEW "NUL"
  #else
	  #ifndef WIN32
		  #define NULDEVICE "/dev/null"
		  #define NULDEVICEW L"/dev/null"
	  #else
		 #define  NULDEVICE "NUL"
		 #define  NULDEVICEW L"NUL"
	  #endif
	#endif
#endif
	int _vscprintf(const char* format, va_list args)
	{
	  FILE* fp = fopen(NULDEVICE,"wt");
	  int result = vfprintf(fp,format,args);
	  fclose(fp);
	  return result;
	}
#ifndef __OLDCPP__
	int _vscwprintf(const wchar_t* format, va_list args)
	{
		FILE* fp = _wfopen(NULDEVICEW,L"wt");
#ifdef UNIX
		fwide(fp, 1);
#endif
		int result = vfwprintf(fp,format,args);
		fclose(fp);
		return result;
	}
#endif
#endif
#ifndef __OLDCPP__
#ifdef CPP11
#include <uchar.h>
#endif
#include <wchar.h>
	BppBuffer str2wcsBuf(const char* str) {
		BppBuffer wbuf;
		if (!str) return wbuf;
	    size_t len = strlen(str)+1;
	    size_t bytes = len * sizeof(wchar_t);
		wbuf.setCapacity(bytes);
	    mbstowcs((wchar_t*) wbuf.data(),str,len);
		return wbuf;		
	}

   BppBuffer wcs2strBuf(const wchar_t* wcs) {
		BppBuffer abuf;
		if (!wcs) return abuf;
		size_t bytes = (wcslen(wcs)+1) * sizeof(wchar_t);
		abuf.setCapacity(bytes);
		wcstombs((char*) abuf.data(),(const wchar_t*) wcs,bytes);
		return abuf;
   }
#endif

   BppUShort c32_to_c16(BppUInt c32)
   {
		return c32 & 0xffff;
   }

   

   BppUInt c32_to_u8(BppUInt c32)
   {
	   BppUInt c = c32;

	   if (c32 > 0x7F) {
		   c = (c32 & 0x000003F)
			   | (c32 & 0x0000FC0) << 2
			   | (c32 & 0x003F000) << 4
			   | (c32 & 0x01C0000) << 6;

		   if (c32 < 0x0000800) c |= 0x0000C080;
		   else if (c32 < 0x0010000) c |= 0x00E08080;
		   else  c |= 0xF0808080;
	   }
	   BppUInt r = 0;
	   while (c) {
		   r <<= 8;
	      r += c & 0xff;
		  c >>= 8;
	   }
	   return r;
   }

   BppUChar c32_to_char(BppUInt c32, unsigned cp)
   {
	   return c16_to_char(c32_to_c16(c32), cp);
   }

   BppUInt c16_to_c32(BppUShort c16)
   {
	   return c16;
   }

   BppUInt c16_to_u8(BppUShort c16)
   {
	   return c32_to_u8(c16);
   }


   BppUInt u8_to_c32(BppUInt u8)
   {
	   BppUInt r = 0;
	   while (u8) {
		   r <<= 8;
	      r += u8 & 0xff;
		  u8 >>= 8;
	   }
	   u8 = r;
	   BppUInt mask;

	   if (u8 > 0x7F) {
		   mask = (u8 <= 0x00EFBFBF) ? 0x000F0000 : 0x003F0000;
		   return ((u8 & 0x07000000) >> 6) |
			   ((u8 & mask) >> 4) |
			   ((u8 & 0x00003F00) >> 2) |
			   (u8 & 0x0000003F);
	   }
	   return u8;	
   }

   BppUShort u8_to_c16(BppUInt u8)
   {
	   return u8_to_c32(u8);
   }

   BppUChar u8_to_char(BppUInt u8, unsigned cp)
   {
	   return c16_to_char(u8_to_c16(u8), cp);
   }

   BppUInt char_to_c32(BppUChar c, unsigned cp)
   {
	   return c16_to_c32(char_to_c16(c, cp));
   }

   BppUInt char_to_u8(BppUChar c, unsigned cp)
   {
	   return c16_to_u8(char_to_c16(c, cp));
   }

   BppChar cbufToBppChar(BppUInt cbuf, unsigned csize, unsigned cp)
   {
	   if (sizeof(BppChar) != csize)
       {
		   switch (sizeof(BppChar)) {
		   case 1:
			   if (csize == 2)
				   return (BppChar)c16_to_char(cbuf,cp);
			   else
				   return (BppChar)c32_to_char(cbuf,cp);
		   case 2:
			   if (csize == 1)
				   return (BppChar)char_to_c16(cbuf,cp);
			   else
				   return (BppChar)c32_to_c16(cbuf);
		   default: // 4
			   if (csize == 1)
				   return (BppChar)char_to_c32(cbuf,cp);
			   else
				   return (BppChar)c16_to_c32(cbuf);
		   }
	   }
	   return (BppChar)cbuf;
   }

   wchar_t cbufToWchar(BppUInt cbuf, unsigned csize, unsigned cp)
   {
	   if (sizeof(wchar_t) != csize)
	   {
		   switch (sizeof(wchar_t)) {
		   case 1:
			   if (csize == 2)
				   return (wchar_t)c16_to_char(cbuf, cp);
			   else
				   return (wchar_t)c32_to_char(cbuf, cp);
		   case 2:
			   if (csize == 1)
				   return (wchar_t)char_to_c16(cbuf, cp);
			   else
				   return (wchar_t)c32_to_c16(cbuf);
		   default: // 4
			   if (csize == 1)
				   return (wchar_t)char_to_c32(cbuf, cp);
			   else
				   return (wchar_t)c16_to_c32(cbuf);
		   }
	   }
	   return (wchar_t)cbuf;
   }
#ifndef NO_CHARCONV
#ifdef __MSDOS__
#define __CTABSMALL__
#endif
#include "codetab.inc"

#ifndef __CTABSMALL__
   static const size_t codetab7size = sizeof(codetab7) / sizeof(CpId_Codes);
#endif
   static const size_t codetabsize = sizeof(codetab) / sizeof(CpId_Codes);

   static int idCmpFunc(const void* k, const void* v) {
	   return (*(int*)k) - (int)((const CpId_Codes*)v)->id;
   }

#endif
   BppUShort char_to_c16(BppUChar c, unsigned cp)
   {
#ifndef NO_CHARCONV
	   if (cp == 0) cp = getActiveCp();
	   if ((c & 0x80) == 0) {
#ifndef __CTABSMALL__
		   const CpId_Codes* item = (const CpId_Codes*)bsearch(&cp, codetab7, codetab7size, sizeof(CpId_Codes), idCmpFunc);
		   if (item) {
			   BppUShort result = item->ucodes[c];
			   if (result == 0xfffd) return 0x20;
			   return result;
		   }
#endif
		   return c;
	   }
	   const CpId_Codes* item = (const CpId_Codes*)bsearch(&cp, codetab, codetabsize, sizeof(CpId_Codes), idCmpFunc);
	   if (!item && cp != DEF_CP) {
		   cp = DEF_CP;
		   item = (const CpId_Codes*)bsearch(&cp, codetab, codetabsize, sizeof(CpId_Codes), idCmpFunc);
	   }
	   if (item) {
		   BppUShort result = item->ucodes[c - 0x80];
		   if (result == 0xfffd) return 0x20;
		   return result;
	   }
#endif //!NO_CHARCONV
	   return c;
   }

   BppUChar c16_to_char(BppUShort c16, unsigned cp)
   {
	   BppUChar c = c16 & 0x7f;
#ifndef NO_CHARCONV
	   if (BppUShort(c) == c16) return c;
	   if (cp == 0) cp = getActiveCp();
#ifdef WIN32
	   char mbs[2] = { 0,0 };
	   wchar_t ws[2] = { 0,0 };
	   ws[0] = c16;
	   if (WideCharToMultiByte(cp, 0, ws, 2, mbs, 2,0,0) == 1)
		   return mbs[0];
#endif
	   const CpId_Codes* item = (const CpId_Codes*)bsearch(&cp, codetab, codetabsize, sizeof(CpId_Codes), idCmpFunc);
	   if (!item && cp != DEF_CP) {
		   cp = DEF_CP;
		   item = (const CpId_Codes*)bsearch(&cp, codetab, codetabsize, sizeof(CpId_Codes), idCmpFunc);
	   }
	   if (item) {
		   for (BppUChar i = 0; i < 128; ++i) {
			   if (item->ucodes[i] == c16)
				   return i | 0x80;
		   }
	   }
#ifndef __CTABSMALL__
	   item = (const CpId_Codes*)bsearch(&cp, codetab7, codetab7size, sizeof(CpId_Codes), idCmpFunc);
	   if (item) {
		   for (BppUChar i = 0; i < 128; ++i) {
			   if (item->ucodes[i] == c16)
				   return i;
		   }
	   }
#endif
#endif //!NO_CHARCONV
	   return c | 0x80;
   }

#ifndef NO_CHARCONV
   struct Name_Id {
	   const BppChar* name;
	   unsigned short id;
   } name_id_tab[] = {
#ifndef __CTABSMALL__
	   {_T("1025"), 21025},
	   //{_T("1026"), 1026},
	   //{_T("1250"), 1250},
	   //{_T("1251"), 1251},
	   //{_T("1252"), 1252},
	   //{_T("1253"), 1253},
	   //{_T("1254"), 1254},
	   //{_T("1255"), 1255},
	   //{_T("1256"), 1256},
	   //{_T("1257"), 1257},
	   //{_T("1258"), 1258},
	   {_T("273"), 20273},
	   {_T("278"), 20278},
	   {_T("280"), 20280},
	   {_T("284"), 20284},
	   {_T("285"), 20285},
	   {_T("290"), 20290},
	   {_T("297"), 20297},
	   {_T("367"), 20127},
	   //{_T("37"), 37},
	   {_T("420"), 20420},
	   {_T("423"), 20423},
	   {_T("424"), 20424},
#endif
	   //{_T("437"), 437},
	   //{_T("500"), 500},
	   //{_T("50227"), 50227},
	   {_T("819"), 28591},
	   //{_T("850"), 850},
	   //{_T("852"), 852},
	   //{_T("855"), 855},
	   //{_T("857"), 857},
	   //{_T("858"), 858},
	   //{_T("860"), 860},
	   //{_T("861"), 861},
	   //{_T("862"), 862},
	   //{_T("863"), 863},
	   //{_T("864"), 864},
	   //{_T("865"), 865},
	   //{_T("866"), 866},
	   //{_T("869"), 869},
	   //{_T("870"), 870},
#ifndef __CTABSMALL__
	   {_T("871"), 20871},
	   //{_T("875"), 875},
	   {_T("880"), 20880},
	   {_T("ansi-x3.4-1968"), 20127},
	   {_T("ansi-x3.4-1986"), 20127},
	   {_T("arabic"), 28596},
	   {_T("ascii"), 20127},
	   {_T("asmo-449"), 709},
	   {_T("asmo-449+"), 709},
	   {_T("asmo-708"), 708},
	   {_T("big5"), 950},
	   {_T("big5-hkscs"), 950},
	   {_T("ccsid00858"), 858},
	   {_T("ccsid00924"), 20924},
	   {_T("ccsid01140"), 1140},
	   {_T("ccsid01141"), 1141},
	   {_T("ccsid01142"), 1142},
	   {_T("ccsid01143"), 1143},
	   {_T("ccsid01144"), 1144},
	   {_T("ccsid01145"), 1145},
	   {_T("ccsid01146"), 1146},
	   {_T("ccsid01147"), 1147},
	   {_T("ccsid01148"), 1148},
	   {_T("ccsid01149"), 1149},
	   {_T("chinese"), 936},
	   {_T("cn-big5"), 950},
	   {_T("cn-gb"), 936},
	   {_T("cp00858"), 858},
	   {_T("cp00924"), 20924},
	   {_T("cp01140"), 1140},
	   {_T("cp01141"), 1141},
	   {_T("cp01142"), 1142},
	   {_T("cp01143"), 1143},
	   {_T("cp01144"), 1144},
	   {_T("cp01145"), 1145},
	   {_T("cp01146"), 1146},
	   {_T("cp01147"), 1147},
	   {_T("cp01148"), 1148},
	   {_T("cp01149"), 1149},
	   {_T("cp037"), 37},
	   {_T("cp1025"), 21025},
	   {_T("cp1026"), 1026},
#endif	
	   {_T("cp1250"), 1250},
	   {_T("cp1251"), 1251},
	   {_T("cp1252"), 1252},
	   {_T("cp1253"), 1253},
	   {_T("cp1254"), 1254},
	   {_T("cp1255"), 1255},
	   {_T("cp1256"), 1256},
	   {_T("cp1257"), 1257},
	   {_T("cp1258"), 1258},
#ifndef __CTABSMALL__
	   {_T("cp273"), 20273},
	   {_T("cp278"), 20278},
	   {_T("cp280"), 20280},
	   {_T("cp284"), 20284},
	   {_T("cp285"), 20285},
	   {_T("cp290"), 20290},
	   {_T("cp297"), 20297},
	   {_T("cp367"), 20127},
	   {_T("cp420"), 20420},
	   {_T("cp423"), 20423},
	   {_T("cp424"), 20424},
#endif
	   {_T("cp437"), 437},
#ifndef __CTABSMALL__
	   {_T("cp500"), 500},
	   {_T("cp50227"), 50227},
#endif
	   {_T("cp819"), 28591},
	   {_T("cp850"), 850},
	   {_T("cp852"), 852},
	   {_T("cp855"), 855},
	   {_T("cp857"), 857},
	   {_T("cp858"), 858},
	   {_T("cp860"), 860},
	   {_T("cp861"), 861},
	   {_T("cp862"), 862},
	   {_T("cp863"), 863},
	   {_T("cp864"), 864},
	   {_T("cp865"), 865},
	   {_T("cp866"), 866},
	   {_T("cp869"), 869},
#ifndef __CTABSMALL__
	   {_T("cp870"), 870},
	   {_T("cp871"), 20871},
	   {_T("cp875"), 875},
	   {_T("cp880"), 20880},
	   {_T("cp905"), 20905},
	   {_T("cp930"), 50930},
	   {_T("cp933"), 50933},
	   {_T("cp935"), 50935},
	   {_T("cp937"), 50937},
	   {_T("cp939"), 50939},
#endif
	   {_T("csascii"), 20127},
	   {_T("csbig5"), 950},
	   {_T("cseuckr"), 51949},
#ifndef __CTABSMALL__
	   {_T("cseucpkdfmtjapanese"), 51932},
	   {_T("csgb2312"), 936},
	   {_T("csgb231280"), 936},
	   {_T("csibm037"), 37},
	   {_T("csibm1026"), 1026},
	   {_T("csibm273"), 20273},
	   {_T("csibm277"), 20277},
	   {_T("csibm278"), 20278},
	   {_T("csibm280"), 20280},
	   {_T("csibm284"), 20284},
	   {_T("csibm285"), 20285},
	   {_T("csibm290"), 20290},
	   {_T("csibm297"), 20297},
	   {_T("csibm420"), 20420},
	   {_T("csibm423"), 20423},
	   {_T("csibm424"), 20424},
	   {_T("csibm500"), 500},
	   {_T("csibm870"), 870},
	   {_T("csibm871"), 20871},
	   {_T("csibm880"), 20880},
	   {_T("csibm905"), 20905},
	   {_T("csibmthai"), 20838},
	   {_T("csiso2022jp"), 50220},
	   {_T("csiso2022kr"), 50225},
	   {_T("csiso58gb231280"), 936},
	   {_T("csisolatin1"), 28591},
	   {_T("csisolatin2"), 28592},
	   {_T("csisolatin3"), 28593},
	   {_T("csisolatin4"), 28594},
	   {_T("csisolatin5"), 28599},
	   {_T("csisolatin9"), 28605},
	   {_T("csisolatinarabic"), 28596},
	   {_T("csisolatincyrillic"), 28595},
	   {_T("csisolatingreek"), 28597},
	   {_T("csisolatinhebrew"), 28598},
	   {_T("cskoi8r"), 20866},
	   {_T("csksc56011987"), 949},
	   {_T("cspc8codepage437"), 437},
	   {_T("csshiftjis"), 932},
	   {_T("csunicode11utf7"), 65000},
	   {_T("cswindows31j"), 932},
	   {_T("cyrillic"), 28595},
	   {_T("din-66003"), 20106},
#endif
	   {_T("dos-720"), 720},
	   {_T("dos-862"), 862},
	   {_T("dos-874"), 874},
#ifndef __CTABSMALL__
	   {_T("ebcdic"), 37},
#endif
	   {_T("ecma-114"), 28596},
	   {_T("ecma-118"), 28597},
	   {_T("elot-928"), 28597},
#ifndef __CTABSMALL__
	   {_T("euc-cn"), 51936},
	   {_T("euc-jp"), 51932},
#endif
	   {_T("euc-kr"), 51949},
#ifndef __CTABSMALL__
	   {_T("euc-tw"), 51950},
#endif
	   {_T("gb-2312-80"), 936},
	   {_T("gb18030"), 936},
	   {_T("gb18030-2000"), 936},
	   {_T("gb2312"), 936},
	   {_T("gb2312-80"), 936},
	   {_T("gbk"), 936},
#ifndef __CTABSMALL__
	   {_T("german"), 20106},
#endif
	   {_T("greek"), 28597},
	   {_T("greek8"), 28597},
	   {_T("hebrew"), 28598},
	   {_T("hz-gb-2312"), 52936},
#ifndef __CTABSMALL__
	   {_T("ibm-thai"), 20838},
#endif
	   {_T("ibm00858"), 858},
#ifndef __CTABSMALL__
	   {_T("ibm00924"), 20924},
	   {_T("ibm01047"), 1047},
	   {_T("ibm01140"), 1140},
	   {_T("ibm01141"), 1141},
	   {_T("ibm01142"), 1142},
	   {_T("ibm01143"), 1143},
	   {_T("ibm01144"), 1144},
	   {_T("ibm01145"), 1145},
	   {_T("ibm01146"), 1146},
	   {_T("ibm01147"), 1147},
	   {_T("ibm01148"), 1148},
	   {_T("ibm01149"), 1149},
	   {_T("ibm037"), 37},
	   {_T("ibm037"), 37},
	   {_T("ibm1026"), 1026},
	   {_T("ibm1026"), 1026},
	   {_T("ibm273"), 20273},
	   {_T("ibm277"), 20277},
	   {_T("ibm278"), 20278},
	   {_T("ibm280"), 20280},
	   {_T("ibm284"), 20284},
	   {_T("ibm285"), 20285},
	   {_T("ibm290"), 20290},
	   {_T("ibm297"), 20297},
#endif
	   {_T("ibm367"), 20127},
#ifndef __CTABSMALL__
	   {_T("ibm420"), 20420},
	   {_T("ibm423"), 20423},
	   {_T("ibm424"), 20424},
#endif
	   {_T("ibm437"), 437},
#ifndef __CTABSMALL__
	   {_T("ibm500"), 500},
#endif
	   {_T("ibm737"), 737},
	   {_T("ibm775"), 775},
	   {_T("ibm819"), 28591},
	   {_T("ibm850"), 850},
	   {_T("ibm852"), 852},
	   {_T("ibm855"), 855},
	   {_T("ibm857"), 857},
	   {_T("ibm860"), 860},
	   {_T("ibm861"), 861},
	   {_T("ibm862"), 862},
	   {_T("ibm863"), 863},
	   {_T("ibm864"), 864},
	   {_T("ibm865"), 865},
	   {_T("ibm866"), 866},
	   {_T("ibm869"), 869},
#ifndef __CTABSMALL__
	   {_T("ibm870"), 870},
	   {_T("ibm871"), 20871},
	   {_T("ibm880"), 20880},
	   {_T("ibm905"), 20905},
	   {_T("irv"), 20105},
#endif
	   {_T("iso-10646-ucs-2"), 1200},
#ifndef __CTABSMALL__
	   {_T("iso-2022-jp"), 50220},
	   {_T("iso-2022-jp-1"), 50221},
	   {_T("iso-2022-jp-2"), 50222},
	   {_T("iso-2022-jpeuc"), 51932},
	   {_T("iso-2022-kr"), 50225},
	   {_T("iso-2022-kr-7"), 50225},
	   {_T("iso-2022-kr-7bit"), 50225},
#endif
	   {_T("iso-2022-kr-8"), 51949},
	   {_T("iso-2022-kr-8bit"), 51949},
	   {_T("iso-646.irv:1991"), 20127},
	   {_T("iso-8859-1"), 28591},
	   {_T("iso-8859-11"), 874},
	   {_T("iso-8859-13"), 28603},
	   {_T("iso-8859-15"), 28605},
	   {_T("iso-8859-1:1987"), 28591},
	   {_T("iso-8859-2"), 28592},
	   {_T("iso-8859-2:1987"), 28592},
	   {_T("iso-8859-3"), 28593},
	   {_T("iso-8859-3:1988"), 28593},
	   {_T("iso-8859-4"), 28594},
	   {_T("iso-8859-4:1988"), 28594},
	   {_T("iso-8859-5"), 28595},
	   {_T("iso-8859-5:1988"), 28595},
	   {_T("iso-8859-6"), 28596},
	   {_T("iso-8859-6:1987"), 28596},
	   {_T("iso-8859-7"), 28597},
	   {_T("iso-8859-7:1987"), 28597},
	   {_T("iso-8859-8 visual"), 28598},
	   {_T("iso-8859-8"), 28598},
	   {_T("iso-8859-8-i"), 38598},
	   {_T("iso-8859-8:1988"), 28598},
	   {_T("iso-8859-9"), 28599},
	   {_T("iso-8859-9:1989"), 28599},
	   {_T("iso-ir-100"), 28591},
	   {_T("iso-ir-101"), 28592},
	   {_T("iso-ir-109"), 28593},
	   {_T("iso-ir-110"), 28594},
	   {_T("iso-ir-126"), 28597},
	   {_T("iso-ir-127"), 28596},
	   {_T("iso-ir-138"), 28598},
	   {_T("iso-ir-144"), 28595},
	   {_T("iso-ir-148"), 28599},
	   {_T("iso-ir-149"), 949},
	   {_T("iso-ir-58"), 936},
	   {_T("iso-ir-6"), 20127},
	   {_T("iso646-us"), 20127},
	   {_T("johab"), 1361},
	   {_T("koi"), 20866},
	   {_T("koi8"), 20866},
	   {_T("koi8-r"), 20866},
	   {_T("koi8-ru"), 21866},
	   {_T("koi8-u"), 21866},
	   {_T("koi8r"), 20866},
	   {_T("korean"), 949},
	   {_T("ks-c-5601"), 949},
	   {_T("ks-c-5601-1987"), 949},
	   {_T("ks-c-5601-1989"), 949},
	   {_T("ks-c5601"), 949},
	   {_T("ksc-5601"), 949},
	   {_T("ksc5601"), 949},
	   {_T("l1"), 28591},
	   {_T("l2"), 28592},
	   {_T("l3"), 28593},
	   {_T("l4"), 28594},
	   {_T("l5"), 28599},
	   {_T("l9"), 28605},
	   {_T("latin-1"), 28591},
	   {_T("latin-2"), 28592},
	   {_T("latin-3"), 28593},
	   {_T("latin-4"), 28594},
	   {_T("latin-5"), 28599},
	   {_T("latin-9"), 28605},
	   {_T("latin1"), 28591},
	   {_T("latin2"), 28592},
	   {_T("latin3"), 28593},
	   {_T("latin4"), 28594},
	   {_T("latin5"), 28599},
	   {_T("latin9"), 28605},
	   {_T("logical"), 28598},
	   {_T("macarabic"), 10004},
	   {_T("maccentraleurope"), 10029},
	   {_T("macchinesesimp"), 10008},
	   {_T("macchinesetrad"), 10002},
	   {_T("maccyrillic"), 10007},
	   {_T("macgreek"), 10006},
	   {_T("machebrew"), 10005},
	   {_T("macicelandic"), 10079},
	   {_T("macintosh"), 10000},
	   {_T("macjapanese"), 10001},
	   {_T("mackorean"), 10003},
	   {_T("macroman"), 10000},
	   {_T("macturkish"), 10081},
	   {_T("ms-kanji"), 932},
#ifndef __CTABSMALL__
	   {_T("norwegian"), 20108},
	   {_T("ns-4551-1"), 20108},
	   {_T("sen-850200-b"), 20107},
#endif
	   {_T("shift-jis"), 932},
	   {_T("shift_jis"), 932},
	   {_T("shiftjis"), 932},
	   {_T("sjis"), 932},
#ifndef __CTABSMALL__
	   {_T("swedish"), 20107},
#endif
	   {_T("tis-620"), 874},
	   {_T("ucs-2"), 1200},
	   {_T("ucs2"), 1200}, 
	   {_T("unicode"), 1200},
	   {_T("unicode-1-1-utf-7"), 65000},
	   {_T("unicode-1-1-utf-8"), 65001},
	   {_T("unicode-2-0-utf-7"), 65000},
	   {_T("unicode-2-0-utf-8"), 65001},
	   {_T("unicodefffe"), 1201},
	   {_T("us"), 20127},
	   {_T("us-ascii"), 20127},
	   {_T("utf-16"), 1200},
	   {_T("utf-16be"), 1201},
	   {_T("utf-16le"), 1200},
	   {_T("utf-32"), 65005},
	   {_T("utf-32be"), 65006},
	   {_T("utf-7"), 65000},
	   {_T("utf-8"), 65001},
	   {_T("utf16"), 1200},
	   {_T("utf16be"), 1201},
	   {_T("utf16le"), 1200},
	   {_T("utf8"), 65001},
	   {_T("visual"), 28598},
	   {_T("windows-1250"), 1250},
	   {_T("windows-1251"), 1251},
	   {_T("windows-1252"), 1252},
	   {_T("windows-1253"), 1253},
	   {_T("windows-1254"), 1254},
	   {_T("windows-1255"), 1255},
	   {_T("windows-1256"), 1256},
	   {_T("windows-1257"), 1257},
	   {_T("windows-1258"), 1258},
	   {_T("windows-874"), 874},
	   {_T("x-ansi"), 1252},
	   {_T("x-chinese-cns"), 20000},
	   {_T("x-chinese-eten"), 20002},
	   {_T("x-cp1250"), 1250},
	   {_T("x-cp1251"), 1251},
	   {_T("x-cp20001"), 20001},
	   {_T("x-cp20003"), 20003},
	   {_T("x-cp20004"), 20004},
	   {_T("x-cp20005"), 20005},
#ifndef __CTABSMALL__
	   {_T("x-cp20261"), 20261},
	   {_T("x-cp20269"), 20269},
#endif
	   {_T("x-cp20936"), 20936},
#ifndef __CTABSMALL__
	   {_T("x-cp20949"), 20949},
	   {_T("x-cp21027"), 21027},
	   {_T("x-cp50227"), 50227},
	   {_T("x-euc"), 51932},
	   {_T("x-euc-cn"), 51936},
	   {_T("x-euc-jp"), 51932},
	   {_T("x-europa"), 29001},
	   {_T("x-ia5"), 20105},
	   {_T("x-ia5-german"), 20106},
	   {_T("x-ia5-norwegian"), 20108},
	   {_T("x-ia5-swedish"), 20107},
#endif
	   {_T("x-iscii-as"), 57006},
	   {_T("x-iscii-be"), 57003},
	   {_T("x-iscii-de"), 57002},
	   {_T("x-iscii-gu"), 57010},
	   {_T("x-iscii-ka"), 57008},
	   {_T("x-iscii-ma"), 57009},
	   {_T("x-iscii-or"), 57007},
	   {_T("x-iscii-pa"), 57011},
	   {_T("x-iscii-ta"), 57004},
	   {_T("x-iscii-te"), 57005},
	   {_T("x-mac-arabic"), 10004},
	   {_T("x-mac-ce"), 10029},
	   {_T("x-mac-chinesesimp"), 10008},
	   {_T("x-mac-chinesetrad"), 10002},
	   {_T("x-mac-croatian"), 10082},
	   {_T("x-mac-cyrillic"), 10007},
	   {_T("x-mac-greek"), 10006},
	   {_T("x-mac-hebrew"), 10005},
	   {_T("x-mac-icelandic"), 10079},
	   {_T("x-mac-japanese"), 10001},
	   {_T("x-mac-korean"), 10003},
	   {_T("x-mac-romanian"), 10010},
#ifndef __CTABSMALL__
	   {_T("x-mac-thai"), 10021},
	   {_T("x-mac-turkish"), 10081},
#endif
	   {_T("x-mac-ukrainian"), 10017},
	   {_T("x-ms-cp932"), 932},
	   {_T("x-sjis"), 932},
	   {_T("x-unicode-1-1-utf-7"), 65000},
	   {_T("x-unicode-1-1-utf-8"), 65001},
	   {_T("x-unicode-2-0-utf-7"), 65000},
	   {_T("x-unicode-2-0-utf-8"), 65001},
	   {_T("x-x-big5"), 950}
   };

 
   static const size_t name_id_tab_size = sizeof(name_id_tab) / sizeof(Name_Id);

   static int nameCmpFunc(const void* k, const void* v) {
	   return _tcsicmp((const BppChar*)k, ((const Name_Id*)v)->name);
   }
#endif //!NO_CHARCONV

   BppUShort getCpIdFromName(const BppChar* name, BppUShort defcp)
   {
#ifndef NO_CHARCONV
	   const Name_Id* item = (const Name_Id*) bsearch(name, name_id_tab, name_id_tab_size, sizeof(Name_Id), nameCmpFunc);
	   if (item) return item->id;
	   BppUShort id = _ttoi(name);
	   if (id) return id;
#endif
	   return defcp;
   }

   //const BppChar* getNameFromCpId(BppUShort cp) {
   //    switch (cp) {
		 //  case BPP_UTF8: 
		 //     return _T("UTF-8");
		 //  case BPP_UTF16:
		 //     return _T("UTF-16");
		 //  case BPP_UTF16be:
		 //     return _T("UTF-16be");
		 //  case BPP_UTF32:
		 //     return _T("UTF-32");
		 //  case BPP_UTF32be:
		 //     return _T("UTF-32be");
	  // }
	  // for (size_t i = 0;i<name_id_tab_size;++i) {
	  //    if (name_id_tab[i].id == cp) return name_id_tab[i].name;
	  // }
	  // return 0;
   //}

   BppUShort getActiveCp(BppUShort defcp)
   {
#ifndef NO_CHARCONV
#ifdef WIN32
	   BppUShort cp = GetACP();
	   if ((cp) && ((cp & 0x8000) == 0)) return cp;
#endif
#ifdef __MSDOS__	   
	   static union REGS _regs;
	   _regs.x.ax = 0x6601;
	   int86(0x21, &_regs, &_regs);
	   if (!_regs.x.cflag) return _regs.x.bx;
#else
	   const BppChar* loc = _tsetlocale(LC_ALL, 0);
	   if (loc) {
		   loc = _tcschr(loc, _T('.'));
		   if (loc) {
		      BppChar buf[100];
			  _tcscpy(buf, loc + 1);
			  for (BppChar* p = _tcschr(buf, _T('_')); p && *p; p = _tcschr(p + 1, _T('_'))) 
				  *p = _T('-');
			  return getCpIdFromName(buf, defcp);
		   }
	   }
#endif
#endif // !NO_CHARCONV
	   return defcp;
   }

   unsigned detectFileCp(FILE* fp)
   {
#ifndef NO_CHARCONV
	   if (!fp) return 0;
	   char cbuf[4];
	   size_t foffset = 0;
	   unsigned result = 0;
	   if (fread(cbuf, 1, 4, fp) != 4) return result;
	   switch ((unsigned char)cbuf[0]) {
	   case 0xef:
		   if (cbuf[1] == 0xbb && cbuf[2] == 0xbf) {
			   result = BPP_UTF8; // utf-8
			   foffset = 3;
		   }
		   break;
	   case 0xfe:
		   if (cbuf[1] == 0xff) {
			   result = BPP_UTF16be; // utf16be
			   foffset = 2;
		   }
		   break;
	   case 0xff:
		   if (cbuf[1] == 0xfe) {
			   if (cbuf[2] == 0 && cbuf[3] == 0) {
				   result = BPP_UTF32; // utf32le
				   foffset = 4;
			   }
			   else {
				   result = BPP_UTF16; // utf16le
				   foffset = 2;
			   }
		   }
		   break;
	   case 0:
		   if (cbuf[1] == 0 && cbuf[2] == 0xfe && cbuf[3] == 0xff) {
			   result = BPP_UTF32be; // utf32be
			   foffset = 4;
		   }
		   else if (cbuf[1] == 0) {
			   result = BPP_UTF32be; // utf32be
			   foffset = 4;
		   }
		   else {
			   result = BPP_UTF16be; // utf32be
			   foffset = 2;
		   }
		   break;
	   default:
		   if (cbuf[1] == 0) {
			   if (cbuf[2] == 0 && cbuf[3] == 0) {
				   result = BPP_UTF32;
				   foffset = 4;
			   }
			   else {
				   result = BPP_UTF16;
				   foffset = 2;
			   }
		   }
	   }
	   if (!result) {
		   result = getActiveCp();
		   if (result == BPP_UTF16 || result == BPP_UTF16be || result == BPP_UTF32 || result == BPP_UTF32be)
			   result = DEF_CP;
	   }
	   fseek(fp, foffset, SEEK_SET);
	   return result;
#else
	   return DEF_CP;
#endif
   }

#if defined(UNIX)
	#include <termios.h>
	#include <unistd.h>
	#include <fcntl.h>
#else
	#if defined(WIN32)
	#include <Windows.h>
	#endif
#endif
   bool keyPressed()
   {
#if defined(UNIX)
	   struct termios oldt, newt;
	   int ch;
	   int oldf;

	   tcgetattr(STDIN_FILENO, &oldt);
	   newt = oldt;
	   newt.c_lflag &= ~(ICANON | ECHO);
	   tcsetattr(STDIN_FILENO, TCSANOW, &newt);
	   oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
	   fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

	   ch = getchar();

	   tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
	   fcntl(STDIN_FILENO, F_SETFL, oldf);

	   if (ch != EOF)
	   {
		   ungetc(ch, stdin);
		   return true;
	   }
	   return false;
#else 
#if defined(WIN32)
  #if !defined(WINCE)
	if (GetConsoleWindow())
		return (_kbhit() != 0);
  #endif
	for (int keyCode = 0x0d; keyCode <= VK_OEM_102; ++keyCode) {
		if ((keyCode != 0x0d && keyCode != VK_SPACE && keyCode < 0x30) ||
			(keyCode > 0x39 && keyCode < 0x41) || (keyCode > 0x5A && keyCode < VK_NUMPAD0) ||
			(keyCode >= VK_F1 && keyCode < VK_OEM_1) || (keyCode > VK_OEM_3 && keyCode < VK_OEM_4) ||
			(keyCode > VK_OEM_8 && keyCode < VK_OEM_102))
			continue;

		// Check if the key with keyCode is currently
		// pressed
		if (GetAsyncKeyState(keyCode) & 0x8000) {
			// Convert the key code to ASCII character
			return true;
		}
   }
	return false;
#else
	   return (kbhit() != 0);
#endif
#endif
   }


#ifdef __OLDCPP__
	int snprintf(char* s, size_t, const char* format, ...) {
		va_list args;
		va_start(args, format);
		int result = vsprintf(s, format, args);
		va_end(args);
		return result;
	}
#endif
#if defined(UNIX)
#include <ctype.h>
#include <unistd.h>
   char* _fullpath(char* absPath,const char* relPath, size_t)
   {
	   return realpath(relPath,absPath);
   }
     wchar_t* _wfullpath(wchar_t* absPath,const wchar_t* relPath, size_t maxLength) {
		 BppBuffer buf(maxLength+1,maxLength);
		 if (!realpath((const char*) wcs2strBuf(relPath).data(),(char*)buf.data())) {
			 *absPath = 0;
			 return 0;
		 }
		 BppBuffer result = str2wcsBuf((const char*)buf.data());
		 wcsncpy(absPath,(const wchar_t*)result.data(),maxLength);
		 return absPath;
	 };

   char* strlwr(char* s)
   {
	   for (unsigned char* p = (unsigned char*)s; *p;++p) *p = tolower(*p);
	   return s;
   }
  
   char* strupr(char* s)
    {
	   for (unsigned char* p = (unsigned char*)s; *p;++p) *p = toupper(*p);
	   return s;
   }
  
   wchar_t* wcslwr(wchar_t* s) {
	   for (wchar_t* p = s; *p;++p) *p = towlower(*p);
	   return s;	   
   }
   wchar_t* wcsupr(wchar_t* s) {
	   for (wchar_t* p = s; *p;++p) *p = towupper(*p);
	   return s;	   
   }

	FILE* _wfopen(const wchar_t* fname, const wchar_t* fmode) {
		BppBuffer name = wcs2strBuf(fname);
		BppBuffer mode = wcs2strBuf(fmode);
		return fopen((const char*)name.data(),(const char*)mode.data());		
	}

    wchar_t* _wctime(const time_t* timer) {
		static wchar_t buf[50];
	    mbstowcs(buf,ctime(timer),50);
		return buf;   
    }
	
	static BppBuffer envbuf;
	wchar_t* wgetenv(const wchar_t* name) {
		static BppBuffer envbuf;
		BppBuffer namebuf = wcs2strBuf(name);
		envbuf = str2wcsBuf(getenv((const char*)namebuf.data()));
		return (wchar_t*) envbuf.data();
	}
	
	int wputenv(wchar_t* val) {
		static BppBuffer valbuf;
		valbuf = wcs2strBuf(val);
		return putenv((char*)valbuf.data());
	}
	
   FILE* _wfreopen(const wchar_t* fname, const wchar_t* fmode, FILE* stream) {
		BppBuffer name = wcs2strBuf(fname);
		BppBuffer mode = wcs2strBuf(fmode);
		return freopen((const char*)name.data(),(const char*)mode.data(),stream);			   
   }
   wchar_t* _wsetlocale(int category, const wchar_t* locale) {
	   static BppBuffer result;
	   result = str2wcsBuf(setlocale(category,locale ? (const char*)wcs2strBuf(locale).data(): 0));
	   return (wchar_t*) result.data();
   }
   int _wsystem(const wchar_t* command) {
	   return(system((const char*)wcs2strBuf(command).data()));
   }
#ifndef NO_DYNLIB
//#define FreeLibrary(hModule) dlclose(hModule)
//#define GetProcAddressA(hModule,lpProcName) dlsym(hModule,lpProcName)
   void* LoadLibraryW(const wchar_t* lpFileName) {
	   return dlopen((const char*)wcs2strBuf(lpFileName).data(),RTLD_NOW);
   }
   
   void* GetProcAddressW(void* hModule,const wchar_t* lpProcName) {
	   return dlsym(hModule,(const char*)wcs2strBuf(lpProcName).data());
   }
#endif
   long wtol(const wchar_t* str) {
	   return wcstol(str,NULL,10);
   }
   int wtoi(const wchar_t* str) {
	   return wcstol(str,NULL,10);
   }
   wchar_t* wstok(wchar_t* str, const wchar_t* delim) {
	   static wchar_t* buf;
	   return wcstok(str,delim,&buf);
   }
    wchar_t* _wgetcwd(wchar_t* buffer, size_t bsize) {
	   
	   char* r = getcwd(0,0);
	   if (!r) return 0;
	   if (!buffer || bsize == 0) {
		   bsize = strlen(r) +1;
		   buffer = (wchar_t*) allocmem(bsize * sizeof(wchar_t));
	   }
	   mbstowcs(buffer,r,bsize);
	   freemem(r);
	   return buffer;
	}
	
     int _wchdir(const wchar_t* dirname) {
	     return chdir((const char*)wcs2strBuf(dirname).data());
	 }
#endif

