#include <float.h>
#include <stdlib.h>
#include "bppvalue.h"
#include "bppstrng.h"
#include "bppclass.h"
#include "bppobjct.h"
#include "bppdict.h"
#include "bppbuf.h"
#include "bppvec.h"
#include "bppconst.h"
#include "bppufunc.h"

/**
@param type Type-ID.
@param getPrefix Flag for using 'BVT_' prefix.
@param noInvalidType Flag for return nullptr insted of 'INVALID TYPE' on invalid type ID.
@return Readable type name. If \e getPrefix is \c true, the return value is same
as constant identifier defined in BppValueType enumeration, otherwise
'BVT_' prefix is removed.
*/
const BppChar* getValueTypeName(BppValueType type, bool getPrefix, bool noInvalidType)
{
	const BppChar* tname = 0;
#define VTSTR(VT) case VT: tname = _T(#VT); break
	switch (type)
	{
		VTSTR(BVT_NULL);
		VTSTR(BVT_VOID);
		VTSTR(BVT_AUTO);
		VTSTR(BVT_NIL);
		VTSTR(BVT_NUMBER);
		VTSTR(BVT_INTNUMBER);
		VTSTR(BVT_UINTNUMBER);
		VTSTR(BVT_INTEGRALTYPE);
		VTSTR(BVT_PTRTYPE);
		VTSTR(BVT_REALNUMBER);
		VTSTR(BVT_REFCNTTYPE);
		VTSTR(BVT_REFERENCE);
		VTSTR(BVT_SBYTE);
		VTSTR(BVT_BOOL);
		VTSTR(BVT_CHAR);
		VTSTR(BVT_BYTE);
		VTSTR(BVT_UCHAR);
		VTSTR(BVT_CHAR8);
		VTSTR(BVT_SHORT);
		VTSTR(BVT_USHORT);
		VTSTR(BVT_CHAR16);
		VTSTR(BVT_WCHAR);
		VTSTR(BVT_INT);
		VTSTR(BVT_INTPTR);
		VTSTR(BVT_UINT);
		VTSTR(BVT_UINTPTR);
		VTSTR(BVT_CHAR32);
		VTSTR(BVT_LONG);
		VTSTR(BVT_ULONG);
		VTSTR(BVT_FLOAT);
		VTSTR(BVT_DOUBLE);
		VTSTR(BVT_FILE);
		VTSTR(BVT_CODE);
		VTSTR(BVT_BYTECODE);
		VTSTR(BVT_LOCALVARINFO);
		VTSTR(BVT_BUFFER);
		VTSTR(BVT_CHARBUFFER);
		VTSTR(BVT_CLASS);
		VTSTR(BVT_OBJECT);
		VTSTR(BVT_STRONGOBJECT);
		VTSTR(BVT_VECTOR);
		VTSTR(BVT_STRING);
		VTSTR(BVT_DICTIONARY);
		VTSTR(BVT_VAR);
		VTSTR(BVT_BYTECODEREF);
		VTSTR(BVT_BUFFERREF);
		VTSTR(BVT_CHARBUFFERREF);
		VTSTR(BVT_CLASSREF);
		VTSTR(BVT_OBJECTREF);
		VTSTR(BVT_VECTORREF);
		VTSTR(BVT_STRINGREF);
		VTSTR(BVT_DICTIONARYREF);
	default:
	   return noInvalidType ? (const BppChar*)0 : _T("INVALID TYPE");
  }	
	return getPrefix ? tname : tname+4;
}

BppValue BppValue::makeBool(bool val)
{
	BppValue  result;
	result._v.v_byte = val ? 1 : 0;
	result.v_type = BVT_BOOL;
	return result;
}

BppValue BppValue::makeChar(BppUInt val, BppValueType vt)
{
	BppValue  result;
	result._v.v_uint = val;
	result.v_type = vt;
	if (!result.isCharacterType()) result.v_type = BVT_TCHAR;
	return result;

}

BppValue BppValue::makeChar(const BppValue& val, BppValueType vt)
{
	BppValue  result;
	result.v_type = vt;
	if (!result.isCharacterType()) {
		result.v_type = BVT_TCHAR;
		vt = BVT_TCHAR;
	}
	vt = mkWcharInternal(vt);
	BppValueType srcT = mkWcharInternal(val.getType());
	switch (srcT) {
	case BVT_CHAR:
	case BVT_UCHAR:
		switch (vt) {
		case BVT_CHAR8:
			result._v.v_uint = char_to_u8(BppChar(val._v.v_byte));
			break;
		case BVT_CHAR16:
			result._v.v_uint = char_to_c16(BppChar(val._v.v_byte));
			// sbbToUtf16
			break;
		case BVT_CHAR32:
			result._v.v_uint = char_to_c32(BppChar(val._v.v_byte));
			// sbToUtf32
			break;
		default:
			result._v.v_uint = (BppByte)val;
		}
		break;
	case BVT_CHAR8:
		switch (vt) {
		case BVT_CHAR8:
			result._v.v_uint = val;
			break;
		case BVT_CHAR16:
			result._v.v_uint = u8_to_c16(val);
			// utf8ToUtf16
			break;
		case BVT_CHAR32:
			result._v.v_uint = u8_to_c32(val);
			// utf8ToUtf16
			break;
		default:
			result._v.v_uint = u8_to_char(val);
			// utf8ToSb;
		}
		break;
	case BVT_CHAR16:
		switch (vt) {
		case BVT_CHAR8:
			//utf16Toutf8;
			result._v.v_uint = c16_to_u8(val);
			break;
		case BVT_CHAR16:
			result._v.v_uint = BppUShort(val);
			break;
		case BVT_CHAR32:
			result._v.v_uint = c16_to_c32(val);
			// utf16ToUtf32
			break;
		default:
			result._v.v_uint = c16_to_char(val);
			// utf16ToSb;
		}
		break;
	case BVT_CHAR32:
		switch (vt) {
		case BVT_CHAR8:
			result._v.v_uint = c32_to_u8(val);
			//utf32Toutf8;
			break;
		case BVT_CHAR16:
			result._v.v_uint = c32_to_c16(val);
			// utf32ToUtf16
			break;
		case BVT_CHAR32:
			result._v.v_uint = val;
			break;
		default:
			result._v.v_uint = c32_to_char(val);
			// utf32ToSb;
		}
		break;
	default:
	   result._v.v_uint = val;
	   switch (vt) {
	   case BVT_CHAR:
	   case BVT_UCHAR:
		   result._v.v_uint &= 0xff;
		   break;
	   case BVT_CHAR16:
		   result._v.v_uint &= 0xffff;
	   }
	}

	return result;
}


/**
Ctor initializes value and value type to be not set.
*/
BppValue::BppValue() BPPNOEXCEPT
:v_exttype(BVT_NULL,0)
{
#ifdef EMULATE_I64
	_v.v_dwords[0] = 0;
	_v.v_dwords[1] = 0;
#else
	_v.v_ulong = 0;
#endif
}

/**
Constructor creates value object of specified type.
@param vtype Type of value to create.
@param flags Additional flags to set for value.
@param createDefaultObject If vtype specifies a string or a container type 
(buffer, charbuffer, vector or dictionary), flag allows creation of empty
container instead of initialization with <tt>null</tt>.
*/
BppValue::BppValue(BppValueType vtype, BppValueFlags flags, bool createDefaultObjects)
:v_exttype(vtype,flags)
{
#ifdef EMULATE_I64
	_v.v_dwords[0] = 0;
	_v.v_dwords[1] = 0;
#else
	_v.v_ulong = 0;
#endif
	if (createDefaultObjects)
	{
	  switch (vtype)
	  {
	  case BVT_BUFFER:
		  _v.v_refObj = new BppBuffer();
		  break;
	  case BVT_CHARBUFFER:
		  _v.v_refObj = new BppCharBuffer();
		  break;
	  case BVT_VECTOR:
		  _v.v_refObj =  new BppVector();
		  break;
	  case BVT_STRING:
		  _v.v_refObj = new BppString();
		  break;
	  case BVT_DICTIONARY:
		  _v.v_refObj = new BppDictionary();
		  break;
	  }
	  if (_v.v_refObj) _v.v_refObj->addRef();
	}
}

/**
@param val Pointer to reference counted object.
@param vtype value type id.
@note This constructor is used intrnally for object deserialization.
@see BppDictionaryIO::readValue
*/
BppValue::BppValue(BppRefObject* val,  BppValueType vtype) BPPNOEXCEPT
:v_exttype(vtype,0)
{
	_v.v_refObj = val;
	if ((vtype < _BVTREFMIN) && (val != 0))
	{
		val->addRef();
	}
	if (vtype == BVT_STRONGOBJECT)
	  v_flags = BVF_STRONGTYPED;
}

/**
@param val Pointer to class value.
@param isRef If flag set, value type is set to BVT_CLASSREF, otherwise
value type id BVT_CLASS and reference counter of <i>val</i> will be incremented.
*/
BppValue::BppValue(BppClass* val, bool isRef) BPPNOEXCEPT
:v_exttype((isRef ? BVT_CLASSREF : BVT_CLASS),0)
{
	_v.v_class = val;
	if ((val != 0) && (!isRef))
		val->addRef();
}

/**
@param val Pointer to object value.
@param isRef If flag set, value type is set to BVT_OBJECTREF, otherwise
value type id BVT_OBJECT and reference counter of <i>val</i> will be incremented.
*/
BppValue::BppValue(BppObject* val, bool isRef) BPPNOEXCEPT
:v_exttype((isRef ? BVT_OBJECTREF : BVT_OBJECT),0)
{
	_v.v_object = val;
	if ((val != 0) && (!isRef))
		val->addRef();
}

/**
@param val Pointer to vector value.
@param isBytecode
@param isRef If flag set, value type is set to BVT_VECTORREF or BVT_BYTECODEREF, otherwise
value type id BVT_VECTOR, BVT_BYTECODE or BVT_LOCALVARINFO and reference counter of <i>val</i> will be incremented.
@note B++ uses vectors for internal representation of bytecode (i.e. user defined functions) and uses
value type attribute to distinguish between them. Furthermore last element of a bytecode vector may contain
additional information about local variables os a function for debugging purpuses, Because such additional
information has an internal structure different from "normal" bytecode, this constructor
sets type BVT_LOCALVARINFO if such structure is found.
*/
BppValue::BppValue(BppVector* val,bool isBytecode, bool isRef) BPPNOEXCEPT
:v_exttype((isBytecode ? (isRef ? BVT_BYTECODEREF : BVT_BYTECODE) : (isRef ? BVT_VECTORREF : BVT_VECTOR)),0)
{
	_v.v_vector = val;
	if ((val != 0) && (!isRef))
		val->addRef();
	else return;
	// Hack: 
	// Ein Bytecode-Vektor wird  ein LOCALVARINFO-Vektor, falls
	// er genau zwei Elemente hat, die wiederum Vektoren sind.
	// "Echter" Bytecode hat nie eine solche Struktur.
	if (v_type == BVT_BYTECODE)
	{
		const BppVector& v = *val;
		if (v.size() == 2)
		{
			if ((v[0].getType() == BVT_VECTOR) && (v[1].getType() == BVT_VECTOR))
				v_type = BVT_LOCALVARINFO;
		}
	}
}

/**
@param val Pointer to string value.
@param isRef If flag set, value type is set to BVT_STRINGREF, otherwise
value type id BVT_STRING and reference counter of <i>val</i> will be incremented.
*/
BppValue::BppValue(BppString* val, bool isRef) BPPNOEXCEPT
:v_exttype((isRef ? BVT_STRINGREF : BVT_STRING),0)
{
	_v.v_string = val;
	if ((val != 0) && (!isRef))
		val->addRef();
}

/**
@param cpy String value.
*/
BppValue::BppValue(const BppString& cpy)
:v_exttype(BVT_STRING, 0)
{
	_v.v_string = new BppString(cpy);
	_v.v_string->addRef();
}

/**
@param val Pointer to byte buffer value.
@param isRef If flag set, value type is set to BVT_BUFFERREF, otherwise
value type id BVT_BUFFER and reference counter of <i>val</i> will be incremented.
*/
BppValue::BppValue(BppBuffer* val, bool isRef) BPPNOEXCEPT
:v_exttype((isRef ? BVT_BUFFERREF : BVT_BUFFER),0)
{
	_v.v_buffer = val;
	if ((val != 0) && (!isRef))
		val->addRef();
}

/**
@param val Pointer to character buffer value.
@param isRef If flag set, value type is set to BVT_CHARBUFFERREF, otherwise
value type id BVT_CHARBUFFER and reference counter of <i>val</i> will be incremented.
*/
BppValue::BppValue(BppCharBuffer* val, bool isRef) BPPNOEXCEPT
:v_exttype((isRef ? BVT_CHARBUFFERREF : BVT_CHARBUFFER),0)
{
	_v.v_charbuffer = val;
	if ((val != 0) && (!isRef))
		val->addRef();
}

/**
@param val Pointer to dictionary value.
@param isRef If flag set, value type is set to BVT_DICTIONARYREF, otherwise
value type id BVT_DICTIONARY and reference counter of <i>val</i> will be incremented.
*/
BppValue::BppValue(BppDictionary* val, bool isRef) BPPNOEXCEPT
:v_exttype((isRef ? BVT_DICTIONARYREF : BVT_DICTIONARY),0)
{
	_v.v_dict = val;
	if ((val != 0) && (!isRef))
		val->addRef();
}

/**
@param val Pointer to variable descriptor (dictionary entry value.
@note Variable descriptors are not reference counted.<br>
Value type is set to BVT_VAR.
*/
BppValue::BppValue(BppDictionaryEntry* val) BPPNOEXCEPT
:v_exttype(BVT_VAR,0)
{
	_v.v_var = val;
}

/**
@param val Pointer to native function.
@note Value type is set to BVT_CODE.
*/
BppValue::BppValue(BPPUSERFUNC val) BPPNOEXCEPT
:v_exttype(BVT_CODE,0)
{
	_v.v_code = val;
}

/**
@param val 8 bits signed integer value.
@note Value type is set to BVT_SBYTE.
*/
BppValue::BppValue(BppSByte val) BPPNOEXCEPT
:v_exttype(BVT_SBYTE,0)
{
	_v.v_sbyte = val;
}  
/**
@param val 8 bits unsigned integer value.
@note Value type is set to BVT_BYTE.
*/
BppValue::BppValue(BppByte val) BPPNOEXCEPT
:v_exttype(BVT_BYTE,0)
{
	_v.v_byte = val;
}

/**
@param val 16 bits signed integer value.
@note Value type is set to BVT_SHORT.
*/
BppValue::BppValue(BppShort val) BPPNOEXCEPT
:v_exttype(BVT_SHORT,0)
{
	_v.v_short = val;
}

/**
@param val 16 bits unsigned integer value.
@note Value type is set to BVT_USHORT.
*/
BppValue::BppValue(BppUShort val) BPPNOEXCEPT
:v_exttype(BVT_USHORT,0)
{
	_v.v_ushort = val;
}

/**
@param val 32 bits signed integer value.
@param isptr specifies if value is a pointer
@note Value type normally is set to BVT_INT. If <i>isptr</i> is <tt>true</tt>,
value type is set to BVT_INTPTR instead.
*/
BppValue::BppValue(BppInt val,bool isptr) BPPNOEXCEPT
:v_exttype((isptr ? BVT_INTPTR : BVT_INT),0)
{
	_v.v_addr = 0;
	_v.v_integer = val;
}

/**
@param val 32 bits unsigned integer value.
@note Value type is set to BVT_UINT.
*/
BppValue::BppValue(BppUInt val, bool isptr) BPPNOEXCEPT
:v_exttype((isptr ? BVT_UINTPTR : BVT_UINT),0)
{
	_v.v_addr = 0;
	_v.v_uint = val;
}

/**
@param val 64 bits signed integer value.
@param isptr specifies if value is a pointer
@note Value type normally is set to BVT_LONG. If <i>isptr</i> is <tt>true</tt>,
value type is set to BVT_INTPTR instead.
*/
BppValue::BppValue(BppLong val,bool isptr) BPPNOEXCEPT
:v_exttype((isptr ? BVT_INTPTR : BVT_LONG),0)
{
	_v.v_long = val;
}

/**
@param val 64 bits unsigned integer value.
@note Value type is set to BVT_ULONG.
*/
BppValue::BppValue(BppULong val, bool isptr) BPPNOEXCEPT
:v_exttype((isptr ? BVT_UINTPTR : BVT_ULONG),0)
{
	_v.v_ulong = val;
}

/**
@param val Single precision real (float) value.
@note Value type is set to BVT_FLOAT.
*/
BppValue::BppValue(float val) BPPNOEXCEPT
:v_exttype(BVT_FLOAT,0)
{
	_v.v_float = val;
}

/**
@param val Single precision real (double) value.
@note Value type is set to BVT_DOUBLE.
*/
BppValue::BppValue(double val) BPPNOEXCEPT
:v_exttype(BVT_DOUBLE,0)
{
	_v.v_double = val;
}

/**
@param val File descriptor pointer.
@note Value type is set to BVT_FILE.
*/
BppValue::BppValue(FILE* val) BPPNOEXCEPT
:v_exttype(BVT_FILE,0)
{
	_v.v_fp = val;
}

/**
Copy constructor behaves as follows:
<ul>
<li>If <i>cpy</i> is of value type, value is copied.</li>
<li>If <i>cpy</i> is of reference type, only reference (pointer) to value is copied.</li>
<li>If <i>cpy</i> is a reference counted value, reference counter is incremented.
<li>Value type attribute and value flags are copied in all cases.</li>
</ul>
@param cpy Value to copy from.
@see BppValueType, BppValueFlags
*/
BppValue::BppValue(const BppValue& cpy, bool clone)
:v_exttype(cpy.v_exttype),_v(cpy._v)
{
	if (_v.v_var == 0)
		return;
	if (!isRefCounted()) return;
	switch (v_type)
	{
	case BVT_CLASS:
		_v.v_class->addRef();
		break;
	case BVT_OBJECT:
		if (clone) _v.v_object = _v.v_object->clone();
		_v.v_object->addRef();
		break;
	case BVT_STRONGOBJECT:
		if (clone) _v.v_vector = _v.v_vector->clone();
		else
		{
			BppClass* c = (*_v.v_vector)[IDX_CLASS];
			BppObject* o = (*_v.v_vector)[IDX_CODE];
			_v.v_vector = new BppVector(2,true);
			  (*_v.v_vector)[IDX_CLASS] = c;
			  (*_v.v_vector)[IDX_CODE] = o;
		}
		_v.v_vector->addRef();
		break;
	case BVT_BUFFER:
		if (clone) _v.v_buffer = _v.v_buffer->clone();
		_v.v_buffer->addRef();
		break;
	case BVT_CHARBUFFER:
		if (clone) _v.v_charbuffer = _v.v_charbuffer->clone();
		_v.v_charbuffer->addRef();
		break;
	case BVT_STRING:
		if (clone) _v.v_string = _v.v_string->clone();
		_v.v_string->addRef();
		break;
	case BVT_DICTIONARY:
		if (clone) _v.v_dict = _v.v_dict->clone();
		_v.v_dict->addRef();
		break;
	case BVT_BYTECODE:
		if (clone) {
			if ((_v.v_vector->_data[IDX_CODE].v_exttype._flags) & BVF_STATIC) {
				_v.v_vector = _v.v_vector->clone();
				_v.v_vector->_data[IDX_FUNCSTATE].dispose();
			}
		}
		 _v.v_vector->addRef();
		break;
	case BVT_LOCALVARINFO:
	case BVT_VECTOR:
		if (clone) _v.v_vector = _v.v_vector->clone();
		_v.v_vector->addRef();
		break;
	}
}
#ifdef CPP11

BppValue::BppValue(BppValue && cpy) BPPNOEXCEPT
:_v(cpy._v), v_exttype(cpy.v_exttype)
{
	//rawcopy(*this, cpy);
	if (v_type < _BVTREFTYPEBASE) return;
	cpy.v_type = BVT_NULL;
	cpy.v_flags = 0;
#ifdef EMULATE_I64
	_v.v_dwords[0] = 0;
	_v.v_dwords[1] = 0;
#else
	_v.v_ulong = 0;
#endif
}
#endif
/**
Destructor simply calls dispose method.
@see dispose
*/
BppValue::~BppValue()
{
	if (v_type >= _BVTREFTYPEBASE) dispose();
}


/**
First, if current value is of a reference counted type, its reference counter is decremented.<br>
Then <i>pref</i> is assigned to current value, If <i>pref</i> is of reference counted type,
its reference counter is incremented.
@param pref Pointer to reference type value to be assigned.
*/
void BppValue:: assignRef(BppRefObject* pref)
{
	if (_v.v_refObj == pref) return;
	if ((_v.v_refObj != 0) && (v_type <= BVT_RCNT_MAX) && (v_type >= BVT_RCNT_MIN))
		_v.v_refObj->release();
	_v.v_refObj = pref;
	if ((pref != 0) && (v_type < _BVTREFMIN))
		pref->addRef();
}



BppVector* BppValue::getStrongObjDesc() const BPPNOEXCEPT
{
	return v_type == BVT_STRONGOBJECT ? _v.v_vector : 0;
}

/**
Assignment operator behaves as follows:
<ul>
<li>First, dispose() is called to release old value.</li>
<li>If <i>assignment</i> is of value type, value is copied.</li>
<li>If <i>assignment</i> is of reference type, only reference (pointer) to value is copied.</li>
<li>If <i>assignment</i> is a reference counted value, reference counter is incremented.
<li>Value type attribute is copied in all cases.</li>
<li>If value type currently is BVT_NULL and BVF_STRONGTYPED flag not set, the BVF_STRONGTYPED
flag of <i>assignment</i> is copied. All other type flags are not copied.</li>
</ul>
@param assignment Value to assign.
@return Self-reference to value object.
@see dispose, BppValueType, BppValueFlags
*/
BppValue& BppValue::operator=(const BppValue& assignment)
{
	if (&assignment == this) return *this;
	if (assignment.isRefCounted() && assignment.v_type != BVT_STRONGOBJECT) {
		if (assignment._v.v_refObj)
		{
			if (_v.v_refObj == assignment._v.v_refObj) return *this;
			assignment._v.v_refObj->addRef();
		}
	}
	unsigned short vflags = v_flags;
	if (v_type == BVT_NULL && (v_flags & BVF_STRONGTYPED) == 0)
		vflags |= assignment.v_flags & BVF_STRONGTYPED;
	if (v_type >= _BVTREFTYPEBASE) dispose();
	v_exttype = assignment.v_exttype;
	v_flags = vflags;
	if (assignment.v_type == BVT_STRONGOBJECT && assignment._v.v_vector != 0)
	{
		BppClass* c = (*assignment._v.v_vector)[IDX_CLASS];
		BppObject* o = (*assignment._v.v_vector)[IDX_CODE];
		_v.v_vector = new BppVector(2,true);
		_v.v_vector->addRef();
		 (*_v.v_vector)[IDX_CLASS] = c;
		 (*_v.v_vector)[IDX_CODE] = o;
	}
	else
		_v = assignment._v;
	return *this;
}
#ifdef CPP11
BppValue & BppValue::operator=(BppValue && assignment)
{
	if (&assignment != this) {
		if (v_type >= _BVTREFTYPEBASE) dispose();
		_v = assignment._v;
		v_exttype = assignment.v_exttype;
		if (v_type < _BVTREFTYPEBASE) return *this;
		assignment.v_type = BVT_NULL;
		assignment.v_flags = 0;
		assignment._v.v_ulong =
#ifdef EMULATE_I64
			UInt64();
#else
			0;
#endif
	}
	return *this;
}
#endif
  /** Type checked assignment.

  @param assignment Value to assign.
  @return true on success
  */
  bool BppValue::assign(const BppValue& assignment)
  {
		if (&assignment == this) return true;
		unsigned short vt = mkWcharInternal(BppValueType(v_type));
		unsigned short vf = v_flags;
		if (v_flags & BVF_CONST)
			return false;
		unsigned short at = mkWcharInternal(BppValueType(assignment.v_type));
		if (at == BVT_VOID) return false;
	  if (((vf & BVF_STRONGTYPED) == 0) || 
		  ((at == vt) &&  vt != BVT_STRONGOBJECT) || 
		  (vt == BVT_NULL) || 
		  (vt == BVT_AUTO))
	  {
		  operator=(assignment);
		  v_flags &= ~BVF_CONST; // delete const flag
		  //if ((vf & BVF_STRONGTYPED) == 0 && vt != BVT_AUTO)
			 // v_flags &= ~((unsigned short)BVF_STRONGTYPED);
		  return true;
	  }
	  if (isCharacterType() && assignment.isCharacterType())
	  {
		  BppValue tmp = BppValue::makeChar(assignment, getType());
		  _v.v_uint = tmp;
		  return true;
	  }
	  if (at >= _BVTINTEGRALBASE && at <= BVT_FILE &&
		  vt >= _BVTINTEGRALBASE && vt <= BVT_FILE) {
		  switch (vt) {
		  case BVT_INT:
			  _v.v_integer = assignment;
			  break;
		  case BVT_CHAR8:
		  case BVT_UINT:
		  case BVT_CHAR32:
			  _v.v_uint = assignment;
			  break;
		  case BVT_FLOAT:
			  _v.v_float = assignment;
			  break;
		  case BVT_DOUBLE:
			  _v.v_double = assignment;
			  break;
		  case BVT_LONG:
			_v.v_long = BppLong(assignment); // ._v.v_long;
			  break;
		  case BVT_ULONG:
			_v.v_ulong = BppULong(assignment); // ._v.v_ulong;
			  break;
		  case BVT_INTPTR:
			_v.v_intptr = assignment; // ._v.v_intptr;
			  break;
		  case BVT_UINTPTR:
			  _v.v_uintptr = assignment; // ._v.v_uintptr;
			  break;
		  case BVT_SHORT:
			_v.v_short = assignment;
			break;
		case BVT_USHORT:
		case BVT_CHAR16:
			_v.v_ushort = assignment;
			break;
		case BVT_SBYTE:
		case BVT_CHAR:
			_v.v_sbyte = assignment;
			break;
		case BVT_BYTE:
		case BVT_UCHAR:
			_v.v_byte = assignment;
			break;
		case BVT_BOOL:
			_v.v_byte = assignment.asBool() ? 1 : 0;
			break;
		case BVT_FILE:
			  if (assignment._v.v_integer) return false;
			  _v.v_fp = assignment;
			  break;
		  default: 
			  return false;
		  }
		  return true;
	  }
		switch (at)
		{
	  case BVT_OBJECT:
		  {
			  BppObject* pobj = assignment;
			  return (pobj && pobj->safeAssignToValue(*this));
		  }
	  case BVT_STRING:
		  switch (v_type)
		  {
		  case BVT_CHARBUFFER:
			  {
			  if (_v.v_charbuffer) _v.v_charbuffer->release();
				  BppString* s = assignment;
			  _v.v_charbuffer = new BppCharBuffer(s ? s->c_str() : _T(""));
			  _v.v_charbuffer->addRef();
			  }
			  break;
		  default:
			  return false;
		  }
		  break;  
		case BVT_CHARBUFFER:
		  switch (v_type)
		  {
		  case BVT_STRING:
			  {
			  if (_v.v_string) _v.v_string->release();
				  BppCharBuffer* b = assignment;
			  _v.v_string = new BppString(b ? b->data() : _T(""));
			  _v.v_string->addRef();
			  }
			  break;
		  default:
			  return false;
		  }
		  break;  
	  case BVT_CODE:
		  if (v_type == BVT_BYTECODE)
		  {
			  if (_v.v_vector) _v.v_vector->release();
			  _v.v_code = assignment;
			  v_type = assignment.v_type;
			  break;
		  }
		  return false;
	  case BVT_BYTECODE:
		  if (v_type == BVT_CODE)
		  {
			  _v.v_code = 0;
			  _v.v_vector = assignment;
			  if (_v.v_vector) _v.v_vector->addRef();
			  v_type = assignment.v_type;
			  break;
		  }
		  return false;
	  case BVT_STRONGOBJECT:
		  {
			  BppObject* pobj = assignment;
			  if (pobj) return pobj->safeAssignToValue(*this);
			  switch (v_type)
				{
					case BVT_OBJECT:
					 if (_v.v_refObj) _v.v_refObj->release();
					 _v.v_refObj = 0;
					 return true;
				 case BVT_STRONGOBJECT:
					 if (_v.v_vector)
						(*_v.v_vector)[IDX_CODE].dispose();
					 return true;
				 case BVT_NULL:
						 _v.v_addr= 0;
					 return true;
			  }
			  return false;
		  }
	  case BVT_NULL:
			 switch (v_type)
		 {
		 case BVT_STRING:
		 case BVT_BUFFER:
		 case BVT_CHARBUFFER:
		 case BVT_OBJECT:
		 case BVT_VECTOR:
		 case BVT_DICTIONARY:
		 case BVT_BYTECODE:
			 if (_v.v_refObj) _v.v_refObj->release();
			 _v.v_refObj = 0;
			 return true;
		 case BVT_STRONGOBJECT:
			 if (_v.v_vector)
				 (*_v.v_vector)[IDX_CODE].dispose();
				 return true;
		 case BVT_FILE:
		 //case BVT_INTEGER:
		 //case BVT_FLOAT:
		 //case BVT_DOUBLE:
		 case BVT_NULL:
		 case BVT_CODE:
		 //case BVT_CODEREF:
		 case BVT_BYTECODEREF:
		 case BVT_BUFFERREF:
		 case BVT_CHARBUFFERREF:
		 case BVT_CLASSREF:
		 case BVT_OBJECTREF:
		 case BVT_VECTORREF:
		 case BVT_STRINGREF:
		 case BVT_DICTIONARYREF:
				 _v.v_addr= 0;
			 return true;
		 }
	  default:
			return false;
		}
		return true;
  }


/**
@return Class pointer if value type is BVT_CLASS or BVT_CLASSREF, otherwise NULL.
*/
BppValue::operator BppClass*() const BPPNOEXCEPT
{
	return ((v_type == BVT_CLASS) || (v_type == BVT_CLASSREF)) ? _v.v_class : 0;
}

/**
@return Object pointer if value type is BVT_OBJECT or BVT_OBJECTREF, otherwise NULL.
*/
BppValue::operator BppObject*() const BPPNOEXCEPT
{
	switch (v_type)
	{
	case BVT_OBJECT:
	case BVT_OBJECTREF:
		return _v.v_object;
	case BVT_STRONGOBJECT:
		if (_v.v_vector)
		{
			return (*_v.v_vector)[IDX_CODE];
		}
	}
	return 0;
}

/**
@return Vector pointer if value type is BVT_VECTOR, BVT_BYTECODE, BVT_LOCALVARINFI, BVT_VECTORREF
or BVT_BYTECODEREFREF, otherwise NULL.
*/
BppValue::operator BppVector*() const BPPNOEXCEPT
{
	if ((v_type & _BVTVECTORBASE) == _BVTVECTORBASE) return _v.v_vector;
	return  0;
}

/**
@return String pointer if value type is BVT_STRING or BVT_STRINGREF, otherwise NULL.
*/
BppValue::operator BppString*() const BPPNOEXCEPT
{
	return ((v_type == BVT_STRING) || (v_type == BVT_STRINGREF)) ? _v.v_string : 0;
}

/**
@return Pointer to byte buffer if value type is BVT_BUFFER or BVT_BUFFERREF, otherwise NULL.
*/
BppValue::operator BppBuffer*() const BPPNOEXCEPT
{
	return ((v_type == BVT_BUFFER) || (v_type == BVT_BUFFERREF)) ? _v.v_buffer : 0; 
}

/**
@return Pointer to character buffer if value type is BVT_CHARBUFFER or BVT_CHARBUFFERREF, otherwise NULL.
*/
BppValue::operator BppCharBuffer*() const BPPNOEXCEPT
{
	return ((v_type == BVT_CHARBUFFER) || (v_type == BVT_CHARBUFFERREF)) ? _v.v_charbuffer : 0; 
}

/**
@return Dictionary pointer if value type is BVT_DICTIONARY or BVT_DICTIONARYREF, otherwise NULL.
*/
BppValue::operator BppDictionary*() const BPPNOEXCEPT
{
	return ((v_type == BVT_DICTIONARY) || (v_type == BVT_DICTIONARYREF)) ? _v.v_dict : 0;
}


/**
@return Pointer to variable descriptor if value type is BVT_VAR, otherwise NULL.
*/
BppValue::operator BppDictionaryEntry*() const BPPNOEXCEPT
{
	return (v_type == BVT_VAR) ? _v.v_var : 0;
}

/**
@return Pointer to native function if value type is BVT_CODE, otherwise NULL.
*/
BppValue::operator BPPUSERFUNC() const BPPNOEXCEPT
{
	return (v_type == BVT_CODE) ? _v.v_code : 0;
}

/**
@return Pointer to file descriptor if value type is BVT_FILE, otherwise NULL.
*/
BppValue::operator FILE*() const BPPNOEXCEPT
{
	return (v_type == BVT_FILE) ? _v.v_fp : 0;
}

/**
@return 8 bits integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to integer is done. Otherwise
current value (independent from type attribute) is interpreted as BppSByte.
*/
BppValue::operator BppSByte() const BPPNOEXCEPT
{
	switch (v_type)
	{
	case BVT_FLOAT:
		return (BppSByte)(BppInt) _v.v_float;
	case BVT_DOUBLE:
		return (BppSByte)(BppInt) _v.v_double;
	}
	 return _v.v_sbyte;
}
/**
@return 8 bits unsigned integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to unsigned integer is done. Otherwise
current value (independent from type attribute) is interpreted as BppByte.
*/
BppValue::operator BppByte() const BPPNOEXCEPT
{
	switch (v_type)
	{
	case BVT_FLOAT:
		return (BppByte)(BppUInt) _v.v_float;
	case BVT_DOUBLE:
		return (BppByte)(BppUInt) _v.v_double;
	}
	 return _v.v_byte;
}

/**
@return 16 bits integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to integer is done. 
Shorter integer types are converted correctly, for all other types
current value (independent from type attribute) is interpreted as BppShort.
*/
BppValue::operator BppShort() const BPPNOEXCEPT
{
	switch (mkWcharInternal(BppValueType(v_type)))
	{
	case BVT_FLOAT:
		return (BppShort)(BppInt) _v.v_float;
	case BVT_DOUBLE:
		  return (BppShort)(BppInt) _v.v_double;
	case BVT_SBYTE:
	case BVT_CHAR:
		  return (BppShort) _v.v_sbyte;
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
		  return (BppShort) _v.v_byte;
	}
	 return _v.v_short;
}

/**
@return 16 bits unsigned integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to unsigned integer is done. 
Shorter integer types are converted correctly, for all other types
current value (independent from type attribute) is interpreted as BppUShort.
*/
BppValue::operator BppUShort() const BPPNOEXCEPT
{
	switch (mkWcharInternal(BppValueType(v_type)))
	{
	case BVT_FLOAT:
		return (BppUShort)(BppUInt) _v.v_float;
	case BVT_DOUBLE:
		return (BppUShort)(BppUInt) _v.v_double;
	case BVT_SBYTE:
	case BVT_CHAR:
		return (BppUShort) _v.v_sbyte;
	case BVT_BOOL:
	case BVT_BYTE:
	case BVT_UCHAR:
		return (BppUShort) _v.v_byte;
	}
	 return _v.v_ushort;
}

/**
@return 32 bits integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to integer is done. 
Shorter integer types are converted correctly, for all other types
current value (independent from type attribute) is interpreted as BppInt.
*/
BppValue::operator BppInt() const BPPNOEXCEPT
{
	BppValueType t = mkWcharInternal(BppValueType(v_type));
	if (t >= BVT_INT && t <= BVT_ULONG) return _v.v_integer;
	switch (t)
	{
	case BVT_FLOAT:
		return (BppInt) _v.v_float;
	case BVT_DOUBLE:
		  return (BppInt) _v.v_double;
	case BVT_SBYTE:
	case BVT_CHAR:
		return (BppInt) _v.v_sbyte;
	case BVT_BOOL:
	case BVT_BYTE:
	case BVT_UCHAR:
		return (BppInt) _v.v_byte;
	case BVT_SHORT:
		  return (BppInt) _v.v_short;
	case BVT_USHORT:
	case BVT_CHAR16:
		  return (BppInt) _v.v_ushort;
	}
	return _v.v_integer;
}

/**
@return 32 bits unsigned integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to unsigned integer is done. 
Shorter integer types are converted correctly, for all other types
current value (independent from type attribute) is interpreted as BppUInt.
*/
BppValue::operator BppUInt() const BPPNOEXCEPT
{
	BppValueType t = mkWcharInternal(BppValueType(v_type));
	if (t >= BVT_INT && t <= BVT_ULONG) return _v.v_uint;
	switch (t)
	{
	case BVT_FLOAT:
		return (BppUInt) _v.v_float;
	case BVT_DOUBLE:
		  return (BppUInt) _v.v_double;
	case BVT_SBYTE:
	case BVT_CHAR:
		  return (BppUInt) _v.v_sbyte;
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
		  return (BppUInt) _v.v_byte;
	case BVT_SHORT:
		  return (BppUInt) _v.v_short;
	case BVT_USHORT:
	case BVT_CHAR16:
		  return (BppUInt) _v.v_ushort;
	}
	return _v.v_uint;
}

/**
@return 64 bits integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to integer is done. 
Shorter integer types are converted correctly, for all other types
current value (independent from type attribute) is interpreted as BppLong.
*/
BppValue::operator BppLong() const BPPNOEXCEPT
{
	switch (mkWcharInternal(BppValueType(v_type)))
	{
	case BVT_FLOAT:
		return BppLong(_v.v_float);
	case BVT_DOUBLE:
		  return BppLong(_v.v_double);
	case BVT_SBYTE:
	case BVT_CHAR:
		  return BppLong((BppInt) _v.v_sbyte);
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
		  return BppLong((BppInt) _v.v_byte);
	case BVT_SHORT:
		  return BppLong(_v.v_short);
	case BVT_USHORT:
	case BVT_CHAR16:
		  return BppLong(_v.v_ushort);
	case BVT_INT:
#ifndef BPP_I64
	case BVT_INTPTR:
#endif
		  return BppLong(_v.v_integer);
	case BVT_UINT:
	case BVT_CHAR8:
	case BVT_CHAR32:
#ifndef BPP_I64
	case BVT_UINTPTR:
#endif
		return BppLong(_v.v_uint);
	}
	return _v.v_long;
}

/**
@return 64 bits unsigned integer
@note If value type is BVT_FLOAT or BVT_DOUBLE a conversion from real number to unsigned integer is done. 
Shorter integer types are converted correctly, for all other types
current value (independent from type attribute) is interpreted as BppULong.
*/
BppValue::operator BppULong() const BPPNOEXCEPT
{
	switch (mkWcharInternal(BppValueType(v_type)))
	{
	case BVT_FLOAT:
		return BppULong(_v.v_float);
	case BVT_DOUBLE:
		  return BppULong(_v.v_double);
	case BVT_SBYTE:
	case BVT_CHAR:
		  return BppULong((BppInt) _v.v_sbyte);
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
		  return BppULong((BppUInt) _v.v_byte);
	case BVT_SHORT:
		  return BppULong(_v.v_short);
	case BVT_USHORT:
	case BVT_CHAR16:
		  return BppULong(_v.v_ushort);
	case BVT_INT:
#ifndef BPP_I64
	case BVT_INTPTR:
#endif
		  return BppULong(_v.v_integer);
	case BVT_UINT:
	case BVT_CHAR8:
	case BVT_CHAR32:
#ifndef BPP_I64
	case BVT_UINTPTR:
#endif
		return BppULong(_v.v_uint);
	}
	return _v.v_ulong;
}

/**
@return Single precision real value (float, 32 bits).
@note If value type is BVT_FLOAT, stored value is returned. If value type is of integral type or type is BVT_DOUBLE 
a conversion to single precision real number is done. Otherwise
result is maximum float value (3.402823466e+38) is returned.
*/
BppValue::operator float() const BPPNOEXCEPT
{
	switch (mkWcharInternal(BppValueType(v_type)))
	{
	case BVT_FLOAT:
		  return _v.v_float;
	case BVT_DOUBLE:
		  return (float) _v.v_double;
	case BVT_INT:
		  return (float) _v.v_integer;
	case BVT_UINT:
	case BVT_CHAR32:
	case BVT_CHAR8:
		  return (float) _v.v_uint;
	case BVT_SHORT:
		  return (float) _v.v_short;
	case BVT_USHORT:
	case BVT_CHAR16:
		  return (float) _v.v_ushort;
	case BVT_SBYTE:
	case BVT_CHAR:
		  return (float) (BppInt) _v.v_sbyte;
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
		  return (float) (BppUInt) _v.v_byte;
	case BVT_LONG:
		  return (float) BppLong(_v.v_long);
	case BVT_ULONG:
		  return (float) BppULong(_v.v_ulong);
	case BVT_INTPTR:
		  return (float) _v.v_intptr;
	case BVT_UINTPTR:
		return (float)_v.v_uintptr;
	}
	return FLT_MAX;
}

/**
@return Double precision real value (double, 64 bits).
@note If value type is BVT_DOUBLE, stored value is returned. If value is of integral type or type is BVT_FLOAT 
a conversion to double precision real number is done. Otherwise
result is maximum double value (1.7976931348623158e+308) is returned.
*/
BppValue::operator double() const BPPNOEXCEPT
{
	switch (mkWcharInternal(BppValueType(v_type)))
	{
	case BVT_FLOAT:
		  return _v.v_float;
	case BVT_DOUBLE:
		  return _v.v_double;
	case BVT_INT:
		  return (double) _v.v_integer;
	case BVT_UINT:
	case BVT_CHAR32:
	case BVT_CHAR8:
		  return (double) _v.v_uint;
	case BVT_SHORT:
		  return (double) _v.v_short;
	case BVT_USHORT:
	case BVT_CHAR16:
		  return (double) _v.v_ushort;
	case BVT_SBYTE:
	case BVT_CHAR:
		  return (double) (BppInt) _v.v_sbyte;
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
		  return (double) (BppUInt) _v.v_byte;
	case BVT_LONG:
		  return (double) BppLong(_v.v_long);
	case BVT_ULONG:
		  return (double) BppULong(_v.v_ulong);
	case BVT_INTPTR:
		  return (double) _v.v_intptr;
	case BVT_UINTPTR:
		return (double)_v.v_uintptr;
	}
	return DBL_MAX;
}

/**
If current value is of any reference counted type, reference counter is decremented.
Independent from type of stored value, value type is set to BVT_NULL, all value flags
are unset and value data are set to zero.
*/
void BppValue::dispose()
{
	if (_v.v_var && isRefCounted())
	{
		if (v_type == BVT_STRONGOBJECT)
			(*(_v.v_vector))[IDX_CODE].dispose();
		 _v.v_refObj->release();
	}
	v_type = BVT_NULL;
	v_flags = 0;
#ifdef EMULATE_I64
	_v.v_dwords[0] = 0;
	_v.v_dwords[1] = 0;
#else
	  _v.v_ulong = 0;
#endif
}

int BppValue::compare(const BppValue & cmp) const
{
//   if (_v.v_addr == cmp._v.v_addr) return 0;
	if (isIntegralType() && cmp.isIntegralType())
	{
		unsigned short vt = v_type > cmp.v_type ? v_type : cmp.v_type;
		vt = mkWcharInternal(BppValueType(vt));
		if (vt <= BVT_INT) {
			BppInt l = *this;
			BppInt r = cmp;
			return l < r ? -1 : r < l ? 1 : 0;
		}
		switch (vt) 
		{
		case BVT_UINT:
		case BVT_CHAR32:
			{
			BppUInt l = *this;
			BppUInt r = cmp;
			return l < r ? -1 : r < l ? 1 : 0;
		}
		case BVT_INTPTR:
		{
			BppIntPtr l = *this;
			BppIntPtr r = cmp;
			return l < r ? -1 : r < l ? 1 : 0;
		}
		case BVT_UINTPTR:
		{
			BppUIntPtr l = *this;
			BppUIntPtr r = cmp;
			return l < r ? -1 : r < l ? 1 : 0;
		}
		case BVT_LONG:
		{
			BppLong l = *this;
			BppLong r = cmp;
			return l < r ? -1 : r < l ? 1 : 0;
		}
		}
		BppULong l = *this;
		BppULong r = cmp;
		return l < r ? -1 : r < l ? 1 : 0;
	}
	if ((isRealType() && cmp.inTypeRange(BVT_INTEGRAL_MIN, BVT_REAL_MAX)) ||
		(cmp.isRealType() && inTypeRange(BVT_INTEGRAL_MIN, BVT_REAL_MAX)))
	{
		double dl = *this;
		double dr = cmp;
		return dl < dr ? -1 : dr < dl ? 1 : 0;
	}
	BppValueType t1 = maskWeakRef((BppValueType)v_type);
	BppValueType t2 = maskWeakRef((BppValueType)cmp.v_type);
	if (t1 == BVT_NULL && cmp.isNull()) t2 = BVT_NULL;
	else if (t2 == BVT_NULL && isNull()) t1 = BVT_NULL;
	if (t1 == BVT_NULL) return t2 == BVT_NULL ? 0 : -1;
	else if (t2 == BVT_NULL) return 1;
	if (t1 == t2)
	{
		switch (t1) {
		case BVT_STRING:
			if (_v.v_string) {
				if (cmp._v.v_string) return _v.v_string->compare(*cmp._v.v_string);
				return 1;
			}
			else return cmp._v.v_string ? -1 : 0;
		case BVT_BUFFER:
			if (_v.v_buffer) {
				if (cmp._v.v_buffer) return _v.v_buffer->compare(*cmp._v.v_buffer);
				return 1;
			}
		  else return cmp._v.v_buffer ? -1 : 0;
		case BVT_CHARBUFFER:
			if (_v.v_charbuffer) {
				if (cmp._v.v_charbuffer) return _v.v_charbuffer->compare(*cmp._v.v_charbuffer);
				return 1;
			}
			else return cmp._v.v_buffer ? -1 : 0;
		case BVT_VECTOR:
			if (_v.v_vector) {
				if (cmp._v.v_vector) return _v.v_vector->compare(*cmp._v.v_vector);
				return 1;
			}
			else return cmp._v.v_vector ? -1 : 0;
		}
	}
	BppIntPtr l = *this;
	BppIntPtr r = cmp;
	return l < r ? -1 : r < l ? 1 : 0;
}


/// Equality operator
  bool BppValue::operator==(const BppValue& cmp) const
  {
	  BppValueType t1 = maskWeakRef((BppValueType)v_type);
	  BppValueType t2 = maskWeakRef((BppValueType)cmp.v_type);
	  if (t1 == BVT_NULL && cmp.isNull()) t2 = BVT_NULL;
	  else if (t2 == BVT_NULL && isNull()) t1 = BVT_NULL;
	  if (t1 !=t2 ) return false;
	  if (t1 == BVT_NULL) return true;
	  if (t1 == BVT_BOOL) return asBool() == cmp.asBool();
	  if (isRealType()) return (((double)*this) == (double)cmp);
	  if ((t1 | BST_UNSIGNED) == BVT_ULONG) return BppULong(_v.v_ulong) == BppULong(cmp._v.v_ulong);
	  if (_v.v_addr == cmp._v.v_addr) return true;
	  if (isIntegralType()) return (((BppUInt)*this) == (BppUInt)cmp);
	  //if (_v.v_addr == 0 || cmp._v.v_addr == 0) return false;
	  switch (t1)
	  {
	  case BVT_STRING:
		  return _v.v_string->compare(*cmp._v.v_string) == 0;
	  case BVT_VECTOR:
	  case BVT_LOCALVARINFO:
	  case BVT_STRONGOBJECT:
			 return (*_v.v_vector)==(*cmp._v.v_vector);
	  case BVT_BYTECODE:
	  {
		//if ((*_v.v_vector)==(*cmp._v.v_vector)) return true;
		size_t size = _v.v_vector->size();
		if (size != cmp._v.v_vector->size()) return false;
		BppValue* p1 = _v.v_vector->data();
		BppValue* p2 = cmp._v.v_vector->data();
		if (p1 == p2) return true;
		for (size_t i=0;i<size;++i,++p1,++p2)
		{
			if (i==IDX_SELFREFERENCE) continue;
			if (*p1 != *p2) return false;
		}
		return true;
	 }
	  case BVT_BUFFER:
		  return (*_v.v_buffer)==(*cmp._v.v_buffer);
	  case BVT_CHARBUFFER:
		  return (*_v.v_charbuffer) ==(*cmp._v.v_charbuffer);
	  case BVT_DICTIONARY:
			 return (*_v.v_dict) ==  (*cmp._v.v_dict);
	  case BVT_VAR:
			 return (*_v.v_var) ==  (*cmp._v.v_var);
	  }
		return false;
  }

  /// Non-equality operator
  bool BppValue::operator!=(const BppValue& cmp) const
  {
		return !(*this == cmp);
  }



/**
@return Address of value data.
@note Function behaves as follows:
<ul>
<li>If value is of a numeric type (integral or real number type), it returns
pointer to (address of) value data.</li>
<li>In case of strong typed value without valid data, result is NULL.</li>
<li>If value is a string (BVT_STRING or BVT_STRINGREF), pointer to character data
of string (i.e. a C-string poiner) is returned.</li>
<li>In case of byte buffer or character buffer (BVT_BUFFER, BVT_BUFFERREF, BVT_CHARBUFFER or BVT_CHARBUFFERREF)
start address of data area is returned.</li>
<li>For all other value types simply ptr() method is called.</li>
</ul>
@see ptr
*/
void* BppValue::addr() const BPPNOEXCEPT
{
	if (inTypeRange(_BVTNUMBERBASE, _BVTVALMAX)) return (void*)&_v;
	void* p = ptr();
	if (p) {
		switch (maskWeakRef((BppValueType)v_type))
		{
		case BVT_STRING:
			return (void*)_v.v_string->c_str();
		case BVT_BUFFER:
			return _v.v_buffer->data();
		case BVT_CHARBUFFER:
			return _v.v_charbuffer->data();
		case BVT_STRONGOBJECT:
			{
				BppVector* desc = getStrongObjDesc();
				if (desc) return (BppObject*)((*desc)[IDX_CODE]);
				return 0;
			}
		}
	}
	return p;
}

bool BppValue::asBool() const BPPNOEXCEPT
{
	if (isNull()) return false;
	if (v_type < BVT_LONG) return INTPTRCAST(*this) != 0U;
	if (v_type == BVT_FLOAT) return _v.v_dwords[0] != 0U;
	if (v_type == BVT_STRONGOBJECT) return ((BppObject*)(*this)) != 0;
	if (v_type > _BVTVALMAX) return INTPTRCAST(*this) != 0U;
	return _v.v_dwords[0] || _v.v_dwords[1];
}

BppChar BppValue::asBppChar() const BPPNOEXCEPT
{
	switch (sizeof(BppChar))
	{
	case 4:
		return (BppUInt)(*this);
	case 2:
		return (BppUShort)(*this);
	}
	return (BppSByte)*this;
}

/**
@param protflags Protection flag mask to be set for value.
@see getProtection
*/
void BppValue::setProtection(short protflags) BPPNOEXCEPT
{
	v_flags &= (0xFF & (~BVF_PUBLIC));
	v_flags |= (protflags & BVF_PUBLIC);
}

