#include "bppdef.h"
#ifdef EMULATE_I64

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "bppufunc.h"
Int64::Int64(const UInt64& val):_high(val._high),_low(val._low){}


Int64::Int64(float val)
:_high(0),_low(0)
{
	static const float divisor = ((float)(unsigned long)0xffffffff) + 1.0f;
	bool neg = (val < 0);
	if (neg) val = -val;
	_low = (unsigned long) fmod(val,divisor);
	_high = (unsigned long) (val / divisor);
	if (neg) (*this) = -(*this);
}

Int64::Int64(double val)
:_high(0),_low(0)
{
	static const double divisor = ((double)(unsigned long)0xffffffff) + 1.0;
	bool neg = (val < 0);
	if (neg) val = -val;
	_low = (unsigned long) fmod(val,divisor);
	_high = (unsigned long) (val / divisor);
	if (neg) (*this) = -(*this);
}

Int64& Int64::operator=(const Int64& val)
{
	if (&val != this)
	{
		_high = val._high;
		_low = val._low;
	}
	return *this;
}

Int64& Int64::operator++()
{
	unsigned long l = _low;
	if (++_low < l) ++_high;
	return *this;
}

const Int64 Int64::pinc()
{
	Int64 result = *this;
	++(*this);
	return result;
}

Int64& Int64::operator--()
{
	unsigned long l = _low;
	if (--_low > l) --_high;
	return *this;
}
const Int64 Int64::pdec()
{
	Int64 result = *this;
	--(*this);
	return result;
}

Int64 Int64::operator~() const
{
	return Int64(~_high,~_low);
}

Int64 Int64::operator-() const
{
	return ++Int64(~_high,~_low);
}

int Int64::compare(const Int64& cmp) const
{
	if (_high < cmp._high) return -1;
	if (_high > cmp._high) return 1;
	if (_low == cmp._low) return 0;
	if (_high >= 0)
	{
		return (_low > cmp._low) ? 1 : -1;
	}
	return (_low < cmp._low) ? 1 : -1;
}

Int64& Int64::operator+=(const Int64& val)
{
	unsigned long l = _low;
	_low += val._low;
	if (_low < l) ++_high;
	_high += val._high;
	return *this;
}

Int64& Int64::operator-=(const Int64& val)
{
	(*this) += -val;
	return *this;
}

Int64& Int64::operator&=(const Int64& val)
{
	_high &= val._high;
	_low &= val._low;
	return *this;
}

Int64& Int64::operator|=(const Int64& val)
{
	_high |= val._high;
	_low |= val._low;
	return *this;
}

Int64& Int64::operator^=(const Int64& val)
{
	_high ^= val._high;
	_low ^= val._low;
	return *this;
}


Int64& Int64::operator<<=(unsigned nbits)
{
	nbits &= 0x3F;
	for (unsigned i=0;i<nbits;++i)
	{
		_high <<= 1;
		if (_low & 0x80000000L)
			_high |=1;
		_low <<= 1;
	}
	return *this;
}

Int64& Int64::operator>>=(unsigned nbits)
{
	nbits &= 0x3F;
	unsigned sign = _high & 0x80000000L;
	for (unsigned i=0;i<nbits;++i)
	{
		_low >>= 1;
		if (_high & 1)
			_low |= 0x80000000L;
		_high >>= 1;
		_high |= sign;
	}
	return *this;
}

Int64 Int64::operator<<(unsigned nbits) const
{
	return Int64(*this) <<= nbits;
}

Int64 Int64::operator>>(unsigned nbits) const
{
	return Int64(*this) >>= nbits;
}

Int64& Int64::operator*=(const Int64& val)
{
	 if ((_high == 0) && (_low == 0))
	  return *this;
	if ((val._high == 0) && (val._low == 0))
	{
		return (*this = val);
	}
	*this = (*this * val);
  //if ((_high == 0) && (_low == 0))
  //   return *this;
  // if ((val._high == 0) && (val._low == 0))
  // {
  //    return (*this = val);
  // }
  // Int64 rop = val;
  // Int64 lop = *this;
  // int sign = 1;
  // if (_high < 0)
  // {
  //    lop = -(*this);
  //    sign *= -1;
  // }
  // if (rop._high < 0)
  // {
  //    rop = -rop;
  //    sign *= -1;
  // }
  // _high = _low = 0;
  // Int64 mask = 1;
  // for (int i=0; i< 63;++i)
  // {
  //    if ((rop._low & mask._low) || (rop._high & mask._high))
  //    {
  //       *this += lop;
  //    }
  //    lop <<= 1;
  //    mask <<= 1;
  // }
  // if (sign < 0) *this = -(*this);
	return *this;
}

bool Int64::_idiv(const Int64& dividend, const Int64& divisor, Int64& q, Int64& r)
{
	if (!(divisor._high || divisor._low)) // divisor is zero
	{
		if (dividend._high < 0)
		{
			q._high = r._high = 0x80000000;
		 }
		else 
		{
			r._high = 0x7fffffff;
			r._low = 0xffffffff;
			if (dividend._high || dividend._low) // dividend is not zero
			{
				q = r;
			}
		}
		return false;
	}
	if (!(dividend._high || dividend._low)) // dividend is zero
	{
		q._high = r._high = 0;
		q._low = r._low = 0;
		return true;
	}
	q._high = r._high = 0;
	q._low = r._low = 0;
	Int64 dend = (dividend._high < 0) ? -dividend : dividend;
	Int64 dor = (divisor._high < 0) ? -divisor : divisor;
	long nh = dividend._high;
	long dh = divisor._high;
  int c = dend.compare(dor);
	if (c == 0)
	{
		q._low = 1;
	}
	else if (c < 0)
	{
		r = dend;
	}
	else
	{
		Int64 qtmp, dd;;
		while (1)
		{
			dd = dor;
			qtmp = 1;
			while ((c = dend.compare(dd <<= 1)) >= 0)
				qtmp <<= 1;
			dd >>= 1; //qtmp >>=1;
			q += qtmp; dend -= dd;
			c = dend.compare(dor);
			if (c == 0)
			{
				++q;
				break;
			}
			else if (c < 0)
			{
				r = /*dor -*/ dend;
				break;
			}      
		}
	}

  ///* the sign of the remainder is the same as the sign of the dividend
  //   and the quotient is negated if the signs of the operands are
  //   opposite */
  //quotient = q;
	if (nh < 0) {
	 r = -r;
	 if (dh > 0)
		q = -q;
  }
  else { /* positive dividend */
	  if (dh < 0)
		 q = -q;
  }
  
	return true;
}

Int64& Int64::operator/=(const Int64& val)
{
	Int64 q,r;
	_idiv(*this, val, q, r);
	*this = q;
	return *this;
}

Int64& Int64::operator%=(const Int64& val)
{
	Int64 q,r;
	_idiv(*this, val, q, r);
	*this = r;
	return *this;
}


Int64::operator float() const
{
	static const float m = ((float)(unsigned long)0xffffffff) + 1.0f;
	if (_high < 0) return -((float) (-(*this)));
	return ((float)_high) * m + (float)_low;
}

Int64::operator double() const
{
	static const double m = ((double)(unsigned long)0xffffffff) + 1.0;
	if (_high < 0) return -((double) (-(*this)));
	return ((double)_high) * m + (double)_low;
}

BppChar* Int64::toa(BppChar* buf, int base, bool upper) const
{
	if (base < 2 || base > 36) return 0;
	if (_high >= 0 || base != 10) return UInt64(*this).toa(buf,base,upper);
	 char tmp[66];
	UInt64(-(*this)).toa(tmp+1,base,upper);
	tmp[0] = _T('-');
	if (buf) return _tcscpy(buf,tmp);
	return dupstr(tmp);
}

Int64 operator+(const Int64& l, const Int64& r)
{
	return Int64(l) += r;
}

Int64 operator-(const Int64& l, const Int64& r)
{
	return Int64(l) -= r;
}

Int64 operator*(const Int64& l, const Int64& r)
{
	int sign = 1;
	if (l._high < 0) sign *= -1;
	if (r._high < 0) sign *= -1;
	UInt64 L = (l._high < 0) ? -l : l;
	UInt64 R = (r._high < 0) ? -r : r;
	UInt64 res = L * R;
	if (sign < 0) return -(*(Int64*) &res);
	return *(Int64*) &res;
//  return Int64(l) *= r;
}

Int64 operator/(const Int64& l, const Int64& r)
{
	return Int64(l) /= r;
}

Int64 operator%(const Int64& l, const Int64& r)
{
	return Int64(l) %= r;
}

Int64 operator&(const Int64& l, const Int64& r)
{
	return Int64(l) &= r;
}

Int64 operator|(const Int64& l, const Int64& r)
{
	return Int64(l) |= r;
}

Int64 operator^(const Int64& l, const Int64& r)
{
	return Int64(l) ^= r;
}
// ==============================




UInt64::UInt64(float val)
:_high(0),_low(0)
{
	static const float divisor = ((float)(unsigned long)0xffffffff) + 1.0f;
	_low = (unsigned long) fmod(val,divisor);
	_high = (unsigned long) (val / divisor);
}

UInt64::UInt64(double val)
:_high(0),_low(0)
{
	static const double divisor = (double)(unsigned long)0xffffffff + 1.0;
	_low = (unsigned long) fmod(val,divisor);
	_high = (unsigned long) (val / divisor);
}


UInt64& UInt64::operator=(const UInt64& val)
{
	if (&val != this)
	{
		_high = val._high;
		_low = val._low;
	}
	return *this;
}

UInt64& UInt64::operator++()
{
	if (++_low == 0) ++_high;
	return *this;
}

const UInt64 UInt64::pinc()
{
	UInt64 result = *this;
	++(*this);
	return result;
}

UInt64& UInt64::operator--()
{
	unsigned long l = _low;
	if (--_low > l) --_high;
	return *this;
}
const UInt64 UInt64::pdec()
{
	UInt64 result = *this;
	--(*this);
	return result;
}

UInt64 UInt64::operator~() const
{
	return UInt64(~_high,~_low);
}

UInt64 UInt64::operator-() const
{
	return ++UInt64(~_high,~_low);
}

int UInt64::compare(const UInt64& cmp) const
{
	if (_high < cmp._high) return -1;
	if (_high > cmp._high) return 1;
	if (_low == cmp._low) return 0;
	return (_low > cmp._low) ? 1 : -1;
}

UInt64& UInt64::operator+=(const UInt64& val)
{
	unsigned long l = _low;
	_low += val._low;
	if (_low < l) ++_high;
	_high += val._high;
	return *this;
}

UInt64& UInt64::operator-=(const UInt64& val)
{
	(*this) += -val;
	return *this;
}

UInt64& UInt64::operator&=(const UInt64& val)
{
	_high &= val._high;
	_low &= val._low;
	return *this;
}

UInt64& UInt64::operator|=(const UInt64& val)
{
	_high |= val._high;
	_low |= val._low;
	return *this;
}

UInt64& UInt64::operator^=(const UInt64& val)
{
	_high ^= val._high;
	_low ^= val._low;
	return *this;
}


UInt64& UInt64::operator<<=(unsigned nbits)
{
	nbits &= 0x3F;
	for (unsigned i=0;i<nbits;++i)
	{
		_high <<= 1;
		if (_low & 0x80000000L)
			_high |=1;
		_low <<= 1;
	}
	return *this;
}

UInt64& UInt64::operator>>=(unsigned nbits)
{
	nbits &= 0x3F;
	for (unsigned i=0;i<nbits;++i)
	{
		_low >>= 1;
		if (_high & 1)
			_low |= 0x80000000L;
		_high >>= 1;
	}
	return *this;
}

UInt64 UInt64::operator<<(unsigned nbits) const
{
	return UInt64(*this) <<= nbits;
}

UInt64 UInt64::operator>>(unsigned nbits) const
{
	return UInt64(*this) >>= nbits;
}

UInt64& UInt64::operator*=(const UInt64& val)
{
	if ((_high == 0) && (_low == 0))
	  return *this;
	if ((val._high == 0) && (val._low == 0))
	{
		return (*this = val);
	}
	*this = *this * val;
///*   if ((_high == 0) && (_low == 0))
//     return *this;
//   if ((val._high == 0) && (val._low == 0))
//   {
//      return (*this = val);
//   }
//   UInt64 lop = *this;
//   _high = _low = 0;
//   UInt64 mask = 1;
//   for (int i=0; i< 63;++i)
//   {
//      if ((val._low & mask._low) || (val._high & mask._high))
//      {
//         *this += lop;
//      }
//      lop <<= 1;
//      mask <<= 1;
// */  }
	return *this;
}

bool UInt64::_div(const UInt64& dividend, const UInt64& divisor, UInt64& q, UInt64& r)
{
	if (!(divisor._high || divisor._low)) // divisor is zero
	{
			r._high = 0xffffffff;
			r._low = 0xffffffff;
			if (dividend._high || dividend._low) // dividend is not zero
			{
				q = r;
			}
		return false;
	}
	if (!(dividend._high || dividend._low)) // dividend is zero
	{
		q._high = r._high = 0;
		q._low = r._low = 0;
		 return true;
	}
	UInt64 dend =   dividend;
	UInt64 dor =  divisor;
	q._high = r._high = 0;
	q._low = r._low = 0;

	
	int c = dend.compare(dor);
	if (c == 0)
	{
		q._low = 1;
	}
	else if (c < 0)
	{
		r = dend;
	}
	else
	{
		UInt64 qtmp, dd;
		while (1)
		{
			dd = dor;
			qtmp = 1;
			bool cf = false;
			while ((c = dend.compare(dd <<= 1)) >= 0)
		 {
				qtmp <<= 1;
			if (!(dd._high || dd._low)) break;
			if (dd._high & 0x10000000) {
				cf = true;
				break;
			}
		 }
			if (!(qtmp._high || qtmp._low)) qtmp._high = 0x80000000;
			if (!(dd._high || dd._low)) dd._high = 0x80000000;
			else if (!cf) dd >>= 1; //qtmp >>=1;
			q += qtmp; dend -= dd;
			c = dend.compare(dor);
			if (c == 0)
			{
				++q;
				break;
			}
			else if (c < 0)
			{
				r = /*dor -*/ dend;
				break;
			}      
		}
	}
  
	return true;
}

UInt64& UInt64::operator/=(const UInt64& val)
{
	UInt64 q,r;
	_div(*this, val, q, r);
	*this = q;
	return *this;
}

UInt64& UInt64::operator%=(const UInt64& val)
{
	UInt64 q,r;
	_div(*this, val, q, r);
	*this = r;
	return *this;
}


UInt64::operator float() const
{
	static const float m = ((float)(unsigned long)0xffffffff) + 1.0f;
	return ((float)_high) * m + (float)_low;
}

UInt64::operator double() const
{
	static const double m = ((double)(unsigned long)0xffffffff) + 1.0;
	return ((double)_high) * m + (double)_low;
}

BppChar* UInt64::toa(BppChar* buf, int base, bool upper) const
{
	if (base < 2 || base > 36) return 0;
	BppChar btmp[65];
	const BppChar a = upper ? 'A' : 'a';
	static const int bufsize = 65;
	if ((!upper) && _high == 0)
	{
		ultoa(_low,btmp,base);
		if (buf)
		{
			return strcpy(buf,btmp);
		}
		return strdup(btmp);
	}
	const UInt64 div = base;
	UInt64 tmp =  *this;
	UInt64 r;
	btmp[bufsize-1] = 0;
	int i;
	for(i=bufsize-2;(tmp._high || tmp._low) && (i>=0);--i)
	{
		_div(tmp,div,tmp,r);
		if (r._low < 10)
			btmp[i] = '0' + (char) r._low;
		else
			btmp[i] = a + (char) (r._low - 10);
	}
	if (buf) return _tcscpy(buf,btmp+i+1);
	return dupstr(btmp+i);
}




UInt64 operator+(const UInt64& l, const UInt64& r)
{
	return UInt64(l) += r;
}

UInt64 operator-(const UInt64& l, const UInt64& r)
{
	return UInt64(l) -= r;
}

UInt64 operator*(const UInt64& l, const UInt64& r)
{
	UInt64 res;
	const unsigned short* L = (const unsigned short*) &l;
	const unsigned short* R = (const unsigned short*) &r;
	for (int i=0;i<4;++i)
	{
		unsigned long li = L[i];
		if (!li) continue;
		for (int j=0;i+j<4 ;++j)
		{
			unsigned long rj = R[j];
			if (!rj) continue;
				res += (UInt64(li*rj)) <<= 16*(i+j);
		}
	}
	return res;
//   return UInt64(l) *= r;
}

UInt64 operator/(const UInt64& l, const UInt64& r)
{
	return UInt64(l) /= r;
}
UInt64 operator%(const UInt64& l, const UInt64& r)
{
	return UInt64(l) %= r;
}

UInt64 operator&(const UInt64& l, const UInt64& r)
{
	return UInt64(l) &= r;
}

UInt64 operator|(const UInt64& l, const UInt64& r)
{
	return UInt64(l) |= r;
}

UInt64 operator^(const UInt64& l, const UInt64& r)
{
	return UInt64(l) ^= r;
}

Int64 _strtoInt64(const char* nptr, char** endptr, int base)
{
	if (!nptr) return Int64();
	for (;(*nptr) > 0 && (*nptr) <= ' ';++nptr);
	if (*nptr == '-') return -Int64(_strtoUInt64(++nptr,endptr,base));
	return Int64(_strtoUInt64(nptr,endptr,base));
}

UInt64 _strtoUInt64(const char* nptr, char** endptr, int base)
{
	UInt64 res;
	if (!nptr) return res;
	unsigned shft = 0;
	for (bool start = true;*nptr;++nptr)
	{
		  unsigned char c = tolower(*nptr);
		if (start)
		{
			if (c <= ' ') continue;
			if (base == 0 && c == '0')
			{
				unsigned char c2 = tolower(nptr[1]);
				switch (c2)
				{
				case 'b':
					base = 2;
					++nptr;
					break;
				case 'x':
					base = 16;
					++nptr;
					break;
				default:
					base = 8;
				}
			}
			else --nptr;
			switch (base)
			{
			case 2:
				shft = 1;
				break;
			case 4:
				shft = 2;
				break;
			case 8:
				shft = 3;
				break;
			case 16:
				shft = 4;
				break;
			case 32:
				shft = 5;
				break;
			}
			base %= 37;
			if (base <= 0) base = 10;
			start = false;
			continue;
		}
		long i = (c >= '0' && c <='9') ? c - '0' :
			(c >= 'a' && c <='z') ? c - 'a' + 10 : 37;
		if (i>=base) break;
		if (shft) { res <<= shft; res |= i; }
		else { res *= base; res += i; }
	}
	if (endptr) *endptr = (char*)nptr;
	return res;
}

#endif // EMULATE_I64
