// collectcodepages.cpp : Diese Datei enthält die Funktion "main". Hier beginnt und endet die Ausführung des Programms.
//
#include <windows.h>
#include <tchar.h>
#include <vector>
#include <string>
#include <stdio.h>
#include <map>

typedef std::basic_string<TCHAR> tstring;
std::map<tstring, unsigned> cpnamemapping;
struct CpInfo{
    bool isascii7 = true;
    bool hasCh32 = false;
    std::vector<unsigned> errors;
    unsigned defreplacement = 0;
    unsigned maxchsize = 0;
    unsigned codes[256] = {};
};
std::map<unsigned, CpInfo> infoMap;
BOOL CALLBACK EnmCodePagesProc(LPTSTR lpCpdePageStr) {
    const TCHAR* p = lpCpdePageStr;
    unsigned cpcode = _ttoi(p);
    if (!cpcode) return TRUE;
    CPINFOEX info;
    if (GetCPInfoEx(cpcode, 0, &info)) {
        cpnamemapping[info.CodePageName] = info.CodePage;
        //if (info.MaxCharSize > 1) return true;
        auto& cpinfo = infoMap[info.CodePage];
        cpinfo.maxchsize = info.MaxCharSize;
        cpinfo.defreplacement = *(unsigned short*)info.DefaultChar;
        unsigned char mb[2] = { 0,0 };
        for (unsigned i = 0; i < 256; ++i) {
            mb[0] = i;
            wchar_t ws[3] = { 0,0,0 };
            auto cnt = MultiByteToWideChar(info.CodePage, 0, (char*)mb, 2, ws, 3);
            if (cnt == 0) {
                cpinfo.errors.push_back(i);
                cpinfo.codes[i] = 0xfffd;
                if (i < 128) cpinfo.isascii7 = false;
                continue;
            }
            unsigned val = *(unsigned*)ws;
            if (val == 0xfffd) cpinfo.errors.push_back(i);
            if (val > 0xffff) cpinfo.hasCh32 = true;
            if (i < 128 && val != i) cpinfo.isascii7 = false;
            cpinfo.codes[i] = val;
        }
    }
    return TRUE;
}

#ifdef _UNICODE
#define FPUTS fputws
#else
#define FPUTS fputs
#endif

int main()
{
    EnumSystemCodePages(EnmCodePagesProc, CP_SUPPORTED);
    tstring codetab7 = _T(R"(struct CpId_Codes {
        unsigned int id;
        unsigned short ucodes[128];
    } codetab7[] = {)");
    tstring codetabsmall = _T(R"(struct CpId_Codes {
        unsigned int id;
        unsigned short ucodes[128];
    } codetab[] = {)");
    tstring codetab = _T(R"(CpId_Codes codetab[] = {)");


    TCHAR buf[20];
    for (auto it = infoMap.begin(); it != infoMap.end(); ++it) {
        const auto& item = it->second;
        if (item.hasCh32) continue;
        if (item.errors.size() == 256) continue;
        if (!item.isascii7) {
            if (item.errors.size() == 256) continue;
            _stprintf(buf, _T("\n{%d, {"), it->first);
            codetab7 += (const TCHAR*)buf;
            for (unsigned i = 0; i < 128; ++i) {
                auto code = item.codes[i];
                if (code > 0xffff) code = 0xfffd;
                _stprintf(buf, _T("0x%02x"), code);
                codetab7 += (const TCHAR*)buf;
                if (i + 1 < 128) codetab7 += _T(", ");
            }
            codetab7 += _T("}},");
        }
        else if (item.errors.size() == 128) continue;
        {
            _stprintf(buf, _T("\n{%d, {"), it->first);
            codetab += (const TCHAR*)buf;
            if (item.isascii7) codetabsmall += (const TCHAR*)buf;
            for (unsigned i = 0; i < 128; ++i) {
                auto code = item.codes[i+128];
                if (code > 0xffff) code = 0xfffd;
                _stprintf(buf, _T("0x%02x"), code);
                codetab += (const TCHAR*)buf;
                if (item.isascii7) codetabsmall += (const TCHAR*)buf;
                if (i + 1 < 128) {
                    codetab += _T(", ");
                    if (item.isascii7) codetabsmall += _T(", ");
                }
            }
            codetab += _T("}},");
            if (item.isascii7) codetabsmall += _T("}},");
        }
    }
    if (codetab7.back() == _T(',')) codetab7.pop_back();
    codetab7 += _T("\n};\n");
    if (codetab.back() == _T(',')) codetab.pop_back();
    codetab += _T("\n};\n");
    if (codetabsmall.back() == _T(',')) codetabsmall.pop_back();
    codetabsmall += _T("\n};\n");
    auto* fp = _tfopen(_T("codetab.inc"), _T("wt"));
    if (fp) {
        FPUTS(_T("#ifdef __CTABSMALL__\n"), fp);
        FPUTS(codetabsmall.c_str(), fp);
        FPUTS(_T("#else\n"), fp);
        FPUTS(codetab7.c_str(), fp);
        FPUTS(codetab.c_str(), fp);
        FPUTS(_T("#endif\n"), fp);
        fclose(fp);
    }
    return 0;
}

