/// Type catagory classification.
enum BppVTypeCategory {
   BTC_PSEUDO = 0,
   BTC_NUMBER = 0x40,
   BTC_WEAKREFLAG = 0x40,
   BTC_REFTYPE = 0x80,
   BTC_WEAKREF = TC_REFTYPE | TC_WEAKREFFLAG
};

enum BppNumberCategory {
   BNC_INTEGRAL = 0,
   BNC_REAL = 0x20
};

enum BppSignedType {
  BST_SIGNED = 0,
  BST_UNSIGNED = 4
};

enum BppIntegralSubtype {
	BIST_NONE = 0,
	BIST_SPECIAL = 1,
	BIST_CHAR = 2
};

enum NumSizeShift {
  BNS_BYTE = 0,
  BNS_WORD = 8,
  BNS_DWORD = 0x10,
  BNS_INTPTR = BNS_DWORD | BIST_SPECIAL,
  BNS_QWORD = 0x18
};

enum BppRefTypeCategory {
	BRTC_PTRTYPE = 0,
	BRTC_REFCOUNTED = 0x20
};

enum BppRefCountedCategory {
	BRCC_SINGLE = 0,
	BRCC_SEQUENCE = 0x10
};

enum BppRefSingleCategory {
   BRSC_OBJECTTYPE = 0,
   BRSC_CLASS = 0x08
}

enum BppRefSequenceCategory {
  BRSQC_SIMPLE = 0,
  BRSQC_VECTOR = 0x04,
  BRSQC_DICT = 0x08
}

enum TypeMasks {
	WEAKMASK = ~TC_WEAKREF | TC_REFTYPE
}


enum Types {
	_BVTMIN = 0,
	BVT_NULL = TC_PSEUDO,
	BVT_VOID,
	BVT_AUTO,
    BVT_NIL,
	/** @name Pseudo types (for internal typechecks only.
	@see BppInterpreter::checkType
	*/
	//@{
	BVT_NUMBER,
	BVT_INTNUMBER,
	BVT_UINTNUMBER,
	BVT_INTEGRALTYPE,
	BVT_PTRTYPE,
	BVT_REALNUMBER,
	BVT_REFCNTTYPE,
	BVT_REFERENCE,
	//@}
	_BVTNUMBERBASE = BTC_NUMBER,
 _BVTINTEGRALBASE = _BVTNUMBERBASE | BNC_INTEGRAL,
 BVT_SBYTE = _BVTINTEGRALBASE | BNS_BYTE | BST_SIGNED, ///< 8 bits integer
 BVT_INTEGRAL_MIN = BVT_SBYTE,
 BVT_CHAR = BVT_SBYTE | BIST_CHAR,
 BVT_BYTE = _BVTINTEGRALBASE | BNS_BYTE | BST_UNSIGNED, ///< 8 bits unsigned integer
 BVT_BOOL = BVT_BYTE | BIST_SPECIAL,
 BVT_UCHAR = BVT_BYTE | BIST_CHAR,
 BVT_SHORT = _BVTINTEGRALBASE | BNS_WORD | BST_SIGNED, ///< 16 bits signed integer
 BVT_USHORT = _BVTINTEGRALBASE | BNS_WORD | BST_UNSIGNED, ///<16 bits unsigned integer
 BVT_WCHAR = BVT_USHORT | BIST_CHAR,
 BVT_INT = _BVTINTEGRALBASE | BNS_DWORD | BST_SIGNED, ///< 32 bits signed integer
 BVT_INTPTR = _BVTINTEGRALBASE | BNS_INTPTR | BST_SIGNED, ///< signed integer of same size as pointer
 BVT_UINT = _BVTINTEGRALBASE | BNS_DWORD | BST_UNSIGNED, ///< 32 bits unsigned integer
 BVT_UINTPTR = _BVTINTEGRALBASE | BNS_INTPTR | BST_UNSIGNED, ///< unsigned integer of same size as pointer
 BVT_LONG = _BVTINTEGRALBASE | BNS_QWORD | BST_SIGNED, ///< 64 bits signed integer
 BVT_ULONG = _BVTINTEGRALBASE | BNS_QWORD | BST_UNSIGNED, ///< 64 bits unsigned integer
 BVT_INTEGRAL_MAX = BVT_ULONG,
 _BVTREALNUMBASE = _BVTNUMBERBASE | BNC_REAL,
 BVT_FLOAT = _BVTREALNUMBASE | BNS_DWORD, ///< single precision real type (32 bits)
 BVT_REAL_MIN = BVT_FLOAT,
 BVT_DOUBLE = _BVTREALNUMBASE | BNS_QWORD, ///< double precision real type (64 bits)
 BVT_REAL_MAX = BVT_DOUBLE,
 _BVTVALMAX = BVT_DOUBLE,
 _BVTREFTYPEBASE = BTC_REFTYPE,
 _BVTREFPTRBASE = _BVTREFTYPEBASE | BRTC_PTRTYPE,
 BVT_FILE = _BVTREFPTRBASE, ///< file descriptor pointer
 BVT_CODE, ///< native code (function) pointer
 BVT_VAR, ///< variable (contained in dictionaries)
 _BVTREFCNTBASE = _BVTREFTYPEBASE | BRTC_REFCOUNTED,
 _BVTREFCNTSINGLEBASE = _BVTREFCNTBASE | BRCC_SINGLE,
 _BVTOBJECTBASE = _BVTREFCNTSINGLEBASE | BRSC_OBJECTTYPE,
 BVT_OBJECT = _BVTOBJECTBASE, ///< B++ object (instance of class)
 BVT_RCNT_MIN = BVT_OBJECT,
 BVT_STRONGOBJECT, ///< strong typed holder of B++ object (instance of class)
 _BVTCLASSBASE = _BVTREFCNTSINGLEBASE | BRSC_CLASS,
 BVT_CLASS = _BVTCLASSBASE,  ///< B++ class descriptor object
 _BVTSEQUENCEBASE = _BVTREFCNTBASE | BRCC_SEQUENCE,
 _BVTSEQSIMPLEBASE = _BVTSEQUENCEBASE | BRSQC_SIMPLE,
 BVT_STRING = _BVTSEQSIMPLEBASE, ///< string value
 BVT_BUFFER, ///< byte buffer
 BVT_CHARBUFFER, ///< character buffer
 _BVTVECTORBASE = _BVTSEQUENCEBASE | BRSQC_VECTOR,
 BVT_VECTOR = _BVTVECTORBASE, ///< vector (dynamic array) of values
 BVT_BYTECODE, ///< bytecode vector
 BVT_LOCALVARINFO, ///< vector of local variable descriptions
 _BVTDICTBASE = _BVTSEQUENCEBASE | BRSQC_DICT,
 BVT_DICTIONARY = _BVTDICTBASE, ///< dictionary (maps key strings to values)
 BVT_RCNT_MAX = BVT_DICTIONARY,
 _BVTMAX = BVT_RCNT_MAX,
 _BVTWEAKREFBASE = _BVTREFCNTBASE | BTC_WEAKREFFLAG, 
  BVT_OBJECTREF = BVT_OBJECT | BTC_WEAKREFFLAG, ///< reference to an object
 _BVTREFMIN = BVT_OBJECTREF,
 BVT_STRONGOBJECTREF = BVT_STRONGOBJECT | BTC_WEAKREFFLAG,
 BVT_CLASSREF = BVT_CLASS | BTC_WEAKREFFLAG, ///< reference to class descriptor
 BVT_STRINGREF = BVT_STRING | BTC_WEAKREFFLAG, ///< reference to string
 BVT_BUFFERREF = BVT_BUFFER | BTC_WEAKREFFLAG, ///< reference to byte buffer
 BVT_CHARBUFFERREF = BVT_CHARBUFFER | BTC_WEAKREFFLAG, ///< reference to character buffer
 BVT_VECTORREF = BVT_VECTOR | BTC_WEAKREFFLAG, ///< reference to vector
 BVT_BYTECODEREF = BVT_BYTECODE | BTC_WEAKREFFLAG, ///< reference to bytecode vector
 BVT_VARINFOREF = BVT_LOCALVARINFO | BTC_WEAKREFFLAG,
 BVT_DICTIONARYREF = BVT_DICTIONARY | BTC_WEAKREFFLAG, ///< dictionary reference
 _BVTREFMAX = BVT_DICTIONARYREF
	
}