#if (!WINCE)

using System;
using System.Runtime.InteropServices;
using System.Text;

namespace bp2builder.InterOp
{
	/// <summary>
	/// Abstract native input class.
	/// </summary>
	public abstract class Bp2Input
	{
		public delegate IntPtr InputStringDelegate();
		public delegate int InputCharDelegate();

		protected Bp2Input()
		{
			_strbuf = IntPtr.Zero;
			_charDelegate = new InputCharDelegate(CBGetChar);
			_stringDelegate = new InputStringDelegate(CBGetString);
			_nativeInput = BppCreateCallbackInput(_charDelegate,
				_stringDelegate);

		}

		~Bp2Input()
		{
			if (IntPtr.Zero != _strbuf)
#warning Won't work for CF 1.0
				Marshal.FreeHGlobal(_strbuf);
			BppReleaseInput(_nativeInput);
		}

		public IntPtr NativeInput
		{
			get { return _nativeInput; }
		}

		public abstract string DoGetString();
		public abstract char DoGetChar();

#region dllimport
#if X64
#if DEBUG
#if _UNICODE
	  const string dllname = "bpplibud_64.dll";
#else
		const string dllname = "bpplibd_64.dll";
#endif
#else
#if _UNICODE
	  const string dllname = "bpplibu_64.dll";
#else
	  const string dllname = "bpplib_64.dll";
#endif
#endif
#else

#if DEBUG
#if _UNICODE
	  const string dllname = "bpplibud.dll";
#else
		const string dllname = "bpplibd.dll";
#endif
#else
#if _UNICODE
	  const string dllname = "bpplibu.dll";
#else
	  const string dllname = "bpplib.dll";
#endif
#endif
#endif
		[DllImport(dllname, EntryPoint = "Bpp_CreateCallbackInput")]
		private static extern IntPtr BppCreateCallbackInput(InputCharDelegate cc, InputStringDelegate cs);

	  [DllImport(dllname, EntryPoint = "Bpp_ReleaseInput")]
		private static extern bool BppReleaseInput(IntPtr pInput);
		#endregion

		private IntPtr CBGetString()
		{
			string s = DoGetString();
			if (null == s)
				s = "";
			if (IntPtr.Zero != _strbuf)
#warning Won't work for CF 1.0
				Marshal.FreeHGlobal(_strbuf);
			_strbuf = Marshal.StringToHGlobalAnsi(s);
//			int n = s.Length;
//			_strbuf = new byte[n+1];
//			for (int i=0; i<n;++i)
//				_strbuf[i] = (byte) s[i];
//			_strbuf[n] = 0;
			return _strbuf;
		}

		private int CBGetChar()
		{
			return (int) DoGetChar();
		}

		private IntPtr _strbuf;
		private IntPtr _nativeInput;
		private InputCharDelegate _charDelegate;
		private InputStringDelegate _stringDelegate;
	}
}
#endif // !WINCE