/*
 * Created by SharpDevelop.
 * User: erik
 * Date: 01.10.2008
 * Time: 08:43
 * 
 */

using System;
using System.Collections;
using System.Text;
using System.IO;
using bp2builder.InterOp;

namespace bp2builder
{
	/// <summary>
	/// Description of Bp2Project.
	/// </summary>
	public class Bp2Project : Bp2Doc
	{
		public Bp2Project()
		{
			DocType = DocTypes.Project;
		}

		public IList Sources
		{
			get { return _sources; }
			set
			{
				_sources = value;
				Changed = true;
			}
		}

		public IList Modules
		{
			get { return _modules; }
			set
			{
				_modules = value;
				Changed = true;
			}
		}

		public string Name
		{
			get { return _name; }
			set
			{
				_name = value;
				Changed = true;
			}
		}

		public string OutDir
		{
			get { return _outDir; }
			set
			{

				_outDir = value.Replace('\\', '/');
				if (_outDir.EndsWith("/"))
					_outDir = _outDir.Substring(0, _outDir.Length - 1);
				Changed = true;
			}
		}

		public string DestFile
		{
			get { return _destFile; }
			set
			{
				_destFile = value.Replace('\\', '/');
				Changed = true;
			}
		}

		public bool BuildModule
		{
			get { return _buildModule; }
			set
			{
				if (value != _buildModule)
				{
					_buildModule = value;
					Changed = true;
				}
			}
		}

		public bool CreateDebugInfo
		{
			get { return _createDebugInfo; }
			set
			{
				if (value != _createDebugInfo)
				{
					_createDebugInfo = value;
					Changed = true;
				}
			}
		}
		private string collectDocListItems(IList lst)
		{
			StringBuilder result = new StringBuilder();
			IEnumerable enumerable = lst as IEnumerable;
			if (enumerable != null)
			{
				IEnumerator it = enumerable.GetEnumerator();
				while (it.MoveNext())
				{
					Bp2Doc doc = it.Current as Bp2Doc;
					if (null == doc) continue;
					result.AppendFormat("{0}\r\n", MakeRelativePath(doc.Location));
				}
			}
			return result.ToString();
		}

		private string MakeRelativePath(string filePath)
		{
			if (filePath.Length == 0) return filePath;
			string loc = LocPath;
			if (loc.Length == 0) return filePath;
			try
			{
				Uri fileUri = new Uri(filePath.Replace('\\', '/'));
				Uri referenceUri = new Uri(LocPath);
				return referenceUri.MakeRelativeUri(fileUri).ToString();
			}
			catch (Exception e) { }
			return filePath;
		}

		private void buildPrjText()
		{
			StringBuilder result = new StringBuilder();
			result.Append("B++ Project\r\n");
			result.Append("[Settings]\r\n");
			result.AppendFormat("Name={0}\r\n", _name);
			result.AppendFormat("OutDir={0}\r\n", MakeRelativePath(_outDir));
			result.AppendFormat("DestFile={0}\r\n", MakeRelativePath(_destFile));
			result.AppendFormat("IncludeDirs={0}\r\n", _incDirs);
			result.AppendFormat("LibDirs={0}\r\n", _libDirs);
			result.AppendFormat("AddPath={0}\r\n", _pathDirs);
			result.AppendFormat("BuildModule={0}\r\n", _buildModule);
			result.AppendFormat("CreateDebugInfo={0}\r\n", _createDebugInfo);
			result.AppendFormat("ShowCode={0}\r\n", _showCode);
			result.AppendFormat("ShowTrace={0}\r\n", _showTrace);
			result.Append("[Defines]\r\n");
			result.AppendFormat("{0}\r\n", _defines);
			result.Append("[Sources]\r\n");
			result.Append(collectDocListItems(_sources));
			result.Append("[Modules]\r\n");
			result.Append(collectDocListItems(_modules));
			_prjText = result.ToString();
		}

		private string MakeAbsPath(string fpath)
		{
			if (fpath.Length == 0) return fpath;
			try
			{
				Uri uri = new Uri(Path.Combine(LocPath, fpath));
				return (Path.GetFullPath(uri.AbsolutePath));
			}
			catch (Exception e) { }
			return fpath;
		}

		private bool parsePrjText()
		{
			if ((null == _sources) || (null == _modules))
				return false;
			if ((null == _prjText) || (!(_prjText.StartsWith("BOB+2 Project\r\n") || _prjText.StartsWith("B++ Project\r\n"))))
				return false;
			_sources.Clear();
			_modules.Clear();
			_name = "";
			_outDir = "";
			_destFile = "";
			_defines = "";
			StringReader reader = new StringReader(_prjText);
			reader.ReadLine();
			bool inSettings = false;
			bool inSources = false;
			bool inModules = false;
			bool inDefines = false;
			for (string s = reader.ReadLine(); null != s; s = reader.ReadLine())
			{
				s = s.Trim();
				if ((s == "") || s.StartsWith(";")) continue;
				if (s.StartsWith("[Defines]"))
				{
					inSettings = false;
					inSources = false;
					inModules = false;
					inDefines = true;
				}
				else if (s.StartsWith("[Settings]"))
				{
					inSettings = true;
					inSources = false;
					inModules = false;
					inDefines = false;
				}
				else if (s.StartsWith("[Sources]"))
				{
					inSettings = false;
					inSources = true;
					inModules = false;
					inDefines = false;
				}
				else if (s.StartsWith("[Modules]"))
				{
					inSettings = false;
					inSources = false;
					inModules = true;
					inDefines = false;
				}
				else
				{
					if (inDefines)
					{
						_defines = _defines + s + "\r\n";
					}
					else if (inSettings)
					{

						if (s.StartsWith("Name="))
							_name = s.Substring(s.IndexOf('=') + 1);
						else if (s.StartsWith("OutDir="))
							_outDir = MakeAbsPath(s.Substring(s.IndexOf('=') + 1));
						else if (s.StartsWith("DestFile="))
							_destFile = s.Substring(s.IndexOf('=') + 1);
						else if (s.StartsWith("IncludeDirs="))
							_incDirs = s.Substring(s.IndexOf('=') + 1);
						else if (s.StartsWith("LibDirs="))
							_libDirs = s.Substring(s.IndexOf('=') + 1);
						else if (s.StartsWith("AddPath="))
							_pathDirs = s.Substring(s.IndexOf('=') + 1);
						else if (s.StartsWith("BuildModule="))
							_buildModule = Boolean.Parse(s.Substring(s.IndexOf('=') + 1));
						else if (s.StartsWith("CreateDebugInfo="))
							_createDebugInfo = Boolean.Parse(s.Substring(s.IndexOf('=') + 1));
						else if (s.StartsWith("ShowCode="))
							_showCode = Boolean.Parse(s.Substring(s.IndexOf('=') + 1));
						else if (s.StartsWith("ShowTrace="))
							_showTrace = Boolean.Parse(s.Substring(s.IndexOf('=') + 1));
					}
					else if (inSources)
					{
						Bp2SourceDoc doc = new Bp2SourceDoc();
						doc.Location = MakeAbsPath(s);
						doc.OwnerProject = this;
						doc.Changed = false;
						_sources.Add(doc);
					}
					else if (inModules)
					{
						Bp2Doc doc = new Bp2SourceDoc();
						doc.Location = MakeAbsPath(s);
						doc.DocType = DocTypes.Module;
						_modules.Add(doc);
					}
				}
			}
			return true;
		}

		public override bool Load()
		{
			try
			{
				using (StreamReader reader = File.OpenText(Location))
				{
					_prjText = reader.ReadToEnd();
				}
				if (!parsePrjText())
				{
					ReportError("Error parsing " + Location);
					return false;
				}
			}
			catch (Exception e)
			{
				ReportError(e.Message);
				return false;
			}
			Changed = false;
			return true;
		}

		public string LocPath
		{
			get
			{
				string loc = Location.Replace('\\', '/');
				int pos = loc.LastIndexOf('/');
				if (pos <= 0) return loc;
				return loc.Substring(0, pos + 1);
			}
		}


		public override bool Save()
		{
			try
			{
				using (StreamWriter writer = new StreamWriter(Location))
				{
					buildPrjText();
					writer.Write(_prjText);
				}
			}
			catch (Exception e)
			{
				ReportError(e.Message);
				return false;
			}
			Changed = false;
			return true;
		}

		public bool Build()
		{
			if (null == _vm)
			{
				ReportError("Bp2Project.Build : Virtual machine not available.\n");
				return false;
			}
			Bp2Printer printer = _vm.DefaultPrinter;
			_vm.Reinit();
			if (Changed)
			{
				if (!Save())
				{
					ReportError("Bp2Project.Build : Error while saving project file " + Location + ".\n");
					return false;
				}
			}
			IEnumerable src = _sources as IEnumerable;
			if (null != src)
			{
				IEnumerator it = src.GetEnumerator();
				while (it.MoveNext())
				{
					Bp2Doc doc = it.Current as Bp2Doc;
					if (doc.Changed)
					{
						if (!doc.Save())
						{
							ReportError("Bp2Project.Build : Error while saving source file " + doc.Location + ".\n");
							return false;
						}
					}
				}
			}
			IEnumerable modules = _modules as IEnumerable;
			if (null != src)
			{
				IEnumerator it = modules.GetEnumerator();
				while (it.MoveNext())
				{
					Bp2Doc doc = it.Current as Bp2Doc;
					if (null != printer)
						printer.DoPrint("Read module " + doc.Location + ".\n");
					if (!_vm.ReadModule(doc.Location))
					{
						ReportError("Bp2Project.Build : Error while loading module " + doc.Location + ".\n");
						return false;
					}
				}
			}
			if (null != src)
			{
				IEnumerator it = src.GetEnumerator();
				_vm.GenerateDebugInfo = CreateDebugInfo;
				while (it.MoveNext())
				{
					Bp2Doc doc = it.Current as Bp2Doc;
					if (null != printer)
						printer.DoPrint("Compile " + doc.Location + ".\n");
                    _vm.PredefinedSymbols = Defines;
					if (!_vm.CompileFile(doc.Location))
					{
						ReportError("Bp2Project.Build : Error while compiling source file " + doc.Location + ".\n");
						return false;
					}
				}
			}

			if (BuildModule && (DestFile.Length > 0))
			{
				string curdir = Directory.GetCurrentDirectory();
				try
				{
					string odir = OutDir;
					if (odir.Length > 0)
					{
						if (!Directory.Exists(OutDir))
							Directory.CreateDirectory(OutDir);
					}
					else
						odir = FileDir;
					Directory.SetCurrentDirectory(odir);
					if (!_vm.WriteModule(DestFile))
					{
						ReportError("Bp2Project.Build : Exception while writing module " + DestFile + ".\n");
						return false;
					}
				}
				catch (System.Exception e)
				{
					ReportError("Bp2Project.Build : Exception while writing module " + DestFile + ".\n");
					ReportError(e.Message + ".\n");
					return false;
				}
				finally
				{
					Directory.SetCurrentDirectory(curdir);
				}
			}
			if (null != printer)
				printer.DoPrint("Build done.\n");
			_needsBuild = false;
			return true;
		}

		public bool NeedsBuild
		{
			get
			{
				if (_needsBuild || Changed) return true;
				IEnumerable src = _sources as IEnumerable;
				if (null != src)
				{
					IEnumerator it = src.GetEnumerator();
					while (it.MoveNext())
					{
						Bp2Doc doc = it.Current as Bp2Doc;
						if (doc.Changed) return true;
					}
				}
				IEnumerable modules = _modules as IEnumerable;
				if (null != src)
				{
					IEnumerator it = src.GetEnumerator();
					while (it.MoveNext())
					{
						Bp2Doc doc = it.Current as Bp2Doc;
						if (doc.Changed) return true;
					}
				}
				return false;
			}
			set { _needsBuild = value; }

		}

		public Bp2VM VM
		{
			get { return _vm; }
			set
			{
				_vm = value;
				_needsBuild = true;
			}
		}

		public string IncDirs
		{
			get { return _incDirs; }
			set
			{
				_incDirs = value;
				_needsBuild = true;
			}
		}

		public string PathDirs
		{
			get { return _pathDirs; }
			set
			{
				_pathDirs = value;
				_needsBuild = true;
			}
		}

		public string LibDirs
		{
			get { return _libDirs; }
			set
			{
				_libDirs = value;
				_needsBuild = true;
			}
		}

		public bool ShowCode
		{
			get { return _showCode; }
			set { _showCode = value; }
		}

		public bool ShowTrace
		{
			get { return _showTrace; }
			set { _showTrace = value; }
		}

		public string Defines
		{
			get
			{
				return _defines;
			}

			set
			{
				_defines = value;
			}
		}

		Bp2VM _vm = null;
		bool _needsBuild = true;
		string _prjText = "";
		string _outDir = "";
		string _destFile = "";
		string _name = "";
		string _defines = "";
		bool _buildModule = false;
		bool _createDebugInfo = false;
		bool _showCode = false;

		bool _showTrace = false;

		string _incDirs = "";

		string _pathDirs = "";

		string _libDirs = "";

		IList _sources;
		IList _modules;
	}
}
