using System;
using System.Collections;
using System.Windows.Forms;
using System.IO;
using System.Text;
using bp2builder.InterOp;

namespace bp2builder
{
	/// <summary>
	/// B++ source document
	/// </summary>
	public class Bp2SourceDoc : Bp2Doc
	{
		public Bp2SourceDoc()
		{
			DocType = DocTypes.Source;
			_breakpoints = new ArrayList();
		}
		/// <summary>
		/// Check changed state of document.
		/// </summary>
		/// <value>true if document was changed</value>
		/// <remarks>
		/// Check is done comparing own text value to
		/// text in connected edit control (if any).
		/// </remarks>
		public override bool Changed
		{
			get 
			{
				if (null == _text) return base.Changed;
				return (base.Changed || (_text != Text)); 
			}
			set { base.Changed = value; }
		}


		public virtual Control Editor
		{
			get { return _editor; }
			set
			{
				_editor = value;
				if (null != _editor)
				{
					if (null != _text)
						_editor.Text = _text;
					else
						_editor.Text = "";
				}
			}
		}

		public virtual string Text
		{
			get 
			{
				if (null != _editor)
					return _editor.Text;
				return (null != _text) ? _text : "";
			}
			set
			{
				_text = (null != value) ? value : "";
				if (null != _editor)
					_editor.Text = _text;
			}
		}

		public override bool Load()
		{
			try
			{
				using (StreamReader reader = new StreamReader(Location,Encoding.GetEncoding("windows-1252"),true))
				{
					Text = reader.ReadToEnd();
				}
				Loaded = true;
			}
			catch (Exception e)
			{
				ReportError(e.Message);
				return false;
			}
			return true;
		}

		public override bool Save()
		{
			try
			{
				using (StreamWriter writer = new StreamWriter(Location,false,Encoding.GetEncoding("windows-1252")))
				{
					writer.Write(Text);
				}
				if (_editor != null)
					_text = _editor.Text;
			}
			catch (Exception e)
			{
				ReportError(e.Message);
				return false;
			}
			Changed = false;
			return true;
		}

		public Bp2Project OwnerProject
		{
			get { return _ownerProject; }
			set { _ownerProject = value; }
		}

		public bool Loaded
		{
			get { return _loaded; }
			set {_loaded = value; }
		}

		public bool ToggleBreakpoint(int line)
		{
			int idx = _breakpoints.IndexOf(line);
			if (idx < 0)
			{
				_breakpoints.Add(line);
				return true;
			}
			_breakpoints.RemoveAt(idx);
			return false;
		}

		bool HasBreakpoint(int line)
		{
			return _breakpoints.Contains(line);
		}

		public ArrayList Breakpoints
		{
			get { return _breakpoints; }
		}
		 
		private Control _editor = null;
		private string _text = null;
		private bool _loaded = false;
		private Bp2Project _ownerProject = null;
		private ArrayList _breakpoints = null;
	}
}
