/*
 * Created by SharpDevelop.
 * User: erik
 * Date: 01.10.2008
 * Time: 08:56
 * 
 */

using System;
using System.Runtime.InteropServices;


namespace bp2builder.InterOp
{
	/// <summary>
	/// Description of bp2vm.
	/// </summary>
	public class Bp2VM
	{
#region dllimport
#if X64
#if DEBUG
#if _UNICODE
	  const string dllname = "bpplibud_64.dll";
#else
		const string dllname = "bpplibd_64.dll";
#endif
#else
#if _UNICODE
	  const string dllname = "bpplibu_64.dll";
#else
	  const string dllname = "bpplib_64.dll";
#endif
#endif
#else

#if DEBUG
#if _UNICODE
	  const string dllname = "bpplibud.dll";
#else
		const string dllname = "bpplibd.dll";
#endif
#else
#if _UNICODE
	  const string dllname = "bpplibu.dll";
#else
	  const string dllname = "bpplib.dll";
#endif
#endif
#endif

	  [DllImport(dllname, EntryPoint = "Bpp_CreateVM")]
		private static extern IntPtr CreateVM(IntPtr pPrinter);

	  [DllImport(dllname, EntryPoint = "Bpp_ReleaseVM")]
		private static extern bool ReleaseVM(IntPtr pVM);

		[DllImport(dllname, EntryPoint = "Bpp_ClearPredefinedSymbols")]
		private static extern void BppClearPredefinedSymbols(IntPtr pVM);

		[DllImport(dllname, EntryPoint = "Bpp_AddPredefinedSymbolW", CharSet = CharSet.Unicode)]
		private static extern bool BppAddPredefinedSymbol(IntPtr pVM,
										 string pcszSymbolName,
										 string pcszSymbolValue );

		[DllImport(dllname, EntryPoint = "Bpp_CompileFileW", CharSet = CharSet.Unicode)]
		private static extern bool BppCompileFile(IntPtr pVM, 
												 string pcszFileName, 
												 bool bShowInfo, 
												 bool bDecode, 
												 IntPtr pErrPrinter, 
												 IntPtr pDbgPrinter);

	  [DllImport(dllname, EntryPoint = "Bpp_CompileTextW", CharSet = CharSet.Unicode)]
		private static extern bool BppCompileText(IntPtr pVM, 
												 string pcszSrc, 
												 bool bShowInfo, 
												 bool bDecode, 
												 IntPtr pErrPrinter, 
												 IntPtr pDbgPrinter);

	  [DllImport(dllname, EntryPoint = "Bpp_ExecuteW", CharSet = CharSet.Unicode)]
		private static extern bool BppExecute(IntPtr pVM, 
											 string pcszFuncName, 
											 int iArgc,
											 [In,Out] string[] ppcszArgv,
											 bool bTrace, 
											 IntPtr pErrPrinter, 
											 IntPtr pDbgPrinter);

	  [DllImport(dllname, EntryPoint = "Bpp_WriteModuleW", CharSet = CharSet.Unicode)]
		private static extern bool BppWriteModule(IntPtr pVM, 
											 string pcszFileName);

	  [DllImport(dllname, EntryPoint = "Bpp_ReadModuleW", CharSet = CharSet.Unicode)]
		private static extern bool BppReadModule(IntPtr pVM, 
											 string pcszFileName);

	  [DllImport(dllname, EntryPoint = "Bpp_GetIntResult")]
		private static extern int BppGetIntResult(IntPtr pVM);

	  [DllImport(dllname, EntryPoint = "Bpp_Reinit")]
		private static extern void BppReinit(IntPtr pVM);

	  [DllImport(dllname, EntryPoint = "Bpp_SetDbgWaitProc")]
		private static extern IntPtr BppSetDbgWaitProc(IntPtr pVM, DbgWaitDelegate dbgWaitProc);

	  [DllImport(dllname, EntryPoint = "Bpp_SetBreakRequestProc")]
		private static extern IntPtr BppSetBreakRequestProc(IntPtr pVM, BreakRequestDelegate dbgWaitProc);

	  [DllImport(dllname, EntryPoint = "Bpp_SetLineChangedProc")]
	  private static extern IntPtr BppSetLineChangedProc(IntPtr pVM, LineChangedDelegate lineChangedProc);

	  [DllImport(dllname, EntryPoint = "Bpp_SetCurrentInput")]
		private static extern IntPtr BppSetCurrentInput(IntPtr pVM, IntPtr pInput);

	  [DllImport(dllname, EntryPoint = "Bpp_SetGenerateLineInfo")]
	  private static extern void BppSetGenerateLineInfo(IntPtr pVM, bool bGenInfo);

	  [DllImport(dllname, EntryPoint = "Bpp_SetGenerateLocalVarInfo")]
	  private static extern void BppSetGenerateLocalVarInfo(IntPtr pVM, bool bGenInfo);

	  [DllImport(dllname, EntryPoint = "Bpp_GetEnvW", CharSet = CharSet.Unicode)]
	  [return: MarshalAs(UnmanagedType.LPWStr)]
	  private static extern string BppGetEnv(string pcszName);

	  [DllImport(dllname, EntryPoint = "Bpp_SetEnvW", CharSet = CharSet.Unicode)]
	  private static extern void BppSetEnv(string pcszName,string pcszValue);

	  [DllImport(dllname, EntryPoint = "Bpp_Cancel")]
		private static extern void BppCancel(IntPtr pVM);
#endregion

		private void _applyPredefinedSymbols()
		{

		   string[] defs =  _predefinedSymbols.Split("\r\n".ToCharArray(),StringSplitOptions.RemoveEmptyEntries);
			int n = defs.Length;
			BppClearPredefinedSymbols(_vm);
			for (int i = 0; i < n; ++i) 
			{
				string s = defs[i].Trim();
				string s2 = null;
				int pos = s.IndexOf('=');
				if (pos > 0)
				{
					s2 = s.Substring(pos + 1).Trim();
					s = s.Substring(0, pos).Trim();
				}
				if (s.Length == 0) continue;
				BppAddPredefinedSymbol(_vm, s, s2);
			}
		}
	   public delegate bool DbgWaitDelegate(IntPtr pVM);
	   public delegate bool BreakRequestDelegate(IntPtr pVM);
	  public delegate bool LineChangedDelegate(IntPtr pVM, IntPtr pcszFName,int lineNumber);
		
		public Bp2VM(Bp2Printer printer)
		{
			IntPtr p = (IntPtr) 0;
			_defaultPrinter = printer;
			if (printer != null)
				p = printer.NativePrinter;
			_vm = CreateVM(p);
		}
		
		~Bp2VM()
		{
			ReleaseVM(_vm);
		}
		
		public bool CompileFile(string fileName)
		{
			IntPtr pPDbg = (IntPtr) 0;
			IntPtr pPErr = (IntPtr) 0;
			if (_debugPrinter != null)
				pPDbg = _debugPrinter.NativePrinter;
			if (_errorPrinter != null)
				pPErr = _errorPrinter.NativePrinter;
		 BppSetGenerateLineInfo(_vm, _genDebugInfo);
		 BppSetGenerateLocalVarInfo(_vm, _genDebugInfo);
			_applyPredefinedSymbols();
		 return BppCompileFile(_vm, fileName, _showInfo, _showDecode, pPErr, pPDbg);
		}
		
		public bool CompileText(string src)
		{
			IntPtr pPDbg = (IntPtr) 0;
			IntPtr pPErr = (IntPtr) 0;
			if (_debugPrinter != null)
				pPDbg = _debugPrinter.NativePrinter;
			if (_errorPrinter != null)
				pPErr = _errorPrinter.NativePrinter;
		 BppSetGenerateLineInfo(_vm,_genDebugInfo);
		 BppSetGenerateLocalVarInfo(_vm,_genDebugInfo);
			_applyPredefinedSymbols();
			return BppCompileText(_vm,src,_showInfo,_showDecode,pPErr,pPDbg);
		}

		public bool Execute(string funcName, string[] args)
		{
			if ((funcName == null) || (funcName == ""))
			{
				funcName = "main";
			}
			int argc = 0;
			if (args != null)
				argc = args.Length;
			IntPtr pPDbg = (IntPtr) 0;
			IntPtr pPErr = (IntPtr) 0;
			if (_debugPrinter != null)
				pPDbg = _debugPrinter.NativePrinter;
			if (_errorPrinter != null)
				pPErr = _errorPrinter.NativePrinter;
			try
			{
				return BppExecute(_vm, funcName, argc, args, _trace, pPErr, pPDbg);
			}
			catch (AccessViolationException e)
			{
				ErrorPrinter.DoPrint(e.Message);
			}
			catch (Exception e)
			{
				ErrorPrinter.DoPrint(e.Message);
			}
			return false;			
		}
		
		public bool Execute(string funcName)
		{
			return Execute(funcName,null);
		}
		
		public bool Execute(string[] args)
		{
			return Execute(null,args);
		}
		
		public bool Execute()
		{
			return Execute(null,null);
		}

	  public void voidExec()
	  {
		 Execute();
	  }

		public bool WriteModule(string fileName)
		{
			return BppWriteModule(_vm,fileName);
		}
		
		public bool ReadModule(string fileName)
		{
			return BppReadModule(_vm,fileName);
		}

		public void Reinit()
		{
			BppReinit(_vm);
		}
		
		public int IntResult
		{
			get { return BppGetIntResult(_vm); }
		}
		
		public Bp2Printer DefaultPrinter
		{
			get { return _defaultPrinter; }
		}
		
		public Bp2Printer DebugPrinter
		{
			get { return _debugPrinter; }
			set	{ _debugPrinter = value; }
		}
		
		public Bp2Printer ErrorPrinter
		{
			get { return _errorPrinter; }
			set	{ _errorPrinter = value; }
		}
		
		public bool ShowInfo
		{
			get { return _showInfo; }
			set { _showInfo = value; }
		}

	  public bool GenerateDebugInfo
	  {
		 get { return _genDebugInfo; }
		 set { _genDebugInfo = value; }
	  }

	  public bool ShowDecode
		{
			get { return _showDecode; }
			set { _showDecode = value; }
		}
		
		public bool Trace
		{
			get { return _trace; }
			set { _trace = value; }
		}

		public string PredefinedSymbols
		{
			get
			{
				return _predefinedSymbols;
			}

			set
			{
				_predefinedSymbols = value;
			}
		}

		public void SetCurrentInput(Bp2Input inp)
		{
			IntPtr pInput = IntPtr.Zero;
			_currentInput = inp;
			if (null != inp)
				pInput = _currentInput.NativeInput;
			BppSetCurrentInput(_vm,pInput);
		}

		public void SetDbgWaitFunc(DbgWaitDelegate waitFunc)
		{
			_dbgWaitFunc = waitFunc;
			BppSetDbgWaitProc(_vm,_dbgWaitFunc);
		}

		public void SetBreakRequestFunc(BreakRequestDelegate breakRequestFunc)
		{
			_breakRequestFunc = breakRequestFunc;
			BppSetBreakRequestProc(_vm,_breakRequestFunc);
		}

	  public void SetLineChangedFunc(LineChangedDelegate lineChangedFunc)
	  {
		 _lineChangedFunc = lineChangedFunc;
		 BppSetLineChangedProc(_vm, _lineChangedFunc);
	  }

		public void Cancel()
		{
		   BppCancel(_vm);
		}

	  
	  public static string getEnv(string name)
	  {
		 return BppGetEnv(name);
	  }

	  public static void setEnv(string name,string val)
	  {
		 BppSetEnv(name,val);
	  }


		private Bp2Printer _defaultPrinter;
		private Bp2Printer _errorPrinter = null;
		private Bp2Printer _debugPrinter = null;
		private Bp2Input _currentInput = null;
		private IntPtr _vm;
		private bool _showInfo = false;
	  private bool _genDebugInfo = false;
	  private bool _showDecode = false;
		private bool _trace = false;
		private string _predefinedSymbols = "";
		private DbgWaitDelegate _dbgWaitFunc = null;
		private BreakRequestDelegate _breakRequestFunc = null;
	  private LineChangedDelegate _lineChangedFunc = null;
	}
}
